"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");

/**
 * ブートストラップ用のユーティリティ
 */
let arBootstrapUtil = {
  /**
   * resource URI に登録する
   *
   * @param   {string} resourceName
   *          リソースのディレクトリ名
   * @param   {object} data
   *          Bootstrap データ
   */
  registerResource: function(resourceName, data) {
    let alias = Services.io.newFileURI(data.installPath);
    if (data.installPath.isDirectory()) {
      alias.spec += 'res/';
    } else {
      alias = Services.io.newURI("jar:" + alias.spec + "!/res/", null, null);
    }
    
    Services.io.getProtocolHandler("resource")
      .QueryInterface(Ci.nsIResProtocolHandler)
      .setSubstitution(resourceName, alias);  
  },
  
  /**
   * resource URI の登録を解除する
   *
   * @param   {string} resourceName
   *          リソースのディレクトリ名
   * @param   {object} data
   *          Bootstrap データ
   */
  unregisterResource: function(resourceName, data) {
    Services.io.getProtocolHandler("resource")
      .QueryInterface(Ci.nsIResProtocolHandler)
      .setSubstitution(resourceName, null);
  }
};

/**
 * 有効な場合の起動時、無効から有効への変更時 (更新含む)
 *
 * @param   {object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            1: APP_STARTUP
 *            3: ADDON_ENABLE
 *            5: ADDON_INSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function startup(data, reason) {
  try {
    arBootstrapUtil.registerResource("unmht", data);
    
    let context = {};
    Cu.import("resource://unmht/modules/startup.jsm", context);
    
    context.startup(data, reason);
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }
}

/**
 * 有効な場合の終了時、有効から無効への変更時 (更新含む)
 *
 * @param   {object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            2: APP_SHUTDOWN
 *            4: ADDON_DISABLE
 *            6: ADDON_UNINSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function shutdown(data, reason) {
  try {
    let context = {};
    Cu.import("resource://unmht/modules/base/arShutdownEvent.jsm", context);
    
    context.arShutdownEvent.broadcast(data, reason);
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }
  
  if (reason == APP_SHUTDOWN) {
    return;
  }
  
  [
          "UnMHT",
    "UnMHTAdditionalDialog",
    "UnMHTCSSUtil",
    "UnMHTCachePage",
    "UnMHTConfirmDialog",
    "UnMHTContentPolicy",
    "UnMHTCustomFilename",
    "UnMHTDLFHandledLocations",
    "UnMHTDLFRegisterer",
    "UnMHTDOMConverter",
    "UnMHTDOMUtil",
    "UnMHTDocumentCategory",
    "UnMHTDocumentLoaderFactory",
    "UnMHTExtractor",
    "UnMHTFailedListDialog",
    "UnMHTFileConverter",
    "UnMHTFileLoadHook",
    "UnMHTFileLoader",
    "UnMHTFilePicker",
    "UnMHTFileSaver",
    "UnMHTFilename",
    "UnMHTFormatter",
    "UnMHTHTMLUtil",
    "UnMHTInfoPanel",
    "UnMHTKeyboardShortcut",
    "UnMHTMAFIntegration",
    "UnMHTMIMEDecoder",
    "UnMHTMIMEEncoder",
    "UnMHTMIMEParam",
    "UnMHTMIMEPart",
    "UnMHTMIMEUtil",
    "UnMHTMail",
    "UnMHTMailPrivacy",
    "UnMHTMaxRunTime",
    "UnMHTMenu",
    "UnMHTMhtmlProtocolHandler",
    "UnMHTNetFileLoader",
    "UnMHTOverlay",
    "UnMHTPageInfo",
    "UnMHTPref",
    "UnMHTPrefWindow",
    "UnMHTProtocolHandler",
    "UnMHTPseudoID",
    "UnMHTSave",
    "UnMHTSaveDialog",
    "UnMHTSaveMulti",
    "UnMHTSaveMultiDialog",
    "UnMHTSaveOptionsDialog",
    "UnMHTSaveProgress",
    "UnMHTScheme",
    "UnMHTScrapBookIntegration",
    "UnMHTStartupDLFRegisterer",
    "UnMHTStartupExtract",
    "UnMHTStartupWindowHook",
    "UnMHTStatusbarProgress",
    "UnMHTStringBundle",
    "UnMHTSystemDirectory",
    "UnMHTToolbarButton",
    "UnMHTTreeScanner",
    "UnMHTVersion",
    "UnMHTWindowUtil",
    "base/arAppInfo",
    "base/arArrayUtil",
    "base/arAsyncProcs",
    "base/arComponentLoader",
    "base/arDialog",
    "base/arFilePicker",
    "base/arFileUtil",
    "base/arListView",
    "base/arLocaleUtil",
    "base/arOverlayLoader",
    "base/arPathUtil",
    "base/arPrefClass",
    "base/arPrototype",
    "base/arShutdownEvent",
    "base/arStringBundle",
    "base/arStylesheetLoader",
    "base/arSystemDirectory",
    "base/arTimer",
    "base/arUconv",
    "base/arWinRegUtil",
    "base/arWindowHook",
    "base/arWindowHookLoader",
    "startup",
  ].forEach(function(module) {
    Cu.unload("resource://unmht/modules/" + module + ".jsm");
  });
  
  try {
    arBootstrapUtil.unregisterResource("unmht", data);
  } catch (e) {
    Cu.reportError(e + "\n"
                   + e.fileName + ":" + e.lineNumber + "\n"
                   + e.stack);
  }
}

/**
 * インストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param   {object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            5: ADDON_INSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function install(data, reason) {
}

/**
 * アンインストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param   {object} data
 *          Bootstrap データ
 * @param   {number} reason
 *          理由
 *            6: ADDON_UNINSTALL
 *            7: ADDON_UPGRADE
 *            8: ADDON_DOWNGRADE
 */
function uninstall(data, reason) {
}
