"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "arLocaleUtil", function() {
  XPCOMUtils.importRelative(this, "base/arLocaleUtil.jsm");
  return arLocaleUtil;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTExtractor", function() {
  XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
  return UnMHTExtractor;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFileLoader", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileLoader.jsm");
  return UnMHTFileLoader;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFileSaver", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileSaver.jsm");
  return UnMHTFileSaver;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTStatusbarProgress", function() {
  XPCOMUtils.importRelative(this, "UnMHTStatusbarProgress.jsm");
  return UnMHTStatusbarProgress;
});

let EXPORTED_SYMBOLS = ["UnMHT"];

/**
 * UI や他のアドオンへのインターフェース
 */
let UnMHT = {
  /**
   * 設定ダイアログを開く
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  showPrefDialog: function(window) {
    let optionsURL = "chrome://unmht/content/pref/pref.xul";
    
    if (UnMHTWindowUtil.focus(optionsURL)) {
      return;
    }
    
    let features = "chrome,toolbar";
    window.openDialog(optionsURL, "", features);
  },
  
  /**
   * キャッシュを開く
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openCache: function(window) {
    this._addTab(window, UnMHTScheme.createCacheURI());
  },
  
  /**
   * キャッシュをクリア
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  clearCache: function(window) {
    if (window.confirm(UnMHTStringBundle.get("cache.clear.confirm"))) {
      UnMHTExtractor.removeOldFileInfo(0);
    }
  },
  
  /**
   * MHT で保存する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  save: function(window) {
    UnMHTFileSaver.save(window);
  },
  
  /**
   * MHT を上書き保存する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  overwrite: function(window) {
    UnMHTFileSaver.overwrite(window);
  },
  
  /**
   * 対象のドキュメントを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   */
  saveDocument: function(window, targetDocument) {
    UnMHTFileSaver.saveDocument(window, targetDocument);
  },
  
  /**
   * 複数のタブを MHT で保存する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  savemulti: function(window) {
    UnMHTFileSaver.savemulti(window);
  },
  
  /**
   * QuickSave
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  quicksave: function(window) {
    UnMHTFileSaver.quicksave(window);
  },
  
  /**
   * 選択範囲を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  saveSelection: function(window) {
    UnMHTFileSaver.saveSelection(window);
  },
  
  /**
   * 対象のドキュメントの選択範囲を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   */
  saveSelectionInDocument: function(window, targetDocument) {
    UnMHTFileSaver.saveSelectionInDocument(window, targetDocument);
  },
  
  /**
   * リンク先を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  saveLink: function(window) {
    UnMHTFileSaver.saveLink(window);
  },
  
  /**
   * ページを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  sendmail: function(window) {
    UnMHTFileSaver.sendmail(window);
  },
  
  /**
   * 対象のドキュメントを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   */
  sendmailDocument: function(window, targetDocument) {
    UnMHTFileSaver.sendmailDocument(window, targetDocument);
  },
  
  /**
   * 展開して表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openWithExtract: function(window) {
    UnMHTFileLoader.openWithExtract(window);
  },
  
  /**
   * 展開せずに表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openWithoutExtract: function(window) {
    UnMHTFileLoader.openWithoutExtract(window);
  },
  
  /**
   * サイトを開く
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openWebsite: function(window) {
    this._addTab(window, arLocaleUtil.isJapanene()
                 ? "http://www.unmht.org/unmht/"
                 : "http://www.unmht.org/unmht/en_index.html");
  },
  
  /**
   * ステータスバーに表示されている保存をキャンセルする
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  cancelSave: function(window) {
    if (UnMHTStatusbarProgressMap.has(window)) {
      UnMHTStatusbarProgressMap.get(window).cancelSave();
    }
  },
  
  /**
   * mht に含まれるファイルの URI を取得する
   *
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          呼び出し元の URI
   *          unmht 形式
   * @returns {nsIURI}
   *          ファイルの URI
   *          存在しなかった場合は null
   */
  getMHTFileURI: function(contentLocation, requestOrigin) {
    return UnMHTExtractor.getMHTFileURI(contentLocation, requestOrigin);
  },
  
  /* ==== private ==== */
  
  /**
   * タブを開いてフォーカスする
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {string} uri
   *          開くページの URI
   */
  _addTab: function(window, uri) {
    let tabbrowser = UnMHTWindowUtil.getBrowser(window);
    tabbrowser.selectedTab = tabbrowser.addTab(uri);
  }
};
