"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");
XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
XPCOMUtils.importRelative(this, "UnMHTFormatter.jsm");
XPCOMUtils.importRelative(this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTCachePage"];

/**
 * キャッシュページ
 */
let UnMHTCachePage = {
  /**
   * キャッシュページを作成する
   *
   * @returns {string}
   *          キャッシュページのソース
   */
  createCachePage: function() {
    let tableRow = function(header, data) {
      return ("  <tr>"
              + "<th>" + header + "</th>"
              + "<td>" + data + "</td>"
              + "</tr>");
    };
    let _ee = function(text) UnMHTHTMLUtil.escapeEntity(text);
    let bundle = function(name) _ee(UnMHTStringBundle.get(name));
    
    let size = 0;
    
    let eFileInfoList = UnMHTExtractor.getFileInfoList();
    
    let bodies
      = eFileInfoList.map(function(eFileInfo) {
        let body = [];
        
        let _row = function(header, data) {
          body.push(tableRow(header, data));
        };
        
        size += eFileInfo.size;
        
        let param = UnMHTScheme.getUnMHTURIParam(eFileInfo.baseURI);
        let original = param.original;
        if (param.scheme == "file") {
          original = arFileUtil.getFileFromURLSpec(original).path;
        }
        
        body.push("<table class=\"file\">");
        
        _row("URI: ",
             "<a href=\"" + _ee(eFileInfo.baseURI) + "\">"
             + _ee(eFileInfo.baseURI)
             + "</a>");
        _row(bundle("info.original_address") + " ",
             _ee(original));
        _row(bundle("cache.opened_time") + " ",
             _ee(UnMHTFormatter.getReadableDate(eFileInfo.extractDate)));
        _row(bundle("cache.last_access_time") + " ",
             _ee(UnMHTFormatter.getReadableDate(eFileInfo.lastAccessDate)));
        _row(bundle("cache.file_count") + " ",
             eFileInfo.parts.length);
        
        if (eFileInfo.date) {
          let savedTime = eFileInfo.date;
          let date = UnMHTMIMEDecoder.decodeDate(savedTime);
          if (date) {
            savedTime = UnMHTFormatter.getReadableDate(date);
          }
          
          _row(bundle("info.saved_date") + " ",
               _ee(savedTime));
        }
        
        if (eFileInfo.subject) {
          _row("Subject: ",
               _ee(eFileInfo.subject));
        }
        
        body.push("</table>");
        
        eFileInfo.parts.forEach(function(part) {
          body.push("<table class=\"part\">");
          _row(bundle("cache.ref_name") + " ",
               "<a href=\""
               + _ee(eFileInfo.baseURI) + _ee(part.eParam.refName)
               + "\">"
               + _ee(part.eParam.refName)
               + "</a>"
               + ((eFileInfo.startPart == part)
                  ? (" (" + bundle("cache.start_part") + ")")
                  : ""));
          if (part.contentLocation) {
            _row("Content-Location: ",
                 _ee(part.contentLocation));
          }
          if (part.eParam.contentBase) {
            _row("Content-Base: ",
                 _ee(part.eParam.contentBase));
          }
          if (part.contentID) {
            _row("Content-ID: ",
                 _ee(part.contentID));
          }
          if (part.contentDispositionType &&
              part.contentDispositionFilename) {
            _row("Content-Disposition: ",
                 _ee(part.contentDispositionType)
                 + "; filename=" + _ee(part.contentDispositionFilename));
          }
          _row("Content-Type: ",
               _ee(part.eParam.mimetype)
               + (part.eParam.charset
                  ? ("; charset=" + _ee(part.eParam.charset))
                  : ""));
          if (part.contentTransferEncoding) {
            _row("Content-Encoding: ",
                 _ee(part.contentTransferEncoding));
          }
          _row(bundle("cache.file_size") + " ",
               UnMHTFormatter.getReadableSize(part.eParam.content.length)
               + " " + bundle("cache.bytes"));
          body.push("</table>");
        });
        
        return body.join("\n");
      });
    
    let data
      = "<!DOCTYPE html>" + "\n"
      + "<html xmlns=\"http://www.w3.org/1999/xhtml\">" + "\n"
      + "<head>" + "\n"
      + "  <title>" + bundle("cache.title") + "</title>" + "\n"
      + "  <style type=\"text/css\">" + "\n"
      + "html {" + "\n"
      + "  background: -moz-Dialog;" + "\n"
      + "  padding: 0 1em;" + "\n"
      + "  font: message-box;" + "\n"
      + "}" + "\n"
      + "\n"
      + "body {" + "\n"
      + "  color: -moz-FieldText;" + "\n"
      + "  min-width: 330px;" + "\n"
      + "  max-width: 80%;" + "\n"
      + "  margin: 4em auto;" + "\n"
      + "  border: 1px solid ThreeDShadow;" + "\n"
      + "  border-radius: 10px;" + "\n"
      + "  padding: 3em;" + "\n"
      + "  -moz-padding-start: 30px;" + "\n"
      + "  background: -moz-Field;" + "\n"
      + "}" + "\n"
      + "\n"
      + "table {" + "\n"
      + "  width: 100%;" + "\n"
      + "  padding: 0.5em;" + "\n"
      + "}" + "\n"
      + "\n"
      + "th {" + "\n"
      + "  width: 12em;" + "\n"
      + "  word-wrap: break-word;" + "\n"
      + "  vertical-align: top;" + "\n"
      + "  text-align: end;" + "\n"
      + "  -moz-padding-end: 0.5em;" + "\n"
      + "}" + "\n"
      + "\n"
      + "td {" + "\n"
      + "  display: block;" + "\n"
      + "  font-family: -moz-fixed;" + "\n"
      + "  white-space: pre-wrap;" + "\n"
      + "}" + "\n"
      + "\n"
      + "#entries {" + "\n"
      + "  border-radius: 10px;" + "\n"
      + "  background-color: rgba(0, 0, 0, .1);" + "\n"
      + "}" + "\n"
      + "\n"
      + "table.file {" + "\n"
      + "  border-radius: 10px 10px 0px 0px;" + "\n"
      + "  background-color: rgba(0, 0, 0, .1);" + "\n"
      + "}" + "\n"
      + "\n"
      + "table.part:nth-child(odd) {" + "\n"
      + "  background-color: rgba(0, 0, 0, .03);" + "\n"
      + "}" + "\n"
      + "\n"
      + "table.part:nth-child(even) {" + "\n"
      + "  background-color: rgba(0, 0, 0, .06);" + "\n"
      + "}" + "\n"
      + "  </style>"
      + "</head>" + "\n"
      + "<body>" + "\n"
      + "<h1>" + bundle("cache.title") + "</h1>" + "\n"
      + "<table id=\"entries\">" + "\n"
    
      + tableRow(bundle("cache.entry") + " ",
                 eFileInfoList.length) + "\n"
    
      + tableRow(bundle("cache.size") + " ",
                 UnMHTFormatter.getReadableSize(size)
                 + " " + bundle("cache.bytes")) + "\n"
    
      + "</table>" + "\n"
      + "<hr />" + "\n"
      + bodies.join("<hr />") + "\n"
      + "</body>" + "\n"
      + "</html>" + "\n";
    
    return arUconv.fromUnicode(data, "utf-8");
  }
};
