"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arDialog.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");

let EXPORTED_SYMBOLS = ["UnMHTConfirmDialogClass"];

/**
 * 汎用確認ダイアログ
 *
 * ウィンドウの引数
 * {
 *   result:   {object}
 *             結果を格納するオブジェクト (オプショナル),
 *
 *   focus:    {string}
 *             デフォルトでフォーカスする対象 (オプショナル)
 *               (yes/no/abort/skip/rename),
 *
 *   title:    {string}
 *             ダイアログのタイトル (必須),
 *   message1: {string}
 *             ダイアログのメッセージ (必須),
 *   message2: {string}
 *             ダイアログのメッセージ 2 行目 (オプショナル),
 *   message3: {string}
 *             ダイアログのメッセージ 3 行目 (オプショナル),
 *
 *   accept:   {string}
 *             OK ボタン [yes] のラベル (必須),
 *   cancel:   {string}
 *             キャンセルボタン [no] のラベル (オプショナル),
 *   extra1:   {string}
 *             スキップボタン [skip] のラベル (オプショナル),
 *   extra2:   {string}
 *             中断ボタン [abort] のラベル (オプショナル),
 *   toall:    {string}
 *             全てに適用チェックのラベル (オプショナル),
 *
 *   rename:   {string}
 *             リネームボタン [rename] のラベル (オプショナル),
 *   file:     {string}
 *             既存のファイル (rename により必須),
 *   exist:    {Function}
 *             ファイルが存在するメッセージのフォーマッタ (rename により必須)
 *               function(leafName)
 *                 *param   {string} leafName
 *                          ファイル名
 *                 *returns {string}
 *                          ファイルが存在するメッセージ
 * }
 *
 * 結果
 * {
 *   result:     {string}
 *               押したボタン (yes/no/abort/skip/rename),
 *   applyToAll: {boolean}
 *               全てに適用するか,
 *   file:       {boolean}
 *               リネームするファイル
 * }
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialog
 * @implements arIDialog
 * @implements arIDialogExtra
 */
function UnMHTConfirmDialogClass(window) {
  arDialog.call(this, window);
}
UnMHTConfirmDialogClass.prototype = {
  /**
   * 各種ボタンを押したイベント
   *
   * @param   {string} type
   *          ボタンの種類 (yes/no/abort/skip/rename)
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onClick: function(type) {
    this.el("message").value = "";
    this.el("message").hidden = true;
    
    if (type == "rename") {
      let filenameBase
        = UnMHTFilename.removeExtension(this.el("filename").value);
      let file
        = arFileUtil.checkAndAppend(this._file.parent, filenameBase, "",
                                    UnMHTFilename.defaultExtension);
      
      if (file.exists()) {
        this.el("message").value = this._existFormatter(file.leafName);
        this.el("message").hidden = false;
        this.el("filename").focus();
        
        return false;
      }
      
      if (this._result) {
        this._result.file = file;
      }
    }
    
    if (this._result) {
      this._result.result = type;
      this._result.applyToAll = this.el("apply_to_all").checked;
    }
    
    if (type == "abort" ||
        type == "skip" ||
        type == "rename") {
      /* ダイアログのボタンではないので手動で閉じる */
      this._window.close();
    }
    
    return true;
  },
  
  /* ==== arIDialog インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._result = null;
    this._file = null;
    arDialog.prototype.destruct.call(this);
  },
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this._window.arguments[0];
    
    let dialog = this.el("unmht_confirm");
    
    if ("result" in arg) {
      this._result = arg.result;
    }
    
    dialog.setAttribute("title", arg.title);
    this.el("title").setAttribute("value", arg.title);
    
    this.el("message1").value = arg.message1;
    if ("message2" in arg && arg.message2) {
      this.el("message2").value = arg.message2;
      this.el("message2").hidden = false;
    }
    if ("message3" in arg && arg.message3) {
      this.el("message3").value = arg.message3;
      this.el("message3").hidden = false;
    }
    
    dialog.getButton("accept").setAttribute("label", arg.accept);
    
    if ("cancel" in arg) {
      dialog.getButton("cancel").setAttribute("label", arg.cancel);
    } else {
      dialog.getButton("cancel").hidden = true;
    }
    
    if ("extra1" in arg) {
      dialog.getButton("extra1").setAttribute("label", arg.extra1);
    } else {
      dialog.getButton("extra1").hidden = true;
    }
    
    if ("extra2" in arg) {
      dialog.getButton("extra2").setAttribute("label", arg.extra2);
    } else {
      dialog.getButton("extra2").hidden = true;
    }
    
    if ("toall" in arg && arg.toall) {
      this.el("apply_to_all").checked = false;
      this.el("apply_to_all").label = arg.toall;
      this.el("apply_to_all").hidden = false;
    }
    
    if ("rename" in arg) {
      this.el("rename").label = arg.rename;
      
      if ("file" in arg) {
        this.el("filename").value = arg.file.leafName;
        this._file = arg.file;
      }
      
      this.el("renamebox").hidden = false;
    }
    
    if ("exist" in arg && arg.exist) {
      this._existFormatter = arg.exist;
    }
    
    if ("focus" in arg) {
      switch (arg.focus) {
        case "yes": {
          dialog.getButton("accept").focus();
          break;
        }
        case "no": {
          dialog.getButton("cancel").focus();
          break;
        }
        case "abort": {
          dialog.getButton("extra2").focus();
          break;
        }
        case "skip": {
          dialog.getButton("extra1").focus();
          break;
        }
        case "rename": {
          this.el("rename").focus();
          break;
        }
      }
    }
    
  },
  
  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    return this.onClick("yes");
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    return this.onClick("no");
  },
  
  /* ==== arIDialogExtra インターフェース ==== */
  
  /**
   * ダイアログの extra1 を押したイベント
   */
  onDialogExtra1: function() {
    this.onClick("skip");
  },
  
  /**
   * ダイアログの extra2 を押したイベント
   */
  onDialogExtra2: function() {
    this.onClick("abort");
  },
  
  /* ==== private ==== */

  _result: null,        /** 結果オブジェクト
                         *  @type {object}}*/
  _file: null,          /** 既存のファイル
                         *  @type {nsIFile} */
  _existFormatter: null /** ファイルが存在するメッセージのフォーマッタ
                         *  @type {Function} */
};
arPrototype.extend(UnMHTConfirmDialogClass, arDialog);
