"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTExtractor", function() {
  XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
  return UnMHTExtractor;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTScheme", function() {
  XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
  return UnMHTScheme;
});

let EXPORTED_SYMBOLS = ["UnMHTContentPolicy"];

/**
 * mht から参照されているファイルのチェック、変換を行う
 *
 * @implements nsIContentPolicy
 */
function UnMHTContentPolicy() {
}
UnMHTContentPolicy.prototype = {
  classID:           Components.ID("{5c21c8ee-2d1e-46c1-9d6f-bc4876c13890}"),
  contractID:        "@unmht.org/unmht-content-policy;1",
  classDescription:  "UnMHT Content Policy JS Component",
  _xpcom_categories: [{ category: "content-policy" }],
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIContentPolicy]),
  
  /* ==== nsIContentPolicy インターフェース ==== */
  
  /**
   * ロードするかどうかのチェック
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          参照元の URI (オプショナル)
   * @param   {nsIDOMElement} context
   *          ロード先 (オプショナル)
   * @param   {string} mimeTypeGuess
   *          予想される MIME-Type (オプショナル)
   * @param   {nsISupports} extra
   *          付加データ (オプショナル)
   */
  shouldLoad: function(contentType, contentLocation,
                       requestOrigin, context,
                       mimeTypeGuess, extra) {
    if (this._shouldLoadPreCheck(contentType, contentLocation,
                                 requestOrigin, context)) {
      return Ci.nsIContentPolicy.ACCEPT;
    }
    
    return Ci.nsIContentPolicy.REJECT_OTHER;
  },
  
  /**
   * 処理するかどうかのチェック
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI (オプショナル)
   * @param   {nsIURI} requestOrigin
   *          参照元の URI (オプショナル)
   * @param   {nsIDOMElement} context
   *          ロード先 (オプショナル)
   * @param   {string} mimeType
   *          MIME-Type
   * @param   {nsISupports} extra
   *          付加データ (オプショナル)
   */
  shouldProcess: function(contentType, contentLocation,
                          requestOrigin, context,
                          mimeTypeGuess, extra) {
    return Ci.nsIContentPolicy.ACCEPT;
  },

  /* ==== private ==== */

  /**
   * shouldLoad のうち UnMHTContentPolicy で行うもの
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          参照元の URI
   * @param   {nsIDOMElement} context
   *          ロード先
   * @returns {boolean}
   *          許可するかどうか
   */
  _shouldLoadPreCheck: function(contentType, contentLocation,
                                requestOrigin, context) {
    if (!requestOrigin) {
      /* 呼び出し元が無い場合チェック不可能なので許可する */
      return true;
    }
    
    if (contentLocation.scheme == "chrome" ||
        contentLocation.scheme == "data") {
      /* chrome, data への参照は対象外 */
      return true;
    }
    
    if (contentLocation.scheme == "unmht" &&
        requestOrigin.scheme != "file" &&
        requestOrigin.scheme != "unmht" &&
        requestOrigin.scheme != "chrome" &&
        requestOrigin.scheme != "moz-nullprincipal") {
      let param = UnMHTScheme.getUnMHTURIParam(contentLocation.spec);
      if (param.scheme == "file") {
        /* file, unmht, chrome, moz-nullprincipal 以外からの
         * unmht を経由した file への参照は拒否する */
        return false;
      }
    }
    
    if (requestOrigin.scheme != "unmht") {
      /* unmht 以外からの参照は対象外 */
      return true;
    }
    /* ---- ここから unmht からの参照 ---- */
    
    if (!context) {
      /* ロード先が無い場合 URI の置換えができないので許可する */
      return true;
    }
    
    if (context instanceof Ci.nsIDOMXULElement) {
      /* タブを開く場合対象外 */
      return true;
    }
    
    /* ---- ここからドキュメント内へのロード ---- */
    
    return UnMHTExtractor.shouldLoad(contentType, contentLocation,
                                     requestOrigin, context);
  }
};
