"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTCustomFilename"];

/**
 * カスタムファイル名フォーマッタ
 */
let UnMHTCustomFilename = {
  /**
   * カスタムファイル名をフォーマットする
   * 設定は自動取得する
   *
   * @param   {string} TITTLE
   *          ページのタイトル
   * @param   {string} FILENAME
   *          ページのファイル名
   * @param   {string} URL
   *          ページの URL
   * @param   {Date} date
   *          日時
   * @returns {string}
   *          フォーマットされた文字列
   */
  format: function(TITLE, FILENAME, URL, date) {
    return this.formatWith(TITLE, FILENAME, URL, date,
                           UnMHTPref.getPref("save.filename.custom"),
                           UnMHTPref.getPref("save.filename.custom.am"),
                           UnMHTPref.getPref("save.filename.custom.pm"));
  },

  /**
   * カスタムファイル名をフォーマットする
   *
   * @param   {string} TITTLE
   *          ページのタイトル
   * @param   {string} FILENAME
   *          ページのファイル名
   * @param   {string} URL
   *          ページの URL
   * @param   {Date} date
   *          日時
   * @param   {string} pattern
   *          カスタムファイル名フォーマット文字列
   * @param   {string} am
   *          a.m. の文字列
   * @param   {string} pm
   *          p.m. の文字列
   * @returns {string}
   *          フォーマットされた文字列
   */
  formatWith: function(TITLE, FILENAME, URL, date, pattern, am, pm) {
    let uri = Services.io.newURI(URL, null, null);
    
    let pad2 = function(n) {
      return ("0" + n).slice(-2);
    };
    
    let port = uri.port;
    if (port == -1) {
      if (uri.scheme == "http") {
        port = 80;
      } else if (uri.scheme == "https") {
        port = 443;
      } else if (uri.scheme == "ftp") {
        port = 21;
      }
    }
    
    let [, path, query, fragment]
      = uri.path.match(/([^\?#]*)(?:\?([^#]*))?(?:#(.*))?/);
    query = query || "";
    fragment = fragment || "";
    
    let decodedURL = URL;
    try {
      decodedURL = decodeURIComponent(URL);
    } catch (e) {
    }
    try {
      query = decodeURIComponent(query);
    } catch (e) {
    }
    try {
      fragment = decodeURIComponent(fragment);
    } catch (e) {
    }
    
    if (UnMHTPref.getPref("save.filename.custom.limit")) {
      let limitCount = UnMHTPref.getPref("save.filename.custom.limit.count");
      TITLE = TITLE.substr(0, limitCount);
      FILENAME = FILENAME.substr(0, limitCount);
    }
    
    let vars = {
      YYYY: date.getFullYear(),
      Y:    date.getFullYear() % 100,
      YY:   pad2(date.getFullYear() % 100),
      M:    date.getMonth() + 1,
      MM:   pad2(date.getMonth() + 1),
      MMM:  this._weekdaysShort[date.getDay()],
      MMMM: this._weekdaysLong[date.getDay()],
      D:    date.getDate(),
      DD:   pad2(date.getDate()),
      H:    date.getHours(),
      HH:   pad2(date.getHours()),
      h:    date.getHours() % 12,
      hh:   pad2(date.getHours() % 12),
      m:    date.getMinutes(),
      mm:   pad2(date.getMinutes()),
      s:    date.getSeconds(),
      ss:   pad2(date.getSeconds()),
      ampm: date.getHours() >= 12 ? pm : am,
      
      TITLE:    TITLE,
      FILENAME: FILENAME,
      URL:      decodedURL,
      URL2:     decodedURL.replace(/^[A-Za-z0-9]+:\/\//, ""),
      SCHEME:   uri.scheme,
      HOST:     uri.host,
      PORT:     port,
      FULLPATH: uri.path,
      PATH:     path,
      QUERY:    query,
      FRAGMENT: fragment
    };
    
    let rexp = new RegExp("%(" + Object.keys(vars).join("|") + ")%", "g");
    
    return pattern.replace(rexp, function(matched, name) vars[name]);
  },
  
  /* ==== private ==== */

  /** 曜日 (省略)
   *  @type {Array.<string>} */
  _weekdaysShort: ["sun", "mon", "tue", "wed", "thu", "fri", "sat"]
    .map(function(n) UnMHTStringBundle.get("weekday." + n + ".short")),
  
  /** 曜日
   *  @type {Array.<string>} */
  _weekdaysLong: ["sun", "mon", "tue", "wed", "thu", "fri", "sat"]
    .map(function(n) UnMHTStringBundle.get("weekday." + n + ".long")),
};
