"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTMailPrivacy.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTPseudoID", function() {
  XPCOMUtils.importRelative(this, "UnMHTPseudoID.jsm");
  return UnMHTPseudoID;
});

let EXPORTED_SYMBOLS = ["UnMHTDOMConverter"];

/**
 * DOM ツリーを文字列に変換する
 */
let UnMHTDOMConverter = {
  /**
   * DOM ツリーを文字列に変換する
   *
   * @param   {string} location
   *          ドキュメントの URI
   * @param   {nsIDOMNode} node
   *          対象の要素
   * @param   {nsISelection} targetSelection
   *          対象の選択範囲もしくは null
   * @param   {object} params
   *            {
   *              noScript:   {boolean}
   *                          スクリプトを削除するか
   *              isMail:     {boolean}
   *                          メールか
   *              saveCanvas: {boolean}
   *                          キャンバスを保存するか
   *              frameIDs:   {Map.<<nsIDOMElement,string>}
   *                          フレームの Content-ID
   *            }
   * @returns {string}
   *          要素全体の HTML
   */
  toPlainText: function(location, node, targetSelection, params) {
    if (!node) {
      return "";
    }
    if (!node.nodeName) {
      return "";
    }
    
    if (targetSelection) {
      if (node instanceof Ci.nsIDOMHTMLHeadElement ||
          node instanceof Ci.nsIDOMHTMLStyleElement ||
          node instanceof Ci.nsIDOMHTMLTitleElement) {
        targetSelection = null;
      }
    }
    
    if (targetSelection) {
      if (!UnMHTDOMUtil.isNodeInSelection(node, targetSelection)) {
        return "";
      }
    }
    
    let text = "";
    
    if ("ignoreNodes" in params &&
        params.ignoreNodes.indexOf(node) != -1) {
      return "";
    }
    
    if (node instanceof Ci.nsIDOMText) {
      text = node.nodeValue;
      
      if (targetSelection) {
        let textList = [];
        for (let i = 0; i < targetSelection.rangeCount; i ++) {
          let t = text;
          let isPartial = false;
          
          let range = targetSelection.getRangeAt(i);
          
          if (range.endContainer == node) {
            try {
              t = t.substr(0, range.endOffset);
              isPartial = true;
            } catch (e) {
            }
          }
          
          if (range.startContainer == node) {
            try {
              t = t.substr(range.startOffset);
              isPartial = true;
            } catch (e) {
            }
          }
          
          if (isPartial) {
            textList.push(t);
          }
        }
        
        if (textList.length > 0) {
          text = textList.join("");
        }
      }
      
      text = UnMHTHTMLUtil.escapeEntity(text);
    } else if (node instanceof Ci.nsIDOMComment) {
      /* コメントをコメントとして展開する */
      text = "<!--" + node.nodeValue + "-->";
    } else if (node instanceof Ci.nsIDOMHTMLBaseElement &&
               "isMail" in params &&
               params.isMail) {
    } else if (node instanceof Ci.nsIDOMHTMLScriptElement &&
               "noScript" in params &&
               params.noScript) {
    } else {
      let attributes = "";
      
      let addStylePre = "";
      let addStylePost = "";
      
      if ("saveCanvas" in params &&
          params.saveCanvas &&
          node instanceof Ci.nsIDOMHTMLCanvasElement) {
        addStylePre
          += "background-image: url(" + node.toDataURL("image/png") + ");";
      }
      
      if (node.attributes) {
        Array.slice(node.attributes)
          .forEach(function(attr) {
            let name = attr.name.toLowerCase();
            let value = attr.value;
            
            if (/^__unmht_/.test(name)) {
              return;
            }
            
            if (node instanceof Ci.nsIDOMHTMLOptionElement) {
              if (name == "selected") {
                return;
              }
            }
            
            if ("noScript" in params &&
                params.noScript &&
                /^on/.test(name)) {
              return;
            }
            
            if (node instanceof Ci.nsIDOMHTMLInputElement) {
              if (node.type == "checkbox" ||
                  node.type == "radio") {
                if (name == "checked") {
                  return;
                }
              } else {
                if (name == "value") {
                  return;
                }
              }
            }
            
            if ((node instanceof Ci.nsIDOMHTMLFrameElement ||
                 node instanceof Ci.nsIDOMHTMLIFrameElement) &&
                name == "src") {
              if ("frameIDs" in params) {
                let id = "";
                if (params.frameIDs.has(node)) {
                  id = params.frameIDs.get(node);
                } else {
                  let size;
                  try {
                    /* COMPAT: Firefox 18 以前 */
                    size = params.frameIDs.size();
                  } catch (e) {
                    size = params.frameIDs.size;
                  }
                  id
                    = "iframe." + size + "." + UnMHTPseudoID.gen()
                    + "@unmht.org";
                  params.frameIDs.set(node, id);
                }
                value = "cid:" + id;
              } else {
                value = node.contentDocument.location.href;
              }
            }
            
            if (/^(src|href|background|action|data)$/.test(name) &&
                "isMail" in params &&
                params.isMail) {
              value = UnMHTMailPrivacy.hideMailURL(value, location);
            }
            
            if (name == "style") {
              if (addStylePre) {
                value = addStylePre + value;
                addStylePre = "";
              }
              if (addStylePost) {
                value = value + ";" + addStylePost;
                addStylePost = "";
              }
            }
            
            value = UnMHTHTMLUtil.escapeEntity(value);
            
            attributes += " " + name +  "=" + "\"" + value + "\"";
          });
      }
      
      if (addStylePre || addStylePost) {
        let name = "style";
        let value = "";
        if (addStylePre) {
          value = addStylePre + value;
          addStylePre = "";
        }
        if (addStylePost) {
          value = value + ";" + addStylePost;
          addStylePost = "";
        }
        attributes += " " + name +  "=" + "\"" + value + "\"";
      }
      
      if (node instanceof Ci.nsIDOMHTMLOptionElement) {
        if (node.selected) {
          let name = "selected";
          let value = "selected";
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }

      if (node instanceof Ci.nsIDOMHTMLInputElement) {
        if (node.type == "checkbox" ||
            node.type == "radio") {
          if (node.checked) {
            let name = "checked";
            let value = "checked";
            attributes += " " + name +  "=" + "\"" + value + "\"";
          }
        } else if (node.type == "file") {
        } else {
          let name = "value";
          let value = UnMHTHTMLUtil.escapeEntity(node.value);
          attributes += " " + name +  "=" + "\"" + value + "\"";
        }
      }

      let nodeName = node.nodeName.toLowerCase();
      
      text += "<" + nodeName + attributes + ">";
      
      /* NOSCRIPT, NOEMBED 要素は専用のインターフェース定義を持たない */
      if (nodeName == "noscript" &&
          node.firstChild &&
          node.firstChild instanceof Ci.nsIDOMText &&
          node.firstChild.nextSibling == null) {
        /* NOSCRIPT 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      } else if (nodeName == "noembed" &&
                 node.firstChild &&
                 node.firstChild instanceof Ci.nsIDOMText &&
                 node.firstChild.nextSibling == null) {
        /* NOEMBED 要素が展開されていない場合 */
        text += node.firstChild.nodeValue;
      } else if (node instanceof Ci.nsIDOMHTMLStyleElement ||
                 node instanceof Ci.nsIDOMHTMLScriptElement) {
        /* noScript の場合には SCRIPT 要素はここまで来ない */
        let node2 = node.firstChild;
        while (node2) {
          text
            += UnMHTHTMLUtil.unescapeEntity(
              this.toPlainText(location, node2, targetSelection, params));
          
          node2 = node2.nextSibling;
        }
      } else if (node instanceof Ci.nsIDOMHTMLTextAreaElement) {
        text += UnMHTHTMLUtil.escapeEntity(node.value);
      } else {
        let node2 = node.firstChild;
        while (node2) {
          text
            += this.toPlainText(location, node2, targetSelection, params);
          
          node2 = node2.nextSibling;
        }
      }
      
      if (!(node instanceof Ci.nsIDOMHTMLBRElement ||
            node instanceof Ci.nsIDOMHTMLBaseElement ||
            node instanceof Ci.nsIDOMHTMLLinkElement ||
            node instanceof Ci.nsIDOMHTMLImageElement ||
            node instanceof Ci.nsIDOMHTMLInputElement ||
            node instanceof Ci.nsIDOMHTMLHRElement ||
            node instanceof Ci.nsIDOMHTMLMetaElement)) {
        text += "</" + nodeName + ">";
      }
    }
    
    return text;
  }
};
