"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTDOMUtil"];

/**
 * DOM 操作ユーティリティ
 */
let UnMHTDOMUtil = {
  /**
   * ノードが選択範囲内かどうかチェックする
   *
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @param   {nsISelection} selection
   *          対象の選択範囲
   * @returns {boolean}
   *          範囲内かどうか
   */
  isNodeInSelection: function(node, selection) {
    let nodeRange = node.ownerDocument.createRange();
    let w = node.ownerDocument.defaultView;
    nodeRange.selectNode(node);
    
    for (let i = 0; i < selection.rangeCount; i ++) {
      let range = selection.getRangeAt(i);
      
      let compareSE = 0;
      let compareES = 0;
      try {
        compareSE
          = range.compareBoundaryPoints(range.START_TO_END, nodeRange);
        compareES
          = range.compareBoundaryPoints(range.END_TO_START, nodeRange);
      } catch (e) {
        /* 選択範囲が別のドキュメント */
        return false;
      }
      
      if (compareSE == -compareES) {
        return true;
      }
    }
    
    return false;
  },
  
  /**
   * ノードを親方向に探す
   * 対象のノードも含む
   *
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @param   {string} nodeName
   *          探すノード名
   * @returns {nsIDOMNode}
   *          見付かったノード
   *          見付からなかった場合 null
   */
  findParentNode: function(node, nodeName) {
    while (node) {
      if (node.nodeName.toLowerCase() == nodeName) {
        return node;
      }
      node = node.parentNode;
    }
    
    return null;
  },
  
  /**
   * コンテンツが選択されているかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @param   {boolean} usePopupNode
   *          コンテキストメニューから情報を取得するか
   * @returns {boolean}
   *          コンテンツが選択されているかどうか
   */
  isContentSelected: function(window, usePopupNode) {
    if (usePopupNode &&
        "gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.isContentSelected;
    } else {
      let targetDocument = null;
      if (usePopupNode && window.document.popupNode) {
        targetDocument = window.document.popupNode.ownerDocument;
      } else {
        targetDocument = UnMHTWindowUtil.getCurrentDocument(window);
      }
      
      if (!targetDocument.defaultView) {
        return false;
      }
      
      try {
        return !targetDocument.defaultView.getSelection().isCollapsed;
      } catch (e) {
        return false;
      }
    }
  },
  
  /**
   * リンク先の URI を取得する
   * 
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @param   {nsIDOMNode} node
   *          対象のノード
   * @returns {string}
   *          リンク先の URI
   */
  getLinkURL: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.linkURL;
    }
    return "";
  },
  
  /**
   * 保存可能なリンクの上に乗っているかどうかを返す
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @returns {boolean}
   *          リンクの上に乗っているかどうか
   */
  onSavableLink: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return window.gContextMenu.onSaveableLink;
    }
    return false;
  },
  
  /**
   * ページ上の何でもない所に乗っているかどうかを返す
   *   (リンクや画像、入力フィールド等でない事)
   * コンテキストメニューを表示してからでないと正しい値を返さない
   *
   * @param   {nsIDOMWindow} window
   *          動作中のウィンドウ
   * @returns {boolean}
   *          ページ上の何でもない所に乗っているかどうか
   */
  onPage: function(window) {
    if ("gContextMenu" in window &&
        window.gContextMenu) {
      return !(window.gContextMenu.onImage ||
               window.gContextMenu.onCanvas ||
               window.gContextMenu.onVideo ||
               window.gContextMenu.onAudio ||
               window.gContextMenu.onLink ||
               window.gContextMenu.onTextInput);
    }
    
    return true;
  }
};
