"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");
XPCOMUtils.importRelative(this, "base/arPathUtil.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");
XPCOMUtils.importRelative(this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");
XPCOMUtils.importRelative(this, "UnMHTMaxRunTime.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTPseudoID.jsm");

let EXPORTED_SYMBOLS = ["UnMHTExtractor"];

/**
 * パートの展開情報
 */
function UnMHTExtractParam() {
  this.cid = "part." + UnMHTPseudoID.gen() + "@unmht.org";
}
UnMHTExtractParam.prototype = {
  charset: "",         /** キャラクタセット
                        *  @type {string} */
  mimetype: "",        /** MIME-Type
                        *  @type {string} */
  
  isHTML: false,       /** HTML かどうか
                        *  @type {boolean} */
  isCSS: false,        /** CSS かどうか
                        *  @type {boolean} */
  
  isMixed: false,      /** multipart/mixed かどうか
                        *  @type {boolean} */
  
  cid: "",             /** Content-ID 代わり
                        *  @type {string} */
  
  location: "",        /** ファイルのアドレス
                        *  @type {string} */
  normalLocation: "",  /** 正規化したファイルのアドレス
                        *  @type {string} 
                        *    フラグメント抜き */
  
  contentBase: "",     /** Content-Base の値
                        *  @type {string} */
  
  baseDir: "",         /** 親ディレクトリ
                        *  @type {string} */
  referredBaseDir: "", /** 参照元のファイルの親ディレクトリ
                        *  @type {string} */
  /*
   * referredBaseDir は
   * IE で保存された MHT ファイルにおける以下の問題を回避するためのもの
   *
   * [状況]
   *   ファイル
   *     http://unmht.org/x.html
   *     http://unmht.org/style/a.css
   *     http://unmht.org/style/b.css
   *     http://unmht.org/style/x.png
   *   x.html 中の記述
   *     <link href="./style/a.css" ... />
   *   a.css 中の記述
   *     @import url("./b.css");
   *   b.css 中の記述
   *     background: url("./x.png");
   *
   * [期待される結果]
   *   相対パス "./b.css" を a.css の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/style/b.css
   *
   *   これにより a.css から b.css が参照できる
   *   また b.css から x.png が参照できる
   *
   * [実際の結果]
   *   相対パス "./b.css" を x.html の URI から解決した URL が
   *   b.css の Content-Location に記述される
   *
   *   Content-Location: http://unmht.org/b.css
   *
   *   これにより a.css から b.css が参照できない
   *   また b.css から x.png が参照できない
   *
   * [対策]
   *   1. すべての X から参照されているすべての Y について
   *      Y.referredBaseDir を X.baseDir とする
   *   2. Z.referredBaseDir が空でない Z について
   *      Z 内の相対パス R による参照について
   *      Z.baseDir から R を解決した URI を A とする
   *      Z.referredBaseDir から R を解決した URI を B とする
   *   3. F.contentLocation が A に一致するような F が存在しない、かつ
   *      G.contentLocation が B に一致するような G が存在する場合
   *      R による参照先を G とする
   *      この時、G.referredBaseDir は F.baseDir となり 2 から再度適用する
   *
   * [例]
   *   0. {x.html}.baseDir = "http://unmht.org/"
   *      {a.css}.baseDir = "http://unmht.org/style/"
   *      {b.css}.contentLocation = "http://unmht.org/b.css"
   *      {x.png}.contentLocation = "http://unmht.org/style/x.png"
   *   1. {a.css}.referredBaseDir = {x.html}.baseDir
   *                                "http://unmht.org/"
   *   2. Z = {a.css}
   *      R = "./b.css"
   *      A = {a.css}.baseDir + R
   *        = "http://unmht.org/style/b.css"
   *      B = {a.css}.referredBaseDir + R
   *        = "http://unmht.org/b.css"
   *   3. A != {b.css}.contentLocation
   *      B == {b.css}.contentLocation
   *      # a.css から b.css が参照できる
   *      {b.css}.referredBaseDir = {a.css}.baseDir
   *                                "http://unmht.org/style/"
   *   2. Z = {b.css}
   *      A = {b.css}.baseDir + R
   *        = "http://unmht.org/" + R
   *      B = {b.css}.referredBaseDir + R
   *        = "http://unmht.org/style/" + R
   *      # b.css から "./x.png" で参照されたファイル x.png が
   *      #   {x.png}.contentLocation = "http://unmht.org/x.png"
   *      # となっていても
   *      #   {x.png}.contentLocation = "http://unmht.org/style/x.png"
   *      # となっていても参照できる
   *
   * [問題]
   *   誤って生成された Content-Location に一致するファイルが
   *   同 MHT 内に存在した場合これらを区別する事はできない
   */
  leafName: "",        /** ファイル名
                        *  @type {string} */
  normalLeafName: "",  /** 正規化したファイル名
                        *  @type {string} */
  
  fragment: "",        /** フラグメント
                        *  @type {string} */
  normalFragment: "",  /** 正規化したフラグメント
                        *  @type {string} */
  
  content: "",         /** 修正済みのファイルの中身
                        *  @type {string} */
  
  refName: ""          /** 参照用のファイル名
                        *  @type {string} */
};

/**
 * ファイル全体の展開情報
 */
function UnMHTExtractFileInfo() {
  this.parts = [];
  this.refParts = new Map();
}
UnMHTExtractFileInfo.prototype = {
  original: "",         /** 展開したファイル名の URI
                         *  @type {string} */
  baseURI: "",          /** 開始ページの URI
                         *  @type {string} */
  
  startPart: null,      /**開始ページのパート
                         *  @type {UnMHTMIMEPart} */
  
  topPart: null,        /** トップレベルのパート
                         *  @type {UnMHTMIMEPart} */
  subject: "",          /** Subject の値
                         *  @type {string} */
  date: "",             /** Date の値
                         *  @type {string} */
  
  parts: null,          /** mht ファイルの中にあるパート
                         *  @type {Array.<UnMHTMIMEPart>} */
  refParts: null,       /** 参照できるパート
                         *  @type {Map.<string,UnMHTMIMEPart>} */
  
  extractDate: null,    /** 展開時刻
                         *  @type {Date} */
  lastAccessDate: null, /** 最終アクセス時刻
                         *  @type {Date} */
  
  size: 0,              /** 全キャッシュのサイズ
                         *  @type {number} */
  
  /**
   * データを開放する
   */
  destruct: function() {
    this.parts
      .forEach(function(part) {
        part.eParam.content = null;
        part.eParam = null;
      });
    this.parts = null;
    
    if (this.topPart) {
      this.topPart.destruct();
      this.topPart = null;
    }
    
    this.startPart = null;
    this.refParts = null;
  }
};

/**
 * permissions.default.image の値
 * @type {number}
 */
let PERM_IMAGE_ALLOW_ALL = 1;
let PERM_IMAGE_BLOCK_ALL = 2;
let PERM_IMAGE_BLOCK_THIRD_PIRTY = 3;

/**
 * ファイルの展開と展開情報の管理
 *
 * @implements arIShutdownEventListener
 */
let UnMHTExtractor = {
  /**
   * 初期化処理
   */
  init: function() {
    UnMHTPref.addListener("open.offline", function() {
      this._enableOffline = UnMHTPref.getPref("open.offline");
    }.bind(this), this, true);
    
    UnMHTPref.addGlobalListener("permissions.default.image", function() {
      this._loadImageType
        = Services.prefs.getIntPref("permissions.default.image");
    }.bind(this), this, true);
    
    arShutdownEvent.addListener(this);
  },
  
  /**
   * 一意性を確保するためにファイル名を正規化する
   *
   * @param   {string} original
   *          ファイル名
   * @returns {string}
   *          正規化したファイル名
   */
  normalizeName: function(original) {
    if (!original) {
      return "";
    }
    
    for (let i = 0; i < 10; i ++) {
      let last = original;
      try {
        original = decodeURIComponent(original);
      } catch (e) {
        original = unescape(original);
      }
      original = UnMHTHTMLUtil.unescapeEntity(original);
      if (last == original) {
        break;
      }
    }
    
    return arPathUtil.unique(original);
  },
  
  /**
   * キャッシュから展開情報を返す
   *
   * @param   {string} originalURISpec
   *          展開したファイル名の URI 表記
   * @returns {UnMHTExtractFileInfo}
   *          展開情報
   *          見付からなければ null
   */
  getFileInfoFromOriginalURISpec: function(originalURISpec) {
    return this._getFileInfo(UnMHTScheme.enUnMHTURI(originalURISpec));
  },
  
  /**
   * mht ファイルを展開する
   *
   * @param   {string} originalURISpec
   *          展開したファイル名の URI 表記
   * @param   {string} text
   *          mht ファイルの内容
   * @param   {boolean} extract
   *          展開するかどうか
   *            false の場合、text で与えられた内容のファイルを作る
   * @returns {UnMHTExtractFileInfo}
   *          展開情報
   */
  extractMHT: function(originalURISpec, text, extract) {
    let eFileInfo = new UnMHTExtractFileInfo();
    
    eFileInfo.original
      = originalURISpec
      .replace(/[\(\)\'\"\[\] ]/g, 
               function(matched) {
                 let hexString
                   = "0" + matched.charCodeAt(0).toString(16).toUpperCase();
                 return "%" + hexString.slice(-2);
               });
    
    eFileInfo.baseURI = UnMHTScheme.enUnMHTURI(eFileInfo.original);
    
    eFileInfo.extractDate = new Date();
    
    UnMHTMaxRunTime.setExtractTime();
    
    if (extract) {
      eFileInfo.topPart
        = UnMHTMIMEDecoder.decodeMessage(text.replace(/\r\n|\r|\n/g, "\r\n"));
    }
    
    if (!eFileInfo.topPart) {
      let createDummyMessage = function(mimetype, body) {
        let data
          = "From: <Created by UnMHT>" + "\r\n"
          + "Content-Type: " + mimetype + "\r\n"
          + "\r\n"
          + body;
        
        return data;
      };
      
      let isHTML= false;
      if (!/\.(eml|mbs)$/i.test(originalURISpec) && /^\s*</.test(text)) {
        isHTML = true;
      }
      
      if (extract && isHTML) {
        eFileInfo.topPart
          = UnMHTMIMEDecoder.decodeMessage(createDummyMessage("text/html; charset=\"UTF-8\"", text));
      } else {
        eFileInfo.topPart
          = UnMHTMIMEDecoder.decodeMessage(createDummyMessage("text/plain; charset=\"UTF-8\"", text));
      }
    }
    
    if (eFileInfo.topPart.fields.has("subject")) {
      let value = eFileInfo.topPart.fields.get("subject");
      eFileInfo.subject = UnMHTMIMEDecoder.decodeWord(value);
    }
    
    if (eFileInfo.topPart.fields.has("date")) {
      eFileInfo.date = eFileInfo.topPart.fields.get("date");
    }
    
    eFileInfo.startPart = eFileInfo.topPart.findStartPart();
    
    this._createExtractParam(eFileInfo, eFileInfo.topPart, null);
    
    this._setRefName(eFileInfo);
    
    this._modifyAttribute(eFileInfo);
    
    this._modifyCSS(eFileInfo);
    
    this._createMixedDocument(eFileInfo);
    
    this._getPartsInfo(eFileInfo);
    
    UnMHTMaxRunTime.restoreTime();
    
    eFileInfo.lastAccessDate = new Date();
    
    this._addFileInfo(eFileInfo);
    
    if (UnMHTPref.getPref("open.cache.expire")) {
      arTimer.setTimeout(function() {
        let expire = UnMHTPref.getPref("open.cache.expire.time");
        if (expire < 30) {
          expire = 30;
        }
        this.removeOldFileInfo(expire * 60000);
      }.bind(this), 100);
    }
    
    return eFileInfo;
  },
  
  /**
   * CSS 中の unmht: 形式の URI を解除する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {string} content
   *          対象のドキュメント
   * @returns {string}
   *          変換したドキュメント
   */
  unmodifyCSS: function(eFileInfo, content) {
    return content
      .replace(this._cssRe, this._unCssFunc.bind(this, eFileInfo));
  },
  
  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {string} content
   *          対象のドキュメント
   * @returns {string}
   *          変換したドキュメント
   */
  unmodifyAttribute: function(eFileInfo, content) {
    return content
      .replace(this._elementRe, this._unElementFunc.bind(this, eFileInfo));
  },
  
  /**
   * Content-Location に対応するパスからパートを探す
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {UnMHTMIMEPart} originPart
   *          呼び出し元のパート
   * @param   {string} path
   *          対象のパス
   * @param   {boolean} referred
   *          referredBaseDir をチェックするか
   * @param   {boolean} useLeafNameCheck
   *          ファイル名のみのチェックを行うか
   * @returns {Array}
   *            [UnMHTMIMEPart 対象のパート, string フラグメント(#つき)]
   *          見付からなければパートが null
   */
  findPart: function(eFileInfo, originPart, path, referred, useLeafNameCheck) {
    let fragment = "";
    let re = path.match(/^([^#]*)#(.*)$/);
    if (re) {
      [, path, fragment] = re;
    }
    if (fragment && path == "") {
      return [originPart, fragment ? ("#" + fragment) : ""];
    }
    
    let normalFragment = this.normalizeName(fragment);
    
    let normalPath = this.normalizeName(path);
    
    let part;
    
    /* フラグメント付きのリクエストなら先にフラグメント付きでチェック */
    if (normalFragment) {
      part
        = arArrayUtil
        .find(eFileInfo.parts,
              function(part) (normalPath == part.eParam.normalLocation &&
                              normalFragment == part.eParam.normalFragment));
      if (part) { return [part, ""]; }
    }
    
    /* フラグメント無しでチェック */
    part
      = arArrayUtil
      .find(eFileInfo.parts,
            function(part) (normalPath == part.eParam.normalLocation));
    if (part) { return [part, fragment ? ("#" + fragment) : ""]; }
    
    if (originPart.eParam.baseDir) {
      /* 相対パスでの参照をチェック */
      let resolvedPath
        = this.normalizeName(arPathUtil.resolve(originPart.eParam.baseDir,
                                                normalPath));
      
      part
        = arArrayUtil
        .find(eFileInfo.parts,
              function(part) (resolvedPath == part.eParam.normalLocation));
      if (part) { return [part, fragment ? ("#" + fragment) : ""]; }
    }
    
    re = normalPath.match(/^cid:(.+)$/);
    if (re) {
      let cid = re[1];
      
      part
        = arArrayUtil
        .find(eFileInfo.parts,
              function(part) (cid == part.contentID));
      if (part) { return [part, fragment ? ("#" + fragment) : ""]; }
    }
    
    
    if (UnMHTScheme.isUnMHTURI(normalPath)) {
      let param = UnMHTScheme.getUnMHTURIParam(normalPath);
      let refName = param.refName;
      
      if (param.original == eFileInfo.original) {
        if (!refName) {
          return [eFileInfo.startPart, fragment ? ("#" + fragment) : ""];
        }
        
        if (eFileInfo.refParts.has(refName)) {
          return [eFileInfo.refParts.get(refName), fragment ? ("#" + fragment) : ""];
        }
        
        return [null, ""];
      }
    }
    
    if (referred && originPart.eParam.referredBaseDir != "") {
      let referredPath
        = this.normalizeName(arPathUtil.resolve(originPart.eParam.referredBaseDir,
                                                normalPath));
      
      part
        = arArrayUtil
        .find(eFileInfo.parts,
              function(part) (referredPath == part.eParam.normalLocation));
      if (part) { return [part, fragment ? ("#" + fragment) : ""]; }
    }
    
    if (useLeafNameCheck) {
      /* ファイル名だけでも同じものがあれば使う
       * あまり良い方法ではないので最小限にする */
      let noQuery = normalPath.replace(/\?.+/, "");
      part
        = arArrayUtil
        .find(eFileInfo.parts, function(part) {
          let leafName = part.eParam.normalLeafName;
          let length = leafName.length;
          return (leafName &&
                  noQuery.length >= length + 1 &&
                  noQuery.substr(noQuery.length - length) == leafName &&
                  /[\/\?\\:]/.test(noQuery.substr(noQuery.length - length - 1, 1)));
        });
      if (part) { return [part, fragment ? ("#" + fragment) : ""]; }
    }
    
    /* 含まれない */
    return [null, ""];
  },
  
  /**
   * unmht 形式の URI から展開情報を取得する
   *
   * @param   {string} uri
   *          unmht: 形式のファイルの URI
   * @returns {Array}
   *          [UnMHTExtractFileInfo, UnMHTMIMEPart]
   *          展開情報
   *          見付からなかった場合はそれぞれ null
   */
  getFileInfoAndPart: function(uri) {
    if (!UnMHTScheme.isUnMHTURI(uri)) {
      return [null, null];
    }
    
    let part = null;
    
    let param = UnMHTScheme.getUnMHTURIParam(uri);
    let baseURI = param.baseURI;
    let refName = param.refName;
    
    let eFileInfo = this._getFileInfo(baseURI);
    if (eFileInfo) {
      eFileInfo.lastAccessDate = new Date();
      
      if (refName) {
        if (eFileInfo.refParts.has(refName)) {
          part = eFileInfo.refParts.get(refName);
        }
      } else {
        part = eFileInfo.startPart;
      }
    }      
    
    return [eFileInfo, part];
  },

  /**
   * unmht 形式の URI からの相対パスを展開する
   *
   * @param   {string} uri
   *          unmht: 形式のファイルの URI
   * @param   {string} spec
   *          相対パス
   * @returns {string}
   *          ファイルの URI
   *          見付からなかった場合は null
   */
  resolveURI: function(uri, spec) {
    if (!spec) {
      return uri;
    }
    
    let [eFileInfo, part] = this.getFileInfoAndPart(uri);
    if (eFileInfo && part) {
      eFileInfo.lastAccessDate = new Date();
      
      let [targetPart, fragment]
        = this.findPart(eFileInfo, part, spec, false, true);
      if (targetPart) {
        return eFileInfo.baseURI + targetPart.eParam.refName + fragment;
      }
      
      if (!/^file/.test(part.eParam.location)) {
        try {
          return Services.io.newURI(part.eParam.location, null, null)
            .resolve(spec);
        } catch (e) {
        }
      }
    }
    
    return null;
  },

  /**
   * mht に含まれるファイルの URI を取得する
   *
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          呼び出し元の URI
   *          unmht 形式
   * @returns {nsIURI}
   *          ファイルの URI
   *          存在しなかった場合は null
   */
  getMHTFileURI: function(contentLocation, requestOrigin) {
    let [eFileInfo, part] = this.getFileInfoAndPart(requestOrigin.spec);
    
    if (eFileInfo && part) {
      eFileInfo.lastAccessDate = new Date();
      
      let [targetPart, fragment]
        = this.findPart(eFileInfo, part, contentLocation.spec, false, false);
      if (targetPart) {
        return Services.io.newURI(eFileInfo.baseURI + targetPart.eParam.refName
                                  + fragment, null, null);
      }
    }
    
    return null;
  },
  
  /**
   * キャッシュから展開情報の配列を返す
   *
   * @returns {Array.<UnMHTExtractFileInfo>}
   *          展開情報
   */
  getFileInfoList: function() {
    return [eFileInfo for ([uri, eFileInfo] of this._eFileInfoMap)]
      .sort(function(x, y) {
        return y.lastAccessDate.getTime() - x.lastAccessDate.getTime();
      });
  },
  
  /**
   * キャッシュから古い展開情報を削除する
   *
   * @param   {number} expire
   *          有効期限 [ms]
   */
  removeOldFileInfo: function(expire) {
    let opened = new Set();
    [window for (window of XPCOMUtils.IterSimpleEnumerator(
      Services.wm.getEnumerator("navigator:browser"),
      Ci.nsIDOMWindow))]
      .forEach(function(window) {
        UnMHTWindowUtil.getDocuments(window)
          .forEach(function(targetDocument) {
            let location;
            try {
              location = targetDocument.location.href;
            } catch (e) {
              /* ドキュメントが生きていない */
              return;
            }
            
            let param = UnMHTScheme.getUnMHTURIParam(location);
            if (param.baseURI) {
              opened.add(param.baseURI);
            }
          }.bind(this));
      });
    
    let now = Date.now();
    for (let [uri, eFileInfo] of this._eFileInfoMap) {
      if (opened.has(eFileInfo.baseURI)) {
        continue;
      }
      
      if (now < eFileInfo.lastAccessDate.getTime() + expire) {
        continue;
      }
      
      eFileInfo.destruct();
      
      this._eFileInfoMap.delete(uri);
    }
  },
  
  /**
   * mht への参照のチェックおよび
   * mht 内の URI を置き換える
   *
   * @param   {number} contentType
   *          コンテントの種類
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @param   {nsIURI} requestOrigin
   *          参照元の URI
   * @param   {nsIDOMElement} context
   *          ロード先
   * @returns {boolean}
   *          許可するかどうか
   */
  shouldLoad: function(contentType, contentLocation, requestOrigin, context) {
    if (this._loadImageType == PERM_IMAGE_BLOCK_ALL &&
        contentType == Ci.nsIContentPolicy.TYPE_IMAGE) {
      /* 画像の読み込みの拒否は Firefox の設定に従う */
      return false;
    }
    
    if (context instanceof Ci.nsIDOMHTMLDocument ||
        context instanceof Ci.nsIDOMHTMLStyleElement) {
      /* CSS によるロードの場合 */
      return this._checkExternal(contentLocation);
    }
    
    /* ---- ここから HTML 要素へのロード ---- */
    
    let [eFileInfo, originPart]
      = this.getFileInfoAndPart(requestOrigin.spec);
    if (!(eFileInfo && originPart)) {
      /* 参照元のファイルが未展開、もしくは存在しないので拒否 */
      return false;
    }
    
    if (UnMHTScheme.isUnMHTURI(contentLocation.spec)) {
      let param = UnMHTScheme.getUnMHTURIParam(contentLocation.spec);
      if (param.original == eFileInfo.original) {
        /* 既に同じ mht ファイル内への参照になっているので許可 */
        return true;
      }
    }
    
    /**
     * 属性を変更する関数
     *
     * @param   {nsIDOMElement} node
     *          対象の要素
     * @param   {string} attr
     *          属性名
     * @param   {string} value
     *          属性値
     */
    let setAttr = function(node, attr, value) {
      if (node instanceof Ci.nsIDOMHTMLEmbedElement ||
          node instanceof Ci.nsIDOMHTMLObjectElement) {
        /* EMBED, OBJECT は属性を変えても効果が無いので
         * 要素を置き換える */
        arTimer.setTimeout(function() {
          try {
            let newNode = node.cloneNode(true);
            newNode.setAttribute("__unmht_" + attr, contentLocation.spec);
            newNode.setAttribute(attr, value);
            node.parentNode.replaceChild(newNode, node);
          } catch (e) {
            /* ノードがドキュメントが生きていない */
          }
        }, 100);
      } else {
        node.setAttribute("__unmht_" + attr, contentLocation.spec);
        node.setAttribute(attr, value);
        
        arTimer.setTimeout(function() {
          try {
            node.setAttribute(attr, value);
          } catch (e) {
            /* ノードがドキュメントが生きていない */
          }
        }, 100);
      }
    };
    
    /**
     * 属性を削除する関数
     *
     * @param   {nsIDOMElement} node
     *          対象の要素
     * @param   {string} attr
     *          属性名
     */
    let removeAttr = function(node, attr) {
      node.removeAttribute(attr);
      arTimer.setTimeout(function() {
        try {
          node.removeAttribute(attr);
        } catch (e) {
          /* ノードがドキュメントが生きていない */
        }
      }, 100);
    };
    
    let attr = "src";
    if (context instanceof Ci.nsIDOMHTMLObjectElement) {
      attr = "data";
    }
    
    let [targetPart, fragment]
      = this.findPart(eFileInfo, originPart, contentLocation.spec, false, false);
    if (targetPart) {
      setAttr(context, attr,
              eFileInfo.baseURI + targetPart.eParam.refName + fragment); 
      return false;
    }
    
    if (contentLocation.scheme == "file") {
      /* file への読み込みは削除 */
      removeAttr(context, attr);
      return false;
    }
    
    if (contentLocation.spec.indexOf(requestOrigin.spec) == 0) {
      /* JavaScript 等で生成された参照元からの相対アドレス */
      /* **** FIXME: パスを遡るものはチェック不可 **** */
      let path = contentLocation.spec.substr(requestOrigin.spec.length);
      
      let origSpec;
      if (path[0] == "?") {
        /* クエリのみ */
        origSpec
          = originPart.eParam.location.replace(/\?.*$/, "") + path;
      } else {
        origSpec
          = arPathUtil.resolve(originPart.eParam.baseDir, path);
      }
      
      let origContentLocation;
      try {
        origContentLocation = Services.io.newURI(origSpec, null, null);
      } catch (e) {
        /* 正しい URI を生成できなかった */
        return this._checkExternal(contentLocation);
      }
      
      if (origContentLocation.spec != contentLocation.spec) {
        /* 本来のアドレスでチェックし直す */
        return this.shouldLoad(contentType, origContentLocation,
                               requestOrigin, context);
      }
    }
    
    return this._checkExternal(contentLocation);
  },

  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    UnMHTPref.removeAllListeners(this);
    
    this.removeOldFileInfo(0);
  },
  
  /* ==== private ==== */
  
  _eFileInfoMap: new Map(), /** 展開情報
                             *  @type {Map.<string,UnMHTExtractFileInfo>}
                             *    <string 正規化した baseURI,
                             *     UnMHTExtractFileInfo 展開情報> */
  
  _loadImageType: PERM_IMAGE_ALLOW_ALL, /** Firefox の画像読み込み設定
                                         *  permissions.default.image の値
                                         *  @type {number} */
  
  _enableOffline: false, /** オフラインモードかどうか
                          *  @type {boolean} */
  
  /**
   * パートの情報を作成する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {UnMHTMIMEPart} part
   *          対象のパート
   * @param   {UnMHTMIMEPart} parentPart
   *          親のパート
   */
  _createExtractParam: function(eFileInfo, part, parentPart) {
    let isMixed = (part.isMultipart &&
                   part.contentSubType == "mixed");
    
    if (part.isMultipart) {
      part.parts
        .forEach(function(p) {
          this._createExtractParam(eFileInfo, p, part);
        }.bind(this));
      
      if (!isMixed) {
        return;
      }
    }
    
    let eParam = new UnMHTExtractParam();
    if (part.contentID) {
      eParam.cid = part.contentID;
    }
    if (isMixed) {
      eParam.isMixed = true;
      eParam.content = "";
      
      eParam.mimetype = "text/html";
      eParam.charset = "utf-8";
      
      eParam.isHTML = true;
      
      eParam.location = "mixed-" + Date.now() + ".html";
    } else {
      eParam.content = part.body;
      
      eParam.mimetype = part.mimetype;
      eParam.charset = part.charset;
      if (part.contentType == "text") {
        eParam.isHTML = true;
        if (part.contentSubType == "css") {
          eParam.isCSS = true;
        }
      }
      
      if (part.fields.has("content-base")) {
        let value = part.fields.get("content-base");
        try {
          value = arUconv.toUnicode(value, "utf-8");
        } catch (e) {
          try {
            value = arUconv.toUnicode(value, "shift_jis");
          } catch (e) {
          }
        }
        eParam.contentBase = value;
      }
      
      if (eParam.contentBase) {
        if (part.contentLocation) {
          eParam.location
            = arPathUtil.resolve(eParam.contentBase, part.contentLocation);
        } else {
          eParam.location = eParam.contentBase;
        }
      } else {
        eParam.location = part.contentLocation;
      }
      
      if (parentPart && parentPart.contentLocation) {
        eParam.location
          = arPathUtil.resolve(arPathUtil.getBaseDir(parentPart.contentLocation),
                               eParam.location);
      }
      
      if (!eParam.location &&
          eParam.mimetype == "text/html") {
        /* 開始ページの Content-Location が
         * 指定されない場合があるので index.html とする */
        eParam.location = "index.html";
      }
    }
    part.eParam = eParam;
    
    this._getPathInfo(part);
    
    eFileInfo.parts.push(part);
    
    if (eParam.mimetype == "application/octet-stream" &&
        part.body.substr(0, 32).indexOf("<") != -1) {
      /* CGI 生成のページ等が application/octet-stream として
       * 保存されている場合がある */
      if (part.body.search(/<html/i) != -1) {
        eParam.mimetype = "text/html";
        eParam.isHTML = true;
      }
    }
  },
  
  /**
   * パスの情報を取得する
   *
   * @param   {UnMHTMIMEPart} part
   *          対象のパート
   */
  _getPathInfo: function(part) {
    let leafNameRe = null;
    let eParam = part.eParam;
    
    let location = eParam.location;
    
    let info = arPathUtil.getPathInfo(location);
    
    if (info.isMHTML) {
      eParam.leafName = info.mhtmlFilename || "";
      eParam.fragment = info.mhtmlFragment || "";
      
      let baseInfo = info.clone();
      baseInfo.mhtmlFilename = undefined;
      baseInfo.mhtmlFragment = undefined;
      eParam.baseDir = arPathUtil.createPath(baseInfo);
    } else if (info.type & arPathUtil.MASK_DATA) {
      eParam.baseDir = location;
      eParam.leafName = "";
      eParam.fragment = "";
    } else {
      if (info.pathComponents.length) {
        eParam.leafName = info.pathComponents[info.pathComponents.length - 1];
      } else {
        eParam.leafName = "";
      }
      eParam.fragment = info.fragment || "";
      
      let baseInfo = info.clone();
      if (baseInfo.pathComponents.length) {
        baseInfo.pathComponents.pop();
      }
      baseInfo.query = undefined;
      baseInfo.fragment = undefined;
      eParam.baseDir = arPathUtil.createPath(baseInfo);
    }
    
    if (part.contentDispositionFilename) {
      eParam.leafName = part.contentDispositionFilename;
    }
    
    eParam.referredBaseDir = "";
    
    let normalInfo = info.clone();
    normalInfo.mhtmlFragment = undefined;
    normalInfo.fragment = undefined;
    
    eParam.normalLocation
      = this.normalizeName(arPathUtil.createPath(normalInfo, false));
    eParam.normalLeafName = this.normalizeName(eParam.leafName);
    eParam.normalFragment = this.normalizeName(eParam.fragment);
  },
  
  /**
   * 参照名を重複しないように設定する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _setRefName: function(eFileInfo) {
    let indices = new Map();
    
    eFileInfo.parts.forEach(function(part, i) {
      let [refBase, extension] = this.getRefBaseAndExt(part, i);
      refBase = encodeURIComponent(refBase).replace(/\//g, "_");
      if (extension) {
        extension = "." + extension;
      }
      
      let refName = refBase + extension;
      
      if (indices.has(refName)) {
        part.eParam.refName
          = refBase + "." + indices.get(refName) + extension;
        indices.set(refName, indices.get(refName) + 1);
      } else {
        part.eParam.refName = refName;
        indices.set(refName, 1);
      }
    }.bind(this));
  },
  
  /**
   * 参照名のベース部分と拡張子を取得する
   *
   * @param   {UnMHTMIMEPart} part
   *          対象のパート
   * @param   {number} i
   *          適切な情報が無い場合に使用するファイルの番号
   * @returns {Array}
   *            [string ベース部分, string 拡張子]
   */
  getRefBaseAndExt: function(part, i) {
    let re = part.eParam.leafName.match(/^([^\?#]+)\.([A-Za-z0-9]+)/);
    if (re) {
      return [re[1], re[2]];
    }
    
    re = part.eParam.leafName.match(/^([^\?#]+)/);
    if (re) {
      return [re[1], ""];
    }
    
    re = part.eParam.leafName.match(/^\?([^#]+)/);
    if (re) {
      let params = new Map();
      re[1].split(/&/).forEach(function(seg) {
        re = seg.match(/^([^=]+)=(.+)$/);
        if (re) {
          params.set(re[1], re[2]);
        }
      });
      if (params.has("filename")) {
        re = params.get("filename").match(/^([^?#]+)\.([A-Za-z0-9]+)/);
        if (re) {
          return [re[1], re[2]];
        }
      }
    }
    
    if (part.contentID) {
      return [part.contentID, ""];
    }
    
    if (part.eParam.mimetype == "text/html") {
      return ["index", "html"];
    }
    
    return i;
  },
  
  /**
   * CSS 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param   {UnMHTMIMEPart} part
   *          対象のパート
   * @param   {string} matched
   *          マッチした全体
   * @param   {string} prev
   * @param   {string} dquote
   * @param   {string} dqpath
   * @param   {string} quote
   * @param   {string} qpath
   * @param   {string} path
   *          以下参照
   *
   *         dquote
   *         -
   * @import "12345.jpg";
   * -------- ---------
   * prev     dqpath
   *
   *      quote
   *      -
   * url ('12345.jpg');
   * ----- ---------
   * prev  qpath
   *
   * @returns {string}
   *          変換した結果
   */
  _cssFunc: function(eFileInfo, part,
                     matched, prev, dquote, dqpath, quote, qpath, path) {
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'");
    }
    quote = quote || dquote || "";
    if (prev.indexOf("(") == -1) {
      quote = quote || "\'";
    }
    
    let [targetPart, fragment]
      = this.findPart(eFileInfo, part, path, part.eParam.isCSS, true);
    if (targetPart) {
      targetPart.eParam.referredBaseDir = part.eParam.baseDir;
      return prev + quote
        + eFileInfo.baseURI + targetPart.eParam.refName + quote;
    }
    
    if (/^file:/.test(path)) {
      return prev + quote + quote;
    }
    
    return matched;
  },
  
  /**
   * CSS 中の unmht: 形式の URI を解除する
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は _cssFunc と同じ
   */
  _unCssFunc: function(eFileInfo,
                       matched, prev, dquote, dqpath, quote, qpath, path) {
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'");
    }
    quote = quote || dquote || "";
    if (prev.indexOf("(") == -1) {
      quote = quote || "\'";
    }
    
    if (!/^unmht:/.test(path)) {
      return matched;
    }
    
    let param = UnMHTScheme.getUnMHTURIParam(path);
    let baseURI = param.baseURI;
    let refName = param.refName;
    let part = null;
    if (refName) {
      if (eFileInfo.refParts.has(refName)) {
        part = eFileInfo.refParts.get(refName);
      }
    } else {
      part = eFileInfo.startPart;
    }
    if (part) {
      let location = part.contentLocation + (param.fragment
                                             ? ("#" + param.fragment)
                                             : "");
      if (!location) {
        location = "cid:" + part.eParam.cid;
      }
      
      return prev + quote + location + quote;
    }
    
    return matched;
  },
  
  /**
   * CSS 中の URI を unmht: 形式に変換する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _modifyCSS: function(eFileInfo) {
    eFileInfo.parts.forEach(function(part) {
      if (part.eParam.isHTML ||
          part.eParam.isCSS) {
        part.eParam.content
          = part.eParam.content
          .replace(this._cssRe, this._cssFunc.bind(this, eFileInfo, part));
      }
    }.bind(this));
  },
  
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * replace に渡す関数
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   * @param   {UnMHTMIMEPart} part
   *          対象のパート
   * @param   {string} matched
   *          マッチした全体
   * @param   {string} prev
   * @param   {string} attrs
   * @param   {string} next
   * @returns {string}
   *          変換した結果
   */
  _elementFunc: function(eFileInfo, part,
                         matched, prev, name, attrs, next) {
    if (/^base$/i.test(name)) {
      return matched;
    }
    let isObject = false;
    if (/^object$/i.test(name) || /^embed$/i.test(name)) {
      isObject = true;
    }
    let isAnchor = false;
    if (/^a$/i.test(name)) {
      isAnchor = true;
    }
    let isLink = false;
    if (/^link$/i.test(name)) {
      isLink = true;
    }
    let isOriginalLink = false;
    
    attrs
      = attrs
      .replace(this._attributeRe,
               function(matched, prev, attr,
                        dquote, dqpath, quote, qpath, path) {
                 if (dqpath) {
                   path = dqpath.replace(/\\\"/g, "\"");
                 } else if (qpath) {
                   path = qpath.replace(/\\\'/g, "\'");
                 }
                 quote = quote || dquote || "\"";
                 
                 if (attr == "class") {
                   if (isAnchor && path == "unmht_link_to_original") {
                     isOriginalLink = true;
                   }
                   
                   return matched;
                 }
                 
                 if (isOriginalLink && attr == "href") {
                   return matched;
                 }
                 
                 if (/^mailto:/.test(path)) {
                   return matched;
                 }
                 
                 if (!isObject && /data/i.test(attr)) {
                   return matched;
                 }
                 
                 let isRelative = false;
                 if (!/^[A-Za-z0-9\-]+:/.test(path)) {
                   isRelative = true;
                 }
                 if (isRelative && !isAnchor && !isLink) {
                   /* 相対パスは UnMHTProtocolHandler.newURI の方で変える */
                   return matched;
                 }
                 
                 let [targetPart, fragment]
                   = this.findPart(eFileInfo, part, path, false, !isAnchor);
                 if (targetPart) {
                   if (isLink) {
                     targetPart.eParam.referredBaseDir = part.eParam.baseDir;
                   }
                   
                   return prev + quote
                     + eFileInfo.baseURI + targetPart.eParam.refName + fragment
                     + quote;
                 }
                 
                 if (/^file:/.test(path) ||
                     /^cid:/.test(path)) {
                   return prev + quote + quote;
                 }
                 
                 if (isRelative && isAnchor) {
                   /* リンク先はあらかじめ展開しておく */
                   try {
                     return prev + quote
                       + Services.io.newURI(part.eParam.location, null, null)
                       .resolve(path) + quote;
                   } catch (e) {
                     /* 展開失敗した場合は諦める */
                   }
                 }
                 
                 return matched;
               }.bind(this));
    
    return prev + attrs + next;
  },
  
  /**
   * HTML 中の unmht: 形式の URI を解除する
   * この際 BASE 要素の内容を baseDir にする
   * replace に渡す関数
   *
   * 引数と返り値は part が無い以外は _elementFunc と同じ
   */
  _unElementFunc: function(eFileInfo, matched, prev, name, attrs, next) {
    let isBase = false;
    if (/^base$/i.test(name)) {
      isBase = true;
    }
    
    attrs
      = attrs
      .replace(this._attributeRe,
               function(matched, prev, attr,
                        dquote, dqpath, quote, qpath, path) {
                 if (dqpath) {
                   path = dqpath.replace(/\\\"/g, "\"");
                 } else if (qpath) {
                   path = qpath.replace(/\\\'/g, "\'");
                 }
                 quote = quote || dquote || "\"";
                 
                 if (!/^unmht:/.test(path)) {
                   return matched;
                 }
                 
                 let param = UnMHTScheme.getUnMHTURIParam(path);
                 let refName = param.refName;
                 let part = null;
                 if (refName) {
                   if (eFileInfo.refParts.has(refName)) {
                     part = eFileInfo.refParts.get(refName);
                   }
                 } else {
                   part = eFileInfo.startPart;
                 }
                 if (part) {
                   let location = part.contentLocation;
                   if (isBase) {
                     location = part.eParam.baseDir;
                   }
                   if (param.fragment) {
                     location += "#" + param.fragment;
                   }
                   if (!location && !isBase) {
                     location = "cid:" + part.eParam.cid;
                   }
                   
                   return prev + quote + location + quote;
                 }
                 
                 return matched;
               }.bind(this));
    
    return prev + attrs + next;
  },
  
  /**
   * HTML 中の BASE の値を収集して削除する
   * replace に渡す関数
   *
   * 引数と返り値は _elementFunc と同じ
   */
  _baseFunc: function(eFileInfo, part,
                      matched, dqpath, qpath, path) {
    if (dqpath) {
      path = dqpath.replace(/\\\"/g, "\"");
    } else if (qpath) {
      path = qpath.replace(/\\\'/g, "\'");
    }
    
    part.eParam.baseDir
      = arPathUtil.resolve(part.eParam.baseDir,
                           arPathUtil.getBaseDir(arUconv.toUnicode(path, part.eParam.charset)));
    
    return "";
  },
  
  /**
   * HTML 中の URI を unmht: 形式に変換する
   * また、BASE 要素と Office 用の要素を削除し
   * 設定されていれば PowerPoint の警告ページをスキップする
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _modifyAttribute: function(eFileInfo) {
    let isPPT = false;
    let framePart = null;
    
    eFileInfo.parts.forEach(function(part) {
      if (part.eParam.refName == "frame.htm") {
        framePart = part;
      }
      
      if (!part.eParam.isCSS && part.eParam.isHTML) {
        part.eParam.content
          = part.eParam.content
          .replace(this._baseRe,
                   this._baseFunc.bind(this, eFileInfo, part))
          .replace(this._elementRe,
                   this._elementFunc.bind(this, eFileInfo, part));
        
        if (part.eParam.content.indexOf("xmlns:v=\"urn:schemas-microsoft-com:vml\"") != -1) {
          part.eParam.content
            = part.eParam.content
            .replace(/<\/?(o|p|oa|v):(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?>/g, "");
          isPPT = true;
        }
      }
    }.bind(this));
    
    if (isPPT && framePart &&
        UnMHTPref.getPref("open.ppt.skip_warning")) {
      eFileInfo.startPart = framePart;
    }
  },
  
  /**
   * multipart/mixed なドキュメントを作成する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _createMixedDocument: function(eFileInfo) {
    let createBody = function(part) {
      if (!part) {
        return "";
      }
      
      let getUTF8Content = function(part) {
        let content = part.eParam.content;
        
        if (!/^utf-8$/i.test(part.eParam.charset)) {
          content
            = arUconv.fromUnicode(
              arUconv.toUnicode(content, part.eParam.charset),
              "utf-8");
        }
        
        return content;
      };
      
      let content;
      if (part.eParam.mimetype == "text/html") {
        content
          = getUTF8Content(part)
          .replace(/<\/?(html|body)(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "")
          .replace(/<![\?A-Za-z0-9_:\-]+(?:\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)?>/ig, "");
      } else if (part.eParam.mimetype == "text/plain") {
        content
          = "<pre>"
          + UnMHTHTMLUtil.escapeEntity(getUTF8Content(part))
          .replace(/[ \t]/g, "&#xa0;")
          + "</pre>";
      } else {
        let href
          = UnMHTHTMLUtil.escapeEntity(eFileInfo.baseURI + part.eParam.refName);
        let filename
          = UnMHTHTMLUtil.escapeEntity(part.eParam.leafName ||
                                       part.contentDispositionFilename ||
                                       "???");
        
        if (part.contentType == "image") {
          content
            = "<div style=\"text-align: center;\">"
            + "<img src=\"" + href + "\" alt=\"" + filename + "\" />"
            + "</div>";
        } else {
          content = "<a href=\"" + href + "\">" + filename + "</a>";
        }
        
        content = toUTF8(content);
      }
      
      if (part.contentDispositionFilename) {
        content
          = "<fieldset style=\"border: 1px solid graytext;\">"
          + "<legend style=\"color: graytext; font-size: 80%;\">"
          + toUTF8(part.contentDispositionFilename)
          + "</legend>"
          + content
          + "</fieldset>";
      }
      return content;
    }.bind(this);
    
    eFileInfo.parts.forEach(function(part) {
      if (part.eParam.isMixed) {
        part.eParam.content
          = part.parts
          .map(function(p) createBody(p.findStartPart()))
          .join("");
      }
    }.bind(this));
  },
  
  /**
   * パートの情報を収集する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _getPartsInfo: function(eFileInfo) {
    eFileInfo.parts.forEach(function(part) {
      eFileInfo.refParts.set(part.eParam.refName, part);
      
      eFileInfo.size += part.eParam.content.length;
    }.bind(this));
  },
  
  /**
   * キャッシュに展開情報を格納する
   *
   * @param   {UnMHTExtractFileInfo} eFileInfo
   *          展開情報
   */
  _addFileInfo: function(eFileInfo) {
    this._eFileInfoMap.set(this.normalizeName(eFileInfo.baseURI), eFileInfo);
  },
  
  /**
   * キャッシュから展開情報を返す
   *
   * @param   {string} baseURI
   *          展開情報の baseURI
   * @returns {UnMHTExtractFileInfo}
   *          展開情報
   *          見付からなければ null
   */
  _getFileInfo: function(baseURI) {
    baseURI = this.normalizeName(baseURI);
    
    if (this._eFileInfoMap.has(baseURI)) {
      return this._eFileInfoMap.get(baseURI);
    }
    
    return null;
  },
  
  /**
   * 外部読み込みをチェックする
   *
   * @param   {nsIURI} contentLocation
   *          対象の URI
   * @returns {boolean}
   *          許可するかどうか
   */
  _checkExternal: function(contentLocation) {
    if (contentLocation.scheme == "unmht") {
      return true;
    }
    
    return !this._enableOffline;
  },

  _cssRe: /(@import\s*(?:url\s*\(\s*)?|[^a-zA-Z0-9_]url\s*\(\s*)(?:(\"|&quot;|&#x22;|&#34;)((?:\\\"|[^\"])*?)\2|(\'|&apos;|&#x27;|&#39;)((?:\\\'|[^\'])*?)\4|([^\"\';\(\)]+))/ig,
  _elementRe: /(<([\?A-Za-z0-9_:\-]+))(\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*)(>)/g,
  _attributeRe: /(\s(class|src|href|background|action|data)\s*=\s*)(?:(\")((?:\\\"|[^\"])*)\"|(\')((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))/ig,
  _baseRe: /<base\s(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*?href\s*=\s*(?:\"((?:\\\"|[^\"])*)\"|\'((?:\\\'|[^\'])*)\'|([^\"\'\\ >]+))(?:\"(?:\\\"|[^\"])*\"|\'(?:\\\'|[^\'])*\'|[^\"\'\\>])*>/ig
};
UnMHTExtractor.init();
