"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");
XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTNetFileLoader", function() {
  XPCOMUtils.importRelative(this, "UnMHTNetFileLoader.jsm");
  return UnMHTNetFileLoader;
});

let EXPORTED_SYMBOLS = ["UnMHTFileLoader"];

/**
 * MHT ファイルのロード
 *
 * @implements arIShutdownEventListener
 */
let UnMHTFileLoader = {
  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  },
  
  /**
   * 展開して表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openWithExtract: function(window) {
    this._loadFromFocusedWindow(window, true);
  },
  
  /**
   * 展開せずに表示
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  openWithoutExtract: function(window) {
    this._loadFromFocusedWindow(window, false);
  },
  
  /* ---- UI イベント ----- */
  
  /**
   * unmht: scheme のドキュメントがロードされたイベント
   * mht ファイルの展開、イベントのフックを行う
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMXULElement} tabbrowser
   *          対象のタブブラウザ
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} location
   *          ファイルの URI
   */
  onUnMHTDocumentLoaded: function(window, tabbrowser,
                                  targetWindow, targetDocument,
                                  location) {
    let param = UnMHTScheme.getUnMHTURIParam(location);
    let original = param.original;
    let refName = param.refName;
    let scheme = param.scheme;
    
    if (scheme == "error") {
      this._hookRetryButton(targetWindow, targetDocument, location);
      return;
    } else if (scheme != "cache" &&
               scheme != "cid" &&
               scheme != "dummy" &&
               scheme != "empty" &&
               scheme != "error") {
      /* 通常のファイル */
      let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(location);
      if (eFileInfo) {
        if (part &&
            part == eFileInfo.startPart) {
          this._hookReload(window, targetWindow);
        }
      } else {
        /* 未展開のファイルだった場合 */
        let uri;
        try {
          uri = Services.io.newURI(original, null, null);
        } catch (e) {
          return;
        }
        
        let loadParam = {
          window:         window,
          uri:            uri,
          targetWindow:   targetWindow,
          hasDocument:    false,
          replaceHistory: false,
          extract:        true,
          refName:        refName,
          useCache:       false
        };
        this.loadWithParam(loadParam);
        
        return;
      }
    }
    
    let action = UnMHTPref.getPref("open.multiple.action");
    
    if (action == "tabs" ||
        action == "first") {
      let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(location);
      if (eFileInfo && part && part == eFileInfo.startPart) {
        if (eFileInfo.topPart.fields.has("x-unmht-tabs")) {
          /* 複数のファイルを含む MHT の
           * インデックスを開いた */
          this._replaceTabs(window, tabbrowser, targetDocument, action);
        }
      }
    }
  },
  
  /**
   * mht ファイルを展開して開く
   *
   * @param   {object} loadParam
   *          ロードのパラメータ
   *            {
   *              window:         {nsIDOMWindow}
   *                              実行中のウィンドウ (必須)
   *              uri:            {nsIURI}
   *                              開く URI (必須)
   *              targetWindow:   {nsIDOMWindow}
   *                              対象のウィンドウ (必須)
   *              hasDocument:    {boolean}
   *                              ドキュメントに内容を持っているか (必須)
   *              replaceHistory: {boolean}
   *                              履歴を置き換えるかどうか (必須)
   *              extract:        {boolean}
   *                              展開するか (必須)
   *              refName:        {string}
   *                              対象のファイル (必須)
   *              useCache:       {boolean}
   *                              キャッシュを使用するか (必須)
   *            }
   */
  loadWithParam: function(loadParam) {
    let { uri } = loadParam;
    
    if (uri.scheme == "file") {
      /* file プロトコルの場合、そのファイルから開く */
      this.loadWithParamFromFile(loadParam,
                                 arFileUtil.getFileFromURLSpec(uri.spec));
    } else {
      /* それ以外の場合、ウェブから開く */
      this.loadWithParamFromWeb(loadParam);
    }
  },
  
  /**
   * ファイルから mht ファイルを展開して開く
   *
   * @param   {object} loadParam
   *          ロードのパラメータ
   * @param   {nsIFile} targetFile
   *          開くファイル
   */
  loadWithParamFromFile: function(loadParam, targetFile) {
    let { window, targetWindow, uri, hasDocument } = loadParam;
    
    let data = arFileUtil.readData(targetFile);
    if (data == "") {
      if (hasDocument) {
        this.loadWithParamFromDocument(loadParam);
      } else {
        this.loadErrorPage(window, targetWindow, uri.spec);
      }
      return;
    }
    
    this.loadWithParamFromData(loadParam, data);
  },
  
  /**
   * データから mht ファイルを展開して開く
   *
   * @param   {object} loadParam
   *          ロードのパラメータ
   * @param   {string} data
   *          対象のデータ
   */
  loadWithParamFromData: function(loadParam, data) {
    let { uri, targetWindow, replaceHistory, extract, refName } = loadParam;
    
    let eFileInfo = UnMHTExtractor.extractMHT(uri.spec, data, extract);
    
    if (!eFileInfo) {
      return;
    }
    
    let urlspec = eFileInfo.baseURI;
    if (refName &&
        eFileInfo.refParts.has(refName)) {
      urlspec += refName;
    }
    
    targetWindow.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIWebNavigation)
      .loadURI(urlspec,
               replaceHistory
               ? Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY
               : 0,
               null, null, null);
  },
  
  /**
   * ドキュメントから mht ファイルを展開して開く
   *
   * @param   {object} loadParam
   *          ロードのパラメータ
   */
  loadWithParamFromDocument: function(loadParam) {
    let { window, uri, targetWindow } = loadParam;
    
    let data = "";
    Array.slice(targetWindow.document.getElementsByTagName("pre"))
      .forEach(function(node) {
        node = node.firstChild;
        while (node) {
          data += node.nodeValue;
          node = node.nextSibling;
        }
      });
    
    if (data == "") {
      this.loadErrorPage(window, targetWindow, uri.spec);
      return;
    }
    
    data = arUconv.fromUnicode(data, "utf-8");
    
    this.loadWithParamFromData(loadParam, data);
  },
  
  /**
   * ウェブのファイルから mht ファイルを展開して開く
   *
   * @param   {object} loadParam
   *          ロードのパラメータ
   */
  loadWithParamFromWeb: function(loadParam) {
    let { window, uri, targetWindow, hasDocument, useCache } = loadParam;
    
    let referrer = null;
    if (hasDocument) {
      try {
        referrer
          = Services.io.newURI(targetWindow.document.location.href, null, null);
      } catch (e) {
      }
    }
    
    UnMHTNetFileLoader.loadData(
      uri, referrer, useCache,
      function(status, data) {
        if (status == UnMHTNetFileLoader.STATUS_NG) {
          if (hasDocument) {
            UnMHTFileLoader.loadWithParamFromDocument(loadParam);
          } else {
            UnMHTFileLoader.loadErrorPage(window, targetWindow, uri.spec);
          }
        } else {
          UnMHTFileLoader.loadWithParamFromData(loadParam, data);
        }
      });
  },
  
  /**
   * エラーメッセージを表示する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {string} location
   *          ファイルの URI
   */
  loadErrorPage: function(window, targetWindow, location) {
    targetWindow.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIWebNavigation)
      .loadURI(UnMHTScheme.enErrorURI(location),
               Ci.nsIWebNavigation.LOAD_FLAGS_REPLACE_HISTORY
               , null, null, null);
  },
  
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this._unloadFunctions.slice(0).forEach(function(f) {
      f();
    });
    this._unloadFunctions = [];
  },
  
  /* ==== private ==== */
  
  _unloadFunctions: [], /** onunload に登録した関数
                         *  @type {Array.<Function>} */
  
  /**
   * インデックスページをタブに置き換える
   * 
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIDOMXULElement} tabbrowser
   *          対象のタブブラウザ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} action
   *          複数のタブを含む MHT を開く時
   *            tabs: 全てのページをタブとして開く
   *            index: インデックスページを表示する
   *            first: 最初のタブを開く
   */
  _replaceTabs: function(window, tabbrowser, targetDocument, action) {
    let nodes = targetDocument.getElementsByTagName("a");
    
    let tabsCount = 0;
    ((nodes.length && action == "first")
     ? [nodes[0]]
     : Array.slice(nodes))
      .forEach(function(node) {
        if ("className" in node &&
            node.className == "unmht-tabs") {
          tabbrowser.addTab(node.href, {relatedToCurrent : true});
          tabsCount ++;
        }
      });
    
    if (tabsCount == 0) {
      /* X-UnMHT-Tabs: true を持っているがタブが無い場合
       * 5.7.7 以前のバグで生成されたファイルの可能性がある */
      return;
    }
    
    let tab = UnMHTWindowUtil.getTabForDocument(window, targetDocument);
    if (tab) {
      tabbrowser.removeTab(tab);
    }
  },
  
  /**
   * リロードをフックする
   * 
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   */
  _hookReload: function(window, targetWindow) {
    let nav = null;
    
    try {
      nav
        = targetWindow.QueryInterface(Ci.nsIInterfaceRequestor)
        .getInterface(Ci.nsIWebNavigation);
    } catch (e) {
      return;
    }
    
    let listener = {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                             Ci.nsISHistoryListener]),
      
      OnHistoryReload: function(reloadURI, reloadFlags) {
        if (!UnMHTScheme.isUnMHTURI(reloadURI.spec)) {
          return true;
        }
        
        let param = UnMHTScheme.getUnMHTURIParam(reloadURI.spec);
        
        let loadParam = {
          window:         window,
          uri:            Services.io.newURI(param.original, null, null),
          targetWindow:   nav,
          hasDocument:    false,
          replaceHistory: true,
          extract:        true,
          refName:        param.refName,
          useCache:       false
        };
        UnMHTFileLoader.loadWithParam(loadParam);
        
        return false;
      },
      
      OnHistoryGoBack: function() { return true; },
      OnHistoryGoForward: function() { return true; },
      OnHistoryGotoIndex: function() { return true; },
      OnHistoryNewEntry: function() { return true; },
      OnHistoryPurge: function() { return true; }
    };
    
    try {
      nav.sessionHistory.addSHistoryListener(listener);
    } catch (e) {
      return;
    }
    
    let onunload = function() {
      try {
        nav.sessionHistory.removeSHistoryListener(listener);
      } catch (e) {
      }
      
      targetWindow.removeEventListener("unload", onunload, true);
      
      this._unloadFunctions
        = this._unloadFunctions.filter(function(f) onunload != f);
    }.bind(this);
    
    targetWindow.addEventListener("unload", onunload, true);
    
    this._unloadFunctions.push(onunload);
  },
  
  /**
   * フォーカスのあるウィンドウから mht ファイルを開く
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {boolean} extract
   *          展開するか
   */
  _loadFromFocusedWindow: function(window, extract) {
    let targetWindow = window.document.commandDispatcher.focusedWindow;
    if (targetWindow == window) {
      let tabbrowser = UnMHTWindowUtil.getBrowser(this._window);
      if (!tabbrowser) {
        return;
      }
      targetWindow = tabbrowser.contentWindow;
    }
    
    let uri
      = Services.io.newURI(targetWindow.document.location.href, null, null);
    
    if (UnMHTScheme.isUnMHTURI(uri.spec)) {
      uri.spec = UnMHTScheme.deUnMHTURI(uri.spec);
    }
    
    let loadParam = {
      window:         window,
      uri:            uri,
      targetWindow:   targetWindow,
      hasDocument:    true,
      replaceHistory: false,
      extract:        extract,
      refName:        "",
      useCache:       true
    };
    this.loadWithParam(loadParam);
  },
  
  /**
   * エラーページのリトライボタンをフックする
   * 
   * @param   {nsIDOMWindow} targetWindow
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {string} location
   *          ファイルの URI
   */
  _hookRetryButton: function(targetWindow, targetDocument, location) {
    let retryButton = targetDocument.getElementById("unmht-retry-button");
    if (!retryButton) {
      return;
    }
    
    let onclick = function() {
      try {
        targetWindow.location.href = UnMHTScheme.deErrorURI(location);
      } catch (e) {
      } 
    };
    
    retryButton.addEventListener("click", onclick, false);
    
    let onunload = function() {
      try {
        retryButton.removeEventListener("click", onclick, false);
      } catch (e) {
      }
      
      targetWindow.removeEventListener("unload", onunload, true);
      
      this._unloadFunctions
        = this._unloadFunctions.filter(function(f) onunload != f);
    }.bind(this);
    
    targetWindow.addEventListener("unload", onunload, true);
    
    this._unloadFunctions.push(onunload);
  }
};
UnMHTFileLoader.init();
