"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTStatusbarProgress.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "arAsyncProcs", function() {
  XPCOMUtils.importRelative(this, "base/arAsyncProcs.jsm");
  return arAsyncProcs;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTMail", function() {
  XPCOMUtils.importRelative(this, "UnMHTMail.jsm");
  return UnMHTMail;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTNetFileLoader", function() {
  XPCOMUtils.importRelative(this, "UnMHTNetFileLoader.jsm");
  return UnMHTNetFileLoader;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTMIMEDecoder", function() {
  XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");
  return UnMHTMIMEDecoder;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFilePicker", function() {
  XPCOMUtils.importRelative(this, "UnMHTFilePicker.jsm");
  return UnMHTFilePicker;
});

let EXPORTED_SYMBOLS = ["UnMHTFileSaver"];

/**
 * MHT で保存のラッパー
 * UI からの要求に応じてデータを取得して MHT での保存を開始する
 */
let UnMHTFileSaver = {
  /** saveWithParam のパラメータ
   *  @type {object} */
  PARAM_AUTO:      {},
  PARAM_CHOOSE:    {},
  PARAM_CURRENT:   {},
  PARAM_DIR:       {},
  PARAM_MAIL:      {},
  PARAM_OVERWRITE: {},
  
  /** エラー定数
   *  @type {number} */
  RESULT_OK:               0x00000000, /* 成功 */
  RESULT_FAIL:             0x10000001, /* 一般エラー */
  RESULT_CANCEL:           0x10000002, /* ユーザによるキャンセル */
  RESULT_NO_WINDOW:        0x20000101, /* window が無い */
  RESULT_NO_DOCUMENT:      0x20000102, /* targetDocument も scrapBookData も
                                        * 無い */
  RESULT_NO_SELECTION:     0x20000103, /* 選択範囲が無い
                                        * (targetSelection が
                                        *  PARAM_CURRENT の場合) */
  RESULT_NO_BASE_DIR:      0x20000104, /* baseDir が無い
                                        * (targetFile が
                                        *  this.PARAM_DIR の場合) */
  RESULT_NO_BROWSER:       0x20000105, /* ブラウザが無い */

  RESULT_NOT_A_FILE:       0x20000201, /* targetFile や baseDir が
                                        * ファイルではない */
  RESULT_NOT_A_LOCAL_FILE: 0x20000202, /* ドキュメントがローカルファイルでは
                                        * ない
                                        * (targetFile が 
                                        *  PARAM_OVERWRITE の場合) */
  RESULT_NOT_A_LINK:       0x20000203, /* リンクに載っていない */
  RESULT_NOT_A_MAIL:       0x20000204, /* ドキュメントがメールではない */
  
  /**
   * 現在のドキュメントを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  save: function(window, callback) {
    let param = {
      window:         window,
      targetDocument: this.PARAM_CURRENT,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * 対象のドキュメントを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  saveDocument: function(window, targetDocument, callback) {
    let param = {
      window:         window,
      targetDocument: targetDocument,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * 現在のドキュメントの選択範囲を保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  saveSelection: function(window, callback) {
    let param = {
      window:          window,
      targetDocument:  this.PARAM_CURRENT,
      targetSelection: this.PARAM_CURRENT,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },

  /**
   * 対象のドキュメントの現在の選択範囲を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  saveSelectionInDocument: function(window, targetDocument, callback) {
    let param = {
      window:          window,
      targetDocument:  targetDocument,
      targetSelection: this.PARAM_CURRENT,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * 現在のドキュメントを MHT に上書き保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  overwrite: function(window, callback) {
    let param = {
      window:          window,
      targetDocument:  this.PARAM_CURRENT,
      targetFile:      this.PARAM_OVERWRITE,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * リンク先を MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  saveLink: function(window, callback) {
    let href = UnMHTDOMUtil.getLinkURL(window);
    if (!href) {
      window.alert(UnMHTStringBundle.get("save.link.error.not_link"));
      if (callback) {
        callback(this.RESULT_NOT_A_LINK);
      }
      return;
    }
    
    let tabbrowser = UnMHTWindowUtil.getBrowser(window);
    if (!tabbrowser) {
      /* ==== FIXME: 警告を表示するべき? ==== */
      if (callback) {
        callback(this.RESULT_NO_BROWSER);
      }
      return;
    }
    
    let iframe = window.document.createElement("iframe");
    iframe.collapsed = true;
    iframe.setAttribute("src", href);
    iframe.setAttribute("type", "content");
    let onload = function() {
      iframe.removeEventListener("load", onload, true);
      
      let param = {
        window:         window,
        targetDocument: iframe.contentDocument,
      };
      this.saveWithParam(param, function(result, param) {
        iframe.parentNode.removeChild(iframe);
        if (callback) {
          callback(result);
        }
      });
    }.bind(this);
    iframe.addEventListener("load", onload, true);
    
    tabbrowser.parentNode.insertBefore(iframe, tabbrowser.nextSibling);
  },
  
  /**
   * 複数のタブを MHT で保存
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   */
  savemulti: function(window) {
    let pageURL = "chrome://unmht/content/savemulti.xul";
    
    if (UnMHTWindowUtil.focus(pageURL)) {
      return;
    }
    
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      targetWindow: window
    };
    window.openDialog(pageURL, "", features, arg);
  },
  
  /**
   * QuickSave
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  quicksave: function(window, callback) {
    let baseDirPath = UnMHTPref.getPref("save.quick.dir");
    let baseDir = baseDirPath
      ? arFileUtil.getFile(baseDirPath)
      : arFileUtil.desktopDirectory.clone();
    if (!baseDir) {
      window.alert(UnMHTStringBundle.get("save.error.illegal_directory"));
      if (callback) {
        callback(this.RESULT_NO_BASE_DIR);
      }
      return;
    }
    if (!baseDir.exists()) {
      try {
        arFileUtil.createDirectory(baseDir);
      } catch (e) {
        window.alert(UnMHTStringBundle.get("save.error.illegal_directory"));
        if (callback) {
          callback(this.RESULT_NO_BASE_DIR);
        }
        return;
      }
    }

    let getBaseDirForLocalFile = function(param) {
      let href = param.targetDocument.location.href;
      let localFile = { value: null };
      if (UnMHTDocumentCategory.isLocalFile(href, localFile)) {
        return localFile.value.parent;
      } else {
        return baseDir;
      }
    };
    let onDup = function(param) {
      let targetFile = param.targetFile;

      let dupAction = UnMHTPref.getPref("save.quick.dup.action");
      
      if (dupAction == "ask") {
        let result
          = UnMHTSave.showOverwriteConfirm(param.window,
                                           param.targetFile,
                                           param.filenameBase, false);
        
        if (result.result == "abort" || result.result == "no") {
          return null;
        }
        
        if (result.result == "rename") {
          targetFile = result.file;
        }
      } else if (dupAction == "rename") {
        targetFile
          = UnMHTSave.getUniqueFile(param.baseDir, param.filenameBase);
      }
      
      return targetFile;
    };
    
    let param = {
      window:           window,
      targetDocument:   this.PARAM_CURRENT,
      targetSelection:  (UnMHTPref.getPref("save.quick.use_selection")
                         ? this.PARAM_AUTO
                         : null),
      targetFile:       this.PARAM_DIR,
      baseDir:          (UnMHTPref.getPref("save.local.same_dir")
                         ? getBaseDirForLocalFile
                         : baseDir),
      additionalURIs:   null,
      showOptionDialog: false,
      saveState:        UnMHTSave.stringToSaveState(UnMHTPref.getPref("save.quick.state")),
      onDup:            onDup
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * 現在のドキュメントを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  sendmail: function(window, callback) {
    let param = {
      window:         window,
      targetDocument: this.PARAM_CURRENT,
      targetFile:     this.PARAM_MAIL,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * 対象のドキュメントを MHT としてメールで送信
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  sendmailDocument: function(window, targetDocument, callback) {
    let param = {
      window:         window,
      targetDocument: targetDocument,
      targetFile:     this.PARAM_MAIL,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },
  
  /**
   * ScrapBook を MHT で保存する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {object} scrapBookData
   *          ScrapBook のデータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(result)
   *              *param   {number} result
   */
  saveScrapBook: function(window, scrapBookData, callback) {
    let param = {
      window:           window,
      scrapBookData:    scrapBookData,
      showOptionDialog: false,
    };
    this.saveWithParam(param, function(result, param) {
      if (callback) {
        callback(result);
      }
    });
  },

  /**
   * パラメータに従ってドキュメントを保存する
   *
   * @param   {object} param
   *          保存パラメータ
   *            {
   *              window: {nsIDOMWindow}
   *                      実行中のウィンドウ (必須)
   *              targetDocument: {nsIDOMDocument}
   *                              対象のドキュメント
   *                              (scrapBookData とどちらか必須)
   *                                * UnMHTFileSaver.PARAM_CURRENT
   *                                    現在のドキュメント
   *                                * nsIDOMDocument
   *                                    指定したドキュメント
   *              scrapBookData: {object}
   *                             ScrapBook のデータ
   *                             (targetDocument とどちらか必須)
   *              date: {Date}
   *                    Dateフィールドの値 (オプショナル)
   *                      * this.PARAM_CURRENT
   *                          現在時刻
   *                      * this.PARAM_MAIL (デフォルト)
   *                          メールならばメールの Date フィールド
   *                          そうでないなら現在時刻
   *                      * Date
   *                          指定した時刻
   *              targetSelection: {nsISelection}
   *                               対象の選択範囲 (オプショナル)
   *                                 * UnMHTFileSaver.PARAM_CURRENT
   *                                     現在の選択範囲
   *                                     なければエラー
   *                                 * UnMHTFileSaver.PARAM_AUTO
   *                                     現在の選択範囲
   *                                     なければなし
   *                                 * nsISelection
   *                                     指定した選択範囲
   *                                 * null (デフォルト)
   *                                     なし
   *              showOptionDialog: {boolean}
   *                                オプションダイアログを表示するか
   *                                (オプショナル)
   *                                指定しない場合設定による
   *              targetFile: {nsIFile}
   *                          保存先 (オプショナル)
   *                            * PARAM_CHOOSE (デフォルト)
   *                                ファイル選択ダイアログを表示
   *                            * PARAM_DIR
   *                                baseDir に保存
   *                            * PARAM_OVERWRITE
   *                                上書き
   *                            * PARAM_MAIL
   *                                メールで送信
   *                            * nsIFile
   *                                指定したファイル
   *              baseDir: {nsIFile}
   *                       保存先のディレクトリ (オプショナル)
   *                         * nsIFile
   *                             指定したディレクトリ
   *                         * Function
   *                             targetDocument が決まってから関数で返す
   *                               function(param)
   *                                 *param   {object} param
   *                                 *returns {nsIFile}
   *                         * null (デフォルト)
   *                             なし
   *              filenameBase: {string}
   *                            ファイル名の基本部分 (オプショナル)
   *                              * string
   *                                  指定した文字列
   *                              * null (デフォルト)
   *                                  自動生成
   *              filenameType: {string}
   *                            ファイル名の種類 (オプショナル)
   *                              * string
   *                                  title/leafname/custom
   *                              * null (デフォルト)
   *                                  設定による
   *              additionalURIs: {Array.<string>}
   *                              追加のファイル (オプショナル)
   *                                * PARAM_CHOOSE (デフォルト)
   *                                    ダイアログを表示
   *                                * Array.<string>
   *                                    指定したファイル
   *                                * null
   *                                    なし
   *              saveState: {number}
   *                         保存状態 (オプショナル)
   *                         指定しない場合は選択する
   *              onDup: {Function }
   *                     同名のファイルがあった場合 (オプショナル)
   *                       function(param)
   *                         *param   {object} param
   *                         *returns {nsIFile}
   *              noRegisterHistory: {boolean}
   *                                 ダウンロード履歴に登録しない (オプショナル)
   *                                 指定しない場合は false
   *            }
   * @param   {Function} f
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  saveWithParam: function(param, callback) {
    let callbackObj = { value: callback };
    
    new arAsyncProcs()
      .next(function(cont, fail) {
        if (!("window" in param)) {
          fail(UnMHTFileSaver.RESULT_NO_WINDOW);
          return;
        }
        
        /* デフォルト値の設定 */
        if (!("targetDocument" in param)) {
          param.targetDocument = null;
        }
        if (!("scrapBookData" in param)) {
          param.scrapBookData = null;
        }
        if (!("date" in param)) {
          param.date = this.PARAM_MAIL;
        }
        if (!("targetSelection" in param)) {
          param.targetSelection = null;
        }
        if (!("showOptionDialog" in param)) {
          param.showOptionDialog
            = (UnMHTPref.getPref("save.show_option_dialog") ||
               param.targetFile == this.PARAM_OVERWRITE);
        }
        if (!("additionalURIs" in param)) {
          param.additionalURIs
            = (param.targetFile == this.PARAM_MAIL)
            ? null
            : this.PARAM_CHOOSE;
        }
        if (!param.additionalURIs) {
          param.additionalURIs = [];
        }
        if (!("filenameType" in param)) {
          param.filenameType = null;
        }
        if (!("filenameBase" in param)) {
          param.filenameBase = null;
        }
        if (!("targetFile" in param)) {
          param.targetFile = null;
        }
        if (!("baseDir" in param)) {
          param.baseDir = null;
        }
        if (!("noRegisterHistory" in param)) {
          param.noRegisterHistory = false;
        }
        if (!("saveState" in param)) {
          param.saveState
            = (param.targetFile == this.PARAM_MAIL)
            ? UnMHTSave.stringToSaveState(UnMHTPref.getPref("save.mail.state"))
            : null;
        }
        
        /* 保存対象 */
        if (param.targetDocument == null &&
            param.scrapBookData == null) {
          param.targetDocument = this.PARAM_CURRENT;
        }
        if (param.targetDocument == this.PARAM_CURRENT) {
          param.targetDocument
            = UnMHTWindowUtil.getCurrentDocument(param.window);
          if (!UnMHTDocumentCategory.isSavable(param.window,
                                               param.targetDocument, false)) {
            fail(this.RESULT_NOT_SAVABLE);
            return;
          }
        }
        if (!(param.targetDocument || param.scrapBookData)) {
          fail(UnMHTFileSaver.RESULT_NO_DOCUMENT);
          return;
        }
        
        /* Date フィールド */
        if (param.date == this.PARAM_MAIL) {
          if (this._needMailDate(param.window)) {
            this._getMailDate(param, function(result, param) {
              if (!(result == this.RESULT_OK ||
                    result == this.RESULT_NOT_A_MAIL)) {
                fail(result);
                return;
              }
              cont();
            }.bind(this));
            return;
          }
          
          param.date = new Date();
        }
        cont();
      }.bind(this))
      .next(function(cont, fail) {
        /* 選択範囲 */
        if (param.targetSelection == this.PARAM_CURRENT ||
            param.targetSelection == this.PARAM_AUTO) {
          let targetSelection
            = param.targetDocument.defaultView.getSelection();
          
          if (targetSelection &&
              !targetSelection.isCollapsed &&
              targetSelection.rangeCount > 0) {
            param.targetSelection = targetSelection;
            
            param.targetDocument
              = targetSelection.getRangeAt(0).startContainer.ownerDocument;
          } else {
            param.targetSelection = null;
            if (param.targetSelection == this.PARAM_CURRENT) {
              fail(UnMHTFileSaver.RESULT_NO_SELECTION);
              return;
            }
          }
        }
        
        /* ファイル名の基本部分 */
        if (param.filenameBase == null) {
          param.filenameBase
            = UnMHTSave.getFilenameBase(param.targetDocument,
                                        param.scrapBookData,
                                        param.filenameType,
                                        param.date);
        }
        
        /* オプションダイアログ */
        if (param.showOptionDialog) {
          param.showOptionDialog = false;
          this._showOptionDialog(param, function(result, param) {
            if (result != this.RESULT_OK) {
              fail(result);
              return;
            }
            cont();
          }.bind(this));
          return;
        }
        cont();
      }.bind(this))
      .next(function(cont, fail) {
        /* 保存先 */
        if (param.targetFile == null ||
            param.targetFile == this.PARAM_CHOOSE) {
          this._showFilePicker(param, function(result, param) {
            if (result != this.RESULT_OK) {
              fail(result);
              return;
            }
            cont();
          }.bind(this));
          return;
        }
        cont();
      }.bind(this))
      .next(function(cont, fail) {
        if (param.targetFile == this.PARAM_DIR) {
          if (param.baseDir == null) {
            fail(UnMHTFileSaver.RESULT_NO_BASE_DIR);
            return;
          }
          
          if (param.baseDir instanceof Function) {
            param.baseDir = param.baseDir(param);
          }
          
          if (!(param.baseDir instanceof Ci.nsIFile)) {
            fail(UnMHTFileSaver.RESULT_NOT_A_FILE);
            return;
          }
          
          param.targetFile
            = arFileUtil.checkAndAppend(param.baseDir, param.filenameBase, "",
                                        UnMHTFilename.defaultExtension);
          
          if (param.targetFile.exists()) {
            if ("onDup" in param) {
              param.targetFile = param.onDup(param);
              if (!param.targetFile) {
                fail(this.RESULT_CANCEL);
                return;
              }
            }
          }
        } else if (param.targetFile == this.PARAM_OVERWRITE) {
          if (!param.targetDocument) {
            fail(UnMHTFileSaver.RESULT_NO_DOCUMENT);
            return;
          }
          
          param.targetFile
            = arFileUtil.getFileFromURLSpec(UnMHTScheme.deUnMHTURI(param.targetDocument.location.href));
          if (!param.targetFile) {
            fail(UnMHTFileSaver.RESULT_NOT_A_LOCAL_FILE);
            return;
          }
        } else if (param.targetFile == this.PARAM_MAIL) {
          let tempDir = arFileUtil.tempDirectory.clone();
          tempDir.append(Date.now());
          arFileUtil.createUniqueDirectory(tempDir);
          
          param.targetFile
            = arFileUtil.checkAndAppend(tempDir, param.filenameBase,
                                        "", UnMHTFilename.defaultExtension);
          
          callbackObj.value = let (origCallback = callbackObj.value)
          function(result, param) {
            if (result != this.RESULT_OK) {
              if (origCallback) {
                origCallback(result, param);
              }
              return;
            }
            
            if (!UnMHTMail.send(param.window, param.targetFile,
                                param.targetDocument.title)) {
              param.window.alert(UnMHTStringBundle.get("save.mail.fail")
                                 + "\n"
                                 + UnMHTStringBundle.get("save.mail.fail2"));
            }
            
            if (origCallback) {
              origCallback(result, param);
            }
          }.bind(this);
        }
        
        if (!(param.targetFile instanceof Ci.nsIFile)) {
          fail(UnMHTFileSaver.RESULT_NOT_A_FILE);
          return;
        }
        
        /* 追加ファイル */
        if (param.additionalURIs == this.PARAM_CHOOSE) {
          this._showAdditionalDialog(param, function(result, param) {
            if (result != this.RESULT_OK) {
              fail(result);
              return;
            }
            cont();
          }.bind(this));
          return;
        }
        cont();
      }.bind(this))
      .next(function(cont, fail) {
        /* 保存開始 */
        if (UnMHTPref.getPref("save.progress.statusbar")) {
          this._startSaveInStatusbar(param, callbackObj.value);
        } else {
          this._startSaveInDialog(param, callbackObj.value);
        }
        cont();
      }.bind(this))
      .onFail(function(result) {
        if (callbackObj.value) {
          callbackObj.value(result, param);
        }
      })
      .run();
  },
  
  /**
   * オプションダイアログを表示する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _showOptionDialog: function(param, callback) {
    let saveStatePref
      = (param.targetFile == this.PARAM_MAIL)
      ? "save.mail.state"
      : "save.single.state";
    
    let saveState
      = UnMHTSave.stringToSaveState(UnMHTPref.getPref(saveStatePref));
    let saveStateMask = UnMHTSave.SAVESTATE_MASK_ALL;
    
    if ("targetFile" in param &&
        param.targetFile == this.PARAM_OVERWRITE) {
      saveStateMask = UnMHTSave.SAVESTATE_MASK_CURRENT;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSave.SAVESTATE_CURRENT;
      }
    } else if (param.scrapBookData) {
      saveStateMask = UnMHTSave.SAVESTATE_MASK_ORIGINAL;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSave.SAVESTATE_ORIGINAL;
      }
    } else if (param.targetSelection) {
      saveStateMask = UnMHTSave.SAVESTATE_MASK_CURRENT;
      if (!(saveState & saveStateMask)) {
        saveState = UnMHTSave.SAVESTATE_CURRENT;
      }
    }
    
    let pageURL = "chrome://unmht/content/saveoptions.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      callback:       function(result, resultParam) {
        if (result != UnMHTSave.RESULT_OK) {
          if (callback) {
            callback(result, param);
          }
          return;
        }
        
        param.filenameBase = resultParam.filenameBase;
        param.saveState = resultParam.saveState;
        param.additionalURIs = resultParam.additionalURIs;
        if (callback) {
          callback(this.RESULT_OK, param);
        }
      }.bind(this),
      filenameBase:   param.filenameBase,
      showFilename:   param.targetFile == this.PARAM_MAIL,
      saveState:      saveState,
      saveStateMask:  saveStateMask,
      saveStatePref:  saveStatePref,
      showInsertLink: param.targetFile == this.PARAM_OVERWRITE,
    };
    param.window.openDialog(pageURL, "", features, arg);
  },
  
  /**
   * ファイル選択ダイアログを表示する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _showFilePicker: function(param, callback) {
    let showAdditionalDialog = false;
    
    let displayDirectory = null;
    let localFile = { value: null };
    if (param.targetDocument &&
        UnMHTPref.getPref("save.local.same_dir") &&
        UnMHTDocumentCategory.isLocalFile(param.targetDocument.location.href,
                                          localFile)) {
      displayDirectory = localFile.value.parent;
    }
    
    let types = [];
    if (param.saveState != null &&
        param.additionalURIs != this.PARAM_CHOOSE) {
      types = [[param.saveState, param.additionalURIs.length > 0]];
    } else {
      let saveStateMask = UnMHTSave.SAVESTATE_MASK_ALL;
      
      if (param.scrapBookData) {
        saveStateMask = UnMHTSave.SAVESTATE_MASK_ORIGINAL;
      }
      if (param.targetSelection) {
        saveStateMask = UnMHTSave.SAVESTATE_MASK_CURRENT;
      }
      
      let stateList = UnMHTSave.getSaveStateList(saveStateMask);
      
      types
        = [...stateList.map(function(state) [state, false]),
           ...stateList.map(function(state) [state, true])];
    }
    
    /* ==== FIXME: 汎用性が無い ==== */
    let lastStateStr = UnMHTPref.getPref("save.single.state");
    let lastAdd = UnMHTPref.getPref("save.single.add");
    
    let filterIndex = -1;
    let filter = [];
    types.forEach(function([saveState, useAdd], i) {
      filter.push([UnMHTSave.makeFiletype(saveState, useAdd), "*.mht"]);
      
      if (UnMHTSave.saveStateToString(saveState) == lastStateStr &&
          useAdd == lastAdd) {
        filterIndex = i;
      }
    });
    
    let params = {
      window:           param.window,
      title:            UnMHTStringBundle.get("save.title"),
      mode:             Ci.nsIFilePicker.modeSave,
      useLastDir:       true,
      displayDirectory: displayDirectory,
      defaultString:    (param.filenameBase
                         + "." + UnMHTFilename.defaultExtension),
      defaultExtension: UnMHTFilename.defaultExtension,
      filter:           filter,
      filterIndex:      filterIndex,
      uri:              (param.targetDocument
                         ? param.targetDocument.location.href
                         : null),
      callback:         function(targetFile, result, filePicker) {
        if (!targetFile) {
          if (callback) {
            callback(this.RESULT_CANCEL, param);
          }
          return;
        }
        
        if (!(param.saveState != null &&
              param.additionalURIs != this.PARAM_CHOOSE)) {
          let [saveState, useAdd] = types[filePicker.filterIndex];
          
          if (UnMHTPref.getPref("save.single.state.remember")) {
            UnMHTPref.setPref("save.single.state", 
                              UnMHTSave.saveStateToString(saveState));
            UnMHTPref.setPref("save.single.add", useAdd);
          }
          
          param.saveState = saveState;
          
          if (!useAdd) {
            param.additionalURIs = [];
          }
        }
        
        let result = { value: true };
        targetFile = UnMHTSave.getValidFile(param.window, targetFile, result);
        if (!result.value) {
          if (callback) {
            callback(this.RESULT_CANCEL, param);
          }
          return;
        }
        
        param.targetFile = targetFile;
        
        if (callback) {
          callback(this.RESULT_OK, param);
        }
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },

  /**
   * 追加ファイルのダイアログを表示する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _showAdditionalDialog: function(param, callback) {
    let pageURL = "chrome://unmht/content/additional.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      callback: function(result, additionalURIs) {
        if (result != UnMHTSave.RESULT_OK) {
          if (callback) {
            callback(result, param);
          }
          return;
        }
        
        param.additionalURIs = additionalURIs;
        if (callback) {
          callback(this.RESULT_OK, param);
        }
      }.bind(this)
    };
    param.window.openDialog(pageURL, "", features, arg);
  },
  
  /**
   * メールの日時が必要かを返す
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {boolean}
   *          メールの日時が必要か
   */
  _needMailDate: function(window) {
    return (UnMHTPref.getPref("save.filename.custom.mail_date") &&
            window &&
            UnMHTWindowUtil.isMailWindow(window) &&
            "gFolderDisplay" in window &&
            "selectedMessageUris" in window.gFolderDisplay &&
            window.gFolderDisplay.selectedMessageUris.length == 1);
  },
  
  /**
   * メールの日時を取得する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _getMailDate: function(param, callback) {
    let window = param.window;
    
    let mailSession
      = Cc["@mozilla.org/messenger/services/session;1"]
      .getService(Ci.nsIMsgMailSession);
    
    let messageURL
      = mailSession.ConvertMsgURIToMsgURL(window.gFolderDisplay.selectedMessageUris[0], window.msgWindow);
    
    let uri;
    try {
      uri = Services.io.newURI(messageURL, null, null);
    } catch (e) {
    }
    
    if (!uri) {
      if (callback) {
        callback(this.RESULT_NOT_A_MAIL, param);
      }
      return;
    }

    UnMHTNetFileLoader.loadData(
      uri, null, true, function(status, data) {
        let date = new Date();
        if (status != UnMHTNetFileLoader.STATUS_NG) {
          let mailPart
            = UnMHTMIMEDecoder.decodeMessage(data.replace(/\r\n|\r|\n/g, "\r\n"), true);
          if (mailPart &&
              mailPart.fields.has("date")) {
            let tmpDate
              = UnMHTMIMEDecoder.decodeDate(mailPart.fields.get("date"));
            if (tmpDate) {
              date = tmpDate;
            }
          }
        }
        
        param.date = date;
        
        if (callback) {
          callback(this.RESULT_OK, param);
        }
      }.bind(this));
  },
  
  /**
   * 保存を開始してステータスバーに登録する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _startSaveInStatusbar: function(param, callback) {
    let prog = UnMHTStatusbarProgressMap.get(param.window);
    let [location, subject]
      = UnMHTSave.getLocationAndTitle(param.targetDocument, param.scrapBookData);
    
    let saveContext
      = UnMHTSave.createContext(param.window, param.targetFile,
                                location, subject, prog,
                                param.noRegisterHistory);
    
    let arg = {
      targetDocument:  param.targetDocument,
      targetSelection: param.targetSelection,
      scrapBookData:   param.scrapBookData,
      saveState:       param.saveState,
      location:        location,
      isHTML:          true,
      isStartPart:     true,
      window:          param.window
    };
    if (!saveContext.addDocument(arg)) {
      saveContext.destruct();
      
      if (callback) {
        callback(UnMHTSave.RESULT_ERROR_ADD_DOCUMENT, param);
      }
      return;
    }
    
    if (param.additionalURIs) {
      param.additionalURIs.forEach(function(file) {
        saveContext.addFile(null, file, false);
      });
    }
    
    prog.addContext(saveContext, function(result) {
      if (callback) {
        callback(result, param);
      }
    });
    
    saveContext.start();
  },
  
  /**
   * ダイアログで保存を開始する
   *
   * @param   {object} param
   *          保存パラメータ
   * @param   {Function} callback
   *          終了時コールバック
   *            function(param, result)
   *              *param   {object} param
   *              *param   {number} result
   */
  _startSaveInDialog: function(param, callback) {
    let pageURL = "chrome://unmht/content/save.xul";
    let features = "chrome,titlebar,centerscreen,resizable";
    let arg = {
      window:            param.window,
      targetFile:        param.targetFile,
      targetDocument:    param.targetDocument,
      saveState:         param.saveState,
      scrapBookData:     param.scrapBookData,
      targetSelection:   param.targetSelection,
      additionalURIs:    param.additionalURIs,
      noRegisterHistory: param.noRegisterHistory,
      callback:          function(result) {
        if (callback) {
          callback(result, param);
        }
      }
    };
    param.window.openDialog(pageURL, "", features, arg);
  }
};
