"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.defineLazyGetter(this, "arFileUtil", function() {
  XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
  return arFileUtil;
});

let EXPORTED_SYMBOLS = ["UnMHTFilename"];

/**
 * mht ファイルのファイル名操作
 */
let UnMHTFilename = {
  /**
   * MHT ファイルの拡張子
   * @type {string}
   */
  get defaultExtension() {
    return "mht";
  },

  /**
   * ファイル名から MHT の拡張子を削除する
   *
   * @param   {string} filename
   *          ファイル名
   * @returns {string}
   *          ファイル名 (拡張子以外)
   */
  removeExtension: function(filename) {
    return filename.replace(/\.mht(ml)?$/i, "");
  },
  

  /**
   * ファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @returns {string}
   *          変換したファイル名
   */
  validateFileName: function(filename) {
    return this.validateFileNameWithOS(filename, Services.appinfo.OS);
  },
  
  /**
   * 指定した OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @param   {string} OS
   *          対象の OS
   * @returns {string}
   *          変換したファイル名
   */
  validateFileNameWithOS: function(filename, OS) {
    let [fromRE, toMap] = this._loadReplaceList();
    if (fromRE && toMap) {
      filename = filename.replace(fromRE, function(mathed) toMap.get(mathed));
    }
    
    if (OS == "WINNT" &&
        !UnMHTPref.getPref("save.filename.replace.ie_bug.disable")) {
      /*
       * ファイル名に使用すると IE が MHT ファイルを開けなくなる文字
       *   #
       *   !
       */
      filename = filename.replace(/[#\!]/g, "_");
    }
    
    return arFileUtil.validateFileNameWithOS(filename, OS);
  },
  
  /**
   * 文字の置換元の正規表現と置換先のマップを返す
   *
   * @returns {Array}
   *            [RegExp  置換元の正規表現,
   *             Map.<string,string>  置換先のマップ]
   *          正しく読み込めなければ [null, null]
   */
  _loadReplaceList: function() {
    let datas;
    try {
      datas = JSON.parse(UnMHTPref.getPref("save.filename.replace.list"));
      if (!datas.every(function(data) ("from" in data && "to" in data))) {
        return [null, null];
      }
    } catch (e) {
      return [null, null];
    }
    
    if (!datas.length) {
      return [null, null];
    }
    
    let validDatas = datas.filter(function(data) data.from != "");
    if (!validDatas.length) {
      return [null, null];
    }
    
    let escapeRegExp = function(text) {
      return text
        .replace(/([\\\[\]\(\)\^\,\.\{\}\|\?\!\-\*\+\$])/g, "\\$1");
    };
    
    let toMap = new Map();
    datas.forEach(function(data) {
      toMap.set(data.from, data.to);
    });
    
    return [new RegExp(validDatas
                       .map(function(data) escapeRegExp(data.from))
                       .join("|"), "g"),
            toMap];
  }
};
