"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTExtractor", function() {
  XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
  return UnMHTExtractor;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFormatter", function() {
  XPCOMUtils.importRelative(this, "UnMHTFormatter.jsm");
  return UnMHTFormatter;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTMIMEDecoder", function() {
  XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");
  return UnMHTMIMEDecoder;
});

let EXPORTED_SYMBOLS = ["UnMHTInfoPanelMap",
                        "UnMHTInfoPanelClass"];

/**
 * window から UnMHTInfoPanel へのマップ
 */
let UnMHTInfoPanelMap = new Map();
let listener = {
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    try {
      UnMHTInfoPanelMap.clear();
    } catch (e) {
      /* COMPAT: Firefox 18 以前 */
      for (let [name,] of UnMHTInfoPanelMap) {
        UnMHTInfoPanelMap.delete(name);
      }
    }
  }
};
arShutdownEvent.addListener(listener);

/**
 * 情報パネル
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements nsISupportsWeakReference
 * @implements nsIWebProgressListener
 * @implements arIWindowHook
 */
function UnMHTInfoPanelClass(window) {
  arWindowHook.call(this, window);
  
  UnMHTInfoPanelMap.set(window, this);
}
UnMHTInfoPanelClass.prototype = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                         Ci.nsIWebProgressListener]),
  
  /**
   * 情報アイコンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onInfoIconClick: function(event) {
    let icon = this.el("unmht-urlbar-icon");
    if (!icon) {
      return;
    }
    
    let targetDocument = UnMHTWindowUtil.getCurrentDocument(this._window);
    
    let location = targetDocument.location.href;
    
    let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(location);
    if (!(eFileInfo && part)) {
      icon.hidden = true;
      return;
    }
    
    this._currentLocation = eFileInfo.original;
    this._currentSavedFrom = "";
    this._linkable = false;
    
    this._panel = this.el("unmht-info-panel");
    
    this._panel.hidden = false;
    this.el("unmht-info-panel").style.width
      = this.el("urlbar").boxObject.width + "px";
    this.el("unmht-info-panel-original").value = this._currentLocation;
    
    if (eFileInfo.startPart &&
        eFileInfo.startPart.contentLocation) {
      this._currentSavedFrom
        = part.contentLocation ||
        eFileInfo.startPart.contentLocation;
      
      this.el("unmht-info-panel-row-saved-from").hidden = false;
      this.el("unmht-info-panel-saved-from").value = this._currentSavedFrom;
      
      this._linkable = /^[A-Za-z0-9\-]+:/.test(this._currentSavedFrom);
      
      if (this._linkable) {
        this.el("unmht-info-panel-saved-from").className = "text-link";
      } else {
        this.el("unmht-info-panel-saved-from").className = "";
      }
    } else {
      this.el("unmht-info-panel-row-saved-from").hidden = true;
    }
    if (eFileInfo.date) {
      let savedTime = eFileInfo.date;
      try {
        let date = UnMHTMIMEDecoder.decodeDate(savedTime);
        if (date) {
          savedTime = UnMHTFormatter.getReadableDate(date);
        }
      } catch (e) {
      }
      
      this.el("unmht-info-panel-row-saved-date").hidden = false;
      this.el("unmht-info-panel-saved-date").value = savedTime;
    } else {
      this.el("unmht-info-panel-row-saved-date").hidden = true;
    }
    
    this._panel.openPopup(icon, "bottomcenter topright");
    
    event.stopPropagation();
  },
  
  /**
   * 情報パネルのコピーボタンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onInfoCopyClick: function(event) {
    Cc["@mozilla.org/widget/clipboardhelper;1"]
      .getService(Ci.nsIClipboardHelper)
      .copyString(this._currentLocation);
  },
  
  /**
   * 情報パネルのコピーボタンがクリックされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onInfoSavedFromClick: function(event) {
    if (this._linkable) {
      this._panel.hidePopup();
      let tabbrowser = UnMHTWindowUtil.getBrowser(this._window);
      tabbrowser.selectedTab = tabbrowser.addTab(this._currentSavedFrom);
    }
  },
  
  /* ==== nsIWebProgressListener インターフェース ==== */
  
  /**
   * 監視ウィンドウのロケーションが変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {nsIURI} location
   *          ロケーション
   */
  onLocationChange: function(webProgress, request, location) {
    this._setInfoIconState();
  },
  
  /**
   * 進捗状況が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} curSelfProgress
   *          対象のリクエストの現在の進捗状況
   * @param   {number} maxSelfProgress
   *          対象のリクエストの最大の進捗状況
   * @param   {number} curTotalProgress
   *          全ての現在の進捗状況
   * @param   {number} maxTotalProgress
   *          全ての最大の進捗状況
   */
  onProgressChange: function(webProgress, request,
                             curSelfProgress, maxSelfProgress,
                             curTotalProgress, maxTotalProgress) {
  },
  
  /**
   * プロトコルのセキュリティ設定が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} state
   *          セキュリティ状態フラグ、セキュリティ強度フラグ
   */
  onSecurityChange: function(webProgress, request, state) {
  },
  
  /**
   * 状況が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} flags
   *          状況フラグ
   *            Ci.nsIWebProgressListener.STATE_*
   * @param   {nsresult} status
   *          ステータス
   */
  onStateChange: function(webProgress, request, flags, status) {
    if (flags & Ci.nsIWebProgressListener.STATE_STOP) {
      this._setInfoIconState();
    }
  },
  
  /**
   * ステータスバーに表示するメッセージが変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {nsresult} status
   *          リクエストのステータス
   * @param   {string} message
   *          ステータスバーのメッセージ
   */
  onStatusChange: function(webProgress, request, status, message) {
  },
  
  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    UnMHTInfoPanelMap.delete(this._window);
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    let browser = UnMHTWindowUtil.getBrowser(this._window);
    if (browser) {
      browser.addProgressListener(this);
    }
    
    this._setInfoIconState();
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    if (this._panel) {
      try {
        this._panel.hidePopup();
      } catch (e) {
      }
      this._panel = null;
    }
    
    let browser = UnMHTWindowUtil.getBrowser(this._window);
    if (browser) {
      browser.removeProgressListener(this);
    }
  },
  
  /* ==== private ==== */
  
  _panel: null,          /** 情報パネル
                          *  @type {nsIDOMXULElement} */
  _currentLocation: "",  /** 情報パネルに表示されている本来のアドレス
                          *  @type {string} */
  _currentSavedFrom: "", /**  情報パネルに表示されている保存元
                          *  @type {string} */
  _linkable: false,      /** 保存元がリンク可能か
                          *  @type {boolean} */
  
  /**
   * 情報アイコンの状態を状況に合わせて切り替える
   */
  _setInfoIconState: function() {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument(this._window);
    
    let icon = this.el("unmht-urlbar-icon");
    if (!icon) {
      return;
    }
    
    let location = targetDocument.location.href;
    if (UnMHTScheme.isUnMHTURI(location)) {
      let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(location);
      if (eFileInfo && part) {
        icon.hidden = false;
        
        /* 開始パートの場合のみロケーションバーに本来の URI を表示する */
        try {
          let param = UnMHTScheme.getUnMHTURIParam(location);
          if (param.refName == "") {
            let urlbar = this.el("urlbar");
            let value = this._window.losslessDecodeURI(Services.io.newURI(param.original, null, null));
            
            urlbar.value = value;
            
            /* ProgressListener の登録が XULBrowserWindow より早かった場合
             * XULBrowserWindow.onLocationChange の方が後から呼ばれ
             * ロケーションバーの表示に反映されないので
             * 非同期にも設定する */
            arTimer.setTimeout(function() {
              urlbar.value = value;
            }, 0);
          }
        } catch (e) {
          /* 仕様変更等で失敗する事もある
           * unmht URI scheme が表示される事自体はそれほど問題ではないので
           * その場合は何もしない */
        }
        return;
      }
    }
    
    icon.hidden = true;
  }
};
arPrototype.extend(UnMHTInfoPanelClass, arWindowHook);
