"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTFileSaver", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileSaver.jsm");
  return UnMHTFileSaver;
});

let EXPORTED_SYMBOLS = ["UnMHTKeyboardShortcut"];

/**
 * キーボードショートカット
 *
 * @param   {string} pref
 *          設定名
 * @param   {Function} func
 *          実行する関数
 */
function UnMHTKeyboardShortcutItem(pref, func) {
  this._pref = pref;
  this._func = func;
}
UnMHTKeyboardShortcutItem.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._func = null;
  },
  
  /**
   * 設定を読み込む
   */
  loadPref: function() {
    this._enabled = UnMHTPref.getPref(this._pref);
    this._keycode
      = Ci.nsIDOMKeyEvent["DOM_" + UnMHTPref.getPref(this._pref + ".keycode")];
    this._modsAlt = UnMHTPref.getPref(this._pref + ".mods.alt");
    this._modsCtrl = UnMHTPref.getPref(this._pref + ".mods.ctrl");
    this._modsMeta = UnMHTPref.getPref(this._pref + ".mods.meta");
    this._modsShift = UnMHTPref.getPref(this._pref + ".mods.shift");
  },
  
  /**
   * 有効かどうかを返す
   *
   * @returns {boolean}
   *          有効かどうか
   */
  isEnabled: function() {
    return this._enabled;
  },
  
  /**
   * ショートカットキーを調べる
   *
   * @param   {Event} event
   *          対象のイベント
   */
  check: function(event) {
    return (this._keycode == event.keyCode &&
            this._modsAlt == event.altKey &&
            this._modsCtrl == event.ctrlKey &&
            this._modsMeta == event.metaKey &&
            this._modsShift == event.shiftKey);
  },
  
  /**
   * 関数を実行する
   */
  fire: function() {
    this._func();
  },
  
  /* ==== private ==== */
  
  _pref: "",         /** 設定名
                      *  @type {string} */
  _func: "",         /** 実行する関数
                      *  @type {Function} */
  
  _enabled: false,   /** 有効
                      *  @type {boolean} */
  _keycode: 0,       /** キーコード
                      *  @type {number} */
  _modsAlt: false,   /** Alt
                      *  @type {boolean} */
  _modsCtrl: false,  /** Ctrl
                      *  @type {boolean} */
  _modsMeta: false,  /** Meta
                      *  @type {boolean} */
  _modsShift: false  /** Shift
                      *  @type {boolean} */
};

/**
 * キーボードショートカットのフック
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 */
function UnMHTKeyboardShortcut(window) {
  arWindowHook.call(this, window);
  
  this._items = [
    new UnMHTKeyboardShortcutItem("shortcut.save.single", function() {
      let targetDocument = UnMHTWindowUtil.getCurrentDocument(window);
      if (!UnMHTDocumentCategory.isSavable(window, targetDocument, false)) {
        return false;
      }
      
      UnMHTFileSaver.save(window);
      return true;
    }),
    new UnMHTKeyboardShortcutItem("shortcut.save.multi", function() {
      UnMHTFileSaver.savemulti(window);
      return true;
    }),
    new UnMHTKeyboardShortcutItem("shortcut.save.quick", function() {
      let targetDocument = UnMHTWindowUtil.getCurrentDocument(window);
      if (!UnMHTDocumentCategory.isSavable(window, targetDocument, false)) {
        return false;
      }
      
      UnMHTFileSaver.quicksave(window);
      return true;
    }),
    new UnMHTKeyboardShortcutItem("shortcut.save.overwrite", function() {
      let targetDocument = UnMHTWindowUtil.getCurrentDocument(window);
      if (!(UnMHTDocumentCategory.isSavable(window, targetDocument, false) &&
            UnMHTDocumentCategory.isLocalMHT(window, targetDocument))) {
        return false;
      }
      
      UnMHTFileSaver.overwrite(window);
      return true;
    })
  ];
  
  this.onKeyDown = this.onKeyDown.bind(this);
}
UnMHTKeyboardShortcut.prototype = {
  /* ---- UI イベント - キーボード ---- */
  
  /**
   * キーが押されたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onKeyDown: function(event) {
    this._enabledItems.some(function(item) {
      if (item.check(event) &&
          item.fire()) {
        event.preventDefault();
        event.stopPropagation();
        return true;
      }
      return false;
    });
  },
  
  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._enabledItems = null;
    this._items.forEach(function(item) item.destruct());
    this._items = null;
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    UnMHTPref.addListener("shortcut.save",
                          function() {
                            this._items.forEach(function(item) item.loadPref());
                            this._hookShortcut();
                          }.bind(this), this, true);
  },  
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    if (this._isHooked) {
      this._window.removeEventListener("keydown", this.onKeyDown, true);
      this._isHooked = false;
    }
    
    UnMHTPref.removeAllListeners(this);
  },
  
  /* ==== private ==== */

  _items: null,        /** ショートカット
                        *  @type {Array.<UnMHTKeyboardShortcutItem>} */
  _enabledItems: null, /** 有効なショートカット
                        *  @type {Array.<UnMHTKeyboardShortcutItem>} */
  
  _isHooked: false,    /** キーボードイベントをフックしたか
                        *  @type {boolean} */
  
  /**
   * 設定に合わせてキーイベントをフックする
   */
  _hookShortcut: function() {
    this._enabledItems = this._items.filter(function(item) item.isEnabled());
    
    if (this._enabledItems.length && !this._isHooked) {
      this._window.addEventListener("keydown", this.onKeyDown, true);
      this._isHooked = true;
    }
    if (!this._enabledItems.length && this._isHooked) {
      this._window.removeEventListener("keydown", this.onKeyDown, true);
      this._isHooked = false;
    }
  }
};
arPrototype.extend(UnMHTKeyboardShortcut, arWindowHook);
