"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTFileSaver", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileSaver.jsm");
  return UnMHTFileSaver;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTSave", function() {
  XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
  return UnMHTSave;
});

let EXPORTED_SYMBOLS = ["UnMHTMAFIntegration"];

/**
 * MAF 組み込みオブジェクト
 *
 * @param   {number} saveState
 *          保存状態
 * @param   {boolean} useAdditionalURIs
 *          追加の URI を使用するか
 * @implements nsICancelable
 * @implements nsIWebBrowserPersist
 */
function UnMHTMAFSavePersist(saveState, useAdditionalURIs) {
  this._saveState = saveState;
  this._useAdditionalURIs = useAdditionalURIs;
  this.currentState = Ci.nsIWebBrowserPersist.PERSIST_STATE_READY;
}
UnMHTMAFSavePersist.prototype = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIWebBrowserPersist,
                                         Ci.nsICancelable]),
  
  /* ==== nsICancelable インターフェース ==== */
  
  /**
   * 保存キャンセル
   *
   * @param   {nsresult} reason
   *          理由
   */
  cancel: function(reason) {
    this.result = reason;
  },
  
  /* ==== nsIWebBrowserPersist インターフェース ==== */
  
  currentState: 0,         /** 状態
                            *  @type {number} */
  persistFlags: 0,         /** データ取得方法フラグ
                            *  @type {number} */
  progressListener: null,  /** 進捗状況リスナ
                            *  @type {nsIWebProgressListener} */
  result: Cr.NS_OK,        /** 理由
                            *  @type {nsresult} */
  
  /**
   * 保存キャンセルイベント
   */
  cancelSave: function() {
    this.cancel(Cr.NS_BINDING_ABORTED);
  },
  
  /**
   * 保存開始イベント
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {nsISupports} targetFileURI
   *          保存先のファイル (nsIFileURI)
   * @param   {nsISupports} targetDataDir
   *          保存先のディレクトリ? (nsIFile/nsIURI)
   * @param   {string} outputContentType
   *          保存先のファイルの MIME-Type
   * @param   {number} encodingFlags
   *          エンコーダに渡すのフラグ
   * @param   {number} wrapColumn
   *          折り返し文字数
   */
  saveDocument: function(targetDocument, targetFileURI,
                         targetDataDir, outputContentType,
                         encodingFlags, wrapColumn) {
    try {
      this.currentState = Ci.nsIWebBrowserPersist
        .PERSIST_STATE_SAVING;
      if (this.progressListener) {
        this.progressListener.onStateChange(
          null, null,
          Ci.nsIWebProgressListener.STATE_START
            | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
          Cr.NS_OK);
      }
      
      let chromeWindow = Services.wm.getMostRecentWindow("navigator:browser");
      
      let result = { value: true };
      let targetFile
        = UnMHTSave.getValidFile(chromeWindow,
                                 targetFileURI.QueryInterface(Ci.nsIFileURL).file,
                                 result);
      if (!result.value) {
        this.result = Cr.NS_OK;
        this._onComplete();
        return;
      }
      
      let param = {
        window:            chromeWindow,
        targetDocument:    targetDocument,
        saveState:         this._saveState,
        additionalURIs:    (this._useAdditionalURIs
                            ? UnMHTFileSaver.PARAM_CHOOSE
                            : null),
        showOptionDialog:  false,
        targetFile:        targetFile,
        noRegisterHistory: true
      };
      
      UnMHTFileSaver.saveWithParam(param, function(result, param) {
        if (result == UnMHTFileSaver.RESULT_OK) {
          this.result = Cr.NS_OK;
        } else {
          this.result = Cr.NS_ERROR_FAILURE;
        }
        
        this._onComplete();
      }.bind(this));
    } catch (e) {
      if (e instanceof Ci.nsIXPCException) {
        this.result = e.result;
      } else {
        this.result = Cr.NS_ERROR_FAILURE;
      }
      
      this._onComplete();
    }
  },
  
  /**
   * チャネルを保存する
   *
   * @param   {nsIChannel} channel
   *          対象のチャネル
   * @param   {nsIFile} file
   *          対象のファイル
   */
  saveChannel: function(channel, file) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * URI を保存する
   *
   * @param   {nsIURI} URI
   *          対象の URI
   * @param   {nsISupports} cacheKey
   *          キャッシュオブジェクト
   * @param   {nsIURI} referrer
   *          リファラ
   * @param   {nsIInputStream} postData
   *          POST データ
   * @param   {string} extraHeaders
   *          拡張ヘッダ
   * @param   {nsIFile} file
   *          対象のファイル
   * @param   {nsILoadContext} privacyContext
   *          プライバシーコンテキスト
   */
  saveURI: function(URI, cacheKey, referrer, postData, extraHeaders, file,
                    privacyContext) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /**
   * プライバシーを意識して URI を保存する
   *
   * @param   {nsIURI} URI
   *          対象の URI
   * @param   {nsISupports} cacheKey
   *          キャッシュオブジェクト
   * @param   {nsIURI} referrer
   *          リファラ
   * @param   {nsIInputStream} postData
   *          POST データ
   * @param   {string} extraHeaders
   *          拡張ヘッダ
   * @param   {nsIFile} file
   *          対象のファイル
   * @param   {boolean} isPrivate
   *          プライベートモードか
   */
  savePrivacyAwareURI: function(URI, cacheKey, referrer, postData,
                                extraHeaders, file, isPrivate) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },
  
  /* ==== private ==== */
  
  _saveState: 0,             /** 保存状態
                              *  @type {number} */
  
  _useAdditionalURIs: false, /** 追加ファイルを使用するか
                              *  @type {boolean} */
  
  /**
   * 保存完了イベント
   */
  _onComplete: function() {
    if (this.currentState == Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED) {
      return;
    }
    
    this.currentState
      = Ci.nsIWebBrowserPersist.PERSIST_STATE_FINISHED;
    
    if (this.progressListener) {
      this.progressListener.onStateChange(
        null, null,
        Ci.nsIWebProgressListener.STATE_STOP
          | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
        this.result);
    }
  }
};

/**
 * MAF 組み込み
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 */
function UnMHTMAFIntegration(window) {
  arWindowHook.call(this, window);
  
  this._behaviors = [];
}
UnMHTMAFIntegration.prototype = {
  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._behaviors = null;
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    if (!("InternalSaveBehavior" in this._window &&
          "gInternalSaveBehaviors" in this._window)) {
      return;
    }
    
    let stateList = UnMHTSave.getSaveStateList(UnMHTSave.SAVESTATE_MASK_ALL);
    
    let types
      = [...stateList.map(function(state) {
        return {
          label:             UnMHTSave.makeFiletype(state, false),
          ext:               "*.mht",
          saveState:         state,
          useAdditionalURIs: false
        };
      }),
         ...stateList.map(function(state) {
           return {
             label:             UnMHTSave.makeFiletype(state, true),
             ext:               "*.mht",
             saveState:         state,
             useAdditionalURIs: true
           };
         })];
    
    this._behaviors
      = types.map(function(type) {
        let newSaveBehavior = new this._window.InternalSaveBehavior();
        
        newSaveBehavior.isComplete = true;
        newSaveBehavior.mandatoryExtension = true;
        newSaveBehavior.isValidForSaveMode = function(saveMode) {
          return saveMode & 0x01;
        };
        
        newSaveBehavior.getFileFilter
          = function(contentType, fileExtension) {
            let result = {
              title: type.label,
              extensionstring: type.ext
            };
            return result;
          };
        newSaveBehavior.getPersistObject
          = function(saveBrowsers) {
            let result = new UnMHTMAFSavePersist(type.saveState, type.useAdditionalURIs);
            
            return result;
          };
        
        /* 末尾は *.* なのでその前に挿入する */
        this._window.gInternalSaveBehaviors
          .splice(this._window.gInternalSaveBehaviors.length - 1, 0,
                  newSaveBehavior);
        
        return newSaveBehavior;
      }.bind(this));
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    if (!("InternalSaveBehavior" in this._window &&
          "gInternalSaveBehaviors" in this._window)) {
      return;
    }

    this._window.gInternalSaveBehaviors
      = this._window.gInternalSaveBehaviors
      .filter((function(b) this._behaviors.indexOf(b) == -1).bind(this));
    
    this._behaviors = [];
  },
  
  /* ==== private ==== */
  
  _behaviors: null /** 追加した組み込み
                    *  @type {Array.<InternalSaveBehavior>} */
};
arPrototype.extend(UnMHTMAFIntegration, arWindowHook);
