"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTMIMEParam", function() {
  XPCOMUtils.importRelative(this, "UnMHTMIMEParam.jsm");
  return UnMHTMIMEParam;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTMIMEPart", function() {
  XPCOMUtils.importRelative(this, "UnMHTMIMEPart.jsm");
  return UnMHTMIMEPart;
});

let EXPORTED_SYMBOLS = ["UnMHTMIMEDecoder"];

/**
 * MIME デコーダ
 */
let UnMHTMIMEDecoder = {
  /**
   * quoted-printable[RFC2045] をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードした文字列
   */
  decodeQ: function(text) {
    return text
      .replace(/=(?:\r?\n|\r|([A-Fa-f0-9]{2}))/g,
               function(matched, octet) {
                 return (octet ? String.fromCharCode(parseInt(octet, 16)) : "");
               });
  },
  
  /**
   * encoded-word をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードした文字列 (Unicode)
   */
  decodeWord: function(text) {
    return text
      .replace(/=\?([^\?]+)\?([^\?])\?([^\?]+)\?=/g,
               function(matched, charset, type, v) {
                 if (type == "Q" || type == "q") {
                   v = this.decodeQ(v);
                 } else if (type == "B" || type == "b") {
                   v = this.decodeBase64(v);
                 }
                 
                 return arUconv.toUnicode(v, charset);
               }.bind(this));
  },
  
  /**
   * Base64 デコードする
   * 改行を含むものや破損したものにも対応
   *
   * @param   {string} text
   *          デコードする文字列
   * @returns {string}
   *          デコードされた文字列
   */
  decodeBase64: function(text) {
    try {
      /* まずは生のデータで試行 */
      return atob(text);
    } catch (e) {
      try {
        /* 不要なデータを削除する */
        text = text.replace(/[^A-Za-z0-9+\/=]+/g, "");
        return atob(text.substr(0, text.length - text.length % 4));
      } catch (e) {
        /* 途中にパディングがある等、壊れているデータ */
        return "";
      }
    }
  },
  
  /**
   * 日時を date-time[RFC5322] 形式等からデコードする
   * 実際は JavaScript の Date まかせなので色々デコードできる
   *
   * @param   {string} text
   *          エンコードされた日時
   * @returns {Date}
   *          デコードされた日時
   *          失敗した場合は null
   */
  decodeDate: function(text) {
    let date;
    try {
      date = new Date(text);
    } catch (e) {
      return null;
    }
    if (isNaN(date.getFullYear())) {
      return null;
    }
    return date;
  },
  
  /**
   * message [RFC5322] をデコードする
   *
   * @param   {string} text
   *          メッセージ
   *          改行コードは CR LF でなければならない
   * @param   {boolean} allowRemain
   *          解析できないフィールドの存在を許可するか
   * @returns {UnMHTMIMEPart}
   *          トップレベルのパート
   *          データが不正ならば null
   */
  decodeMessage: function(text, allowRemain) {
    let fields = "";
    let body = "";
    
    let pos = text.indexOf("\r\n\r\n");
    if (pos != -1) {
      fields = text.substr(0, pos + 2);
      body = text.substr(pos + 4);
    } else {
      fields = text;
    }
    
    let part = this.decodeFields(fields, allowRemain);
    if (!part) {
      return null;
    }

    if (part.isMultipart) {
      part.parts = this._decodeMultipart(body, part.boundary);
    } else {
      if (part.contentTransferEncoding == "quoted-printable") {
        part.body = this.decodeQ(body);
      } else if (part.contentTransferEncoding == "base64") {
        part.body = this.decodeBase64(body);
      } else if (part.format == "flowed") {
        part.body = this.decodeFlowed(body, part.delsp);
      } else {
        part.body = body;
      }
    }
    
    return part;
  },
  
  /**
   * fields[RFC5322] をデコードする
   *
   * @param   {string} fieldsString
   *          フィールド文字列
   *          改行コードは CR LF でなければならない
   * @param   {boolean} allowRemain
   *          解析できないフィールドの存在を許可するか
   * @returns {UnMHTMIMEPart}
   *          パート
   *          データが不正ならば null
   */
  decodeFields: function(fieldsString, allowRemain) {
    let part = new UnMHTMIMEPart();
    
    let decodeFunc = function(text) {
      return text
        .replace(/([\x21-\x39\x3B-\x7E]+)[ \t]*:[ \t]*([^\r\n]*)\r\n((?:[\t ]+[^\r\n]*\r\n)*)/g,
                 function(matched, name, value, restLines) {
                   restLines
                     .replace(/[\t ]*([^\r\n]*)\r\n/g,
                              function(matched, v) {
                                value += v;
                                
                                return "";
                              });
                   
                   part.addField(name, value);
                   
                   return "";
                 });
    };
    
    let remain = decodeFunc(fieldsString);
    if (remain) {
      let re = remain.match(/^(From)[ \t]+(.*)$/);
      if (re) {
        /* RFC4155: application/mbox
         * From の行の timestamp をフィールド扱いしている可能性があるので
         * 改めて解析する */
        part = new UnMHTMIMEPart();
        remain
          = decodeFunc(fieldsString.replace(/^From[^\r\n]+\r\n/, ""));
      }
    }
    if (!allowRemain && remain) {
      return null;
    }
    
    if (part.fields.has("content-type")) {
      let value = part.fields.get("content-type");
      part.contentTypeMap = this.decodeParam(value);
      part.contentType    = part.contentTypeMap.type.toLowerCase();
      part.contentSubType = part.contentTypeMap.subtype.toLowerCase();
      part.mimetype       = part.contentType + "/" + part.contentSubType;
      if (part.contentTypeMap.params.has("charset")) {
        part.charset = part.contentTypeMap.params.get("charset");
      }
      if (part.contentTypeMap.params.has("format")) {
        part.format = part.contentTypeMap.params.get("format");
      }
      if (part.contentTypeMap.params.has("delsp")) {
        part.delsp = part.contentTypeMap.params.get("delsp") == "yes";
      }
    }
    
    if (part.fields.has("content-location")) {
      part.contentLocation
        = this._tryConvertToUnicode(part.fields.get("content-location"));
    }
    
    if (part.fields.has("content-disposition")) {
      let value = part.fields.get("content-disposition");
      part.contentDispositionMap = this.decodeParam(value);
      part.contentDispositionType
        = part.contentDispositionMap.type.toLowerCase();
      if (part.contentDispositionMap.params.has("filename")) {
        part.contentDispositionFilename
          = this._tryConvertToUnicode(part.contentDispositionMap.params.get("filename"));
      }
    }
    
    if (part.fields.has("content-id")) {
      let value = part.fields.get("content-id");
      part.contentID = value.replace(/^<([^>]+)>$/, "$1");
    }
    
    if (part.contentType == "multipart") {
      part.isMultipart = true;
      if (part.contentTypeMap.params.has("boundary")) {
        part.boundary = part.contentTypeMap.params.get("boundary");
      } else {
        return null;
      }
      
      if (part.contentTypeMap.params.has("start")) {
        let value = part.contentTypeMap.params.get("start");
        part.start = value.replace(/^<([^>]+)>$/, "$1");
      }
      
      if (part.contentTypeMap.params.has("type")) {
        part.startMimetype = part.contentTypeMap.params.get("type");
      }
    } else {
      if (part.fields.has("content-transfer-encoding")) {
        let value = part.fields.get("content-transfer-encoding");
        part.contentTransferEncoding = value.toLowerCase();
      }
    }
    
    return part;
  },
  
  /**
   * content[RFC2045] や disposition[RFC2183] の値を
   * parameter[RFC2231] に従ってデコードする
   * 
   * ただし disposition-type[RFC2183] や type/subtype[RFC2045] の位置が
   * 先頭でない事例が存在するのでこれらのデコードも同時に行う
   *
   * @param   {string} value
   *          contentRFC2045] や disposition[RFC2183] の値
   * @returns {UnMHTMIMEParam}
   *          パラメータ
   */
  decodeParam: function(value) {
    let tmpParams = new Map();
    let params = new Map();
    let type = "";
    let subtype = "";
    
    value
      .replace(this._parameterRe,
               function(matched, name, section, extended,
                        charset, lang, evalue,
                        dqvalue, value,
                        ctype, csubtype, dtype) {
                 if (dtype && !type) {
                   type = dtype;
                 } else if (ctype && !type) {
                   type = ctype;
                   subtype = csubtype;
                 } else if (name) {
                   if (!value && dqvalue) {
                     value = dqvalue.replace(/\\\"/g, "\"");
                   }
                   
                   let param = null;
                   if (tmpParams.has(name)) {
                     param = tmpParams.get(name);
                   } else {
                     param = {
                       values:  new Map(),
                       charset: "",
                       lang:    ""
                     };
                     tmpParams.set(name, param);
                   }
                   
                   if (!section) {
                     section = "0";
                   }
                   
                   if (extended && evalue) {
                     param.values.set(section, { v: evalue, e: true });
                     param.charset = charset;
                     param.lang = lang;
                   } else {
                     /* 実際にはここに encoded-word は使えないが使用事例がある */
                     value = this.decodeWord(value);
                     
                     param.values.set(section, { v: value, e: extended != "" });
                   }
                 }
               }.bind(this));
    
    for (let [name, param] of tmpParams) {
      let value
        = [x for (x of param.values)]
        .sort(function(a, b) parseInt(a[0], 10) - parseInt(b[0], 10))
        .map(function([section, v]) {
          if (v.e) {
            if (param.charset) {
              return arUconv.toUnicode(unescape(v.v), param.charset);
            } else {
              return unescape(v.v);
            }
          } else {
            return v.v;
          }
        })
        .join("");
      params.set(name, value);
    }
    
    let param = new UnMHTMIMEParam();
    param.type = type;
    param.subtype = subtype;
    param.params = params;
    
    return param;
  },
  
  /**
   * format=flowed をデコードする
   *
   * @param   {string} text
   *          デコードする文字列
   * @param   {boolean} delsp
   *          delsp の値
   * @returns {string}
   *          デコードされた文字列
   */
  decodeFlowed: function(text, delsp) {
    let currentFlow = false;
    return text.split(/\r\n/)
      .map(function(line) {
        let re = line.match(/^(>+)?(?:(.*?)|([^ ].*?))( )?$/);
        if (re) {
          let quote = re[1] || "";
          let stuffed = re[2];
          let unstuffed = re[3];
          let flow = re[4];
          if (unstuffed == "--" && flow) {
            /* sig-sep[RFC3676] */
            flow = "";
          }
          if (flow) {
            if (currentFlow) {
              return (stuffed || unstuffed) + (delsp ? "" : flow);
            } else {
              currentFlow = true;
              return quote + (stuffed || unstuffed) + (delsp ? "" : flow);
            }
          } else {
            if (currentFlow) {
              currentFlow = false;
              return (stuffed || unstuffed) + "\r\n";
            } else {
              return quote + (stuffed || unstuffed) + "\r\n";
            }
          }
        } else {
          return line + "\r\n";
        }
      }).join("");
  },
  
  /* ==== private ==== */
  
  /**
   * multipart-body[RFC2046] をデコードする
   *
   * @param   {string} body
   *          multipart なメッセージボディ
   *          改行コードは CR LF でなければならない
   * @param   {string} boundary
   *          バウンダリ文字列
   * @returns {Array.<UnMHTMIMEPart>}
   *          パートの配列
   */
  _decodeMultipart: function(body, boundary) {
    let beginPos = 0;
    let lastPos = 0;
    let endPos = 0;
    
    let dash_boundary = "--" + boundary;
    let dash_boundary_len = dash_boundary.length;
    
    /* skip preamble and first dash-boundary */
    
    endPos = body.indexOf(dash_boundary, lastPos);
    let p = endPos + dash_boundary_len;
    if (endPos == -1) {
      /* not a multipart */
      return [];
    }
    /* check transport-padding */
    while (/[ \t]/.test(body[p])) {
      p ++;
    }
    
    /* check CRLF */
    if (body.substr(p, 2) == "\r\n") {
      beginPos = p + 2;
      lastPos = beginPos;
    } else {
      /* not a multipart */
      return [];
    }
    
    let datas = [];
    let isClose = false;
    
    let delimiter = "\r\n--" + boundary;
    let delimiter_len = delimiter.length;
    
    /* parse each body-part */
    for (;;) {
      endPos = body.indexOf(delimiter, lastPos);
      if (endPos == -1) {
        break;
      }
      
      let p = endPos + delimiter_len;
      
      /* check close-delimiter */
      if (body.substr(p, 2) == "--") {
        isClose = true;
        p += 2;
      }
      
      /* check transport-padding */
      while (let (c = body[p]) c == " " || c == "\t") {
        p ++;
      }
      
      /* check CRLF */
      if (body.substr(p, 2) == "\r\n") {
        datas.push(body.substring(beginPos, endPos));
        
        if (isClose) {
          break;
        } else {
          beginPos = p + 2;
          lastPos = beginPos;
        }
      } else {
        if (isClose) {
          /* close-delimiter does not deed CRLF */
          
          datas.push(body.substring(beginPos, endPos));
          break;
        } else {
          /*  delimiter needs CRLF */
          
          lastPos = endPos + 1;
        }
      }
    }
    
    return datas
      .map((function(data) this.decodeMessage(data)).bind(this))
      .filter(function(part) part != null);
  },
  
  /**
   * Unicode への変換を試みる
   *
   * @param   {string} value
   *          対象の文字列
   * @returns {string}
   *          成功したら変換した文字列
   *          失敗したら元の文字列
   */
  _tryConvertToUnicode: function(value) {
    try {
      return arUconv.toUnicode(value, "utf-8");
    } catch (e) {
    }
    try {
      return arUconv.toUnicode(value, "shift_jis");
    } catch (e) {
    }
    return value;
  },

  /*
   * ( 1*attribute-char [section] ["*"] "="
   *   ( ( [token] "'" [token] "'" 1*<attribute-char and %> ) /
   *     quoted-string / token ) ) /
   * ( token "/" token ) /
   * token
   */
  _parameterRe: /([^\t\r\n \*\'%\(\)<>@,;:\\\"\/\[\]\?=]+)(?:\*([0-9]+))?(\*)?=(?:([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]*)\'([^\t\r\n \*\'\(\)<>@,;:\\\"\/\[\]\?=]+)|\"((?:\\.|[^\"])*)\"|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+))|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)\/([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)|([^\t\r\n \(\)<>@,;:\\\"\/\[\]\?=]+)/g
};
