"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMIMEEncoder"];

/**
 * MIME エンコーダ
 */
let UnMHTMIMEEncoder = {
  /**
   * quoted-printable[RFC2045] にエンコードする
   *
   * @param   {string} text
   *          エンコードする文字列
   * @param   {boolean} wrap
   *          76 文字で折り返す
   * @returns {string}
   *          エンコードした文字列
   */
  encodeQ: function(text, wrap) {
    text
      = text
      .replace(/[^\t\r\n \x21-\x3C\x3E-\x7E]/g,
               function(matched) {
                 let hexString
                   = "0" + matched.charCodeAt(0).toString(16).toUpperCase();
                 return "=" + hexString.slice(-2);
               });
    
    if (!wrap) {
      return text;
    }
    
    return (text + "\r\n")
      .replace(/([^\r\n]{0,73})(?:(=..|[ \t]|[^\r\n=]{2}[^\t\r\n =])($|\r?\n|\r)|([^\t\r\n =]{0,2})(\r?\n?))/g,
               function(matched, aMain,
                        aLastThree, aLastThreeEOL,
                        aLastTwo, aLastTwoEOL) {
                 let line = aMain + aLastThree + aLastTwo;
                 if (aLastThreeEOL || aLastTwoEOL) {
                   if (line) {
                     let lastChar = line [line.length - 1];
                     if (lastChar === " " ||
                         lastChar === "\t") {
                       line
                         = line.slice(0, -1)
                         + (lastChar === " " ? "=20" : "=09");
                     }
                   }
                   return line + "\r\n";
                 }
                 return line ? (line + "=\r\n") : "";
               }).slice(0, -2);
  },
  
  /**
   * encoded-word にエンコードする
   *
   * @param   {string} text
   *          エンコードする文字列 (Unicode)
   * @returns {string}
   *          エンコードした文字列
   */
  encodeWord: function(charset, type, text) {
    if (text == "") {
      /* 空の文字列はエンコードできない */
      return "";
    }
    
    text = arUconv.fromUnicode(text, charset);
    
    if (type == "B") {
      text = this.encodeBase64(text, false);
    } else if (type == "Q") {
      text
        = this.encodeQ(text, false)
        .replace(/\?/g, "=3F")
        .replace(/ /g, "=20");
    }
    
    return ("=?" + charset + "?" + type + "?" + text + "?=");
  },
  
  /**
   * Base64 エンコードする
   *
   * @param   {string} text
   *          エンコードする文字列
   * @param   {boolean} wrap
   *          76 文字で折り返す
   * @returns {string}
   *          エンコードされた文字列
   */
  encodeBase64: function(text, wrap) {
    text = btoa(text);
    if (!wrap) {
      return text;
    }
    
    return text.replace(/.{76}/g, "$&\r\n");
  },

  /**
   * 日時を date-time[RFC5322] 形式にエンコードする
   *
   * @param   {Date} date
   *          対象の日時
   * @returns {string}
   *          エンコードされた日時
   */
  encodeDate: function(date) {
    let weekDay = [
      "Sun",
      "Mon",
      "Tue",
      "Wed",
      "Thu",
      "Fri",
      "Sat"
    ];
    let month = [
      "Jan",
      "Feb",
      "Mar",
      "Apr",
      "May",
      "Jun",
      "Jul",
      "Aug",
      "Sep",
      "Oct",
      "Nov",
      "Dec"
    ];
    
    let pad = function(x, n) {
      return ("0000" + n).slice(-x);
    };
    
    let sign = "-";
    let zone = date.getTimezoneOffset() * 100 / 60;
    if (zone < 0) {
      zone = -zone;
      sign = "+";
    }
    
    return (weekDay[date.getDay()] + ", "
            + pad(2, date.getDate()) + " "
            + month[date.getMonth()] + " "
            + date.getFullYear() + " "
            + pad(2, date.getHours()) + ":"
            + pad(2, date.getMinutes()) + ":"
            + pad(2, date.getSeconds()) + " "
            + sign + pad(4, zone));
  },
  
  /**
   * message [RFC5322] にエンコードする
   *
   * @param   {UnMHTMIMEPart} part
   *          トップレベルのパート
   * @param   {nsIOutputStream} stream
   *          メッセージ出力用ストリーム
   *          改行コードは CR LF である
   */
  encodeMessage: function(part, stream) {
    this.encodeFields(part, stream);
    
    let data = "\r\n";
    stream.write(data, data.length);

    if (part.isMultipart) {
      part.parts.forEach(function(p) {
        let data = "--" + part.boundary + "\r\n";
        stream.write(data, data.length);
        this.encodeMessage(p, stream);
        data = "\r\n";
        stream.write(data, data.length);
      }.bind(this));
      
      let data = "--" + part.boundary + "--\r\n";
      stream.write(data, data.length);
      
      return;
    }
    
    if (part.contentTransferEncoding == "quoted-printable") {
      data = this.encodeQ(part.body, true);
    } else if (part.contentTransferEncoding == "base64") {
      data = this.encodeBase64(part.body, true) + "\r\n";
    } else {
      data = part.body;
    }
    stream.write(data, data.length);
  },
  
  /**
   * fields[RFC5322] にエンコードする
   *
   * @param   {UnMHTMIMEPart} part
   *          トップレベルのパート
   * @param   {nsIOutputStream} stream
   *          メッセージ出力用ストリーム
   *          改行コードは CR LF である
   */
  encodeFields: function(part, stream) {
    let order = [
      "from",
      "to",
      "subject",
      "date",
      "mime-version",
      "content-location",
      "content-base",
      "content-id",
      "content-transfer-encoding"
    ];
    let PLAIN        = 1;
    let ENCODED_WORD = 2;
    let UNICODE_EW   = 3;
    let types = {
      "from":                      PLAIN,
      "to":                        PLAIN,
      "subject":                   ENCODED_WORD,
      "date":                      PLAIN,
      "mime-version":              PLAIN,
      "content-transfer-encoding": PLAIN,
      "content-location":          UNICODE_EW,
      "content-base":              UNICODE_EW,
      "content-id":                PLAIN
    };
    
    part.fieldList
      .filter(function([name,])
              !(name.toLowerCase() == "content-type" ||
                name.toLowerCase() == "content-disposition"))
      .sort(function([name1,], [name2,]) {
        let p1 = order.indexOf(name1.toLowerCase());
        let p2 = order.indexOf(name2.toLowerCase());
        if (p1 >= 0) {
          if (p2 >= 0) {
            return p1 - p2;
          } else {
            return -1;
          }
        } else {
          if (p2 >= 0) {
            return 1;
          } else {
            return ((name1 < name2) ? -1
                    : (name1 > name2) ? 1
                    : 0);
          }
        }
      })
      .map(function([name, value]) {
        let type = PLAIN;
        let lname = name.toLowerCase();
        if (lname in types) {
          type = types[lname];
        };
        
        if (type == UNICODE_EW) {
          try {
            value = arUconv.fromUnicode(value, "utf-8");
          } catch (e) {
          }
          if (!/^[\x20-\x7E]*$/.test(value)) {
            value = this.encodeWord("utf-8", "B", value);
          } else if (!/^[\x21-\x7E]*$/.test(value)) {
            value = this.encodeWord("us-ascii", "B", value);
          }
        } else if (type == ENCODED_WORD) {
          if (!/^[\x20-\x7E]*$/.test(value)) {
            value = this.encodeWord("utf-8", "B", value);
          }
        }
        
        return name + ": " + value + "\r\n";
      }.bind(this))
      .forEach(function(data) stream.write(data, data.length));
    
    if (part.contentTypeMap) {
      let name = "Content-Type";
      let value = this.encodeParam(part.contentTypeMap, true, true);
      let data = name + ": " + value + "\r\n";
      stream.write(data, data.length);
    }
    if (part.contentDispositionMap) {
      let name = "Content-Disposition";
      let value = this.encodeParam(part.contentDispositionMap, true, true);
      let data = name + ": " + value + "\r\n";
      stream.write(data, data.length);
    }
  },

  /**
   * UnMHTMIMEParam を文字列にエンコードする
   *
   * @param   {UnMHTMIMEParam} param
   *          パラメータ
   * @param   {boolean} wrap
   *          パラメータごとに改行するか
   * @param   {boolean} ignoreFormatDelSP
   *          format, delsp を無視するか
   * @returns {string}
   *          文字列表現
   */
  encodeParam: function(param, wrap, ignoreFormatDelSP) {
    let value = "";
    if (param.subtype) {
      value = param.type + "/" + param.subtype;
    } else {
      value = param.type;
    }
    
    let encodeSingleParam = function(pName, pValue) {
      if (/^[\x21-\x26\x28-\x7E]*$/.test(pValue)) {
        return pName + "=\"" + pValue.replace(/\"/g, "\\\"") + "\"";
      } else {
        return pName + "*=utf-8\'\'"
          + encodeURIComponent(pValue).replace(/\'/g, "%27");
      }
    };
    
    return [value,
            ...[encodeSingleParam(pName, pValue)
                for ([pName, pValue] of param.params)
                  if (!(ignoreFormatDelSP &&
                        (pName == "format" ||
                         pName == "delsp")))]]
      .join(wrap ? ";\r\n\t": ";");
  }
};
