"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMIMEPart"];

/**
 * MIME の各パート情報
 */
function UnMHTMIMEPart() {
  this.fieldList = [];
  this.fields = new Map();
}
UnMHTMIMEPart.prototype = {
  fieldList: null, /** フィールドの配列
                    *  @type {Array.<[string,string]>}
                    *  **encode 時に必要**
                    *    [[string フィールド名, string フィールド値], ...] */
  fields: null,    /** フィールドのマップ
                    *  @type {Map.<string,string>}
                    *  解析用 (フィールド名は小文字)
                    *  **encode 時に必要**
                    *  **(addField か createFieldsFromList で作る)**
                    *    <string フィールド名, string フィールド値> */
  
  /* Content-Type: (content[RFC2045]) */
  contentTypeMap: null, /** 解析結果
                         *  @type {UnMHTMIMEParam}
                         *  **encode 時に必要** */
  contentType: "",      /** type[RFC2045]
                         *  @type {string} */
  contentSubType: "",   /** subtype[RFC2045]
                         *  @type {string} */
  mimetype: "",         /** MIME-Type
                         *  @type {string} */
  charset: "",          /** charset[???]
                         *  @type {string} */
  format: "",           /** Format Parameter[RFC3676]
                         *  @type {string} */
  delsp: false,         /** DelSp Parameter[RFC3676]
                         *  @type {string} */
  
  isMultipart: false,   /** マルチパートか
                         *  @type {boolean}
                         *  **encode 時に必要** */
  boundary: "",         /** boundary[RFC2046]
                         *  @type {string}
                         *  **encode 時に必要** */
  start: "",            /** "start"[RFC2387]
                         *  @type {string} */
  startMimetype: "",    /** "type"[RFC2387]
                         *  @type {string} */
  
  /* Content-Location: (content-location[RFC2557]) */
  contentLocation: "", /** URI[RFC2557]
                        *  @type {string} */
  
  /* Content-Disposition: (disposition[RFC2138]) */
  contentDispositionMap: null,     /** 解析結果
                                    *  @type {UnMHTMIMEParam}
                                    *  **encode 時に必要** */
  contentDispositionType: "",      /** disposition-type[RFC2138]
                                    *  @type {string} */
  contentDispositionFilename: "",  /** filename-parm[RFC2138]
                                    *  @type {string} */
  
  /* Content-ID: (id[RFC2045]) */
  contentID: "", /** msg-id[RFC2045]
                  *  @type {string} */
  
  /* Content-Transfer-Encoding: (encoding[RFC2045]) */
  contentTransferEncoding: "", /** mechanism[RFC2045]
                                *  @type {string}
                                *  **encode 時に必要** */
  
  body: "", /** body[RFC5322]
             *  @type {string}
             *  **encode 時に必要** */
  
  parts: null, /** マルチパート
                *  @type {Array.<UnMHTMIMEPart>}
                *  **encode 時に必要**
                *    [UnMHTMIMEPart body-part[RFC2046], ...] */
  
  eParam: null, /** 展開で使用するパラメータ
                 *  @type {UnMHTExtractParam} */
  sParam: null, /** 保存で使用するパラメータ
                 *  @type {UnMHTSaveParam} */

  /**
   * 開始パートを探す
   *
   * @returns {UnMHTMIMEPart}
   *          開始パート
   *          適切なパートが無ければ null
   */
  findStartPart: function() {
    if (this.isMultipart) {
      if (this.contentSubType == "mixed") {
        return this;
      }
      
      if (this.start) {
        let startContentID = this.start;
        let start
          = arArrayUtil
          .find(this.parts,
                function(p) p.contentID == startContentID);
        if (start) {
          return start.findStartPart();
        }
      }
      
      if (this.contentSubType == "related") {
        let startMimetype = this.startMimetype;
        let start
          = arArrayUtil
          .find(this.parts,
                function(p) p.mimetype == startMimetype);
        if (start) {
          return start.findStartPart();
        }
      }
      
      if (this.contentSubType == "alternative") {
        let start
          = arArrayUtil
          .find(this.parts,
                function(p) p.mimetype == "text/html");
        if (start) {
          return start;
        }
        
        start
          = arArrayUtil
          .find(this.parts,
                function(p) {
                  if (p.isMultipart) {
                    let startPart = p.findStartPart();
                    if (startPart && startPart.mimetype == "text/html") {
                      return true;
                    }
                  }
                  return false;
                });
        if (start) {
          return start.findStartPart();
        }
      }
      
      if (this.parts.length) {
        return this.parts[0].findStartPart();
      }
      
      return null;
    }
    
    return this;
  },
  
  /**
   * 複製する
   *
   * @param   {object} findPart
   *          検索するパート
   *            {
   *              from: {UnMHTMIMEPart}
   *                    検索するパート
   *              to:   {UnMHTMIMEPart}
   *                    (出力) from に対応する複製先のパート
   *            }
   * @returns {UnMHTMIMEPart}
   *          複製
   */
  clone: function(findPart) {
    if (findPart && this == findPart.from) {
      findPart.to = newPart;
    }
    
    let newPart = new UnMHTMIMEPart();
    
    newPart.fieldList
      = this.fieldList.map(function([name, value]) [name, value]);
    newPart.createFieldsFromList();
    if (this.contentTypeMap) {
      newPart.contentTypeMap = this.contentTypeMap.clone();
    }
    newPart.contentType = this.contentType;
    newPart.contentSubType = this.contentSubType;
    newPart.mimetype = this.mimetype;
    newPart.charset = this.charset;
    newPart.format = this.format;
    newPart.delsp = this.delsp;
    newPart.isMultipart = this.isMultipart;
    newPart.boundary = this.boundary;
    newPart.start = this.start;
    newPart.startMimetype = this.startMimetype;
    newPart.contentLocation = this.contentLocation;
    if (this.contentDispositionMap) {
      newPart.contentDispositionMap = this.contentDispositionMap.clone();
    }
    newPart.contentDispositionType = this.contentDispositionType;
    newPart.contentDispositionFilename = this.contentDispositionFilename;
    newPart.contentID = this.contentID;
    newPart.contentTransferEncoding = this.contentTransferEncoding;
    newPart.body = this.body;
    if (this.parts) {
      newPart.parts = this.parts.map(function(p) p.clone(findPart));
    }
    
    return newPart;
  },
  
  /**
   * 破棄
   */
  destruct: function() {
    if (this.parts) {
      this.parts.forEach(function(p) {
        p.destruct();
      }.bind(this));
      this.parts = null;
    }
    this.fieldList = null;
    this.fields = null;
    if (this.contentTypeMap) {
      this.contentTypeMap.destruct();
      this.contentTypeMap = null;
    }
    if (this.contentDispositionMap) {
      this.contentDispositionMap.destruct();
      this.contentDispositionMap = null;
    }
    this.body = "";
  },
  
  /**
   * フィールドを追加する
   *
   * @param   {string} name
   *          フィールド名
   * @param   {string} value
   *          フィールド値
   */
  addField: function(name, value) {
    this.fieldList.push([name, value]);
    this.fields.set(name.toLowerCase(), value);
  },
  
  /**
   * Content-Type を設定する
   *
   * @param   {UnMHTMIMEParam} param
   *          Content-Type
   */
  setContentType: function(param) {
    this.contentTypeMap = param;
  },
  
  /**
   * fieldList から fields を作成する
   */
  createFieldsFromList: function() {
    this.fields = new Map(this.fieldList.map(function([name, value])
                                             [name.toLowerCase(), value]));
  }
};
