"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arAppInfo.jsm");
XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTSystemDirectory", function() {
  XPCOMUtils.importRelative(this, "UnMHTSystemDirectory.jsm");
  return UnMHTSystemDirectory;
});
XPCOMUtils.defineLazyGetter(this, "arWinRegUtil", function() {
  XPCOMUtils.importRelative(this, "base/arWinRegUtil.jsm");
  return arWinRegUtil;
});

let EXPORTED_SYMBOLS = ["UnMHTMail"];

/**
 * メール送信
 */
let UnMHTMail = {
  /** アプリケーションの引数
   *  @type {string} */
  thunderbirdArg: UnMHTPref.thunderbirdArg,
  seamonkeyArg:   UnMHTPref.seamonkeyArg,
  macmailArg:     UnMHTPref.macmailArg,
  
  /**
   * メールを送信する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @param   {nsIFile} targetFile
   *          添付ファイル
   * @param   {string} title
   *          ページのタイトル
   * @returns {boolean}
   *          成功したか
   */
  send: function(window, targetFile, title) {
    if (UnMHTPref.getPref("save.mail.zip")) {
      targetFile = this._createZipFile(targetFile);
    }
    
    if (UnMHTPref.getPref("save.mail.app.type") == "seamonkey" &&
        arAppInfo.isSeaMonkey) {
      return this._openComposeWindow(targetFile, title);
    }
    
    let [app, arg] = this._getApplicationAndArgument();
    arg = this._replaceVars(arg, targetFile, title);
    let args = this.argToArray(arg);
    
    if (UnMHTPref.getPref("save.mail.debug")) {
      window.alert("[Debug]\n"
                   + "Command line for mail\n"
                   + "\n"
                   + (app ? app.path : "???") + " " + arg + "\n"
                   + "\n"
                   + "Arguments will be passed as following array\n"
                   + "\n"
                   + args.join("\n"));
    }
    
    try {
      let process
        = Cc["@mozilla.org/process/util;1"].createInstance(Ci.nsIProcess);
      process.init(app);
      process.runw(false, args, args.length);
    } catch (e) {
      return false;
    }
    
    return true;
  },
  
  /* ==== private ==== */
  
  /**
   * スペース区切りの引数を引数配列に変換する
   *
   * @param   {string} argString
   *          スペース区切りの引数
   * @returns {Array.<string>}
   *          引数配列
   */
  argToArray: function(argString) {
    let STATE_NORMAL          = 0;
    let STATE_ESCAPED         = 1;
    let STATE_DQUOTED         = 10;
    let STATE_DQUOTED_ESCAPED = 11;
    let STATE_QUOTED          = 20;
    let STATE_QUOTED_ESCAPED  = 21;
    
    let args = [];
    let currentArg = "";
    let state = STATE_NORMAL;
    
    Array
      .forEach(argString,
               function(c) {
                 switch (state) {
                   case STATE_NORMAL: {
                     if (c == "\\") {
                       state = STATE_ESCAPED;
                     } else if (currentArg == "" && c == "\"") {
                       state = STATE_DQUOTED;
                     } else if (currentArg == "" && c == "\'") {
                       state = STATE_QUOTED;
                     } else if (c == " ") {
                       if (currentArg != "") {
                         args.push(currentArg);
                       }
                       currentArg = "";
                     } else {
                       currentArg += c;
                     }
                     break;
                   }
                   case STATE_ESCAPED: {
                     currentArg += c;
                     state = STATE_NORMAL;
                     break;
                   }
                   case STATE_DQUOTED: {
                     if (c == "\"") {
                       state = STATE_NORMAL;
                     } else if (c == "\\") {
                       state = STATE_DQUOTED_ESCAPED;
                     } else {
                       currentArg += c;
                     }
                     break;
                   }
                   case STATE_DQUOTED_ESCAPED: {
                     currentArg += c;
                     state = STATE_DQUOTED;
                     break;
                   }
                   case STATE_QUOTED: {
                     if (c == "\'") {
                       state = STATE_NORMAL;
                     } else if (c == "\\") {
                       state = STATE_QUOTED_ESCAPED;
                     } else {
                       currentArg += c;
                     }
                     break;
                   }
                   case STATE_QUOTED_ESCAPED: {
                     currentArg += c;
                     state = STATE_QUOTED;
                     break;
                   }
                 }
               });
    if (currentArg != "") {
      args.push(currentArg);
    }
    
    return args;
  },
  
  /**
   * SeaMonkey でメール作成画面を開く
   *
   * @param   {nsIFile} targetFile
   *          添付ファイル
   * @param   {string} title
   *          ページのタイトル
   * @returns {boolean}
   *          成功したか
   */
  _openComposeWindow: function(targetFile, title) {
    let params
      = Cc["@mozilla.org/messengercompose/composeparams;1"]
      .createInstance(Ci.nsIMsgComposeParams);
    
    params.composeFields
      = Cc["@mozilla.org/messengercompose/composefields;1"]
      .createInstance(Ci.nsIMsgCompFields);
    params.composeFields.subject = title;
    
    let attachmentData
      = Cc["@mozilla.org/messengercompose/attachment;1"]
      .createInstance(Ci.nsIMsgAttachment);
    attachmentData.url = arFileUtil.getURLSpecFromFile(targetFile);
    
    params.composeFields.addAttachment(attachmentData);
    
    let composeService
      = Cc["@mozilla.org/messengercompose;1"]
      .getService(Ci.nsIMsgComposeService);
    
    try {
      params.identity = composeService.defaultIdentity;
    } catch (e) {
      params.identity = null;
    }
    
    composeService.OpenComposeWindowWithParams(null, params);
    
    return true;
  },
  
  /**
   * 使用するアプリケーションと引数を取得する
   *
   * @returns {Array}
   *            [nsIFile アプリケーション
   *                     見付からない場合は null,
   *             string 引数]
   */
  _getApplicationAndArgument: function() {
    let arg;
    switch (UnMHTPref.getPref("save.mail.arg.type")) {
      case "thunderbird": {
        arg = this.thunderbirdArg;
        break;
      }
      case "seamonkey": {
        arg = this.seamonkeyArg;
        break;
      }
      case "macmail": {
        arg = this.macmailArg;
        break;
      }
      default: {
        arg = UnMHTPref.getPref("save.mail.arg.other");
      }
    }
    
    let app;
    switch (UnMHTPref.getPref("save.mail.app.type")) {
      case "thunderbird": {
        app = this._getThunderbird();
        break;
      }
      case "seamonkey": {
        app = this._getSeamonkey();
        break;
      }
      case "macmail": {
        app = this._generateMacMailScript();
        break;
      }
      default: {
        let path = UnMHTPref.getPref("save.mail.app");
        if (path) {
          app = arFileUtil.getFile(path);
        } else {
          app = this._getThunderbird();
          if (app) {
            arg = this.thunderbirdArg;
          }
        }
      }
    }
    
    return [app, arg];
  },
  
  /**
   * 引数中の変数を置換する
   *
   * @param   {string} arg
   *          引数
   * @param   {nsIFile} targetFile
   *          添付ファイル
   * @param   {string} title
   *          ページのタイトル
   * @returns {string}
   *          置換した引数
   */
  _replaceVars: function(arg, targetFile, title) {
    let targetPath = targetFile.path;
    
    let targetURI
      = arFileUtil.getURLSpecFromFile(targetFile)
      .replace(/,/g, "%2C");
    
    let vars = {
      "FILE":       targetPath.replace(/\\/g, "\\\\"),
      "URI":        targetURI.replace(/\\/g, "\\\\"),
      "TITLE":      (title
                     .replace(/\\/g, "\\\\")
                     .replace(/[\'\"]/g, "")),
      "TITLE_NOSP": (title
                     .replace(/\\/g, "\\\\")
                     .replace(/[\'\"]/g, "")
                     .replace(/[ \t\r\n]/g, "_"))
    };
    
    let rexp = new RegExp("%(" + Object.keys(vars).join("|") + ")%", "g");
    return arg.replace(rexp, function(matched, name) vars[name]);
  },
  
  /**
   * Thunderbird の実行ファイルを探す
   *
   * @returns {nsIFile}
   *          Thunderbird の実行ファイル
   *          見付からない場合は null
   */
  _getThunderbird: function() {
    let list;
    switch (Services.appinfo.OS) {
      case "Darwin": {
        list = [
          [arFileUtil.applicationsDirectory.clone(),
           "Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"],
          [arFileUtil.userApplicationsDirectory.clone(),
           "Thunderbird.app", "Contents", "MacOS", "thunderbird-bin"]
        ];
        break;
      }
      case "WINNT": {
        let app = null;
        
        let currentVersion
          = arWinRegUtil.getHKLMValue(["SOFTWARE", "Mozilla", "Mozilla Thunderbird"], "CurrentVersion");
        if (currentVersion) {
          let PathToExe
            = arWinRegUtil.getHKLMValue(["SOFTWARE", "Mozilla", "Mozilla Thunderbird", currentVersion, "Main"], "PathToExe");
          if (PathToExe) {
            app = arFileUtil.getFile(PathToExe);
          }
        }
        
        list = [
          [app],
          [arFileUtil.programFilesDirectory.clone(),
           "Mozilla Thunderbird", "thunderbird.exe"],
          [arFileUtil.programFilesDirectory.clone(),
           "thunderbird", "thunderbird.exe"]
        ];
        break;
      }
      default: {
        list = [];
      }
    }
    
    return this._getApplication(list);
  },

  /**
   * SeaMonkey の実行ファイルを探す
   *
   * @returns {nsIFile}
   *          SeaMonkey の実行ファイル
   *          見付からない場合は null
   */
  _getSeamonkey: function() {
    let list;
    switch (Services.appinfo.OS) {
      case "Darwin": {
        list = [
          [arFileUtil.applicationsDirectory.clone(),
           "SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"],
          [arFileUtil.userApplicationsDirectory.clone(),
           "SeaMonkey.app", "Contents", "MacOS", "seamonkey-bin"]
        ];
        break;
      }
      case "WINNT": {
        let app = null;
        
        let currentVersion
          = arWinRegUtil.getHKLMValue(["SOFTWARE", "Mozilla", "SeaMonkey"], "CurrentVersion");
        if (currentVersion) {
          let PathToExe
            = arWinRegUtil.getHKLMValue(["SOFTWARE", "Mozilla", "SeaMonkey", currentVersion, "Main"], "PathToExe");
          if (PathToExe) {
            app = arFileUtil.getFile(PathToExe);
          }
        }
        
        list = [
          [app],
          [arFileUtil.programFilesDirectory.clone(),
           "mozilla.org", "SeaMonkey", "seamonkey.exe"],
          [arFileUtil.programFilesDirectory.clone(),
           "SeaMonkey", "seamonkey.exe"]
        ];
        break;
      }
      default: {
        list = [];
      }
    }
    
    return this._getApplication(list);
  },
  
  /**
   * 存在するアプリケーションを返す
   *
   * @param   {Array} list
   *          アプリケーションのパスの候補
   *            [[nsIFile 親ディレクトリ, string サブディレクトリ, ...], ...]
   * @returns {nsIFile}
   *          アプリケーションのファイル
   *          見付からない場合は null
   */
  _getApplication: function(list) {
    return arArrayUtil
      .find(list.map(function(paths) paths.slice(1).reduce(function(file, path) (file ? (file.append(path), file) : null), paths[0])),
            function(file) file && file.exists()) || null;
  },
  
  /**
   * Mac で Mail を起動するスクリプトを作成する
   *
   * @returns {nsIFile}
   *          Mail を起動するスクリプト
   */
  _generateMacMailScript: function() {
    let data
      = "#!/usr/bin/osascript" + "\n"
      + "on run argv" + "\n"
      + "  tell application \"Mail\"" + "\n"
      + "    tell content of (make new outgoing message with properties {visible:true, subject:(item 1 of argv)})" + "\n"
      + "      make new attachment with properties {file name:(item 2 of argv)} at after the last paragraph" + "\n"
      + "    end tell" + "\n"
      + "    activate" + "\n"
      + "  end tell" + "\n"
      + "end run" + "\n";
    
    let macmailFile = UnMHTSystemDirectory.cacheDirectory.clone();
    macmailFile.append("macmailscript");
    
    let fstream
      = arFileUtil.openOutputStream(macmailFile, arFileUtil.PERM_EXEC_FILE);
    fstream.write(data, data.length);
    fstream.close();
    
    return macmailFile;
  },
  
  /**
   * ファイルを ZIP で圧縮する
   *
   * @param   {nsIFile} targetFile
   *          圧縮するファイル
   * @returns {nsIFile}
   *          圧縮したファイル
   */
  _createZipFile: function(targetFile) {
    let zipFile = targetFile.parent.clone();
    zipFile.append(UnMHTFilename.removeExtension(targetFile.leafName)
                   + ".zip");
    
    let zip = arFileUtil.openZipWriter(zipFile);
    
    zip.addEntryFile(targetFile.leafName,
                     Ci.nsIZipWriter.COMPRESSION_DEFAULT,
                     targetFile,
                     false);
    
    zip.close();
    
    return zipFile;
  }
};
