"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTMenuMap",
                        "UnMHTMenuClass"];

/**
 * window から UnMHTMenu へのマップ
 */
let UnMHTMenuMap = new Map();
let listener = {
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    try {
      UnMHTMenuMap.clear();
    } catch (e) {
      /* COMPAT: Firefox 18 以前 */
      for (let [name,] of UnMHTMenuMap) {
        UnMHTMenuMap.delete(name);
      }
    }
  }
};
arShutdownEvent.addListener(listener);

/**
 * メニュー管理
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 */
function UnMHTMenuClass(window) {
  arWindowHook.call(this, window);
  
  UnMHTMenuMap.set(window, this);
}
UnMHTMenuClass.prototype = {
  /**
   * メニューが開かれるイベント
   * メニューの項目の表示／非表示を設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onContextMenuShowing: function(event) {
    this._setMenuItem("context", true);
    this._setMenuItem("frame-context", true,
                      this._window.document.popupNode.ownerDocument);
  },
  
  /**
   * ファイルメニューを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onFileMenuShowing: function(event) {
    this._setMenuItem("filemenu");
  },
  
  /**
   * ツールバーのメニューを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onToolbarMenuShowing: function(event) {
    this._setMenuItem("toolbar");
  },

  /**
   * ツールバーのメニュー (右クリック用) を設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onToolbarMenu2Showing: function(event) {
    this._setMenuItem("toolbar2");
  },
  
  /**
   * ツールメニューを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onToolMenuShowing: function(event) {
    this._setMenuItem("toolmenu");
  },

  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    UnMHTMenuMap.delete(this._window);
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    UnMHTPref.addListener("menu.context",
                          function() {
                            this._enableContextMenuSave
                              = UnMHTPref.getPref("menu.context.save");
                            this._enableContextMenuQuickSave
                              = UnMHTPref.getPref("menu.context.save.quick");
                            this._enableContextMenuSendmail
                              = UnMHTPref.getPref("menu.context.save.mail");
                          }.bind(this), this, true);
    
    UnMHTPref.addListener("menu.all",
                          function() {
                            this._enableMenuAllCache
                              = UnMHTPref.getPref("menu.all.cache");
                            this._enableMenuAllExtract
                              = UnMHTPref.getPref("menu.all.extract");
                          }.bind(this), this, true);
    
    let menu = this.el("unmht-tools-preferences");
    if (menu) {
      UnMHTPref.addListener("menu.tool",
                            function() {
                              menu.hidden = !UnMHTPref.getPref("menu.tool");
                            }, this, true);
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    UnMHTPref.removeAllListeners(this);
  },
  
  /* ==== private ==== */
  
  /* メニューを表示するか */
  _enableContextMenuSave: false,      /** MHT で保存
                                       *  @type {boolean} */
  _enableContextMenuQuickSave: false, /** QuickSave
                                       *  @type {boolean} */
  _enableContextMenuSendmail: false,  /** メール送信
                                       *  @type {boolean} */
  _enableMenuAllCache: false,         /** キャッシュを表示/キャッシュをクリア
                                       *  @type {boolean} */
  _enableMenuAllExtract: false,       /** 展開して表示/ソースを表示
                                       *  @type {boolean} */
  
  /**
   * メニューを設定する
   *
   * @param   {string} id
   *          メニューの種類
   * @param   {boolean} isContext
   *          コンテキストメニューか
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   *          null の場合現在開いているドキュメント
   */
  _setMenuItem: function(id, isContext, targetDocument) {
    if (!targetDocument) {
      targetDocument = UnMHTWindowUtil.getCurrentDocument(this._window);
    }
    
    let savable
      = UnMHTDocumentCategory.isSavable(this._window, targetDocument, false);
    let isLocalMHT
      = UnMHTDocumentCategory.isLocalMHT(this._window, targetDocument);
    
    let showMenu = function(type, shown) {
      let menuitem = this.el("unmht-" + id + "-menuitem-" + type);
      if (menuitem) {
        menuitem.hidden = !shown;
      }
      
      return shown;
    }.bind(this);
    
    showMenu("cache-show",      this._enableMenuAllCache);
    showMenu("cache-clear",     this._enableMenuAllCache);
    showMenu("separator-cache", this._enableMenuAllCache);
    
    showMenu("extract",           this._enableMenuAllExtract);
    showMenu("unextract",         this._enableMenuAllExtract);
    showMenu("separator-extract", this._enableMenuAllExtract);
    
    let onPage = isContext ? UnMHTDOMUtil.onPage(this._window) : true;
    let isContentSelected
      = UnMHTDOMUtil.isContentSelected(this._window, isContext);
    let onSavableLink
      = isContext ? UnMHTDOMUtil.onSavableLink(this._window) : false;
    
    let shown = [
      showMenu("save",
               savable &&
               (!isContext ||
                (onPage && this._enableContextMenuSave))),
      showMenu("overwrite",
               savable &&
               isLocalMHT &&
               (!isContext ||
                (onPage && this._enableContextMenuSave))),
      showMenu("save-selection",
               savable &&
               isContentSelected &&
               (!isContext ||
                (onPage && this._enableContextMenuSave))),
      showMenu("save-link",
               onSavableLink &&
               (!isContext ||
                this._enableContextMenuSave)),
      showMenu("quicksave",
               savable &&
               (!isContext ||
                (onPage && this._enableContextMenuQuickSave))),
      showMenu("sendmail",
               savable &&
               (!isContext ||
                (onPage && this._enableContextMenuSendmail)))
    ];
    
    showMenu("separator", shown.some(function(b) b));
  }
};
arPrototype.extend(UnMHTMenuClass, arWindowHook);
