"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");

let EXPORTED_SYMBOLS = ["UnMHTNetFileLoader"];

/**
 * ファイルロードコンテキスト
 *
 * @param   {nsIURI} location
 *          対象の URI
 * @param   {nsIURI} referrer
 *          リファラ
 *          必要なければ null
 * @param   {boolean} useCache
 *          キャッシュを使用するか
 * @param   {IUnMHTNetFileLoaderListener} listener
 *          イベントリスナ
 * @implements nsIRequestObserver
 * @implements nsIStreamListener
 * @implements nsICacheListener
 */
function UnMHTNetFileLoadContext(location, referrer, useCache, listener) {
  this._location = location;
  this._referrer = referrer;
  this._useCache = useCache;
  this._listener = listener;
}
UnMHTNetFileLoadContext.prototype = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIStreamListener,
                                         Ci.nsIRequestObserver,
                                         Ci.nsICacheListener]),
  
  /**
   * 破棄
   * 自分で呼ぶ
   */
  destruct: function() {
    this._location = null;
    this._referrer = null;
    this._listener = null;
    this._channel = null;
  },
  
  /**
   * 取得を開始する
   */
  start: function() {
    this._channel = null;
    
    if (this._useCache) {
      this._startCache();
    } else {
      this._startNet();
    }
  },
  
  /**
   * 取得をキャンセルする
   */
  cancel: function() {
    this._status = UnMHTNetFileLoader.STATUS_CANCEL;
    
    if (this._channel) {
      try {
        this._channel.cancel(Cr.NS_BINDING_ABORTED);
      } catch (e) {
      }
    }
    
    this.destruct();
  },
  
  /**
   * キャッシュ取得をスキップする
   */
  skipCache: function() {
    if (this._status == UnMHTNetFileLoader.STATUS_CACHE) {
      this._startNet();
    }
  },

  /* ==== nsIRequestObserver インターフェース ==== */
  
  /**
   * リクエスト開始のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {nsISupports} context
   *          ユーザ定義
   */
  onStartRequest: function(request, context) {
    this._dataLength = 0;
    
    let size = 0;
    if (this._channel) {
      let httpChannel = null;
      try {
        httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);
        
        /* アクセス可能かチェックする */
        httpChannel.responseStatus;
      } catch (e) {
        httpChannel = null;
      }
      
      if (httpChannel) {
        if (httpChannel.responseStatus >= 400) {
          this._status = UnMHTNetFileLoader.STATUS_NG;
          if (this._listener) {
            this._listener.onStop(this, this._status, null);
          }
          this.destruct();
          return;
        }
        
        try {
          size = parseInt(httpChannel.getResponseHeader("Content-Length"), 10);
        } catch (e) {
        }
      }
    }
    
    if (this._listener) {
      this._listener.onStart(this, this._status, this._location, size);
    }
  },
  
  /**
   * リクエスト終了のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {nsISupports} context
   *          ユーザ定義
   * @param   {number} statusCode
   *          終了コード
   */
  onStopRequest: function(request, context, statusCode) {
    if (this._status == UnMHTNetFileLoader.STATUS_CANCEL ||
        this._status == UnMHTNetFileLoader.STATUS_NG) {
      return;
    }
    
    if (this._channel) {
      let httpChannel = null;
      try {
        httpChannel = this._channel.QueryInterface(Ci.nsIHttpChannel);
      } catch (e) {
      }
      
      if (httpChannel) {
        let location = "";
        try {
          location = httpChannel.getResponseHeader("Location");
        } catch (e) {
        }
        if (location) {
          try {
            this._location = Services.io.newURI(location, null, null);
          } catch (e) {
            if (this._listener) {
              this._listener.onStop(this, UnMHTNetFileLoader.STATUS_NG, null);
            }
            this.destruct();
            return;
          }
          this.start();
          return;
        }
      }
    }
    
    if (this._status == UnMHTNetFileLoader.STATUS_CACHE &&
        this._dataLength == 0) {
      this._startNet();
      return;
    }
    
    if (this._listener) {
      this._listener.onStop(this, this._status, this._channel);
    }
    this.destruct();
  },
  
  /* ==== nsIStreamListener インターフェース ==== */
  
  /**
   * データ到着のイベント
   *
   * @param   {nsIRequest} request
   *          対象のリクエスト
   * @param   {nsISupports} context
   *          ユーザ定義
   * @param   {nsIInputStream} inputStream
   *          データを取得するストリーム
   * @param   {number} offset
   *          データの位置
   * @param   {number} count 
   *          データの長さ
   */
  onDataAvailable: function(request, context, inputStream, offset, count) {
    if (this._status == UnMHTNetFileLoader.STATUS_CANCEL ||
        this._status == UnMHTNetFileLoader.STATUS_NG) {
      return;
    }
    
    let bstream
      = Cc["@mozilla.org/binaryinputstream;1"]
      .createInstance(Ci.nsIBinaryInputStream);
    bstream.setInputStream(inputStream);
    this._dataLength += count;
    if (this._listener) {
      this._listener.onData(this, this._status,
                            bstream.readBytes(count), count);
    }
  },
  
  /* ==== nsICacheListener インターフェース ==== */
  
  /**
   * キャッシュエントリが使用可能になったイベント
   *
   * @param   {nsICacheEntryDescriptor} descriptor
   *          キャッシュの情報
   * @param   {nsCacheAccessMode} accessGranted
   *          アクセス権限
   * @param   {nsresult} status
   *          キャッシュエントリの状態
   */
  onCacheEntryAvailable: function(descriptor, accessGranted, status) {
    if (this._status != UnMHTNetFileLoader.STATUS_CACHE) {
      return;
    }
    
    if (!accessGranted) {
      this._startNet();
      return;
    }
    
    if (this._applyResponseHead(descriptor)) {
      return;
    }
    
    this.onStartRequest(null, null);
    
    let istream = descriptor.openInputStream(0);
    this.onDataAvailable(null, null, istream, 0, descriptor.dataSize);
    istream.close();
    
    this.onStopRequest(null, null, Cr.NS_OK);
  },
  
  /* ==== private ==== */
  
  _status: 0,        /** 取得状況
                      *  @type {number}
                      *    UnMHTNetFileLoader.STATUS_* */
  
  _location: null,   /** 対象の URI
                      *  @type {nsIURI} */
  _referrer: null,   /** リファラ
                      *  @type {nsIURI} */
  _useCache: false,  /** キャッシュを使用するか
                      *  @type {boolean} */
  _listener: null,   /** イベントリスナ
                      *  @type {IUnMHTNetFileLoaderListener} */
  
  _dataLength: 0,    /** 取得したデータのサイズ
                      *  @type {number} */
  _channel: null,    /** 取得中のチャネル
                      *  @type {nsIChannel} */
  
  /**
   * キャッシュから取得を開始する
   */
  _startCache: function() {
    this._status = UnMHTNetFileLoader.STATUS_CACHE;
    if (this._listener) {
      this._listener.onStatusChange(this, this._status);
    }
    
    let cacheSession
      = Services.cache.createSession("HTTP", Ci.nsICache.STORE_ON_DISK, true);
    cacheSession.doomEntriesIfExpired = false;
    
    try {
      cacheSession
        .asyncOpenCacheEntry(this._location.spec,
                             Ci.nsICache.ACCESS_READ, this);
    } catch (e) {
      this._startNet();
    }
  },  
  
  /**
   * HTTP のレスポンスヘッダを必要であれば適用する
   *
   * @param   {nsICacheEntryDescriptor} descriptor
   *          キャッシュの情報
   * @returns {string}
   *          適用したか
   */
  _applyResponseHead: function(descriptor) {
    let head = "";
    try {
      head = descriptor.getMetaDataElement("response-head");
    } catch (e) {
      return false;
    }
    
    let re = head.match(/^HTTP\/[0-9\.]+ ([0-9]+)[^\r\n]*\r?\n((?:.|[\r\n])*)$/);
    if (!re) {
      return false;
    }
    
    let [, code, fields] = re;
    if (code >= 400) {
      this._startNet();
      return true;
    }
    
    let part = UnMHTMIMEDecoder.decodeFields(fields);
    if (!part) {
      return false;
    }
    
    if (part.fields.has("location")) {
      try {
        this._location
          = Services.io.newURI(part.fields.get("location"), null, null);
      } catch (e) {
        this._startNet();
        return true;
      }
      this.start();
      return true;
    }
    
    if (part.fields.has("content-encoding")) {
      let encoding = part.fields.get("content-encoding");
      
      let converter = null;
      try {
        converter
          = Cc["@mozilla.org/streamconv;1?from=" + encoding + "&to=uncompressed"]
          .createInstance(Ci.nsIStreamConverter);
        converter.asyncConvertData(encoding, "uncompressed", this, null);
      } catch (e) {
        this._startNet();
        return true;
      }
      
      let listener = converter.QueryInterface(Ci.nsIStreamListener);
      
      listener.onStartRequest(null, null);
      
      let istream = descriptor.openInputStream(0);
      listener.onDataAvailable(null, null, istream, 0, descriptor.dataSize);
      istream.close();
      
      listener.onStopRequest(null, null, Cr.NS_OK);
      
      return true;
    }
    
    return false;
  },
  
  /**
   * ネットワークから取得を開始する
   */
  _startNet: function() {
    this._status = UnMHTNetFileLoader.STATUS_NET;
    if (this._listener) {
      this._listener.onStatusChange(this, this._status);
    }
    
    let channel;
    try {
      channel = Services.io.newChannelFromURI(this._location);
    } catch (e) {
      /* NS_ERROR_UNKNOWN_PROTOCOL 等が投げられる */
      if (this._listener) {
        this._listener.onStop(this, UnMHTNetFileLoader.STATUS_NG, null);
      }
      this.destruct();
      return;
    }
    
    let httpChannel = null;
    try {
      httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
    } catch (e) {
    }
    if (httpChannel) {
      httpChannel.setRequestHeader("Accept", "*/*", false);
      if (this._referrer) {
        httpChannel.referrer = this._referrer;
      }
    }
    
    this._channel = channel;
    try {
      channel.asyncOpen(this, null);
    } catch (e) {
      if (this._listener) {
        this._listener.onStop(this, UnMHTNetFileLoader.STATUS_NG, null);
      }
      this.destruct();
    }
  }
};

/**
 * ネットワークからのファイル取得
 */
let UnMHTNetFileLoader = {
  /**
   * 取得状況
   * @type {number}
   */
  STATUS_NA:     0x0000, /* 未取得 */
  STATUS_CACHE:  0x0001, /* キャッシュからの取得 */
  STATUS_NET:    0x0002, /* ネットワークからの取得 */
  STATUS_NG:     0x0011, /* 取得失敗 */
  STATUS_CANCEL: 0x0012, /* 取得キャンセル */
  
  /**
   * データを取得してコールバックを呼ぶ
   *
   * @param   {nsIURI} location
   *          対象の URI
   * @param   {nsIURI} referrer
   *          リファラ
   *          必要なければ null
   * @param   {boolean} useCache
   *          キャッシュを使用するか
   * @param   {Function} callback
   *          完了時コールバック
   *            function(status, data)
   *              *param   {number} status
   *              *param   {string} data
   */
  loadData: function(location, referrer, useCache, callback) {
    let listener = {
      onStatusChange: function(context, status) {
      },
      onStart: function(context, status, location, size) {
        this._data = "";
      },
      onStop: function(context, status, channel) {
        callback(status, this._data);
        this._data = "";
      },
      onData: function(context, status, data, count) {
        this._data += data;
      },
      _data: "",
    };
    this.createContext(location, referrer, useCache, listener).start();
  },
  
  /**
   * ファイルロードコンテキストを作成する
   *
   * @param   {nsIURI} location
   *          対象の URI
   * @param   {nsIURI} referrer
   *          リファラ
   *          必要なければ null
   * @param   {boolean} useCache
   *          キャッシュを使用するか
   * @param   {IUnMHTNetFileLoaderListener} listener
   *          イベントリスナ
   * @returns {UnMHTNetFileLoadContext}
   *          ファイルロードコンテキスト
   */
  createContext: function(location, referrer, useCache, listener) {
    return new UnMHTNetFileLoadContext(location, referrer, useCache, listener);
  }
};
