"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTExtractor", function() {
  XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
  return UnMHTExtractor;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFormatter", function() {
  XPCOMUtils.importRelative(this, "UnMHTFormatter.jsm");
  return UnMHTFormatter;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTMIMEDecoder", function() {
  XPCOMUtils.importRelative(this, "UnMHTMIMEDecoder.jsm");
  return UnMHTMIMEDecoder;
});

let EXPORTED_SYMBOLS = ["UnMHTPageInfo"];

/**
 * ページ情報ダイアログ
 * MHT に関する情報を追加する
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 */
function UnMHTPageInfo(window) {
  arWindowHook.call(this, window);
  
  this._rows = [];
}
UnMHTPageInfo.prototype = {
  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._rows = null;
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    if (!("gDocument" in this._window &&
          this._window.gDocument)) {
      arTimer.setTimeout(function() {
        this.load(isStartup, reason);
      }.bind(this), 100);
      return;
    }
    
    let urltext = this.el("urltext");
    if (!urltext) {
      return;
    }
    
    let urlRow = UnMHTDOMUtil.findParentNode(urltext, "row");
    if (!urlRow) {
      return;
    }
    
    let url = this._window.gDocument.location.toString();

    if (!UnMHTScheme.isUnMHTURI(url)) {
      return;
    }
    
    let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(url);
    if (!(eFileInfo && part)) {
      return;
    }
    
    let nextSibling = urlRow.nextSibling;
    
    let addRow = function(labelText, valueText) {
      let row = this._window.document.createElement("row");
      
      let label = this._window.document.createElement("label");
      label.setAttribute("control", "unmht-original-url");
      label.setAttribute("value", labelText);
      row.appendChild(label);
      
      let sep = this._window.document.createElement("separator");
      row.appendChild(sep);
      
      let textbox = this._window.document.createElement("textbox");
      textbox.setAttribute("readonly", "unmht-original-url");
      textbox.setAttribute("id", "unmht-original-url");
      textbox.setAttribute("value", valueText);
      row.appendChild(textbox);
      
      urlRow.parentNode.insertBefore(row, nextSibling);
      
      this._rows.push(row);
    }.bind(this);
    
    addRow(UnMHTStringBundle.get("info.original_address"),
           eFileInfo.original);
    if (eFileInfo.startPart &&
        eFileInfo.startPart.contentLocation) {
      addRow(UnMHTStringBundle.get("info.saved_from"),
             part.contentLocation ||
             eFileInfo.startPart.contentLocation);
    }
    if (eFileInfo.date) {
      let savedTime = eFileInfo.date;
      try {
        let date = UnMHTMIMEDecoder.decodeDate(savedTime);
        if (date) {
          savedTime = UnMHTFormatter.getReadableDate(date);
        }
      } catch (e) {
      }
      
      addRow(UnMHTStringBundle.get("info.saved_date"),
             savedTime);
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    if (this._rows) {
      this._rows.forEach(function(row) {
        if (row.parentNode) {
          row.parentNode.removeChild(row);
        }
      }.bind(this));
      this._rows = null;
    }
  },
  
  /* ==== private ==== */
  
  _rows: null    /** 追加した行
                  *  @type {Array.<nsIDOMXULElement>} */
};
arPrototype.extend(UnMHTPageInfo, arWindowHook);
