"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arAppInfo.jsm");
XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "UnMHTDOMUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "arListView", function() {
  XPCOMUtils.importRelative(this, "base/arListView.jsm");
  return arListView;
});
XPCOMUtils.defineLazyGetter(this, "arLocaleUtil", function() {
  XPCOMUtils.importRelative(this, "base/arLocaleUtil.jsm");
  return arLocaleUtil;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTCustomFilename", function() {
  XPCOMUtils.importRelative(this, "UnMHTCustomFilename.jsm");
  return UnMHTCustomFilename;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFileConverter", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileConverter.jsm");
  return UnMHTFileConverter;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTFilePicker", function() {
  XPCOMUtils.importRelative(this, "UnMHTFilePicker.jsm");
  return UnMHTFilePicker;
});

let EXPORTED_SYMBOLS = ["UnMHTPrefWindowClass"];

/**
 * HTML ファイルから MHT への変換器
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @implements IUnMHTFileConverterListener
 */
function UnMHTPrefWindowFileConverter(window) {
  this._window = window;
  this._contexts = new Set();
  this._dirs = [];
  this._total = 0;
  this._failed = 0;
  this._no_res = 0;
  this._resDirs = new Set();
}
UnMHTPrefWindowFileConverter.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    if (this._contexts) {
      for (let context of this._contexts) {
        context.abort();
      }
      this._contexts = null;
    }
  },
  
  /**
   * ファイルを変換する
   *
   * @param   {Array.<nsIFile>} files
   *          対象のファイル
   * @param   {string} suffix
   *          フォルダ名の末尾
   * @param   {boolean} searchSub
   *          サブフォルダを検索するか
   */
  convertFiles: function(files, suffix, searchSub) {
    if (searchSub) {
      this._dirs
        = this._dirs.concat(files
                            .filter(function(f) f.isDirectory())
                            .map(function(d) [d, suffix]));
    }
    
    let context = UnMHTFileConverter.createContext(files.filter(function(f) f.isFile()), suffix, this);
    this._contexts.add(context);
    this._updateButton();
    context.start();
  },
  
  /**
   * 全ての変換を中断する
   */
  abort: function() {
    this._dirs = [];
    this._resDirs = new Set();
    for (let context of this._contexts) {
      context.abort();
    }
  },
  
  /**
   * ログを追加する
   *
   * @param   {string} s
   *          ログ
   */
  log: function(s) {
    let logBox = this.el("convert_log");
    logBox.value += s;
    logBox.inputField.scrollTop = logBox.inputField.scrollTopMax;
  },
  
  /**
   * ログを改行付きで追加する
   *
   * @param   {string} s
   *          ログ
   */
  logln: function(s) {
    this.log(s + "\n");
  },
  
  /**
   * 変換ログを消去
   */
  clearLog: function() {
    this.el("convert_log").value = "";
  },
  
  /* ==== IUnMHTFileConverterListener インターフェース ==== */
  
  /**
   * ファイル変換開始イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {nsIFile} mhtFile
   *          出力先の MHT ファイル
   * @param   {nsIFile} file
   *          入力の HTML ファイル
   * @param   {nsIFile} resDir
   *          リソースディレクトリ
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onStart: function(context, mhtFile, file, resDir, current, total) {
    this.log(UnMHTStringBundle.get("convert.log.start", [file.path]));
    this.el("convert_progressbar").value
      = Math.floor((this._total + current) / (this._total + total) * 100);
    this.el("convert_progressbar").style.visibility = "visible";
  },

  /**
   * ファイル変換完了イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {nsIFile} mhtFile
   *          出力先の MHT ファイル
   * @param   {nsIFile} file
   *          入力の HTML ファイル
   * @param   {nsIFile} resDir
   *          リソースディレクトリ
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onFinish: function(context, mhtFile, file, resDir, current, total) {
    this._resDirs.add(resDir.path);
    this.logln(UnMHTStringBundle.get("convert.log.finish"));
    this.logln("  " + UnMHTStringBundle.get("convert.log.input", [file.path]));
    this.logln("  " + UnMHTStringBundle.get("convert.log.res", [resDir.path]));
    this.logln("  " + UnMHTStringBundle.get("convert.log.output", [mhtFile.path]));
  },
  
  /**
   * ファイル変換失敗イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {nsIFile} mhtFile
   *          出力先の MHT ファイル
   * @param   {nsIFile} file
   *          入力の HTML ファイル
   * @param   {nsIFile} resDir
   *          リソースディレクトリ
   * @param   {number} current
   *          現在のファイルの位置
   * @param   {number} total
   *          総ファイル数
   */
  onFail: function(context, mhtFile, file, resDir, current, total) {
    this.logln(UnMHTStringBundle.get("convert.log.failed"));
    this._failed ++;
  },
  
  /**
   * リソースファイルが無いファイル
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {nsIFile} file
   *          入力の HTML ファイル
   */
  onNoResDir: function(context, file) {
    this.logln(UnMHTStringBundle.get("convert.log.no_res", [file.path]));
    this._no_res ++;
  },
  
  /**
   * 全ファイル変換完了イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   * @param   {number} count
   *          変換したファイル数
   */
  onComplete: function(context, count) {
    this._contexts.delete(context);
    this._updateButton();
    
    this._total += count;
    
    while (this._dirs.length) {
      let [dir, suffix] = this._dirs.shift();
      if (!this._resDirs.has(dir.path)) {
        this.convertFiles(arFileUtil.getDirectoryEntries(dir), suffix, true);
        return;
      }
    }
    
    this.logln(UnMHTStringBundle.get("convert.log.complete"));
    this.logln("  " + UnMHTStringBundle.get("convert.log.summary.ok", [this._total]));
    this.logln("  " + UnMHTStringBundle.get("convert.log.summary.failed", [this._failed]));
    this.logln("  " + UnMHTStringBundle.get("convert.log.summary.no_res", [this._no_res]));
    this.el("convert_progressbar").value = 100;
    this.el("convert_progressbar").style.visibility = "hidden";
    
    this._total = 0;
    this._failed = 0;
    this._no_res = 0;
  },
  
  /**
   * 中断イベント
   *
   * @param   {UnMHTFileConvertContext} context
   *          変換コンテキスト
   */
  onAbort: function(context, count) {
    this._contexts.delete(context);
    this._updateButton();
    this.logln(UnMHTStringBundle.get("convert.log.abort"));
    this.el("convert_progressbar").style.visibility = "hidden";
  },
  
  /* ==== private ==== */

  _window: null, /** 対象のウィンドウ
                  *  @type {nsIDOMWindow} */
  
  _contexts: null, /** 実行中の変換コンテキスト
                    *  @type {Set.<UnMHTFileConvertContext>} */
  
  _dirs: null,    /** 検索する予定のフォルダ
                   *  @type {Array.<nsIFile>} */
  _resDirs: null, /** リソースフォルダのパス
                   *  この中は検索しない
                   *  @type {Set.<string>} */
  _total: 0,      /** 変換したファイル数の合計
                   *  @type {number} */
  _failed: 0,     /** 失敗したファイル数の合計
                   *  @type {number} */
  _no_res: 0,     /** リソースファイルが無いファイルの合計
                   *  @type {number} */
  
  /**
   * 中断ボタンの状態を更新する
   */
  _updateButton: function() {
    let size;
    try {
      /* COMPAT: Firefox 18 以前 */
      size = this._contexts.size();
    } catch (e) {
      size = this._contexts.size;
    }
    let size2 = this._dirs.length;
    this.el("convert_abort").disabled = (size == 0 && size2 == 0);
  },
  
  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  el: function(id) {
    return this._window.document.getElementById(id);
  }
};

/**
 * 設定ウィンドウ
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 */
function UnMHTPrefWindowClass(window) {
  this._window = window;
  this._dragService
    = Cc["@mozilla.org/widget/dragservice;1"].getService(Ci.nsIDragService);
  this._converter = new UnMHTPrefWindowFileConverter(window);
}
UnMHTPrefWindowClass.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    this._dragService = null;
    if (this._converter) {
      this._converter.destruct();
      this._converter = null;
    }
  },
  
  /* ---- UI 関係イベント ---- */
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    this._instantApply = this.el("unmht_pref_window").instantApply;
    
    this._disableItems();
    
    this._updateUI();
    
    this._window.sizeToContent();
  },

  /**
   * アンロードイベント
   */
  onUnload: function() {
  },
  
  /**
   * 各パネルがロードされたイベント
   */
  onPaneLoad: function() {
    this._disableItems();
    
    this._updateUI();
  },
  
  /**
   * ヘルブボタンがクリックされたイベント
   */
  onHelp: function(event) {
    this._openURI((arLocaleUtil.isJapanene()
                   ? "http://www.unmht.org/unmht/usage-6.4.0.html"
                   : "http://www.unmht.org/unmht/en_usage-6.4.0.html")
                  + "#" + this.el("unmht_pref_window").currentPane.helpTopic);
  },
  
  /**
   * サイトへのリンクがクリックされたイベント
   * 現在未使用
   */
  onClickWebsiteLink: function() {
    this._openURI(arLocaleUtil.isJapanene()
                  ? "http://www.unmht.org/unmht/"
                  : "http://www.unmht.org/unmht/en_index.html");
  },
  
  /**
   * ショートカットキー入力欄でキーを押したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onShortcutKeyDown: function(event, id) {
    let elem = function(type) {
      return this.el("shortcut.save." + id + "." + type);
    }.bind(this);
    
    let keycode = "";
    for (let attribute of Object.keys(Ci.nsIDOMKeyEvent)) {
      let re = attribute.match(/^DOM_(VK_.+)$/);
      if (!re) {
        continue;
      }
      let code = re[1];
      if (event.keyCode == Ci.nsIDOMKeyEvent[attribute]) {
        keycode = code;
      }
    }
    if (keycode == "VK_SHIFT" ||
        keycode == "VK_CONTROL" ||
        keycode == "VK_ALT" ||
        keycode == "VK_META") {
      return;
    }
    
    if (keycode != "") {
      elem("keycode").value = keycode;
      elem("mods.alt").value = event.altKey;
      elem("mods.ctrl").value = event.ctrlKey;
      elem("mods.meta").value = event.metaKey;
      elem("mods.shift").value = event.shiftKey;
    }
    
    event.preventDefault();
    event.stopPropagation();
  },
  
  /**
   * メール送信アプリケーションの変更ボタンを押したイベント
   */
  onClickSelectSendMailApp: function() {
    let node = this.el("save.mail.app");
    
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("save.mail.app.title"),
      mode:             Ci.nsIFilePicker.modeOpen,
      useLastDir:       false,
      displayDirectory: arFileUtil.getFile(node.value),
      uri:              null,
      callback:         function(file, result, filePicker) {
        if (!file) {
          return;
        }
        
        node.value = file.path;
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * メール送信アプリケーションの引数を変更したイベント
   */
  onChangeSendMailArgType: function() {
    let textbox = this.el("ui-save.mail.arg.other");
    let value = "";
    
    switch (this.el("save.mail.arg.type").value) {
      case "thunderbird": {
        value = UnMHTPref.thunderbirdArg;
        textbox.disabled = true;
        break;
      }
      case "seamonkey": {
        value = UnMHTPref.seamonkeyArg;
        textbox.disabled = true;
        break;
      }
      case "macmail": {
        value = UnMHTPref.macmailArg;
        textbox.disabled = true;
        break;
      }
      default: {
        textbox.disabled = false;
        return;
      }
    }
    
    this.el("save.mail.arg.other").value = value;
  },
  
  /**
   * ディレクトリの変更ボタンを押したイベント
   *
   * @param   {string} id
   *          対象の設定
   */
  onClickSelectBase: function(id) {
    let node = this.el("save." + id + ".dir");
    
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("save." + id + ".base.title"),
      mode:             Ci.nsIFilePicker.modeGetFolder,
      useLastDir:       false,
      displayDirectory: arFileUtil.getFile(node.value),
      uri:              null,
      callback:         function(dir, result, filePicker) {
        if (!dir) {
          return;
        }
        
        node.value = dir.path;
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * カスタムファイル名のサンプルの更新が必要になったイベント
   */
  onNeedUpdateCustomFilename: function() {
    this.el("save_filename_custom_sample").value
      = UnMHTFilename.validateFileName(
        UnMHTCustomFilename.formatWith(
          "no title",
          "index.html",
          "http://localhost/index.html?query#fragment",
          new Date(),
          this.el("save.filename.custom").value,
          this.el("save.filename.custom.am").value,
          this.el("save.filename.custom.pm").value));
  },
  
  /**
   * ファイル名の文字の置換のリストの設定値が変更されたイベント
   */
  onFilenameReplaceListPrefChange: function() {
    if (this._updating) {
      return;
    }
    
    let node = this.el("save.filename.replace.list");
    
    let datas;
    try {
      datas = JSON.parse(node.value);
      if (!datas.every(function(data) ("from" in data && "to" in data))) {
        datas = [];
      }
    } catch (e) {
      datas = [];
    }
    
    this._filenameReplaceListView.setDatas(datas);
  },
  
  /**
   * HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertFiles: function() {
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("convert.title"),
      mode:             Ci.nsIFilePicker.modeOpenMultiple,
      useLastDir:       true,
      displayDirectory: null,
      uri:              null,
      callback:         function(files, result, filePicker) {
        if (!files) {
          return;
        }
        
        this._converter.convertFiles(files,
                                     this.el("convert.dir.suffix").value,
                                     this.el("ui-convert.dir.sub").checked);
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * ディレクトリ内の HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertDir: function() {
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("convert.title"),
      mode:             Ci.nsIFilePicker.modeGetFolder,
      useLastDir:       true,
      displayDirectory: null,
      uri:              null,
      callback:         function(dir, result, filePicker) {
        if (!dir) {
          return;
        }
        
        this._converter.convertFiles(arFileUtil.getDirectoryEntries(dir),
                                     this.el("convert.dir.suffix").value,
                                     this.el("ui-convert.dir.sub").checked);
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * 変換ログを消去するボタンを押したイベント
   */
  onClickClearConvertLog: function() {
    this._converter.clearLog();
  },
  
  /**
   * 変換パネルにファイルがドラッグされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDragOverConvertFiles: function(event) {
    if (this._dragService.getCurrentSession()
        .isDataFlavorSupported("application/x-moz-file")) {
      event.stopPropagation();
      event.preventDefault();
    }
  },
  
  /**
   * 変換パネルにファイルがドロップされたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDragDropConvertFiles: function(event) {
    let dataTransfer = event.dataTransfer;
    
    let files = [];
    for (let i of arArrayUtil.range(0, dataTransfer.mozItemCount)) {
      let types = dataTransfer.mozTypesAt(i);
      if (Array.indexOf(types, "application/x-moz-file") == -1) {
        continue;
      }
      
      let data = dataTransfer.mozGetDataAt("application/x-moz-file", i);
      try {
        let file = data.QueryInterface(Ci.nsIFile);
        
        if (file.isDirectory()) {
          files
            = files.concat(arFileUtil.getDirectoryEntries(file));
        } else if (file.isFile()) {
          files.push(file);
        }
      } catch (e) {
      }
    }
    
    this._converter.convertFiles(files,
                                 this.el("convert.dir.suffix").value,
                                 this.el("ui-convert.dir.sub").checked);
  },
  
  /**
   * ディレクトリ名の末尾を初期化するボタンを押したイベント
   */
  onClickInitConvertDirSuffix: function() {
    let node = this.el("convert.dir.suffix");
    node.value = "_files";
  },
  
  /**
   * 変換中断ボタンを押したイベント
   */
  onClickAbortConvert: function() {
    this._converter.abort();
  },

  /**
   * 設定のインポートボタンを押したイベント
   */
  onClickImportPrefs: function() {
    let setStatus = function(type) {
      this.el("import_export_message").value
        = UnMHTStringBundle.get("pref.import." + type);
    }.bind(this);
    
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("pref.import.title"),
      mode:             Ci.nsIFilePicker.modeOpen,
      useLastDir:       false,
      displayDirectory: null,
      defaultString:    "unmhtConfig.txt",
      filter:           [["Text", "*.txt"], [Ci.nsIFilePicker.filterAll]],
      uri:              null,
      callback:         function(file, result, filePicker) {
        if (!file) {
          setStatus("abort");
          return;
        }
        
        let input = UnMHTPref.initPrefFromFile(file);
        if (input) {
          let f = function() {
            try {
              this._loadPrefs(input);
              UnMHTPref.closePrefFromFile(input);
              setStatus("done");
            } catch (e) {
              Cu.reportError(e + "\n" + e.stack);
              setStatus("fail");
            }
          }.bind(this);
          if (this._instantApply) {
            f();
          } else {
            this._showAllPanesAnd(f);
          }
        } else {
          setStatus("illegal");
        }
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * 設定のエクスポートボタンを押したイベント
   */
  onClickExportPrefs: function() {
    let setStatus = function(type) {
      this.el("import_export_message").value
        = UnMHTStringBundle.get("pref.export." + type);
    }.bind(this);
    
    let params = {
      window:           this._window,
      title:            UnMHTStringBundle.get("pref.export.title"),
      mode:             Ci.nsIFilePicker.modeSave,
      useLastDir:       false,
      displayDirectory: null,
      defaultString:    "unmhtConfig.txt",
      filter:           [["Text", "*.txt"], [Ci.nsIFilePicker.filterAll]],
      uri:              null,
      callback:         function(file, result, filePicker) {
        if (!file) {
          setStatus("abort");
          return;
        }
        
        let output = UnMHTPref.initPrefToFile(file);
        if (output) {
          let f = function() {
            try {
              this._savePrefs(output);
              UnMHTPref.closePrefToFile(output);
              setStatus("done");
            } catch (e) {
              Cu.reportError(e + "\n" + e.stack);
              setStatus("fail");
            }
          }.bind(this);
          if (this._instantApply) {
            f();
          } else {
            this._showAllPanesAnd(f);
          }
        } else {
          setStatus("fail");
        }
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /**
   * 設定の初期化ボタンを押したイベント
   */
  onClickInitPrefs: function() {
    let setStatus = function(type) {
      this.el("import_export_message").value
        = UnMHTStringBundle.get("pref.reset." + type);
    }.bind(this);
    
    let input = UnMHTPref.initPrefFromFile(null);
    let f = function() {
      this._loadPrefs(input);
      UnMHTPref.closePrefFromFile(input);
      setStatus("done");
    }.bind(this);
    if (this._instantApply) {
      f();
    } else {
      this._showAllPanesAnd(f);
    }
  },
  
  /**
   * カスタムファイル名に変数を挿入する
   *
   * @param   {string} text
   *          挿入するテキスト
   */
  insertCustom: function(text) {
    let textbox = this.el("ui-save.filename.custom");
    let value = textbox.value;
    let start = textbox.selectionStart;
    let end = textbox.selectionEnd;
    
    textbox.value = value.substr(0, start) + text + value.substr(end);
    this.el("save.filename.custom").value = textbox.value;
    
    textbox.focus();
    
    textbox.selectionStart = textbox.selectionEnd = start + text.length;
    
    this.onNeedUpdateCustomFilename();
  },
  
  /**
   * スピンボタンで数値を上げる
   *
   * @param   {string} id
   *          対象の要素の ID
   * @param   {number} max
   *          数値の最大値
   */
  onUpSpin: function(id, max) {
    let node = this.el(id);
    
    let n = parseInt(node.value, 10) + 1;
    if (n > max) {
      n = max;
    }
    
    node.value = n;
  },
  
  /**
   * スピンボタンで数値を下げる
   *
   * @param   {string} id
   *          対象の要素の ID
   * @param   {number} max
   *          数値の最小値
   */
  onDownSpin: function(id, min) {
    let node = this.el(id);
    
    let n = parseInt(node.value, 10) - 1;
    if (n < min) {
      n = min;
    }
    
    node.value = n;
  },
  
  /* ---- UI イベント - 依存関係 ---- */
  
  onChangeSaveLocalSameDir: function() {
    this._setDisabled(this.el("box_save_local_same_dir"),
                      !this.el("save.local.same_dir").value);
  },
  onChangeSaveInsertLink: function() {
    this._setDisabled(this.el("box_save_insert_link_1"),
                      !this.el("save.insert.link").value);
    this._setDisabled(this.el("box_save_insert_link_2"),
                      !this.el("save.insert.link").value);
  },
  onChangeSaveMultiIndexStyle: function() {
    this._setDisabled(this.el("ui-save.multi.index.style.content"),
                      !this.el("save.multi.index.style").value);
  },
  onChangeSaveMailAppType: function() {
    this._setDisabled(this.el("box_save_mail_app"),
                      this.el("save.mail.app.type").value != "other");
    
    if (this.el("save.mail.app.type").value != "other") {
      let argtype = this.el("save.mail.arg.type");
      
      argtype.value = this.el("save.mail.app.type").value;
      
      this.onChangeSendMailArgType();
    }
  },
  onChangeShortcut: function(id) {
    this._setDisabled(this.el("box_shortcut_" + id.replace(/\./g, "_")),
                      !this.el("shortcut." + id).value);
    this._setDisabled(this.el("field_shortcut_" + id.replace(/\./g, "_")),
                      !this.el("shortcut." + id).value);
  },
  onChangeOpenMaxRuntime: function() {
    this._setDisabled(this.el("subbox_open_max_run_time"),
                      !this.el("open.max_run_time").value);
  },
  onChangeSaveMaxRuntime: function() {
    this._setDisabled(this.el("subbox_save_max_run_time"),
                      !this.el("save.max_run_time").value);
  },
  onChangeOpenCacheExpire: function() {
    this._setDisabled(this.el("subbox_open_cache_expire"),
                      !this.el("open.cache.expire").value);
  },
  
  /* ==== private ==== */
  
  _converter: null, /** 変換器
                     *  @type {UnMHTPrefWindowFileConverter} */
  
  _updating: false, /** 設定を手動で更新中か
                     *  @type {boolean} */
  
  _replaceHooked: false, /** ファイル名の置換のリストの
                          *  ドラッグアンドドロップをフックしたか
                          *  @type {boolean} */
  
  _instantApply: false, /** prefwindow.instantApply の値
                         *  @type {boolean} */
  
  _window: null,        /** 対象のウィンドウ
                         *  @type {nsIDOMWindow} */
  _dragService: null,   /** ドラッグ管理
                         *  @type {nsIDragService} */
  
  /**
   * 設定を読み込む
   *
   * @param   {UnMHTPrefFileInputObject} input
   *          設定入力オブジェクト
   */
  _loadPrefs: function(input) {
    UnMHTPref.names
      .forEach(function(name) {
        let value = UnMHTPref.getPrefFromFile(input, name);
        
        if (UnMHTPref.JSONDataNames.indexOf(name) != -1) {
          value = JSON.stringify(value);
        }
        
        if (this._instantApply) {
          UnMHTPref.setPref(name, value);
        } else {
          let node = this.el(name);
          if (!node) {
            return;
          }
          node.value = value;
        }
      }.bind(this));
    
    this._updateUI();
  },
  
  /**
   * 全ての設定を保存する
   *
   * @param   {UnMHTPrefFileOutputObject} output
   *          設定出力オブジェクト
   */
  _savePrefs: function(output) {
    UnMHTPref.names
      .forEach(function(name) {
        let value;
        if (this._instantApply) {
          value = UnMHTPref.getPref(name);
        } else {
          let node = this.el(name);
          if (!node) {
            return;
          }
          
          value = node.value;
        }
        
        if (UnMHTPref.JSONDataNames.indexOf(name) != -1) {
          value = JSON.parse(value);
        }
        
        UnMHTPref.setPrefToFile(output, name, value);
      }.bind(this));
  },
  
  /**
   * 全てのパネルを表示してから関数を実行する
   * パネルは選択されるまでロードされないので
   * インポート、エクスポート、初期化時に必要
   *
   * @param   {Function} func
   *          実行する関数
   */
  _showAllPanesAnd: function(func) {
    let panes = this._window.document.getElementsByTagName("prefpane");
    let defaultPane = this.el("pane_general");
    let prefWindow = this.el("unmht_pref_window");
    let i = 0;
    
    let f = function() {
      if (i == panes.length) {
        prefWindow.showPane(defaultPane);
        func();
        return;
      }
      
      if (!panes[i].loaded) {
        prefWindow.showPane(panes[i]);
      } else {
        i ++;
      }
      
      arTimer.setTimeout(f, 100);
    }.bind(this);
    
    f();
  },
  
  /**
   * チェックボックスの状態等を更新する
   */
  _updateUI: function() {
    if (this.el("pane_save").loaded) {
      this.onChangeSaveLocalSameDir();
      this.onChangeSaveInsertLink();
    }
    if (this.el("pane_savemulti").loaded) {
      this.onChangeSaveMultiIndexStyle();
    }
    if (this.el("pane_mail").loaded) {
      this.onChangeSaveMailAppType();
      this.onChangeSendMailArgType();
    }
    if (this.el("pane_filename").loaded) {
      this.onNeedUpdateCustomFilename();
      
      if (!this._replaceHooked) {
        this._replaceHooked = true;
        
        let dataHandler = {
          /**
           * データを作成する
           *
           * @returns {ANY}
           *          datas に追加する要素
           */
          newData: function() {
            return { from: "/", to: "_" };
          },
          
          /**
           * データを複製する
           *
           * @param   {ANY} origItem
           *          複製元の要素
           * @returns {ANY}
           *          複製した要素
           */
          dupData: function(origItem) {
            return { from: origItem.from, to: origItem.to };
          },
          
          /**
           * データが更新されたイベント
           *
           * @param   {Array} datas
           *          更新されたデータ
           */
          onModifyDatas: function(datas) {
            let node = this.el("save.filename.replace.list");
            
            this._updating = true;
            node.value = JSON.stringify(datas);
            this._updating = false;
            
            this.onNeedUpdateCustomFilename();
          }.bind(this),
          
          /**
           * 要素をリストの行に変換する
           *
           * @param   {ANY} item
           *          対象の要素
           * @returns {Array}
           *          行のデータ
           *            [object セルのデータ, ...]
           */
          dataToRow: function(item) {
            let row = [];
            
            row.push({ label: item.from });
            row.push({ label: item.to });
            
            return row;
          },
          
          /**
           * リストの行を要素に変換する
           *
           * @param   {Array} row
           *          行のラベル
           *            [{ label: string ラベル }, ...]
           * @param   {ANY} oldItem
           *          過去の要素
           * @returns {ANY}
           *          対象の要素
           */
          rowToData: function(row, oldItem) {
            let item = {
              from: row[0].label,
              to: row[1].label
            };
            
            return item;
          },
          
          /**
           * リストの要素を選択したイベント
           *
           * @param   {Array} items
           *           [[ANY  選択した要素, number  行のインデックス], ...]
           */
          onSelect: function(items) {
            /* 直接編集するので何もしない */
          },
          
          /**
           * ソートのためのデータを返す
           *
           * @param   {ANY} item
           *          対象の要素
           * @param   {number} pos
           *          カラムの位置
           * @returns {comparable}
           *          ソート時の比較に使うデータ
           */
          getSortKeyFor: function(item, pos) {
            /* ソートは不要なので何もしない */
          }
        };
        
        this._filenameReplaceListView
          = new arListView(this.el("filename_replace_list"),
                           null,
                           this.el("filename_replace_list_children"),
                           this.el("filename_replace_list_new"),
                           this.el("filename_replace_list_dup"),
                           this.el("filename_replace_list_del"),
                           "application/x-unmht-filename-replace",
                           [],
                           true,
                           [],
                           dataHandler);
        
        this.onFilenameReplaceListPrefChange();
      }
    }
    if (this.el("pane_keyboard").loaded) {
      if (Services.appinfo.OS == "Darwin") {
        this.el("ui-shortcut.save.single.mods.alt").label
          = this.el("ui-shortcut.save.multi.mods.alt").label
          = this.el("ui-shortcut.save.quick.mods.alt").label
          = this.el("ui-shortcut.save.overwrite.mods.alt").label
          = this.el("keyboard_mods_option").value;
        
        this.el("ui-shortcut.save.single.mods.meta").label
          = this.el("ui-shortcut.save.multi.mods.meta").label
          = this.el("ui-shortcut.save.quick.mods.meta").label
          = this.el("ui-shortcut.save.overwrite.mods.meta").label
          = this.el("keyboard_mods_command").value;
      }
      
      this.onChangeShortcut("save.single");
      this.onChangeShortcut("save.multi");
      this.onChangeShortcut("save.quick");
      this.onChangeShortcut("save.overwrite");
    }
    if (this.el("pane_advanced").loaded) {
      this.onChangeOpenMaxRuntime();
      this.onChangeSaveMaxRuntime();
      this.onChangeOpenCacheExpire();
    }
  },
  
  /**
   * ノードの状態を変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象のノード
   * @param   {boolean} disabled
   *          無効か
   * @param   {boolean} uncheck
   *          チェックを外すか
   */
  _setDisabled: function(node, disabled, uncheck) {
    try {
      node.disabled = disabled;
    } catch (e) {
    }
    
    if (uncheck) {
      try {
        node.checked = false;
      } catch (e) {
      }
    }
    
    if ("childNodes" in node) {
      Array.slice(node.childNodes)
        .forEach(function(node) {
          this._setDisabled(node, disabled, uncheck);
        }.bind(this));
    }
  },
  
  /**
   * 環境によって使用できない設定を無効にする
   */
  _disableItems: function() {
    if (arAppInfo.isThunderbird) {
      [
        "group_open",
        
        "group_addonbar",
        
        "box_insert",
        "box_insert_url",
        "ui-save.local.same_dir",
        "box_save_local_same_dir",
        
        "group_save_multi",
        "group_save_multi_index",
        "group_save_mail",
        
        "save_filename_leafname",
        
        "group_menu_context",
        "group_shortcut",
        
        "box_open_max_run_time",
        "group_menu_all",
        "group_cache",
        "group_advanced_mail",
        "group_other"
      ].forEach(function(id) {
        let node = this.el(id);
        if (node) {
          this._setDisabled(node, true, true);
        }
      }.bind(this));
    }
    
    if (arAppInfo.isFirefox) {
      [
        "ui-save.header.mail.original_field",
        "ui-save.filename.custom.mail_date",
        "ui-save.header.mail.show",
        "ui-save.header.mail.show-desc"
      ].forEach(function(id) {
        let node = this.el(id);
        if (node) {
          this._setDisabled(node, true, true);
        }
      }.bind(this));
    }
    
    if (Services.appinfo.OS != "Darwin") {
      [
        "save_mail_app_type_macmail",
        "save_mail_arg_type_macmail"
      ].forEach(function(id) {
        let node = this.el(id);
        if (node) {
          this._setDisabled(node, true, true);
        }
      }.bind(this));
    }
    if (Services.appinfo.OS != "WINNT") {
      [
        "group_advanced_filename",
      ].forEach(function(id) {
        let node = this.el(id);
        if (node) {
          this._setDisabled(node, true, true);
        }
      }.bind(this));
    }
  },
  
  /**
   * ページをタブか新しいウィンドウに開く
   *
   * @param   {string} uri
   *          アドレス
   */
  _openURI: function(uri) {
    let chromeWindow = Services.wm.getMostRecentWindow("navigator:browser");
    if (chromeWindow) {
      let tabbrowser = UnMHTWindowUtil.getBrowser(chromeWindow);
      tabbrowser.selectedTab = tabbrowser.addTab(uri);
    } else {
      this._window.open(uri);
    }
  },
  
  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  el: function(id) {
    return this._window.document.getElementById(id);
  }
};
