"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTScheme.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.defineLazyGetter(this, "arUconv", function() {
  XPCOMUtils.importRelative(this, "base/arUconv.jsm");
  return arUconv;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTCachePage", function() {
  XPCOMUtils.importRelative(this, "UnMHTCachePage.jsm");
  return UnMHTCachePage;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTExtractor", function() {
  XPCOMUtils.importRelative(this, "UnMHTExtractor.jsm");
  return UnMHTExtractor;
});
XPCOMUtils.defineLazyGetter(this, "UnMHTHTMLUtil", function() {
  XPCOMUtils.importRelative(this, "UnMHTHTMLUtil.jsm");
  return UnMHTHTMLUtil;
});

let EXPORTED_SYMBOLS = ["UnMHTProtocolHandler"];

/**
 * unmht プロトコルからチャネルを生成する
 *
 * @implements nsIProtocolHandler
 */
function UnMHTProtocolHandler() {
}
UnMHTProtocolHandler.prototype = {
  classID:          Components.ID("{d605b959-0688-4846-887c-f6658725f7f1}"),
  contractID:       "@mozilla.org/network/protocol;1?name=unmht",
  classDescription: "UnMHT Protocol Handler JS Component",
  QueryInterface:   XPCOMUtils.generateQI([Ci.nsIProtocolHandler]),
  
  /* ==== nsIProtocolHandler インターフェース ==== */
  
  scheme: "unmht", /** プロトコルスキーム
                    *  @type {string} */
  defaultPort: -1, /** ポート番号
                    *  ネットワークベースではないので -1
                    :  @type {number} */
  
  /** プロトコルの詳細
   *  file と同様のプロトコル
   *  @type {number} */
  protocolFlags: (Ci.nsIProtocolHandler.URI_NOAUTH
                  | Ci.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE),
  
  /**
   * ブラックリストのポートを上書きするか
   *
   * @param   {number} port
   *          ポート番号
   * @param   {string} scheme
   *          プロトコルスキーム
   * @returns {boolean}
   *          上書きするか
   */
  allowPort: function(port, scheme) {
    return false;
  },
  
  /**
   * URI を作成する
   *
   * @param   {string} spec
   *          対象の URI
   * @param   {string} charset
   *          対象の文字コード
   * @param   {nsIURI} baseURI
   *          読み込み元の URI
   * @returns {nsIURI}
   *          作成した URI
   */
  newURI: function(spec, charset, baseURI) {
    let uri
      = Cc["@mozilla.org/network/standard-url;1"].createInstance(Ci.nsIURI);
    
    if (!baseURI) {
      uri.spec = spec;
      return uri;
    }
    
    if (UnMHTScheme.isUnMHTURI(baseURI.spec)) {
      let resolved = UnMHTExtractor.resolveURI(baseURI.spec, spec);
      if (resolved) {
        uri.spec = resolved;
        return uri;
      }
    }
    
    uri.spec = baseURI.resolve(spec);
    return uri;
  },
  
  /**
   * チャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  newChannel: function(uri) {
    let param = UnMHTScheme.getUnMHTURIParam(uri.spec);
    
    try {
      if (!("original" in param &&
            param.original)) {
        /* 不正な URI */
        return this._createIllegalChannel(uri);
      } else if (param.scheme == "cache") {
        return this._createCacheChannel(uri);
      } else if (param.scheme == "cid" ||
                 param.scheme == "dummy" ||
                 param.scheme == "empty") {
        return this._createEmptyChannel(uri);
      } else if (param.scheme == "error") {
        return this._createNotFoundChannel(uri);
      } else {
        /* mht 内のファイル */
        return this._createFileChannel(uri, param);
      }
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
    
    return this._createEmptyChannel(uri);
  },
  
  /* ==== private ==== */
  
  /**
   * mht 内のファイルのチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI (umht:// 形式)
   * @param   {UnMHTURIParam} param
   *          対象の URI の情報
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createFileChannel: function(uri, param) {
    let [eFileInfo, part] = UnMHTExtractor.getFileInfoAndPart(uri.spec);
    if (!eFileInfo) {
      try {
        /* URI が不正でないかチェック */
        Services.io.newURI(param.original, null, null);
      } catch (e) {
        return this._createIllegalChannel(uri);
      }
      
      return this._createEmptyChannel(uri); 
    }
    
    if (!part) {
      return this._createEmptyChannel(uri);
    }
    
    let pipe = Cc["@mozilla.org/pipe;1"].createInstance(Ci.nsIPipe);
    pipe.init(false, false, part.eParam.content.length, 1, null);
    pipe.outputStream.write(part.eParam.content, part.eParam.content.length);
    pipe.outputStream.close();
    
    return this._createInputStreamChannel(uri,
                                          part.eParam.charset,
                                          part.eParam.mimetype,
                                          pipe.inputStream,
                                          part.eParam.content.length);
  },
  
  /**
   * キャッシュ情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createCacheChannel: function(uri) {
    let data = UnMHTCachePage.createCachePage();
    
    return this._createStringInputStreamChannel(uri, "utf-8", "text/html", 
                                                data, data.length);
  },
  
  /**
   * ファイルが無い情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createNotFoundChannel: function(uri) {
    let bundle = function(name) {
      return UnMHTHTMLUtil.escapeEntity(UnMHTStringBundle.get(name));
    };
    
    let message
      = UnMHTHTMLUtil.escapeEntity(UnMHTScheme.deErrorURI(uri.spec));
    return this._createErrorChannel(uri, bundle("error.file_not_found"),
                                    message, true);
  },
  
  /**
   * 不正な URI 情報のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createIllegalChannel: function(uri) {
    let bundle = function(name) {
      return UnMHTHTMLUtil.escapeEntity(UnMHTStringBundle.get(name));
    };
    
    let message
      = UnMHTHTMLUtil.escapeEntity(uri.spec);
    return this._createErrorChannel(uri, bundle("error.illegal_address"),
                                    message, false);
  },
  
  /**
   * エラーメッセージを表示するチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} title
   *          タイトル (HTML)
   * @param   {string} message
   *          表示するメッセージ (HTML)
   * @param   {boolean} showRetryButton
   *          リトライボタンを表示する
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createErrorChannel: function(uri, title, message, showRetryButton) {
    let bundle = function(name) {
      return UnMHTHTMLUtil.escapeEntity(UnMHTStringBundle.get(name));
    };
    
    let data
      = "<!DOCTYPE html>" + "\n"
      + "<html xmlns=\"http://www.w3.org/1999/xhtml\">" + "\n"
      + "<head>" + "\n"
      + "  <title>" + bundle("error.title") + "</title>" + "\n"
      + "  <style type=\"text/css\">" + "\n"
      + "html {" + "\n"
      + "  background: -moz-Dialog;" + "\n"
      + "}" + "\n"
      + "\n"
      + "body {" + "\n"
      + "  margin: 0;" + "\n"
      + "  padding: 0 1em;" + "\n"
      + "  color: -moz-FieldText;" + "\n"
      + "  font: message-box;" + "\n"
      + "}" + "\n"
      + "\n"
      + "h1 {" + "\n"
      + "  margin: 0 0 .6em 0;" + "\n"
      + "  border-bottom: 1px solid ThreeDLightShadow;" + "\n"
      + "  font-size: 160%;" + "\n"
      + "}" + "\n"
      + "\n"
      + "h2 {" + "\n"
      + "  margin: 0 0 .6em 0;" + "\n"
      + "  border-bottom: 1px solid ThreeDLightShadow;" + "\n"
      + "  font-size: 140%;" + "\n"
      + "}" + "\n"
      + "\n"
      + "#errorPageContainer {" + "\n"
      + "  position: relative;" + "\n"
      + "  min-width: 13em;" + "\n"
      + "  max-width: 52em;" + "\n"
      + "  margin: 4em auto;" + "\n"
      + "  border: 1px solid ThreeDShadow;" + "\n"
      + "  border-radius: 10px;" + "\n"
      + "  padding: 3em;" + "\n"
      + "  -moz-padding-start: 30px;" + "\n"
      + "  background: -moz-Field;" + "\n"
      + "  background-origin: content-box;" + "\n"
      + "}" + "\n"
      + "\n"
      + "#errorDesc > p {" + "\n"
      + "  overflow: auto;" + "\n"
      + "  font-size: 130%;" + "\n"
      + "  white-space: pre-wrap;" + "\n"
      + "}" + "\n"
      + "\n"
      + "#unmht-retry-button {" + "\n"
      + "  margin-top: 2em;" + "\n"
      + "}" + "\n"
      + "  </style>" + "\n"
      + "</head>" + "\n"
      + "<body>" + "\n"
      + "  <div id=\"errorPageContainer\">" + "\n"
      + "    <h1>" + bundle("error.title") + "</h1>" + "\n"
      + "    <h2>" + title + "</h2>" + "\n"
      + "    <div>" + "\n"
      + "      <div id=\"errorDesc\">" + "\n"
      + "        <p>" + message + "</p>" + "\n"
      + "      </div>" + "\n"
      + "    </div>" + "\n"
      + (showRetryButton
         ? ("    <button id=\"unmht-retry-button\" autocomplete=\"off\" autofocus=\"true\">"
            + bundle("error.retry")
            + "</button>" + "\n")
         : "")
      + "  </div>" + "\n"
      + "</body>" + "\n"
      + "</html>" + "\n";
    
    data = arUconv.fromUnicode(data, "utf-8");
    
    return this._createStringInputStreamChannel(uri, "utf-8", "text/html", 
                                                data, data.length);
  },

  /**
   * 空のチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createEmptyChannel: function(uri) {
    return this._createStringInputStreamChannel(uri, "utf-8", "text/html", 
                                                "", 0);
  },
  
  /**
   * 文字列を返すチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} charset
   *          文字セット
   * @param   {string} type
   *          MIME-Type
   * @param   {string} data
   *          文字列
   * @param   {number} length
   *          文字列の長さ
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createStringInputStreamChannel: function(uri, charset, type,
                                            data, length) {
    let sstream
      = Cc["@mozilla.org/io/string-input-stream;1"]
      .createInstance(Ci.nsIStringInputStream);
    sstream.setData(data, length);
    
    return this._createInputStreamChannel(uri, charset, type,
                                          sstream, length);
  },
  
  /**
   * ストリームの内容を返すチャネルを作成する
   *
   * @param   {nsIURI} uri
   *          対象の URI
   * @param   {string} charset
   *          文字セット
   * @param   {string} type
   *          MIME-Type
   * @param   {nsIInputStream} stream
   *          ストリーム
   * @param   {number} length
   *          ストリームの長さ
   * @returns {nsIChannel}
   *          作成したチャネル
   */
  _createInputStreamChannel: function(uri, charset, type, stream, length) {
    let inputStreamChannel
      = Cc["@mozilla.org/network/input-stream-channel;1"]
      .createInstance(Ci.nsIInputStreamChannel)
      .QueryInterface(Ci.nsIChannel);
    
    inputStreamChannel.setURI(uri);
    inputStreamChannel.contentCharset = charset;
    inputStreamChannel.contentType    = type;
    inputStreamChannel.contentStream  = stream;
    inputStreamChannel.contentLength  = length;
    
    return inputStreamChannel;
  }
};
