"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arDialog.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
XPCOMUtils.importRelative(this, "UnMHTSaveProgress.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveDialogClass"];

/**
 * 保存ダイアログ
 *
 * ウィンドウの引数
 * {
 *   saveState:       {number}
 *                    保存状態 (必須)
 *
 *   scrapBookData:   {object}
 *                    ScrapBook のデータ (必須)
 *   targetDocument:  {nsIDOMDocument}
 *                    対象のドキュメント (必須)
 *   targetSelection: {nsISelection}
 *                    選択範囲 (必須)
 *
 *   targetFile:      {nsIFile}
 *                    保存先のファイル (必須)
 *   additionalURIs:  {Array.<string>}
 *                    追加ファイルの URI (必須)
 *   noRegisterHistory: {boolean}
 *                    ダウンロード履歴に登録しない (必須)
 *
 *   callback:        {Function}
 *                    保存完了/中断時のコールバック (オプショナル)
 *                      function(result)
 *                        *param   {number] result
 *                                 結果
 *                                   UnMHTSave.RESULT_OK
 *                                   UnMHTSave.RESULT_CANCEL
 *                                   UnMHTSave.RESULT_FAIL_ADD_DOCUMENT
 *                                   UnMHTSave.RESULT_ERROR_WRITE_FILE
 *
 *   window:          {nsIDOMWindow}
 *                    ドキュメントを持つブラウザを実行中のウィンドウ
 *                    メールの添付ファイル取得に使う (オプショナル)
 * }
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialog
 * @implements arIDialog
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveDialogClass(window) {
  arDialog.call(this, window);
}
UnMHTSaveDialogClass.prototype = {
  /* ==== arIDialog インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._callback = null;
    this._saveContext = null;
    arDialog.prototype.destruct.call(this);
  },
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this._window.arguments[0];
    
    let title = UnMHTStringBundle.get("save.title");
    switch (arg.saveState) {
      case UnMHTSave.SAVESTATE_CURRENT: {
        title += " " + UnMHTStringBundle.get("save.state.current");
        break;
      }
      case UnMHTSave.SAVESTATE_CURRENT_NOJS: {
        title += " " + UnMHTStringBundle.get("save.state.current")
          + " (" + UnMHTStringBundle.get("save.nojs") + ")";
        break;
      }
      case UnMHTSave.SAVESTATE_ORIGINAL: {
        title += " " + UnMHTStringBundle.get("save.state.original");
        break;
      }
    }

    let dialog = this.el("unmht_save_dialog");
    dialog.setAttribute("title", title);
    
    dialog.getButton("accept").disabled = true;
    
    let [location, subject]
      = UnMHTSave.getLocationAndTitle(arg.targetDocument, arg.scrapBookData);
    
    if ("callback" in arg) {
      this._callback = arg.callback;
    }
    
    let saveContext
      = UnMHTSave.createContext(this._window, arg.targetFile, location, subject,
                                new UnMHTSaveProgress(this._window, this),
                                arg.noRegisterHistory);
    
    let arg2 = {
      targetDocument: arg.targetDocument,
      targetSelection: arg.targetSelection,
      scrapBookData: arg.scrapBookData,
      saveState: arg.saveState,
      location: location,
      isHTML: true,
      isStartPart: true,
      window: arg.window
    };
    if (!saveContext.addDocument(arg2)) {
      saveContext.destruct();
      
      if (this._callback) {
        this._callback(UnMHTSave.RESULT_FAIL_ADD_DOCUMENT);
      }
      
      return;
    }
    
    if (arg.additionalURIs) {
      arg.additionalURIs.forEach(function(uri) {
        saveContext.addFile(null, uri, false);
      });
    }
    
    this._saveContext = saveContext;
    
    saveContext.start();
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function(context) {
    if (!this._saveContext) {
      return true;
    }
    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;
      
      return false;
    }
    
    if (this._callback) {
      this._callback(UnMHTSave.RESULT_CANCEL);
    }
    return true;
  },
  
  /* ==== IUnMHTSaveProgressOwner インターフェース ===== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(progress, context) {
    if (this._callback) {
      this._callback(UnMHTSave.RESULT_CANCEL);
    }
    this._window.close();
  },
  
  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {number} reason
   *          エラーの理由
   * @param   {Exception} e
   *          例外
   */
  onError: function(progress, context, reason, e) {
    if (reason == UnMHTSave.ERROR_WRITE_FILE) {
      if (this._callback) {
        this._callback(UnMHTSave.RESULT_ERROR_WRITE_FILE);
      }
    }
  },
  
  /**
   * 保存完了
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(progress, context, failedList) {
    let dialog = this.el("unmht_save_dialog");
    
    dialog.getButton("accept").disabled = false;
    dialog.getButton("accept").focus();
    dialog.getButton("cancel").disabled = true;
    
    if (this._callback) {
      this._callback(UnMHTSave.RESULT_OK);
    }
    
    if (UnMHTPref.getPref("save.show_errors") && failedList.length > 0) {
      progress.showFail(failedList);
    } else {
      let window = this._window;
      arTimer.setTimeout(function() {
        window.close();
      }, 1000);
    }
  },
  
  /* ==== private ==== */
  
  _callback: null,   /** 終了時コールバック
                      *  @type {Function} */
  
  _saveContext: null /** 保存コンテキスト
                      *  @type {UnMHTSaveContext} */
};
arPrototype.extend(UnMHTSaveDialogClass, arDialog);
