"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arUconv.jsm");
XPCOMUtils.importRelative(this, "UnMHTPseudoID.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.importRelative(this, "UnMHTHTMLUtil.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
XPCOMUtils.importRelative(this, "UnMHTSaveProgress.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveMulti"];

/**
 * 複数のタブの保存コンテキスト
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @param   {IUnMHTSaveMultiListener} listener
 *          イベントリスナ
 * @implements IUnMHTSaveProgressOwner
 */
function UnMHTSaveMultiContext(window, listener) {
  this._window = window;
  this._listener = listener;
  
  this._targets = [];
  this._additionalURIs = [];
  this._failedList = [];
  
  this._dupAction = UnMHTPref.getPref("save.multi.dup.action");
}
UnMHTSaveMultiContext.prototype = {
  /**
   * 破棄
   * start 後は自分で呼ぶ
   */
  destruct: function() {
    this._saveContext = null;
    this._failedList = null;
  },
  
  /**
   * 保存対象を追加する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {ANY} id
   *          保存対象の ID
   * @param   {object} param
   *          パラメータ
   *            {
   *              filenameBase: {string}
   *                            ファイル名の拡張子以外
   *              title:        {string}
   *                            インデックスページのタイトル
   *              desc:         {string}
   *                            インデックスページの説明
   *            }
   */
  addTarget: function(targetDocument, id, param) {
    this._targets.push([targetDocument, id, param]);
  },
  
  /**
   * ファイルを追加する
   *
   * @param   {string} url
   *          ファイルのアドレス
   */
  addFile: function(url) {
    this._additionalURIs.push(url);
  },
  
  /* ==== IUnMHTSaveProgressOwner インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(progress, context) {
    this._saveContext = null;
    if (this._isCanceled) {
      this._listener.onCancel(this);
      this.destruct();
    }
  },
  
  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {number} reason
   *          エラーの理由
   * @param   {Exception} e
   *          例外
   */
  onError: function(progress, context, reason, e) {
    let [, id, ] = this._targets[this._currentIndex];
    this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_NG);
    
    this._saveContext = null;
    switch (reason) {
      case UnMHTSave.ERROR_WRITE_FILE: {
        if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
          this._saveNext();
        } else {
          this.destruct();
        }
        break;
      }
    }
  },
  
  /**
   * 保存完了
   *
   * @param   {UnMHTSaveProgress} progress
   *          進捗状況表示オブジェクト
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>>} failedList
   *          失敗したファイル名
   */
  onSave: function(progress, context, failedList) {
    this._saveContext = null;
    
    this._failedList = this._failedList.concat(failedList);
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      let [, id, ] = this._targets[this._currentIndex];
      this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_OK);
      
      this._saveNext();
    } else {
      this._listener.onSave(this, this._failedList);
      this.destruct();
    }
  },
  
  /**
   * 保存を開始する
   */
  start: function() {
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      this._currentIndex = -1;
      this._saveNext();
    } else {
      this._currentIndex = 0;
      this._saveSingle();
    }
  },
  
  /**
   * 現在のファイルをスキップする
   */
  skip: function() {
    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;
    }
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED) {
      if (this._currentIndex < this._targets.length) {
        let [, id, ] = this._targets[this._currentIndex];
        this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_NG);
      }
      
      this._saveNext();
    } else {
      this._listener.onAborted(this);
      this.destruct();
    }
  },
  
  /**
   * 保存をキャンセルする
   */
  cancel: function() {
    this._isCanceled = true;
    
    if (this._saveContext) {
      this._saveContext.cancel();
      this._saveContext = null;
    } else {
      this._listener.onCancel(this);
      this.destruct();
    }
  },
  
  /* ==== private ==== */
  
  _window: null,          /** 対象のウィンドウ
                           *  @type {nsIDOMWindow} */
  _listener: null,        /** イベントリスナ
                           *  @type {IUnMHTSaveMultiListener} */
  
  _targets: null,         /** 対象のドキュメントリスト
                           *  @type {Array}
                           *   [[nsIDOMDocument, ANY ID, object param], ...] */
  _additionalURIs: null,  /** 追加 URI
                           *  @type {Array.<string>} */
  
  _dupAction: "",         /** 同名のファイルがあった場合にどうするか
                           *  @type {string}
                           *    "replace":  上書きする
                           *    "rename":  名前を変える
                           *    "ask":  ダイアログを表示する */
  
  _currentIndex: 0,       /** 現在保存中のドキュメントの位置
                           *  @type {number} */
  
  _saveContext: null,     /** 保存コンテキスト
                           *  @type {UnMHTSaveContext} */
  
  _failedList: null,      /** 失敗したファイル名
                           *  @type {Array.<string>} */
  
  _isCanceled: false,     /** キャンセルされたか
                           *  @type {boolean} */
  
  /**
   * 次のタブを保存する
   * 別々のファイル用
   */
  _saveNext: function() {
    this._currentIndex ++;
    
    if (this._currentIndex >= this._targets.length) {
      this._listener.onSave(this, this._failedList);
      return;
    }
    
    let [targetDocument, id, param] = this._targets[this._currentIndex];
    
    if (!UnMHTDocumentCategory.isAlive(targetDocument)) {
      this.skip();
      return;
    }
    
    this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_CURRENT);
    
    let [file, isAborted] = this._getFile(param.filenameBase);
    if (file == null) {
      if (isAborted) {
        this._listener.onAborted(this);
        this.destruct();
      } else {
        this.skip();
      }
      return;
    }
    
    let location;
    let title;
    
    try {
      location = targetDocument.location.href;
      title = targetDocument.title;
    } catch (e) {
      if (UnMHTDocumentCategory.isAlive(targetDocument)) {
        Cu.reportError(e + "\n" + e.stack);
      }
      this.skip();
      return;
    }
    
    let saveContext
      = UnMHTSave.createContext(this._window, file, location, title,
                                new UnMHTSaveProgress(this._window, this),
                                false);
    
    let arg = {
      targetDocument: targetDocument,
      saveState: this.saveState,
      location: location,
      isHTML: true,
      isStartPart: true
    };
    
    if (!saveContext.addDocument(arg)) {
      saveContext.destruct(); 
      this.skip();
      return;
    }
    
    this._saveContext = saveContext;
    
    this._listener.onStart();
    this._saveContext.start();
  },
  
  /**
   * 単一のファイルとして保存する
   */
  _saveSingle: function() {
    let [file,] = this._getFile(this.filenameBase);
    if (file == null) {
      this._listener.onAborted(this);
      this.destruct();
      return;
    }
    
    let saveContext
      = UnMHTSave.createContext(this._window,
                                file, this.location, this.filenameBase,
                                new UnMHTSaveProgress(this._window, this),
                                false);
    
    let isStartPart = (this.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE_NOINDEX);
    
    let tabs
      = this._targets
      .map(function([targetDocument, id, param], index) {
        if (!UnMHTDocumentCategory.isAlive(targetDocument)) {
          this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_NG);
          return null;
        }
        
        let savedFrom;
        let isHTML;
        try {
          savedFrom = targetDocument.location.href;
          isHTML = UnMHTSaveMulti.isHTML(targetDocument);
        } catch (e) {
          if (UnMHTDocumentCategory.isAlive(targetDocument)) {
            Cu.reportError(e + "\n" + e.stack);
          }
          
          this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_NG);
          return null;
        }
        
        let arg = {
          targetDocument: targetDocument,
          saveState: this.saveState,
          location: savedFrom,
          savedFrom: savedFrom,
          isHTML: isHTML,
          isStartPart: isStartPart,
          doNotStop: true
        };
        
        let part = saveContext.addDocument(arg);
        if (!part) {
          this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_NG);
          return null;
        }
        
        if (!part.sParam.cid) {
          part.sParam.cid
            = "tab." + index + "." + UnMHTPseudoID.gen() + "@unmht.org";
          part.sParam.storeContentID = true;
        }
        
        isStartPart = false;
        
        let title
          = ("title" in param)
          ? param.title
          : (targetDocument.title || targetDocument.location.href);
        
        let desc
          = ("desc" in param)
          ? param.desc
          : "";
        
        this._listener.onStatusChange(this, id, UnMHTSaveMulti.STATUS_OK);
        
        return ["cid:" + part.sParam.cid, title, desc];
      }.bind(this))
      .filter(function(tab) tab);
    
    if (this.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE) {
      let n = Math.floor(Math.random() * 1000000);
      let indexFilename = "unmht-index-" + n + ".html";
      let cssFilename = "unmht-index-" + n + ".css";
      
      saveContext.addDataDocument(
        UnMHTSaveMulti.createIndexDocument(tabs, this._additionalURIs,
                                           cssFilename),
        indexFilename, "UTF-8", "text/html", true);
      
      saveContext.addDataDocument(
        UnMHTSaveMulti.createIndexCSS(true),
        cssFilename, "UTF-8", "text/css", false);
      
      saveContext.setTabs(true);
    }
    
    this._additionalURIs.forEach(function(uri) {
      saveContext.addFile(null, uri, false);
    }.bind(this));
    
    this._saveContext = saveContext;
    this._listener.onStart();
    this._saveContext.start();
  },
  
  /**
   * 保存先のファイルを返す
   *
   * @param   {string} filenameBase
   *          ファイル名の拡張子以外
   * @returns {Array}
   *            [nsIFile 保存先のファイル,
   *             boolean 中断されたか]
   */
  _getFile: function(filenameBase) {
    if (!this.baseDir) {
      this._listener.onError(this, UnMHTSaveMulti.ERROR_ILLEGAL_DIRECTORY);
      
      return [null, true];
    }
    
    if (!this.baseDir.exists()) {
      try {
        arFileUtil.createDirectory(this.baseDir);
      } catch (e) {
        this._listener.onError(this, UnMHTSaveMulti.ERROR_ILLEGAL_DIRECTORY);
        
        return [null, true];
      }
    }
    
    let targetFile
      = arFileUtil.checkAndAppend(this.baseDir, filenameBase,
                                  "", UnMHTFilename.defaultExtension);
    
    if (targetFile.exists()) {
      if (this._dupAction == "ask") {
        let result
          = UnMHTSave.showOverwriteConfirm(
            this._window, targetFile, filenameBase,
            this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED,
            this.saveFormat == UnMHTSaveMulti.FORMAT_SEPARATED);
        
        if (result.result == "skip") {
          if (result.applyToAll) {
            this._dupAction = result.result;
          }
          return [null, false];
        } else if (result.result == "abort" || result.result == "no") {
          return [null, true];
        }
        
        if (result.applyToAll) {
          this._dupAction = result.result;
        }
        
        if (result.result == "rename") {
          targetFile = result.file;
        }
      } else if (this._dupAction == "rename") {
        targetFile = UnMHTSave.getUniqueFile(this.baseDir, filenameBase);
      } else if (this._dupAction == "skip") {
        return [null, false];
      }
    }
    
    return [targetFile, false];
  }
};

/**
 * 複数のタブを MHT で保存
 */
let UnMHTSaveMulti = {
  /**
   * 保存対象の状況
   * @type {number}
   */
  STATUS_NA:      0x0001,
  STATUS_CURRENT: 0x0002,
  STATUS_OK:      0x0003,
  STATUS_NG:      0x0004,
  
  /**
   * 保存形式
   * @type {number}
   */
  FORMAT_SEPARATED:      0x0001,
  FORMAT_SINGLE:         0x0011,
  FORMAT_SINGLE_NOINDEX: 0x0012,
  
  /**
   * エラー
   * @type {number}
   */
  ERROR_ILLEGAL_DIRECTORY: 0x0001,
  
  /**
   * 保存コンテキストを作成する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {IUnMHTSaveMultiListener} listener
   *          イベントリスナ
   * @returns {UnMHTSaveMultiContext}
   *          保存コンテキスト
   */
  createContext: function(window, listener) {
    return new UnMHTSaveMultiContext(window, listener);
  },

  /**
   * インデックスページの内容を生成する
   *
   * @param   {Array} tabs
   *          タブのリスト
   *            [[string URI, string タイトル, string 説明], ...]
   * @param   {Array.<string>} additionalURIs
   *          追加 URI のリスト
   * @param   {string} cssFilename
   *          CSS ファイルのファイル名
   * @returns {string}
   *          インデックスページの内容 (UTF-8)
   */
  createIndexDocument: function(tabs, additionalURIs, cssFilename) {
    let content
      = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
      + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\""
      + "          \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
      + "<html lang=\"ja\" xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\">\n"
      + "<head>\n"
      + "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n"
      + "<title>index</title>\n"
      + "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\" />"
      + "<link rel=\"stylesheet\" type=\"text/css\" href=\""
      + cssFilename
      + "\" />"
      + "</head>\n"
      + "<body>\n"
      + "<ul class=\"unmht-tabs-list\">\n"
      + tabs.map(function([location, title, desc]) 
                 ("<li class=\"unmht-tabs-item\">"
                  + "<a class=\"unmht-tabs\" href=\""
                  + UnMHTHTMLUtil.escapeEntity(location)
                  + "\">"
                  + UnMHTHTMLUtil.escapeEntity(title)
                  + "</a>"
                  + (desc ? ("<div class=\"unmht-tabs-desc\">"
                             + UnMHTHTMLUtil.escapeEntity(desc)
                             + "</div>") : "")
                  + "</li>\n")).join("")
      + additionalURIs.map(function(uri)
                           ("<li class=\"unmht-additional-item\">"
                            + "<a class=\"unmht-additional\" href=\""
                            + UnMHTHTMLUtil.escapeEntity(uri)
                            + "\">"
                            + UnMHTHTMLUtil.escapeEntity(uri)
                            + "</a></li>\n")).join("")
      + "</ul>\n"
      + "</body>\n"
      + "</html>\n";
    
    return arUconv.fromUnicode(content, "UTF-8");
  },
  
  /**
   * インデックスページの CSS を生成する
   *
   * @param   {boolean} useCustom
   *          カスタマイズを使用するか
   * @returns {string}
   *          インデックスページの CSS (UTF-8)
   */
  createIndexCSS: function(useCustom) {
    let content
      = ".unmht-tabs-desc {\n"
      + "  margin-left: 1em;\n"
      + "}\n";
    
    if (useCustom && UnMHTPref.getPref("save.multi.index.style")) {
      content += UnMHTPref.getPref("save.multi.index.style.content");
    }
    
    return arUconv.fromUnicode(content, "UTF-8");
  },
  
  /**
   * HTML かどうかを調べる
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {boolean}
   *          HTML かどうか
   */
  isHTML: function(targetDocument) {
    return (/^text/.test(targetDocument.contentType) ||
            /\/x(ht)?ml(\+.+)$/.test(targetDocument.contentType));
  }
};
