"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");
XPCOMUtils.importRelative(this, "base/arDialog.jsm");
XPCOMUtils.importRelative(this, "base/arFileUtil.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
XPCOMUtils.importRelative(this, "UnMHTSaveMulti.jsm");
XPCOMUtils.importRelative(this, "UnMHTSaveProgress.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTFilePicker", function() {
  XPCOMUtils.importRelative(this, "UnMHTFilePicker.jsm");
  return UnMHTFilePicker;
});

let EXPORTED_SYMBOLS = ["UnMHTSaveMultiDialogClass"];

/**
 * パネルのインデックス
 * @type {number}
 */
let STATUS_BOX_CONTROL  = 0;
let STATUS_BOX_PROGRESS = 1;
let STATUS_BOX_RESULT   = 2;

/**
 * 複数のタブを MHT で保存ダイアログ
 *
 * ウィンドウの引数
 * {
 *   targetWindow: {nsIDOMWindow}
 *                 現在のウィンドウ (必須)
 * }
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialog
 * @implements arIDialog
 * @implements IUnMHTSaveMultiListener
 */
function UnMHTSaveMultiDialogClass(window) {
  arDialog.call(this, window);
}
UnMHTSaveMultiDialogClass.prototype = {
  /* ---- UI イベント - 一般 ---- */
  
  /**
   * テキストフィールドでキーが押されたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTextFieldKeyPress: function(event) {
    if (event.keyCode == 13) {
      this.onDialogAccept();
    }
  },
  
  /**
   * 対象を変更したイベント
   */
  onTargetChange: function() {
    this._updateTargetList();
  },
  
  /**
   * すべてを選択/解除ボタンがクリックされたイベント
   */
  onCheckAllClick: function() {
    if (!this._targetDocuments.length) {
      return;
    }
    
    let save = !this._isSave(0);
    this._targetDocuments
      .forEach((function(d, i) this._changeSave(i, save)).bind(this));
    
    this._updateCount();
    this._updateSingleFilename();
  },
  
  /**
   * インデックスページのタイトルと説明を変更するのチェックを変更したイベント
   */
  onChangeTitleChange: function() {
    this._updateTargetListState();
  },
  
  /**
   * ファイルの URI を追加のチェックを変更したイベント
   */
  onAddFilesChange: function() {
    this.el("add_uris").hidden = !this.el("use_add_uris").checked;
  },
  
  /**
   * ファイル名のドロップダウンメニューを変更したイベント
   */
  onFilenameTypeChange: function() {
    this._updateFilenameBase();
  },
  
  /**
   * 保存形式を変更したイベント
   */
  onFormatChange: function() {
    this.el("changetitle").disabled
      = !(this.el("saveformat").value == "single");
    
    this.el("filename_sequential").disabled
      = !(this.el("saveformat").value == "separated");
    
    this.el("use_add_uris").disabled
      = !(this.el("saveformat").value == "single" ||
          this.el("saveformat").value == "single_noindex");
    
    if (this.el("saveformat").value == "separated") {
      this.el("use_add_uris").checked = false;
      this.onAddFilesChange();
    }
    
    this._updateFilenameBase();
    
    this._updateTargetList();
  },
  
  /**
   * ディレクトリ選択ボタンがクリックされたイベント
   */
  onBaseDirSelectClick: function() {
    let baseDir = arFileUtil.getFile(this.el("basedir").value);
    
    let params = {
      window:           this._window,
      title:            "",
      mode:             Ci.nsIFilePicker.modeGetFolder,
      useLastDir:       false,
      displayDirectory: arFileUtil.getFile(this.el("basedir").value),
      uri:              null,
      callback:         function(dir, result, filePicker) {
        if (!dir) {
          return;
        }
        
        this.el("basedir").value = dir.path;
      }.bind(this)
    };
    UnMHTFilePicker.asyncOpen(params);
  },
  
  /* ---- UI イベント - 保存制御 ---- */
  
  /**
   * スキップボタンをクリックしたイベント
   */
  onSkipClick: function() {
    if (this._savemultiContext) {
      this._savemultiContext.skip();
    }
  },
  
  /**
   * 戻るボタンのイベント
   */
  onBackClick: function() {
    let dialog = this.el("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = false;
    dialog.getButton("cancel").label
      = UnMHTStringBundle.get("save.multi.cancel");
    
    this.el("status_box").selectedIndex = STATUS_BOX_CONTROL;
  },
  
  /* ---- UI イベント - 対象リスト ---- */
  
  /**
   * マウスが出たイベント
   * 対象リストのドラッグを終了する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDialogMouseOut: function(event) {
    if (event.explicitOriginalTarget == this.dialog) {
      this._lastDragNode = null;
    }
  },
  
  /**
   * マウスのボタンを開放したイベント
   * 対象リストのドラッグを終了する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDialogMouseUp: function(event) {
    this._lastDragNode = null;
  },
  
  /**
   * 対象リストが変更されたイベント
   * タイトルの変更をチェックする
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListChange: function(event) {
    if (this.el("filename").value == "title") {
      this._updateSingleFilename();
    }
  },
  
  /**
   * 対象リストでマウスを動かしたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListMouseMove: function(event) {
    let checkNode = this._getCheckNode(event);
    if (!checkNode) {
      return;
    }
    
    this.onTargetListCheckMouseMove(checkNode);
  },
  
  /**
   * 対象リストでマウスを押下したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTargetListMouseDown: function(event) {
    let checkNode = this._getCheckNode(event);
    if (!checkNode) {
      return;
    }
    
    this.onTargetListCheckMouseDown(checkNode);
  },
  
  /**
   * 対象リストで対象のチェック上でマウスを押下したイベント
   * チェックを変更する
   *
   * @param   {nsIDOMXULElement} checkNode
   *          対象の要素
   */
  onTargetListCheckMouseDown: function(checkNode) {
    let save = this._isSaveNode(checkNode);
    
    this._lastDragType = save;
    this._lastDragNode = checkNode;
    
    this._changeSaveNode(checkNode, !save);
    
    this._updateCount();
    this._updateSingleFilename();
  },
  
  /**
   * 対象リストで対象のチェック上でマウスを動かしたイベント
   * ドラッグ時にチェックを変更する
   *
   * @param   {nsIDOMXULElement} checkNode
   *          対象の要素
   */
  onTargetListCheckMouseMove: function(checkNode) {
    if (this._lastDragNode == null) {
      return;
    }
    
    if (this._lastDragNode == checkNode) {
      return;
    }
    
    this._changeSaveNode(checkNode, !this._lastDragType);
    
    this._updateCount();
    this._updateSingleFilename();
  },
  
  /* ---- UI イベント - ファイル名フィールド ---- */
  
  /**
   * ファイル名のフィールドがフォーカスを得たイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onFilenameBaseFocus: function(event) {
    this._lastFilename = this.el("filename_base").value;
  },
  
  /**
   * ファイル名のフィールドがフォーカスを失なったイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onFilenameBaseBlur: function(event) {
    if (this._lastFilename != this.el("filename_base").value) {
      this._isFilenameChanged = true;
    }
  },
  
  /* ==== arIDialog インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    this._targetWindow = null;
    
    this._targetDocuments = null;

    this._lastDragNode = null;
    
    arDialog.prototype.destruct.call(this);
  },
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this._window.arguments[0];
    
    this._targetWindow = arg.targetWindow;
    
    this._targetDocuments = [];
    
    let dialog = this.el("unmht_savemulti_dialog");
    dialog.getButton("accept").setAttribute("accesskey", "s");
    dialog.getButton("accept").label
      = UnMHTStringBundle.get("save.multi.save");
    dialog.getButton("cancel").label
      = UnMHTStringBundle.get("save.multi.cancel");
    
    this.el("basedir").value
      = (UnMHTPref.getPref("save.multi.dir") ||
         arFileUtil.desktopDirectory.path);
    
    this.el("target").value = "this";
    this.el("filename").value = UnMHTPref.getPref("save.filename");
    this.el("saveformat").value = UnMHTPref.getPref("save.multi.format");
    this.el("savestate").value = UnMHTPref.getPref("save.multi.state");
    this.el("use_add_uris").checked = false;
    
    this.onFormatChange();
    this.onAddFilesChange();
    this._updateSingleFilename();
  },
  
  /**
   * ダイアログを OK したイベント
   * 保存を開始する
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    this._startSave();
    
    return false;
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   * 保存中であればキャンセルする
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    if (this._savemultiContext) {
      this._savemultiContext.cancel();
      
      let dialog = this.el("unmht_savemulti_dialog");
      dialog.getButton("accept").disabled = false;
      
      this.el("status_box").selectedIndex = STATUS_BOX_CONTROL;
      
      return false;
    }
    
    return true;
  },

  /* ==== IUnMHTSaveMultiListener インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(context) {
    this._savemultiContext = null;
  },
  
  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   */
  onError: function(context, reason, e) {
    switch (reason) {
      case UnMHTSaveMulti.ERROR_ILLEGAL_DIRECTORY: {
        this.failedLabel1.value
          = UnMHTStringBundle.get("save.error.illegal_directory");
        this.failedLabel1.hidden = false;
        break;
      }
    }
    this.el("skip").disabled = true;
    this._savemultiContext = null;
  },
  
  /**
   * 実際に保存が開始したイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          対象の保存コンテキスト
   */
  onStart: function(context) {
    this.el("skip").disabled = false;
  },
  
  /**
   * 保存状況が変わったイベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   * @param   {ANY} id
   *          保存対象の ID
   * @param   {number} status
   *          状況 UnMHTSaveMulti.STATUS_*
   */
  onStatusChange: function(context, id, status) {
    switch (status) {
      case UnMHTSaveMulti.STATUS_NA: {
        this._changeStatus(id, "");
        break;
      }
      case UnMHTSaveMulti.STATUS_OK: {
        this._changeStatus(id, "o");
        break;
      }
      case UnMHTSaveMulti.STATUS_CURRENT: {
        this._changeStatus(id, "current");
        break;
      }
      case UnMHTSaveMulti.STATUS_NG: {
        this._changeStatus(id, "x");
        break;
      }
    }
  },
  
  /**
   * 全ファイル保存完了イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(context, failedList) {
    this.el("saved_label").value = UnMHTStringBundle.get("save.done");
    this._showResult(failedList, true);
    this._savemultiContext = null;
    
    if (!(failedList &&
          failedList.length > 0 &&
          UnMHTPref.getPref("save.show_errors")) &&
        UnMHTPref.getPref("save.multi.tabs.autoclose")) {
      this._targetDocuments.forEach(function(targetDocument, i) {
        if (!this._getStatus(i)) {
          return;
        }
        
        this._getTargetWindows().forEach(function(window) {
          let tab
            = UnMHTWindowUtil.getTabForDocument(window, targetDocument);
          if (tab) {
            let tabbrowser = UnMHTWindowUtil.getBrowser(window);
            tabbrowser.removeTab(tab);
          }
        });
      }.bind(this));
    }
  },
  
  /**
   * 保存中断イベント
   *
   * @param   {UnMHTSaveMultiContext} context
   *          保存コンテキスト
   */
  onAborted: function(context) {
    this.el("saved_label").value = UnMHTStringBundle.get("save.aborted");
    this._showResult(null, false);
    this._savemultiContext = null;
  },
  
  /* ==== private ==== */
  
  _targetWindow: null,    /** 現在のウィンドウ
                           *  @type {nsIWindow} */
  _targetDocuments: null, /** 対象のドキュメント
                           *  @type {Array.<nsIDOMDocument>} */
  
  _lastDragType: false, /** ドラッグ開始した項目の選択の状態
                         *  @type {boolean} */
  _lastDragNode: null,  /** ドラッグ開始した項目
                         *  @type {nsIDOMXULElement} */
  
  _isFilenameChanged: false, /** ファイル名のフィールドが変更されたか
                              *  @type {boolean} */
  _lastFilename: "",         /** ファイル名のフィールドのフォーカス前の値
                              *  @type {string} */
  
  savemultiContext: null,    /** 保存コンテキスト
                              *  @type {UnMHTSaveMultiContext} */
  
  
  /* ---- 保存関係メソッド ---- */
  
  /**
   * 保存を開始する
   *
   * @returns {boolean}
   *          開始したか
   */
  _startSave: function() {
    if (!this._targetDocuments
        .some((function(d, i)
               this._isSave(i) && UnMHTDocumentCategory.isAlive(d))
              .bind(this))) {
      /* 保存できるものが 1 つも無いので中断 */
      this._targetDocuments
        .filter((function(d) !UnMHTDocumentCategory.isAlive(d)).bind(this))
        .forEach((function(d, i) this._changeStatus(i, "x")).bind(this));
      
      return false;
    }
    
    UnMHTPref.setPref("save.multi.dir",    this.el("basedir").value);
    UnMHTPref.setPref("save.multi.format", this.el("saveformat").value);
    if (UnMHTPref.getPref("save.multi.state.remember")) {
      UnMHTPref.setPref("save.multi.state",  this.el("savestate").value);
    }
    
    this.el("status_box").selectedIndex = STATUS_BOX_PROGRESS;
    
    if (this.el("saveformat").value == "single" ||
        this.el("saveformat").value == "single_noindex") {
      if (this.el("filename_base").value == "") {
        this.el("status_box").selectedIndex = STATUS_BOX_CONTROL;
        this.el("filename_base").focus();
        
        return false;
      }
    }
    
    let dialog = this.el("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = true;
    
    let filenameBase
      = UnMHTFilename.validateFileName(this.el("filename_base").value);
    
    let savemultiContext
      = UnMHTSaveMulti.createContext(this._window, this);
    
    let isSingle = false;
    switch (this.el("saveformat").value) {
      case "separated": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SEPARATED;
        break;
      }
      case "single": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SINGLE;
        isSingle = true;
        break;
      }
      case "single_noindex": {
        savemultiContext.saveFormat = UnMHTSaveMulti.FORMAT_SINGLE_NOINDEX;
        isSingle = true;
        break;
      }
    }
    
    savemultiContext.baseDir
      = arFileUtil.getFile(this.el("basedir").value);
    savemultiContext.saveState
      = UnMHTSave.stringToSaveState(this.el("savestate").value);
    savemultiContext.filenameBase = filenameBase;
    
    if (isSingle) {
      UnMHTSave.parseAdditionalURIs(this.el("add_uris").value)
        .forEach(function(uri) {
          savemultiContext.addFile(uri);
        });
    }
    
    let isChangeTitle
      = (savemultiContext.saveFormat == UnMHTSaveMulti.FORMAT_SINGLE &&
         this.el("changetitle").checked);
    
    let first = true;
    this._targetDocuments.forEach(function(targetDocument, i) {
      if (!this._isSave(i)) {
        return;
      }
      
      if (!UnMHTDocumentCategory.isAlive(targetDocument)) {
        return;
      }
      
      let param = {};
      
      if (isSingle) {
        if (first) {
          first = false;
          savemultiContext.location = targetDocument.location.href;
        }
      } else {
        if (this.el("filename").value == "leafname" ||
            this.el("filename").value == "title" ||
            this.el("filename").value == "custom") {
          param.filenameBase = this._getFilenameBase(targetDocument);
        } else if (this.el("filename").value == "sequential") {
          param.filenameBase = filenameBase + "-" + (i + 1);
        }
      }
      
      if (isChangeTitle) {
        param.title = this._getTitle(i);
        param.desc = this._getDesc(i);
      }
      
      savemultiContext.addTarget(targetDocument, i, param);
    }.bind(this));
    
    this._savemultiContext = savemultiContext;
    
    this._savemultiContext.start();
    
    return true;
  },
  
  /**
   * 保存結果を表示する
   *
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   * @param   {boolean} isClosable
   *          閉じてもよいか
   */
  _showResult: function(failedList, isClosable) {
    let dialog = this.el("unmht_savemulti_dialog");
    dialog.getButton("accept").disabled = true;
    dialog.getButton("cancel").label
      = UnMHTStringBundle.get("save.multi.close");
    dialog.getButton("cancel").focus ();
    
    this.el("status_box").selectedIndex = STATUS_BOX_RESULT;
    
    if (failedList &&
        failedList.length > 0 &&
        UnMHTPref.getPref("save.show_errors")) {
      let progress = new UnMHTSaveProgress(this._window, this);
      progress.showFail(failedList);
      progress.destruct();
    } else if (isClosable && UnMHTPref.getPref("save.multi.autoclose")) {
      arTimer.setTimeout(function() {
        this._window.close();
      }.bind(this), 1000);
    }
  },
  
  /**
   * 保存可能かどうかを調べる
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @param   {boolean} asSingle
   *          単一のファイルとして保存するかどうか
   * @returns {boolean}
   *          保存可能かどうか
   */
  _isSavable: function(targetDocument, asSingle) {
    return UnMHTDocumentCategory.isSavable(null, targetDocument, asSingle);
  },
  
  /**
   * 保存する際のファイル名を取得する
   *
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {string}
   *          ファイル名 (拡張子以外)
   */
  _getFilenameBase: function(targetDocument) {
    return UnMHTSave.getFilenameBase(targetDocument, null,
                                     this.el("filename").value, new Date());
  },
  
  /* ---- 対象リスト関連メソッド ---- */
  
  /**
   * 対象リストの選択している数を更新する
   */
  _updateCount: function() {
    this.el("count_label").value
      = this._targetDocuments
      .filter((function(d, i) this._isSave(i)).bind(this))
      .length
      + "/" + this._targetDocuments.length;
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {boolean}
   *          対象のタブが保存対象かどうか
   */
  _isSave: function(index) {
    return this._isSaveNode(this.el("savecheck_" + index));
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @returns {boolean}
   *          対象のタブが保存対象かどうか
   */
  _isSaveNode: function(node) {
    return node.getAttribute("value") == "save";
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {boolean} save
   *          対象のタブが保存対象かどうか
   */
  _changeSave: function(index, save) {
    return this._changeSaveNode(this.el("savecheck_" + index), save);
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @param   {boolean} save
   *          対象のタブが保存対象かどうか
   */
  _changeSaveNode: function(node, save) {
    node.setAttribute("value", save ? "save" : "ignore");
    node.setAttribute("src",
                      save
                      ? "resource://unmht/images/check_v.png"
                      : "resource://unmht/images/check_dot.png");
  },
  
  /**
   * 対象のタブの保存状況を取得する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {boolean} status
   *          対象のタブの保存に成功したか
   */
  _getStatus: function(index) {
    return this._getStatusNode(this.el("savestatus_" + index));
  },
  
  /**
   * 対象のタブの保存状況を取得する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @returns {boolean} status
   *          対象のタブの保存に成功したか
   */
  _getStatusNode: function(node) {
    return node.getAttribute("value") == "saved";
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {boolean} status
   *          対象のタブの保存に成功したか
   */
  _changeStatus: function(index, status) {
    return this._changeStatusNode(this.el("savestatus_" + index), status);
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param   {nsIDOMXULElement} node
   *          対象リストのチェックボックス
   * @param   {boolean} status
   *          対象のタブの保存に成功したか
   */
  _changeStatusNode: function(node, status) {
    node.setAttribute("src",
                      status ?
                      ("resource://unmht/images/check_" + status + ".png")
                      : "");
    node.setAttribute("value", status ? "saved" : "notsaved");
  },
  
  /**
   * イベントの対象のチェックボックスを取得する
   *
   * @param   {Event} event
   *          対象のイベント
   * @returns {nsIDOMXULElement}
   *          イベントの対象のチェックボックス
   */
  _getCheckNode: function(event) {
    let row = {}, col = {}, child = {};
    
    this.el("target_list").treeBoxObject
      .getCellAt(event.clientX, event.clientY, row, col, child);    
    
    if (row.value == -1) {
      return null;
    }
    
    if (col.value == null) {
      return null;
    }
    
    if (col.value.index != 0) {
      return null;
    }
    
    let listbox = this.el("target_list_children");
    
    return listbox.childNodes[row.value].firstChild.firstChild;
  },
  
  /**
   * 対象リストを空にする
   */
  _clearTargetList: function() {
    let listbox = this.el("target_list_children");
    while (listbox.firstChild) {
      listbox.removeChild(listbox.firstChild);
    }
  },
  
  /**
   * 対象リストのアイテムを追加する
   * 
   * @param   {number} index
   *          対象のタブのインデックス
   * @param   {string} title
   *          タブのタイトル
   */
  _addTargetListItem: function(index, title) {
    let listbox = this.el("target_list_children");
    
    let listitem = this._window.document.createElement("treeitem");
    listbox.appendChild(listitem);
    
    let listrow = this._window.document.createElement("treerow");
    listitem.appendChild(listrow);
    
    let checkNode = this._window.document.createElement("treecell");
    listrow.appendChild(checkNode);
    checkNode.id = "savecheck_" + index;
    checkNode.className = "save-iconic";
    checkNode.setAttribute("editable", "false");
    this._changeSaveNode(checkNode,
                         !UnMHTPref.getPref("save.multi.check_none"));
    
    let statusNode = this._window.document.createElement("treecell");
    listrow.appendChild(statusNode);
    statusNode.id = "savestatus_" + index;
    statusNode.setAttribute("editable", "false");
    statusNode.className = "save-iconic";
    
    let titleNode = this._window.document.createElement("treecell");
    listrow.appendChild(titleNode);
    titleNode.id = "savetitle_" + index;
    titleNode.setAttribute("editable", "true");
    titleNode.setAttribute("label", title);
    titleNode.setAttribute("value", title);
    titleNode.flex = 1;
    
    let descNode = this._window.document.createElement("treecell");
    listrow.appendChild(descNode);
    descNode.id = "savedesc_" + index;
    descNode.setAttribute("editable", "true");
    descNode.flex = 1;
    
    this._updateTargetListItemState(index);
  },
  
  /**
   * 対象リストのアイテムの状態を更新する
   *
   * @param   {number} index
   *          対象のアイテムのインデックス
   */
  _updateTargetListItemState: function(index) {
    let titleNode = this.el("savetitle_" + index);
    let descNode = this.el("savedesc_" + index);
    
    let isEditable
      = (this.el("saveformat").value == "single" &&
         this.el("changetitle").checked);
    
    if (!isEditable) {
      titleNode.setAttribute("label", titleNode.getAttribute("value"));
    }
    
    titleNode.setAttribute("editable", isEditable);
    descNode.setAttribute("editable", isEditable);
    descNode.hidden = !isEditable;
  },

  /**
   * 対象のドキュメントリストを更新する
   */
  _updateTargetList: function() {
    let asSingle
      = (this.el("saveformat").value == "single" ||
         this.el("saveformat").value == "single_noindex");
    
    this._targetDocuments = [];
    this._getTargetWindows().forEach(function(window) {
      UnMHTWindowUtil.getDocuments(window)
        .forEach(function(targetDocument) {
          if (this._isSavable(targetDocument, asSingle)) {
            this._targetDocuments.push(targetDocument);
          }
        }.bind(this));
    }.bind(this));
    
    this._clearTargetList();
    this._targetDocuments.forEach(function(targetDocument, i) {
      this._addTargetListItem(i,
                              targetDocument.title ||
                              targetDocument.location.href);
    }.bind(this));
    
    this._updateTargetListState();
    this._updateCount();
  },
  
  /**
   * 対象リストの状態を更新する
   */
  _updateTargetListState: function() {
    this._targetDocuments
      .forEach((function(d, i) this._updateTargetListItemState(i)).bind(this));
    
    this.el("target_list_desc").hidden
      = !(this.el("saveformat").value == "single" &&
          this.el("changetitle").checked);
    
    this._updateSingleFilename();
  },
  
  /**
   * 対象になるウィンドウのリストを返す
   *
   * @returns {Array.<nsIDOMWindow>}
   */
  _getTargetWindows: function() {
    if (this.el("target").value == "this") {
      return [this._targetWindow];
    } else {
      return [window for (window of XPCOMUtils.IterSimpleEnumerator(
        Services.wm.getEnumerator("navigator:browser"),
        Ci.nsIDOMWindow))];
    }
  },
  
  /**
   * 対象のタブのタイトルを取得する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {string}
   *          対象のタブのタイトル
   */
  _getTitle: function(index) {
    return this.el("savetitle_" + index).getAttribute("label");
  },
  
  /**
   * 対象のタブのタイトルを説明する
   *
   * @param   {number} index
   *          対象のタブのインデックス
   * @returns {string}
   *          対象のタブの説明
   */
  _getDesc: function(index) {
    return this.el("savedesc_" + index).getAttribute("label");
  },
  
  /* ---- ファイル名フィールド関連メソッド ---- */
  
  /**
   * 必要であれば単一のファイルのファイル名を更新する
   */
  _updateSingleFilename: function() {
    if (!(this.el("saveformat").value == "single" ||
          this.el("saveformat").value == "single_noindex")) {
      return;
    }
    
    if (this._isFilenameChanged && this.el("filename_base").value != "") {
      return;
    }
    
    let [targetDocument, i]
      = (arArrayUtil
         .find(this._targetDocuments.map(function(d, i) [d, i]),
               (function([d, i])
                this._isSave(i) && UnMHTDocumentCategory.isAlive(d)).bind(this)) ||
         [null, -1]);
    if (!targetDocument) {
      return;
    }
    
    let filenameBase = "";
    if (this.el("changetitle").checked &&
        this.el("filename").value == "title") {
      filenameBase = this._getTitle(i);
    } else {
      filenameBase = this._getFilenameBase(targetDocument);
    }
    
    this.el("filename_base").value = filenameBase;
    this._lastFilename = this.el("filename_base").value;
    this._isFilenameChanged = false;
  },
  
  /**
   * ファイル名フィールドの状態を更新する
   */
  _updateFilenameBase: function() {
    this.el("filename_base").disabled
      = !(this.el("filename").value == "sequential" ||
          this.el("saveformat").value == "single" ||
          this.el("saveformat").value == "single_noindex");
    
    if (this.el("filename").value == "sequential") {
      this.el("filename_base")
        .setAttribute("placeholder",
                      UnMHTStringBundle.get("save.multi.sequential.placeholder"));
      this.el("filename_base").value = "";
      this._lastFilename = this.el("filename_base").value;
      this._isFilenameChanged = false;
    } else {
      this.el("filename_base").setAttribute("placeholder", "");
    }
    
    this._updateSingleFilename();
  }
};
arPrototype.extend(UnMHTSaveMultiDialogClass, arDialog);
