"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arDialog.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "UnMHTFilename.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveOptionsDialogClass"];

/**
 * 保存オプション
 *
 * ウィンドウの引数
 * {
 *   showFilename:   {boolean}
 *                   ファイル名を表示し、修正可能にするか (必須)
 *   filenameBase:   {string}
 *                   ファイル名 (拡張子以外) (必須)
 *
 *   saveStateMask:  {number}
 *                   リスト表示する保存状態 (必須)
 *   saveState:      {number}
 *                   デフォルトの保存状態 (必須)
 *   saveStatePref:  {string}
 *                   選択された保存状態を格納する設定名 (必須)
 *   showInsertLink: {boolean}
 *                   元ページへのリンクを挿入を表示し、修正可能にするか (必須)
 *
 *   callback:       {Function}
 *                   終了時コールバック (オプショナル)
 * }
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arDialog
 * @implements arIDialog
 */
function UnMHTSaveOptionsDialogClass(window) {
  arDialog.call(this, window);
}
UnMHTSaveOptionsDialogClass.prototype = {
  /* ==== arIDialog インターフェース ==== */
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    let arg = this._window.arguments[0];
    
    this._showFilename = arg.showFilename;
    this._filenameBase = arg.filenameBase;
    if (this._showFilename) {
      this.el("filename").value
        = this._filenameBase + "." + UnMHTFilename.defaultExtension;
      this.el("filename_box").hidden = false;
    }
    
    UnMHTSave.getSaveStateList(arg.saveStateMask)
      .map(function(state) UnMHTSave.saveStateToString(state))
      .forEach(function(stateStr) {
        this.el("savestate_" + stateStr).hidden = false;
      }.bind(this));
    this.el("savestate").value
      = UnMHTSave.saveStateToString(arg.saveState);
    
    this._saveStatePref = arg.saveStatePref;
    
    if ("callback" in arg) {
      this._callback = arg.callback;
    }
    
    this._showInsertLink = arg.showInsertLink;
    if (this._showInsertLink) {
      this.el("save_insertlink").checked
        = UnMHTPref.getPref("save.insert.link");
    } else {
      this.el("save_insertlink").hidden = true;
    }
  },
  
  /**
   * ダイアログを OK したイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogAccept: function() {
    if (UnMHTPref.getPref(this._saveStatePref + ".remember")) {
      UnMHTPref.setPref(this._saveStatePref, this.el("savestate").value);
    }
    if (this._showInsertLink) {
      UnMHTPref.setPref("save.insert.link",
                        this.el("save_insertlink").checked);
    }
    
    let resultParam = {
      filenameBase:   (this._showFilename
                       ? this.el("filename").value.replace(/\.mht/i, "")
                       : this._filenameBase),
      saveState:      UnMHTSave.stringToSaveState(this.el("savestate").value),
      additionalURIs: (UnMHTSave.parseAdditionalURIs(this.el("add_uris").value))
    };
    
    if (this._callback) {
      let callback = this._callback;
      arTimer.setTimeout(function() {
        callback(UnMHTSave.RESULT_OK, resultParam);
      }.bind(this), 100);
    }
    
    return true;
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   *
   * @returns {boolean}
   *          ダイアログを閉じるか
   */
  onDialogCancel: function() {
    if (this._callback) {
      let callback = this._callback;
      arTimer.setTimeout(function() {
        callback(UnMHTSave.RESULT_CANCEL, null);
      }.bind(this), 100);
    }
    
    return true;
  },
  
  /* ==== private ==== */
  
  _showFilename: false,   /** ファイル名を表示し、修正可能にするか
                           *  @type {boolean} */
  _filenameBase: "",      /** ファイル名 (拡張子以外)
                           *  @type {string} */
  
  _saveStatePref: "",     /** 選択された保存状態を格納する設定名
                           *  @type {string} */
  
  _showInsertLink: false, /** 元ページへのリンクを挿入を表示し、
                           *  修正可能にするか
                           *  @type {boolean} */
  
  _callback: null         /** 終了時コールバック
                           *  @type {Function} */
};
arPrototype.extend(UnMHTSaveOptionsDialogClass, arDialog);
