"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveProgress"];

/**
 * 保存の進捗状況表示
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @param   {UnMHTSaveProgressOwner} owner
 *          オーナー
 * @implements IUnMHTSaveListener
 */
function UnMHTSaveProgress(window, owner) {
  this._window = window;
  this._owner = owner;
  
  this._progressLabel = this.el("progress_label");
  this._progressLabel.value = UnMHTStringBundle.get("save.begin");
  this._progressCurrent = this.el("progress_current");
  this._progressCurrentSize = this.el("progress_current_size");
  this._progress = this.el("progress");
  this._progress.value = 0;
  
  this._failedLabel1 = this.el("failed_label1");
  this._failedLabel2 = this.el("failed_label2");
  this._failedLabel3 = this.el("failed_label3");
  this._failedList = this.el("failed");
  
  this._failedLabel1.hidden = true;
  this._failedLabel2.hidden = true;
  this._failedLabel3.hidden = true;
  this._failedList.hidden = true;
}
UnMHTSaveProgress.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    
    this._progressLabel = null;
    this._progressCurrent = null;
    this._progressCurrentSize = null;
    this._progress = null;
    
    this._failedLabel1 = null;
    this._failedLabel2 = null;
    this._failedLabel3 = null;
    this._failedList = null;
    
    this._owner = null;
  },
  
  /* ==== IUnMHTSaveListener インターフェース ===== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(context) {
    this._owner.onCancel(this, context);
  },
  
  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {number} reason
   *          エラーの理由
   * @param   {Exception} e
   *          例外
   */
  onError: function(context, reason, e) {
    switch (reason) {
      case UnMHTSave.ERROR_WRITE_FILE: {
        this._progressLabel.value = UnMHTStringBundle.get("save.error.open");
        break;
      }
    }
    
    this._progressCurrent.value = e ? e.toString() : "";
    
    this._owner.onError(this, context, reason, e);
  },

  /**
   * 進捗状況変化イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {string} currentFile
   *          現在取得のファイル名
   * @param   {string} currentSize
   *          取得したバイト数 / ファイルサイズ
   * @param   {number} ok
   *          取得完了した数
   * @param   {number} na
   *          取得中の数
   * @param   {number} ng
   *          取得失敗した数
   */
  onProgress: function(context, currentFile, currentSize, ok, na, ng) {
    this._progressCurrent.value = currentFile;
    this._progressCurrentSize.value = currentSize;
    
    this._progressLabel.value
      = UnMHTStringBundle.get("save.begin")
      + " " + (ok + ng) + "/" + (ok + na + ng);
    
    if (context.progress == 1) {
      this._progress.mode = "undetermined";
    } else {
      this._progress.mode = "determined";
      this._progress.value = Math.floor(context.progress * 100);
    }
  },
  
  /**
   * 保存完了
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(context, failedList) {
    this._progress.mode = "determined";
    this._progress.value = 100;
    
    this._progressLabel.value = UnMHTStringBundle.get("save.done");
    
    this._owner.onSave(this, context, failedList);
  },
  
  /**
   * 保存コンテキストの破棄
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onDestruct: function(context) {
    this.destruct();
  },
  
  /* ==== private ==== */

  _window: null, /** 対象のウィンドウ
                  *  @type {nsIDOMWindow} */
  
  _owner: null,  /** オーナー
                  *  @type {nsIDOMWindow} */
  
  _progressLabel: null,       /** 取得状況のラベル
                               *  @type {nsIDOMXULElement} */
  _progressCurrent: null,     /** 現在取得しているファイル
                               *  @type {nsIDOMXULElement} */
  _progressCurrentSize: null, /** 現在取得しているファイルのサイズ
                               *  @type {nsIDOMXULElement} */
  _progress: null,            /** 取得状況のバー
                               *  @type {nsIDOMXULElement} */
  
  _failedLabel1: null,        /** 失敗したファイルのラベル
                               *  @type {nsIDOMXULElement} */
  _failedLabel2: null,        /** 失敗したファイルのラベル2
                               *  @type {nsIDOMXULElement} */
  _failedLabel3: null,        /** 失敗したファイルのラベル3
                               *  @type {nsIDOMXULElement} */
  _failedList: null,          /** 失敗したファイル
                               *  @type {nsIDOMXULElement} */
  
  /**
   * 保存結果を表示する
   */
  showFail: function(failedList) {
    if (UnMHTPref.getPref("save.show_errors") && failedList.length > 0) {
      this._failedList.value = failedList.join("\n");
      this._failedList.hidden = false;
      
      if (this._failedLabel1 && this._failedLabel1.value != "") {
        this._failedLabel1.hidden = false;
      }
      if (this._failedLabel2 && this._failedLabel2.value != "") {
        this._failedLabel2.hidden = false;
      }
      if (this._failedLabel3 && this._failedLabel3.value != "") {
        this._failedLabel3.hidden = false;
      }
      
      this._window.sizeToContent();
    }
  },
  
  /**
   * getElementById の省略
   *
   * @param   {string} id
   *          要素の ID
   * @returns {nsIDOMElement}
   *          id に対応する要素
   *          見付からなければ null
   */
  el: function(id) {
    return this._window.document.getElementById(id);
  }
};
