"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

let EXPORTED_SYMBOLS = ["UnMHTScheme"];

/**
 * URI の情報
 */
function UnMHTURIParam() {
}
UnMHTURIParam.prototype = {
  original: "",   /** 本来の URI
                   *  @type {string} */
  scheme: "",     /** 本来の URI の scheme
                   *  @type {string} */
  fragment: "",   /** 本来の URI のフラグメント
                   *  @type {string} */
  
  refName: "",    /** 参照名
                   *  @type {string} */
  baseURI: ""     /** 開始パートの URI
                   *  @type {string} */
};
/**
 * unmht: スキーム操作
 */
let UnMHTScheme = {
  /**
   * unmht: スキームの URI にする
   *
   * @param   {string} uri
   *          URI
   * @returns {string}
   *          unmht: スキームの URI
   */
  enUnMHTURI: function(uri) {
    let re = uri.match(/^([A-Za-z0-9\-]+):(\/\/)?([^\/]*)(\/\/)?(\/)?(.*)$/);
    if (re) {
      let [, scheme, sep1, host, sep2, sep3, path] = re;
      
      let sep = 0;
      if (sep1) {
        sep |= 1;
      }
      if (sep2) {
        sep |= 2;
      }
      if (sep3) {
        sep |= 4;
      }
      
      if (scheme.match(/^data$/i) && sep == 4) {
        path = host + "/" + path;
        host = "";
        sep = 0;
      }
      
      let port = "";
      re = host.match(/^([^:]+):([0-9]+)$/);
      if (re) {
        host = re[1];
        port = re[2];
      }
      
      uri
        = "unmht://" + host + (host ? "." : "") + "unmht"
        + (port ? (":" + port) : "")
        + "/"
        + scheme + "." + sep + "/" + path + "/";
    }
    
    return uri;
  },
  
  /**
   * unmht: スキームの URI かどうかを返す
   *
   * @param   {string} uri
   *          URI
   * @returns {boolean}
   *          unmht: スキームの URI かどうか
   */
  isUnMHTURI: function(uri) {
    let param = UnMHTScheme.getUnMHTURIParam(uri);
    if ("original" in param &&
        param.original) {
      return true;
    }
    
    return false;
  },
  
  /**
   * unmht: スキームの URI を元に戻す
   *
   * @param   {string} uri
   *          unmht: スキームの URI
   * @returns {string}
   *          URI
   */
  deUnMHTURI: function(uri) {
    let param = UnMHTScheme.getUnMHTURIParam(uri);
    if ("original" in param &&
        param.original) {
      return param.original;
    }
    
    return uri;
  },
  
  /**
   * unmht: スキームの URI の情報を取得する
   *
   * @param   {string} uri
   *          unmht: スキームの URI
   * @returns {UnMHTURIParam}
   *          URI の情報
   */
  getUnMHTURIParam: function(uri) {
    let param = new UnMHTURIParam();
    
    let re = uri.match(/^unmht:\/\/(?:([^\/]+)\.)?unmht(?::([0-9]+))?\/([A-Za-z0-9\-]+)\.([0-9]+)\/(.*)\/([^\/#]*)(?:#(.*))?$/);
    if (!re) {
      return param;
    }
    
    let idn
      = Cc["@mozilla.org/network/idn-service;1"]
      .getService(Ci.nsIIDNService);
    
    param.host     = re[1] ? (idn.convertUTF8toACE(re[1])) : "";
    param.port     = re[2];
    param.scheme   = re[3].toLowerCase();
    let sep        = parseInt(re[4], 10);
    param.path     = re[5];
    param.refName  = re[6];
    param.fragment = re[7];
    
    let sep1 = (sep & 1) ? "//" : "";
    let sep2 = (sep & 2) ? "//" : "";
    let sep3 = (sep & 4) ? "/" : "";
    
    param.original
      = param.scheme + ":" + sep1 + param.host
      + (param.port ? (":" + param.port) : "")
      + sep2 + sep3
      + param.path;
    
    param.baseURI
      = "unmht://" + param.host + (param.host ? "." : "") + "unmht"
      + (param.port ? (":" + param.port) : "")
      + "/"
      + param.scheme + "." + sep + "/" + param.path + "/";
    
    return param;
  },
  
  /**
   * Ver.5.7.x 以前の URI をデコードする
   *
   * @param   {string} uri
   *          対象の URI
   * @returns {string}
   *          デコードした URI
   *          デコードできなければ null
   */
  deUnMHTURI570: function(uri) {
    let re = uri.match(/^unmht:\/\/(?:([^\/]+))?(?::([0-9]+))?\/([A-Za-z0-9\-]+)\.([0-9]+)\/(.*)\/([^\/#]*)(?:#(.*))?$/);
    if (!re) {
      return null;
    }
    
    let idn
      = Cc["@mozilla.org/network/idn-service;1"]
      .getService(Ci.nsIIDNService);
    
    let host     = re[1] ? (idn.convertUTF8toACE(re[1])) : "";
    let port     = re[2];
    let scheme   = re[3].toLowerCase();
    let sep        = parseInt(re[4], 10);
    let path     = re[5];
    let refName  = re[6];
    let fragment = re[7];
    
    let sep1 = (sep & 1) ? "//" : "";
    let sep2 = (sep & 2) ? "//" : "";
    let sep3 = (sep & 4) ? "/" : "";
    
    let original
      = scheme + ":" + sep1 + host
      + (port ? (":" + port) : "")
      + sep2 + sep3
      + path;
    
    return original;
  },
  
  /**
   * エラーページの URI を返す
   *
   * @param   {string} uri
   *          URI
   * @returns {string}
   *          エラーページの URI
   */
  enErrorURI: function(uri) {
    uri = this.enUnMHTURI(uri);
    uri = uri.replace(/^unmht:/, "error:");
    uri = this.enUnMHTURI(uri);
    
    return uri;
  },
  
  /**
   * エラーページの URI を元に戻す
   *
   * @param   {string} uri
   *          エラーページの URI
   * @returns {string}
   *          URI
   */
  deErrorURI: function(uri) {
    uri = this.deUnMHTURI(uri);
    uri = uri.replace(/^error:/, "unmht:");
    uri = this.deUnMHTURI(uri);
    
    return uri;
  },
  
  /**
   * 空のページの URI を返す
   *
   * @returns {string}
   *          空のページの URI
   */
  createEmptyURI: function() {
    return this.enUnMHTURI("empty:");
  },
  
  /**
   * キャッシュページの URI を返す
   *
   * @returns {string}
   *          キャッシュページの URI
   */
  createCacheURI: function() {
    return this.enUnMHTURI("cache:");
  }
};
