"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTFileSaver", function() {
  XPCOMUtils.importRelative(this, "UnMHTFileSaver.jsm");
  return UnMHTFileSaver;
});

let EXPORTED_SYMBOLS = ["UnMHTScrapBookIntegration"];

/**
 * ScrapBook サイトバー組み込み
 *
 * @param   {nsIDocument} sidebarDocument
 *          対象のサイドバーのドキュメント
 * @param   {UnMHTScrapBookIntegration} parent
 *          生成者
 */
function UnMHTScrapBookSidebarIntegration(sidebarDocument, parent) {
  this._sidebarDocument = sidebarDocument;
  this._parent = parent;
  
  this.onPopupShowing = this.onPopupShowing.bind(this);
  this.onSave = this.onSave.bind(this);
  this.onUnload = this.onUnload.bind(this);
}
UnMHTScrapBookSidebarIntegration.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._popup = null;
    this._menuitem = null;
    this._sidebarDocument = null;
    
    this._scrapBookData = null;
    this._parent = null;
  },
  
  /**
   * ロードイベント
   */
  onLoad: function() {
    this._sidebarDocument.addEventListener("unload", this.onUnload, false);
    
    this._createMenuItem();
  },
  
  /**
   * アンロードイベント
   */
  onUnload: function() {
    if (this._popup) {
      try {
        this._popup.removeEventListener("popupshowing", this.onPopupShowing, false);
      } catch (e) {
      }
      this._popup = null;
    }
    
    if (this._menuitem) {
      try {
        this._menuitem.removeEventListener("command", this.onSave, false);
        this._menuitem.parentNode.removeChild(this._menuitem);
      } catch (e) {
      }
      this._menuitem = null;
    }
    
    if (this._sidebarDocument) {
      try {
        this._sidebarDocument.removeEventListener("unload", this.onUnload, false);
      } catch (e) {
      }
      this._sidebarDocument = null;
    }
    
    this._parent.removeIntegration(this);
  },
  
  /**
   * ScrapBook のポップアップを設定する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onPopupShowing: function(event) {
    let sidebarWindow = this._sidebarDocument.defaultView;
    
    let sbData;
    let sbUtils;
    let sbResource;
    if ("ScrapBookData" in sidebarWindow) {
      sbData = sidebarWindow.ScrapBookData;
    } else if ("sbDataSource" in sidebarWindow) {
      sbData = sidebarWindow.sbDataSource;
    } else {
      return;
    }
    if ("ScrapBookUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.ScrapBookUtils;
    } else if ("sbCommonUtils" in sidebarWindow) {
      sbUtils = sidebarWindow.sbCommonUtils;
    } else {
      return;
    }
    if ("sbTreeUI" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeUI.resource;
    } else if ("sbTreeHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbTreeHandler.resource;
    } else if ("sbListHandler" in sidebarWindow) {
      sbResource = sidebarWindow.sbListHandler.resource;
    } else {
      return;
    }
    
    let id = sbData.getProperty(sbResource, "id");
    
    let scrapBookData = {
      source:  sbData.getProperty(sbResource, "source"),
      title:   sbData.getProperty(sbResource, "title"),
      id:      id,
      dir:     sbUtils.getContentDir(id),
      charset: sbData.getProperty(sbResource, "chars")
    };
    
    if (scrapBookData["source"]) {
      this._menuitem.hidden = false;
    } else {
      this._menuitem.hidden = true;
    }
    
    this._scrapBookData = scrapBookData;
  },
  
  /**
   * 保存
   */
  onSave: function() {
    if (!this._scrapBookData) {
      return;
    }
    
    let window = this._sidebarDocument.defaultView;
    if (window.parent) {
      window = window.parent;
    }
    
    UnMHTFileSaver.saveScrapBook(window, this._scrapBookData);
  },
  
  /* ==== private ==== */
  
  _sidebarDocument: null, /* nsIDocument  対象のサイドバーのドキュメント */
  _popup: null,           /* nsIDOMXULElement  ScrapBook のポップアップ */
  _menuitem: null,        /* nsIDOMXULElement  追加したメニューアイテム */
  _scrapBookData: null,   /* Object  ScrapBook のデータ */
  
  _parent: null,          /* UnMHTScrapBookIntegration  ScrapBook 組み込み */
  
  /**
   * メニューアイテムを作成する
   */
  _createMenuItem: function() {
    let popup = this._sidebarDocument.getElementById("sbPopup");
    let pos = this._sidebarDocument.getElementById("sbPopupTools");
    
    if (popup && pos) {
      let item = this._sidebarDocument.createElement("menuitem");
      item.id = "unmht-scrapbook-save";
      item.className = "menuitem-iconic";
      item.setAttribute("label", UnMHTStringBundle.get("save.title"));
      item.setAttribute("image", "resource://unmht/images/icon_small.png");
      item.addEventListener("command", this.onSave, false);
      pos.parentNode.insertBefore(item, pos);
      
      popup.addEventListener("popupshowing", this.onPopupShowing, false);
      
      this._popup = popup;
      this._menuitem = item;
    }
  }
};

/**
 * ScrapBook 組み込み
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 */
function UnMHTScrapBookIntegration(window) {
  arWindowHook.call(this, window);
  
  this._integs = [];

  this.onLoadSidebar = this.onLoadSidebar.bind(this);
}
UnMHTScrapBookIntegration.prototype = {
  /**
   * サイドバーの内容が ScrapBook かどうかチェック
   * ScrapBook ならばメニューの項目を追加する
   *
   * @param   {Event} event
   *          対象のイベント
   * @param   {nsIDocument} sidebarDocument
   *          対象のサイドバーのドキュメント
   */
  onLoadSidebar: function(event, sidebarDocument) {
    if (!sidebarDocument) {
      sidebarDocument = event.target.defaultView.document;
    }
    
    if (sidebarDocument.documentURI
        != "chrome://scrapbook/content/scrapbook.xul") {
      return;
    }
    
    let integ = new UnMHTScrapBookSidebarIntegration(sidebarDocument, this);
    integ.onLoad();
    this._integs.push(integ);
  },
  
  /**
   * 組み込みを解除する
   *
   * @param   {UnMHTScrapBookIntegration} integration
   *          解除する組み込み
   */
  removeIntegration: function(integration) {
    this._integs
      = this._integs
      .filter(function(integ) {
        if (integ == integration) {
          integ.destruct();
          return false;
        }
        return true;
      });
  },

  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    let sidebar = this._window.document.getElementById("sidebar");
    if (sidebar) {
      sidebar.addEventListener("DOMContentLoaded", this.onLoadSidebar, false);
      
      this.onLoadSidebar(null, sidebar.contentDocument);
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    let sidebar = this._window.document.getElementById("sidebar");
    if (sidebar) {
      sidebar.removeEventListener("DOMContentLoaded",
                                  this.onLoadSidebar, false);
    }
    
    this._integs
      .forEach(function(integ) {
        integ.onUnload();
      });
    this._integs = [];
    
    this._integs = null;
  },
  
  /* ==== private ==== */

  _integs: null  /** 組み込みリスト
                  *  @type {Array.<UnMHTScrapBookSidebarIntegration>} */
};
arPrototype.extend(UnMHTScrapBookIntegration, arWindowHook);
