"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "UnMHTDLFRegisterer.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");

let EXPORTED_SYMBOLS = ["UnMHTStartupDLFRegisterer"];

/**
 * 起動時に Document Loader Factory を登録する
 *
 * @implements nsIObserver
 * @implements arIStartupEventListener
 * @implements arIShutdownEventListener
 */
let UnMHTStartupDLFRegisterer = {
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIObserver]),
  
  /* ==== nsIObserver インターフェース ==== */
  
  /**
   * 通知を受けたイベント
   *
   * @param   {nsISupports} subject
   *          変更やアクションのあったオブジェクト
   * @param   {string} topic
   *          イベントの種類
   * @param   {string} data
   *          イベントのデータ
   */
  observe: function(subject, topic, data) {
    if (topic == "profile-after-change") {
      this._registerDLF();
    }
  },
  
  /* ==== arIStartupEventListener インターフェース ==== */
  
  /**
   * スタートアップ時のイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    let APP_STARTUP = 1;
    
    if (reason == APP_STARTUP) {
      /* profile-after-change イベント前に
       * nsIMIMEService.getTypeFromExtension を呼ぶと
       * いくつかのプラグインがクラッシュする */
      Services.obs.addObserver(this, "profile-after-change", false);
      this._observing = true;
    } else {
      this._registerDLF();
    }
  },
  
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    if (this._observing) {
      Services.obs.removeObserver(this, "profile-after-change");
      this._observing = false;
    }
    
    this._unregisterDLF();
  },
  
  /* ==== private ==== */
  
  _registerer: null,  /** 登録用
                       *  @type {UnMHTDLFRegisterer} */
  
  _observing: false,  /** プロファイル変更イベント監視中か
                       *  @type {boolean} */
  
  /**
   * Document Loader Factory を登録する
   */
  _registerDLF: function() {
    this._registerer
      = new UnMHTDLFRegisterer("@unmht.org/unmht-document-loader-factory;1");
    
    this._getMimeTypeList().forEach(function(type) {
      this._registerer.addMimeType(type, true);
    }.bind(this));

    this._getMimeExtensionList().forEach(function([ext, type]) {
      this._registerer.addFileExtension(ext, type, true);
    }.bind(this));
    
    try {
      let mime = Cc["@mozilla.org/mime;1"].getService(Ci.nsIMIMEService);
      this._registerer.addMimeType(mime.getTypeFromExtension("mht"), true);
    } catch (e) {
    }
    
    this._registerer.register();
    
    arShutdownEvent.addListener(this);
  },
  
  /**
   * Document Loader Factory を登録解除する
   */
  _unregisterDLF: function() {
    if (this._registerer) {
      this._registerer.unregister();
      this._registerer = null;
    }
  },
  
  /**
   * ハンドルする MIME-Type のリストを返す
   *
   * @returns {Array.<string>}
   *          ハンドルする MIME-Type のリスト
   */
  _getMimeTypeList: function() {
    let tmp = UnMHTPref.getPref("open.mime.types");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(function(n) typeof(n) == "string")) {
        list = [];
      }
    } catch (e) {
    }
    
    return list;
  },
  
  /**
   * ハンドルする拡張子と MIME-Type のリストを返す
   *
   * @returns {Array.<[string,string]>}
   *          ハンドルする拡張子と MIME-Type のリスト
   */
  _getMimeExtensionList: function() {
    let tmp = UnMHTPref.getPref("open.mime.extensions");
    let list = [];
    try {
      list = JSON.parse(tmp);
      if (!list.every(function(n) typeof(n[0]) == "string" && typeof(n[1]) == "string")) {
        list = [];
      }
    } catch (e) {
    }
    
    return list;
  }
};
