"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arShutdownEvent.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTStringBundle.jsm");
XPCOMUtils.defineLazyGetter(this, "UnMHTSave", function() {
  XPCOMUtils.importRelative(this, "UnMHTSave.jsm");
  return UnMHTSave;
});

let EXPORTED_SYMBOLS = ["UnMHTStatusbarProgressMap",
                        "UnMHTStatusbarProgress"];

/**
 * window から UnMHTStatusbarProgress へのマップ
 */
let UnMHTStatusbarProgressMap = new Map();
let listener = {
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    try {
      UnMHTStatusbarProgressMap.clear();
    } catch (e) {
      /* COMPAT: Firefox 18 以前 */
      for (let [name,] of UnMHTStatusbarProgressMap) {
        UnMHTStatusbarProgressMap.delete(name);
      }
    }
  }
};
arShutdownEvent.addListener(listener);

/**
 * 保存状況をステータスバーに表示するリスナ
 *
 * @param   {nsIDOMWindow} window
 *          実行中のウィンドウ
 * @extends    arWindowHook
 * @implements arIWindowHook
 * @implements IUnMHTSaveListener
 */
function UnMHTStatusbarProgress(window) {
  arWindowHook.call(this, window);
  
  this._contexts = [];
  
  UnMHTStatusbarProgressMap.set(window, this);
}
UnMHTStatusbarProgress.prototype = {
  /**
   * 保存コンテキストを監視対象に追加
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Function} callback
   *          保存完了/中断時のコールバック
   *            function(result)
   *              *param   {number} result
   */
  addContext: function(context, callback) {
    this._contexts.push([context, callback]);
    
    this._getProgressUI();
  },
  
  /**
   * 保存コンテキストを監視対象から削除
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  removeContext: function(context) {
    this._contexts = this._contexts.filter(function([c, f]) c != context);
    context.destruct();
  },
  
  /**
   * 保存をキャンセルする
   */
  cancelSave: function() {
    this._contexts.forEach(function([context, callback]) context.cancel());
    this._contexts = [];
    
    this._updateProgress();
  },
  
  /* ==== arIWindowHook インターフェース ==== */

  /**
   * 破棄
   */
  destruct: function() {
    UnMHTStatusbarProgressMap.delete(this._window);
    
    if (this._contexts) {
      this._contexts
        .forEach(function([context, callback]) {
          context.cancel();
        });
      this._contexts = [];
    }
    
    this._progmeter = null;
    this._proglabel = null;
    
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    if (isStartup) {
      /* スタートアップイベント中に insertItem を呼ぶと
       * ツールバーが強制的に表示されるため遅延させる */
      arTimer.setTimeout(function() {
        this._getProgressUI();
      }.bind(this), 100);
    } else {
      this._getProgressUI();
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    this._progmeter = this.el("unmht-toolbarbutton-progress-meter");
    if (this._progmeter) {
      this._progmeter.style.display = "";
    }
    this._proglabel = this.el("unmht-toolbarbutton-progress-label");
    if (this._proglabel) {
      this._proglabel.style.display = "";
    }
  },
  
  /* ==== IUnMHTSaveListener インターフェース ==== */
  
  /**
   * キャンセルが完了したイベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onCancel: function(context) {
    this._contexts
      .filter(function([c, callback]) c == context)
      .forEach(function([c, callback]) {
        if (callback) {
          callback(UnMHTSave.RESULT_CANCEL);
        }
      });
    
    this.removeContext(context);
  },
  
  /**
   * 保存失敗イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {number} reason
   *          エラーの理由
   * @param   {Exception} e
   *          例外
   */
  onError: function(context, reason, e) {
    let mes = "";
    switch (reason) { 
      case UnMHTSave.ERROR_WRITE_FILE: {
        mes = UnMHTStringBundle.get("save.error.open");
        break;
      }
    }
    
    let confirmURL = "chrome://unmht/content/confirm.xul";
    let features = "chrome,dialog,modal,centerscreen";
    let arg = {
      title: UnMHTStringBundle.get("error.title"),
      message1: mes,
      message2: e ? e.toString() : "",
      accept: UnMHTStringBundle.get("ok"),
      focus: "yes"
    };
    this._window.openDialog(confirmURL, "", features, arg);
    
    let result
      = (reason == UnMHTSave.ERROR_WRITE_FILE)
      ? UnMHTSave.RESULT_ERROR_WRITE_FILE
      : UnMHTSave.RESULT_FAIL;
    
    this._contexts
      .filter(function([c, callback]) c == context)
      .forEach(function([c, callback]) {
        if (callback) {
          callback(result);
        }
      });
    
    this.removeContext(context);
  },
  
  /**
   * 進捗状況変化イベント
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {string} currentFile
   *          現在取得のファイル名
   * @param   {string} currentSize
   *          取得したバイト数 / ファイルサイズ
   * @param   {number} ok
   *          取得完了した数
   * @param   {number} na
   *          取得中の数
   * @param   {number} ng
   *          取得失敗した数
   */
  onProgress: function(context, currentFile, currentSize, ok, na, ng) {
    this._updateProgress();
  },
  
  /**
   * 保存完了
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   * @param   {Array.<string>} failedList
   *          失敗したファイル名
   */
  onSave: function(context, failedList) {
    if (UnMHTPref.getPref("save.show_errors") && failedList.length > 0) {
      let failedlistURL = "chrome://unmht/content/failedlist.xul";
      let features = "chrome,titlebar,toolbar,centerscreen,resizable";
      let arg = {
        failedList: failedList
      };
      this._window.openDialog(failedlistURL, "", features, arg);
    }
    
    this._contexts
      .filter(function([c, callback]) c == context)
      .forEach(function([c, callback]) {
        if (callback) {
          callback(false);
        }
      });
    
    this._updateProgress();
    this.removeContext(context);
  },
  
  /**
   * 保存コンテキストの破棄
   *
   * @param   {UnMHTSaveContext} context
   *          対象の保存コンテキスト
   */
  onDestruct: function(context) {
  },
  
  /* ==== private ==== */
  
  _contexts: null,  /** 保存コンテキストとコールバックのマップ
                     *  @type {Array}
                     *    [[UnMHTSaveContext, Function], ...] */
  
  _progmeter: null, /** 進捗状況のプログレスメーター
                     *  @type {nsIDOMXULElement} */
  _proglabel: null, /** 進捗状況のラベル
                     *  @type {nsIDOMXULElement} */
  
  /**
   * ステータスバーの進捗状況を更新する
   */
  _updateProgress: function() {
    if (!(this._progmeter && this._proglabel)) {
      return;
    }

    let na = 0;
    let ok = 0;
    let progress = 0;
    
    this._contexts.forEach(function([context, callback]) {
      if (context.progress == 1) {
        ok ++;
      } else {
        na ++;
      }
      progress += context.progress;
    });
    progress /= this._contexts.length;
    
    if (na == 0) {
      if (ok > 0) {
        this._progmeter.style.display = "";
        this._progmeter.setAttribute("value", "0");
        this._proglabel.style.display = "-moz-box";
        this._proglabel.setAttribute("value",
                                     UnMHTStringBundle.get("save.done"));
        
        arTimer.setTimeout(function() {
          if (this._contexts.length == 0) {
            this._progmeter.style.display = "";
            this._progmeter.setAttribute("value", "0");
            this._proglabel.style.display = "";
            this._proglabel.setAttribute("value", "");
          }
        }.bind(this), 3000);
      }
    } else {
      /* 保存中 */
      
      this._progmeter.style.display = "-moz-box";
      this._progmeter.setAttribute("value", Math.floor(progress * 100));
      this._proglabel.style.display = "";
    }
  },
  
  /**
   * プログレスバーの要素を取得する
   * 無ければ追加する
   */
  _getProgressUI: function() {
    this._progmeter = this.el("unmht-toolbarbutton-progress-meter");
    this._proglabel = this.el("unmht-toolbarbutton-progress-label");
    if (!this._progmeter) {
      let addonbar = this.el("addon-bar");
      if (!addonbar) {
        return;
      }
      addonbar.insertItem("unmht-toolbarbutton-progress");
      
      this._progmeter = this.el("unmht-toolbarbutton-progress-meter");
      this._proglabel = this.el("unmht-toolbarbutton-progress-label");
    }
  }
};
arPrototype.extend(UnMHTStatusbarProgress, arWindowHook);
