"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "base/arPrototype.jsm");
XPCOMUtils.importRelative(this, "base/arTimer.jsm");
XPCOMUtils.importRelative(this, "base/arWindowHook.jsm");
XPCOMUtils.importRelative(this, "UnMHTDocumentCategory.jsm");
XPCOMUtils.importRelative(this, "UnMHTPref.jsm");
XPCOMUtils.importRelative(this, "UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTToolbarButtonClass"];

/**
 * ツールバーのボタン管理
 *
 * @param   {nsIDOMWindow} window
 *          対象のウィンドウ
 * @extends    arWindowHook
 * @implements nsISupportsWeakReference
 * @implements nsIWebProgressListener
 * @implements arIWindowHook
 */
function UnMHTToolbarButtonClass(window) {
  arWindowHook.call(this, window);
}
UnMHTToolbarButtonClass.prototype = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsISupportsWeakReference,
                                         Ci.nsIWebProgressListener]),
  
  /* ==== nsIWebProgressListener インターフェース ==== */
  
  /**
   * 監視ウィンドウのロケーションが変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {nsIURI} location
   *          ロケーション
   */
  onLocationChange: function(webProgress, request, location) {
    this._setStatus();
  },
  
  /**
   * 進捗状況が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} curSelfProgress
   *          対象のリクエストの現在の進捗状況
   * @param   {number} maxSelfProgress
   *          対象のリクエストの最大の進捗状況
   * @param   {number} curTotalProgress
   *          全ての現在の進捗状況
   * @param   {number} maxTotalProgress
   *          全ての最大の進捗状況
   */
  onProgressChange: function(webProgress, request,
                             curSelfProgress, maxSelfProgress,
                             curTotalProgress, maxTotalProgress) {
  },
  
  /**
   * プロトコルのセキュリティ設定が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} state
   *          セキュリティ状態フラグ、セキュリティ強度フラグ
   */
  onSecurityChange: function(webProgress, request, state) {
  },
  
  /**
   * 状況が変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {number} flags
   *          状況フラグ
   *            Ci.nsIWebProgressListener.STATE_*
   * @param   {nsresult} status
   *          ステータス
   */
  onStateChange: function(webProgress, request, flags, status) {
  },

  /**
   * ステータスバーに表示するメッセージが変わった時のイベント
   *
   * @param   {nsIWebProgress} webProgress
   *          呼び出し元
   * @param   {nsIRequest} request
   *          状況の変わったリクエスト
   * @param   {nsresult} status
   *          リクエストのステータス
   * @param   {string} message
   *          ステータスバーのメッセージ
   */
  onStatusChange: function(webProgress, request, status, message) {
  },
  
  /* ==== arIWindowHook インターフェース ==== */
  
  /**
   * 破棄
   */
  destruct: function() {
    arWindowHook.prototype.destruct.call(this);
  },
  
  /**
   * ウィンドウにロードする
   *
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  load: function(isStartup, reason) {
    let target = UnMHTWindowUtil.getBrowser(this._window);
    if (target) {
      target.addProgressListener(this);
    }
    
    this._setStatus();
    
    if (isStartup) {
      /* スタートアップイベント中に insertItem を呼ぶと
       * ツールバーが強制的に表示されるため遅延させる */
      arTimer.setTimeout(function() {
        this._insertAddonBarButton();
      }.bind(this), 100);
    } else {
      this._insertAddonBarButton();
    }
  },
  
  /**
   * ウィンドウからアンロードする
   *
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  unload: function(isShutdown, reason) {
    let target = UnMHTWindowUtil.getBrowser(this._window);
    if (target) {
      target.removeProgressListener(this);
    }
  },
  
  /* ==== private ==== */

  /**
   * ツールバーのボタンの状況に合わせて切り替える
   */
  _setStatus: function() {
    let targetDocument = UnMHTWindowUtil.getCurrentDocument(this._window);
    
    let savable
      = UnMHTDocumentCategory.isSavable(this._window, targetDocument, false);
    
    let enablePanel = function(type, enabled) {
      let panel = this.el("unmht-toolbarbutton-" + type);
      if (panel) {
        panel.disabled = !savable;
      }
    }.bind(this);
    
    enablePanel("save", !savable);
    enablePanel("quicksave", !savable);
  },
  
  /**
   * アドオンバーにボタンを配置する
   * 他に配置されていれば何もしない
   */
  _insertAddonBarButton: function() {
    if (!UnMHTPref.getPref("statusbar.icon")) {
      return;
    }
    
    let toolbars = this._window.document.getElementsByTagName("toolbar");
    let id = "unmht-toolbarbutton-preferences";
    
    if (Array.some(toolbars,
                   function(toolbar) 
                   toolbar.getAttribute("currentset").split(",").indexOf(id) != -1)) {
      return;
    }
    
    let addonbar = this.el("addon-bar");
    if (!addonbar) {
      return;
    }
    
    addonbar.insertItem(id);
  }
};
arPrototype.extend(UnMHTToolbarButtonClass, arWindowHook);
