"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.defineLazyGetter(this, "arArrayUtil", function() {
  XPCOMUtils.importRelative(this, "base/arArrayUtil.jsm");
  return arArrayUtil;
});

let EXPORTED_SYMBOLS = ["UnMHTWindowUtil"];

/**
 * Window 操作ユーティリティ
 */
let UnMHTWindowUtil = {
  /**
   * ブラウザを取得する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {nsIDOMXULElement}
   *          ブラウザ (tabbrowser もしくは browser)
   *          見付からない場合は null
   */
  getBrowser: function(window) {
    if (this.isMailWindow(window)) {
      return window.document.getElementById("messagepane");
    }
    
    return window.document.getElementById("content");
  },
  
  /**
   * 現在開いているドキュメントを取得する
   *
   * @param   {nsIDOMWindow} window
   *          実行中のウィンドウ
   * @returns {nsIDOMDocument}
   *          現在開いているドキュメント
   *          見付からない場合は null
   */
  getCurrentDocument: function(window) {
    let browser = this.getBrowser(window);
    if (!browser) {
      return null;
    }
    return browser.contentDocument;
  },
  
  /**
   * ウィンドウのタブの配列を取得する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {Array.<nsIDOMXULElement>}
   *          タブの配列
   */
  getTabs: function(window) {
    let tabbrowser = window.document.getElementById("content");
    if (!tabbrowser) {
      return [];
    }
    
    return ("visibleTabs" in tabbrowser)
      ? tabbrowser.visibleTabs
      : tabbrowser.tabs;
  },

  /**
   * ウィンドウのブラウザの配列を取得する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {Array.<nsIDOMXULElement>}
   *          ブラウザの配列
   */
  getBrowsers: function(window) {
    let tabbrowser = window.document.getElementById("content");
    if (!tabbrowser) {
      return [];
    }
    
    return Array
      .map(this.getTabs(window),
           function(tab) tabbrowser.getBrowserForTab(tab));
  },

  /**
   * ウィンドウのドキュメントの配列を取得する
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {Array.<nsIDOMDocument>}
   *          ドキュメントの配列
   */
  getDocuments: function(window) {
    return Array
      .map(this.getBrowsers(window),
           function(browser) browser.contentDocument);
  },
  
  /**
   * ドキュメントに対応するタブを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @param   {nsIDOMDocument} targetDocument
   *          対象のドキュメント
   * @returns {nsIDOMXULElement}
   *          対応するタブ
   *          見付からなかったら null
   */
  getTabForDocument: function(window, targetDocument) {
    let tabbrowser = this.getBrowser(window);
    if (!tabbrowser) {
      return null;
    }
    
    return arArrayUtil
      .find(tabbrowser.tabs,
            function(tab) (tab.linkedBrowser &&
                           tab.linkedBrowser.contentDocument == targetDocument)) || null;
  },
  
  /**
   * メッセージパネルが閉じているかどうか調べる
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          メッセージパネルが閉じているか
   *          メッセージパネルが無い場合は false
   */
  isMessagePaneCollapsed: function(window) {
    if (!this.isMailWindow(window)) {
      return false;
    }
    
    let node;
    node = window.document.getElementById("messagepanebox");
    if (node && node.collapsed) {
      return true;
    }
    node = window.document.getElementById("messagepaneboxwrapper");
    if (node && node.collapsed) {
      return true;
    }
    
    return false;
  },
  
  /**
   * メールのウィンドウかどうかを返す
   *
   * @param   {nsIDOMWindow} window
   *          対象のウィンドウ
   * @returns {boolean}
   *          メールのウィンドウかどうか
   */
  isMailWindow: function(window) {
    return (window.document.documentURI
            == "chrome://messenger/content/messenger.xul");
  },
  
  /**
   * URL で指定されたウィンドウを探す
   *
   * @param   {string} url
   *          探すウィンドウの URL
   * @returns {nsIDOMWindow}
   *          URL に一致するウィンドウ
   */
  find: function(url) {
    for (let window of XPCOMUtils.IterSimpleEnumerator
         (Services.wm.getEnumerator(null), Ci.nsIDOMWindow)) {
      if (window.document.documentURI == url) {
        return window;
      }
    }
    
    return null;
  },
  
  /**
   * URL で指定されたウィンドウにフォーカスする
   *
   * @param   {string} url
   *          フォーカスするウィンドウの URL
   * @returns {boolean}
   *          見付かったか
   */
  focus: function(url) {
    let window = this.find(url);
    if (window) {
      window.focus();
      return true;
    }
    
    return false;
  }
};
