"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

let EXPORTED_SYMBOLS = ["arArrayUtil"];

/**
 * 配列ユーティリティ
 */
let arArrayUtil = {
  /**
   * Array.prototype.find みたいなヤツ
   *
   * @param   {Array.<ANY>} a
   *          対象の配列
   * @param   {Function} pred
   *          述語
   *            function(item, index, array)
   *              *param   {ANY} item
   *              *param   {number} index
   *              *param   {Array.<ANY>} array
   *              *returns {boolean}
   * @returns {ANY}
   *          a の要素もしくは undefined
   */
  find: function(a, pred) {
    for (let i = 0, length = a.length; i < length; i ++) {
      let x = a[i];
      if (pred(x, i, a)) {
        return x;
      }
    }
    
    return undefined;
  },
  
  /**
   * 配列 a の要素 x に f を適用した y について
   * 最初に pred を満たした y を返す
   * どの y も pred を満たさない場合は undefined を返す
   *
   * 速度が必要な場面で map と filter の代わりに使う
   * f と p が副作用を持たない場合に限り下の 2 つは等価
   * [y] = a.map(f).filter(p);
   * y = arArrayUtil.mapfind(a, f, p);
   * 
   * @param   {Array.<ANY>} a
   *          対象の配列
   * @param   {Function} f
   *          a の要素に適用する関数
   *            function(item, index, array)
   *              *param   {ANY} item
   *              *param   {number} index
   *              *param   {Array.<ANY>} array
   *              *returns {ANY}
   * @param   {Function} pred
   *          述語
   *            function(item)
   *              *param   {ANY} item
   *              *param   {number} index
   *              *param   {Array.<ANY>} array
   *              *returns {boolean}
   * @returns {ANY}
   *          f(a の要素) もしくは undefined
   */
  mapfind: function(a, f, pred) {
    for (let i = 0, length = a.length; i < length; i ++) {
      let y = f(a[i], i, a);
      if (pred(y, i, a)) {
        return y;
      }
    }
    
    return undefined;
  },
  
  /**
   * Array.reduce の途中結果も含めて配列で返す
   * Haskell の Data.List の scanl
   *
   * @param   {Array.<ANY>} a
   *          対象の配列
   * @param   {Function} func
   *          適用する関数   
   *            function(accum, item)
   *              *param   {ANY} accum
   *              *param   {ANY} item
   *              *returns {ANY}
   * @param   {ANY} init
   *          初期値
   * @returns {Array}
   *          結果の配列
   */
  scanl: function(a, func, init) {
    let accum = init;
    return a.map(function(item) {
      return (accum = func(accum, item));
    });
  },
  
  /**
   * a の各要素 [x, y] を x と y で別々にリストにして返す
   *
   * @param   {Array} a
   *            [[x, y], ...]
   * @returns {Array}
   *            [[x, ...], [y, ...]]
   */
  unzip: function(a) {
    let xs = [], ys = [];
    a.forEach(function([x, y]) {
      xs.push(x);
      ys.push(y);
    });
    return [xs, ys];
  },
  
  /**
   * [begin, end) のジェネレータ
   *
   * @param   {number} begin
   *          始値
   * @param   {number} end
   *          終値
   * @returns {Generator}
   *          [begin, end) のジェネレータ
   */
  range: function(begin, end) {
    if (begin < end) {
      for (let i = begin; i < end; i ++) {
        yield i;
      }
    } else {
      for (let i = begin; i > end; i --) {
        yield i;
      }
    }
  }
};
