"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu,
      manager: Cm } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arShutdownEvent.jsm");

let EXPORTED_SYMBOLS = ["arComponentLoader"];

/**
 * XPCOM のロードとアンロード
 *
 * @implements arIShutdownEventListener
 */
let arComponentLoader = {
  /**
   * XPCOM を登録してアンロード時のリスナを登録も行う
   *
   * @param   {nsISupport} component
   *          登録するコンポーネント
   */
  loadComponent: function(component) {
    let factory = {
      /* ==== nsIFactory インターフェース ==== */
      
      /**
       * インスタンスを生成する
       *
       * @param   {nsISupport} outer
       *          統合する対象
       * @param   {nsIIDRef} iid
       *          生成する対象のインターフェース ID
       * @returns {nsISupports}
       *          生成したインスタンス
       * @throws  {Cr.NS_ERROR_NO_AGGREGATION}
       */
      createInstance: function(outer, iid) {
        if (outer) {
          throw Cr.NS_ERROR_NO_AGGREGATION;
        }
        return (new component()).QueryInterface(iid);
      }
    };
    
    this._registerComponent(component, factory);
    this._components.push([component, factory]);
  },
  
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this._unregisterComponents();
  },
  
  /* ==== private ==== */
  
  _components: [], /** コンポーネントとファクトリー
                    *  @type {Array.<[nsISupports,nsIFactory]> */
  
  /**
   * 初期化処理
   */
  init: function() {
    arShutdownEvent.addListener(this);
  },
  
  /**
   * XPCOM を登録する
   *
   * @param   {nsISupport} component
   *          登録するコンポーネント
   * @param   {nsIFactory} factory
   *          登録するコンポーネントの Factory
   */
  _registerComponent: function(component, factory) {
    Cm.QueryInterface(Ci.nsIComponentRegistrar).registerFactory(
      component.prototype.classID,
      component.prototype.classDescription,
      component.prototype.contractID,
      factory);
    
    if ("_xpcom_categories" in component.prototype) {
      component.prototype._xpcom_categories.forEach(function(cat) {
        XPCOMUtils.categoryManager.addCategoryEntry(
          cat.category,
          component.prototype.contractID,
          component.prototype.contractID,
          false, true);
      });
    }
  },

  /**
   * 全ての XPCOM の登録を解除する
   */
  _unregisterComponents: function() {
    let factoryInfos
      = this._components.map(function([component, factory]) {
        if ("_xpcom_categories" in component.prototype) {
          component.prototype._xpcom_categories.forEach(function(cat) {
            XPCOMUtils.categoryManager.deleteCategoryEntry(
              cat.category,
              component.prototype.contractID,
              false);
          });
        }
        return [component.prototype.classID, factory];
      });
    this._components = [];
    
    let manager = Cm.QueryInterface(Ci.nsIComponentRegistrar);
    Services.tm.mainThread.dispatch(function() {
      /* ここはモジュールのアンロード後に動く別スレッドなので
       * 必要な変数を解決してからキャプチャする */
      factoryInfos.forEach(function([classID, factory]) {
        manager.unregisterFactory(classID, factory);
      });
    }, Ci.nsIEventTarget.DISPATCH_NORMAL);
  }
};
arComponentLoader.init();
