"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

let EXPORTED_SYMBOLS = ["arFilePicker", "arFilePickerClass"];

/**
 * ファイル選択ダイアログ
 */
function arFilePickerClass() {
}
arFilePickerClass.prototype = {
  /**
   * ファイル選択ダイアログを非同期で開く
   *
   * @param   {object} params
   *          パラメータ
   *            {
   *              window: {nsIDOMWindow}
   *                      親ウィンドウ (必須),
   *              title: {string}
   *                     ダイアログのタイトル (オプショナル),
   *              mode: {number}
   *                    nsIFilePicker のモード (オプショナル)
   *                    デフォルトは Ci.nsIFilePicker.modeOpen,
   *              useLastDir: {boolean}
   *                      最後に使用したディレクトリを使用/保存する
   *                      (オプショナル)
   *                      デフォルトは false,
   *              displayDirectory: {nsIFile}
   *                      useLastDir が false の際に表示するディレクトリ
   *                      (オプショナル)
   *                      ファイルを指している場合は親ディレクトリになる,
   *              uri: {string}
   *                   保存対象の URI (オプショナル)
   *                   サイトごとに記憶する際に使用する
   *                   デフォルトは null,
   *              defaultString: {string}
   *                             デフォルトファイル名 (オプショナル),
   *              defaultextension: {string}
   *                                デフォルト拡張子 (オプショナル),
   *              filter: {Array.<([string,string]|number)>}
   *                      ファイルタイプ (オプショナル)
   *                        [type, ...]
   *                          type: [string 名前, string 拡張子]
   *                                または
   *                                number filterMask
   *              filterIndex: {number}
   *                           デフォルトのファイルタイプの
   *                           インデックス (オプショナル),
   *              callback: {Function}
   *                        終了時コールバック (オプショナル)
   *                          function(file, result, filePicker)
   *                            *param   {(nsIFile|Array.<nsIFile>)} file
   *                                     選択されたファイル
   *                                     mode が
   *                                     Ci.nsIFilePicker.modeOpenMultiple
   *                                     の場合 Array.<nsIFile>
   *                                     中断した場合は null
   *                                     それ以外の場合は nsIFile
   *                            *param   {number} result
   *                            *param   {nsIFilePicker} filePicker
   *            }
   * @throws  {Error}
   *          params.window が無い場合
   */
  asyncOpen: function(params) {
    if (!("window" in params
          && params.window)) {
      throw new Error("params.window must not be null");
    }
    let window = params.window;
    
    let getParam = function(name, defValue) {
      return (name in params) ? params[name] : defValue;
    };
    
    let title            = getParam("title", "");
    let mode             = getParam("mode", Ci.nsIFilePicker.modeOpen);
    let useLastDir       = getParam("useLastDir", false);
    let displayDirectory = getParam("displayDirectory", null);
    let uri              = getParam("uri", null);
    let defaultString    = getParam("defaultString", null);
    let defaultExtension = getParam("defaultExtension", null);
    let filter           = getParam("filter", null);
    let filterIndex      = getParam("filterIndex", -1);
    let callback         = getParam("callback", null);
    
    let filePicker
      = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);
    
    if (uri) {
      try {
        uri = Services.io.newURI(uri, null, null);
      } catch (e) {
        uri = null;
      }
    }
    
    filePicker.init(window, title, mode);
    
    let dir = null;
    if (useLastDir) {
      if (mode == Ci.nsIFilePicker.modeSave) {
        dir = this.getLastSaveDir(window, uri);
      } else {
        dir = this.getLastOpenDir(window);
      }
    } else if (displayDirectory) {
      dir = displayDirectory.clone();
      if (dir && dir.exists()) {
        if (!dir.isDirectory()) {
          dir = dir.parent;
        }
      } else {
        dir = null;
      }
    }
    if (dir) {
      filePicker.displayDirectory = dir;
    }
    
    if (filter) {
      filter.forEach(function(type) {
        if (type.length == 1) {
          let [mask] = type;
          filePicker.appendFilters(mask);
        } else {
          let [name, ext] = type;
          filePicker.appendFilter(name, ext);
        }
      });
    }
    if (filterIndex >= 0 && filterIndex < filter.length) {
      filePicker.filterIndex = filterIndex;
    }
    
    if (defaultString) {
      filePicker.defaultString = defaultString;
    }
    if (defaultExtension) {
      filePicker.defaultExtension = defaultExtension;
    }
    
    let self = this;
    
    let callbackObject = {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsIFilePickerShownCallback]),
      
      done: function(result) {
        if (!(result == Ci.nsIFilePicker.returnOK ||
              result == Ci.nsIFilePicker.returnReplace)) {
          if (callback) {
            callback(null, result, filePicker);
          }
          
          return;
        }
        
        switch (mode) {
          case Ci.nsIFilePicker.modeOpenMultiple: {
            let files
              = [file for (file of XPCOMUtils.IterSimpleEnumerator(
                filePicker.files, Ci.nsIFile))];
            if (files.length) {
              self.setLastOpenDir(window, files[0].parent);
            }
            
            if (callback) {
              callback(files, result, filePicker);
            }
            break;
          }
          case Ci.nsIFilePicker.modeGetFolder: {
            let dir = filePicker.file;
            
            self.setLastOpenDir(window, dir);
            
            if (callback) {
              callback(dir, result, filePicker);
            }
            break;
          }
          case Ci.nsIFilePicker.modeSave: {
            let file = filePicker.file;
            
            self.setLastSaveDir(window, uri, file.parent);
            
            if (callback) {
              callback(file, result, filePicker);
            }
            break;
          }
          default: {
            let file = filePicker.file;
            
            self.setLastOpenDir(window, file.parent);
            
            if (callback) {
              callback(file, result, filePicker);
            }
          }
        }
      }
    };
    
    filePicker.open(callbackObject);
  },
  
  /**
   * 開くのに最後に使用したディレクトリを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @returns {nsIFile}
   *          最後に使用したディレクトリ
   *          無い場合は null
   */
  getLastOpenDir: function(window) {
    let name = "browser.open.lastDir";
    if (Services.prefs.getPrefType(name) != 0) {
      return Services.prefs.getComplexValue(name, Ci.nsILocalFile);
    }
    
    return null;
  },
  
  /**
   * 開くのに最後に使用したディレクトリを設定する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {nsIFile} dir
   *          最後に使用したディレクトリ
   */
  setLastOpenDir: function(window, dir) {
    Services.prefs.setComplexValue("browser.open.lastDir",
                                   Ci.nsILocalFile, dir);
  },
  
  /**
   * 保存するのに最後に使用したディレクトリを取得する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {string} uri
   *          保存対象の URI
   *          不要ならば null
   * @returns {nsIFile}
   *          最後に使用したディレクトリ
   *          無い場合は null
   */
  getLastSaveDir: function(window, uri) {
    let context = {};
    Cu.import("resource://gre/modules/DownloadLastDir.jsm", context);
    
    if ("DownloadLastDir" in context) {
      return (new context.DownloadLastDir(window)).getFile(uri);
    } else if ("gDownloadLastDir" in context) {
      /* COMPAT: Firefox 18 以前 */
      return context.gDownloadLastDir.getFile(uri);
    }
    
    return null;
  },
  
  /**
   * 保存するのに最後に使用したディレクトリを設定する
   *
   * @param   {nsIDOMWindow} window
   *          親ウィンドウ
   * @param   {string} uri
   *          保存対象の URI
   *          不要ならば null
   * @param   {nsIFile} dir
   *          最後に使用したディレクトリ
   */
  setLastSaveDir: function(window, uri, dir) {
    let context = {};
    Cu.import("resource://gre/modules/DownloadLastDir.jsm", context);
    
    if ("DownloadLastDir" in context) {
      new context.DownloadLastDir(window).setFile(uri, dir);
    } else if ("gDownloadLastDir" in context) {
      /* COMPAT: Firefox 18 以前 */
      context.gDownloadLastDir.setFile(uri, dir);
    }
  }
};

let arFilePicker = new arFilePickerClass();
