"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

let EXPORTED_SYMBOLS = ["arFileUtil"];

/** PR_Open の定数
 *  @type {number} */
let PR_RDONLY      = 0x01;
let PR_WRONLY      = 0x02;
let PR_CREATE_FILE = 0x08;
let PR_APPEND      = 0x10;
let PR_TRUNCATE    = 0x20;
let PR_SYNC        = 0x40;
let PR_EXCL        = 0x80;

/**
 * ファイル操作ユーティリティ
 */
let arFileUtil = {
  /** パーミッション
   *  @type {number} */
  PERM_NORMAL_FILE: parseInt("0644", 8),
  PERM_EXEC_FILE:   parseInt("0755", 8),
  PERM_DIRECTORY:   parseInt("0755", 8),
  
  /**
   * ファイルを取得する
   *
   * @param   {string} path
   *          ファイルのパス
   * @returns {nsIFile}
   *          指定されたパスのファイル
   *          失敗した場合は null
   */
  getFile: function(path) {
    let file = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
    try {
      file.initWithPath(path);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      return null;
    }
    
    return file;
  },

  /**
   * ファイル名の長さをチェックしてからパスに追加する
   *
   * @param   {nsIFile} baseDir
   *          親ディレクトリ
   * @param   {string} filenameBase
   *          ファイル名 (拡張子以外)
   *          省略される可能性あり
   * @param   {string} suffix
   *          ファイル名の末尾
   *          ASCII のみ
   *          必ず使用される
   * @param   {string} extension
   *          拡張子
   *          ASCII のみ
   *          必ず使用される
   * @returns {nsIFile}
   *          ファイル
   */
  checkAndAppend: function(baseDir, filenameBase, suffix, extension) {
    if (!suffix) {
      suffix = "";
    }
    if (extension) {
      suffix += "." + extension;
    }
    
    let tempFile = this.tempDirectory.clone();
    tempFile.append(filenameBase + suffix);
    /* ファイル名の短縮における長さの算出は nsIFile.createUnique に任せる
     * ただし、実際に短縮されたファイル名は suffix を含まない可能性があるので
     * 使用しない */
    tempFile.createUnique(Ci.nsIFile.NORMAL_FILE_TYPE, this.PERM_NORMAL_FILE);
    tempFile.remove(false);
    if (tempFile.leafName.length < filenameBase.length) {
      /* 短かくなったのでその長さを参考に縮める
       * 場合によっては余分に縮める事にもなるが
       * suffix と extension が ASCII であれば縮め足りない事はない */
      filenameBase
        = filenameBase.substr(0, tempFile.leafName.length - suffix.length);
    }
    
    let file = baseDir.clone();
    file.append(filenameBase + suffix);
    return file;
  },
  
  /**
   * ファイルの入力ストリームを開く
   *
   * @param   {nsIFile} file
   *          対象のファイル
   * @returns {nsIFileInputStream}
   *          ファイルの入力ストリーム
   *          失敗したら null
   */
  openInputStream: function(file) {
    let fstream
      = Cc["@mozilla.org/network/file-input-stream;1"]
      .createInstance(Ci.nsIFileInputStream);
    
    try {
      fstream.init(file, -1, 0, 0);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      return null;
    }
    
    return fstream;
  },
  
  /**
   * ファイルの出力ストリームを開く
   *
   * @param   {nsIFile} file
   *          対象のファイル
   * @param   {number} permission
   *          パーミッション (オプショナル)
   *          undefined の場合はデフォルト
   * @returns {nsIFileOutputStream}
   *          ファイルの出力ストリーム
   *          失敗したら null
   */
  openOutputStream: function(file, permission) {
    let fstream
      = Cc["@mozilla.org/network/file-output-stream;1"]
      .createInstance(Ci.nsIFileOutputStream);
    
    try {
      fstream.init(file, -1, permission || this.PERM_NORMAL_FILE, 0);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      return null;
    }
    
    return fstream;
  },
  
  /**
   * Zip 書き込みを作成する
   *
   * @param   {nsIFile} file
   *          対象のファイル
   * @returns {nsIZipWriter}
   *          ZIP 書き込み
   *          失敗した場合は null
   */
  openZipWriter: function(file) {
    let zip = Cc["@mozilla.org/zipwriter;1"].createInstance(Ci.nsIZipWriter);
    
    try {
      zip.open(file, PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
      return null;
    }
    
    return zip;
  },
  
  /**
   * ファイルからデータを読み込む
   *
   * @param   {nsIFile} file
   *          対象のファイル
   * @returns {string}
   *          ファイルのデータ
   */
  readData: function(file) {
    let fstream = this.openInputStream(file);
    if (!fstream) {
      return "";
    }
    
    let bstream
      = Cc["@mozilla.org/binaryinputstream;1"]
      .createInstance(Ci.nsIBinaryInputStream);
    bstream.setInputStream(fstream);
    
    let data = bstream.readBytes(file.fileSize);
    
    bstream.close();
    
    return data;
  },
  
  /**
   * ディレクトリの内容を配列で返す
   *
   * @param   {nsIFile} dir
   *          対象のディレクトリ
   * @returns {Array}
   *          ディレクトリの内容のファイル
   *            [nsIFile, ...]
   */
  getDirectoryEntries: function(dir) {
    return [file for (file of XPCOMUtils.IterSimpleEnumerator(
      dir.directoryEntries, Ci.nsIFile))];
  },
  
  /**
   * 現在の OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @returns {string}
   *          変換したファイル名
   */
  validateFileName: function(filename) {
    return this.validateFileNameWithOS(filename, Services.appinfo.OS);
  },
  
  /**
   * 指定した OS でファイル名に使えない文字を変換する
   *
   * @param   {string} filename
   *          元のファイル名
   * @param   {string} OS
   *          対象の OS
   * @returns {string}
   *          変換したファイル名
   */
  validateFileNameWithOS: function(filename, OS) {
    let sepRe;
    if (OS == "WINNT") {
      sepRe = /[\\\/]/g;
      
      /* ファイル名に使えない文字
       *   |
       *   *
       *   :
       *   ?
       *   "
       *   <
       *   >
       */
      
      filename = filename.replace(/\"/g, "'")
        .replace(/[\*\:\?\|]/g, "_")
        .replace(/</g, "(")
        .replace(/>/g, ")");
    } else if (OS == "Darwin") {
      /* ファイル名に使えない文字
       *   :
       *   /
       */
      sepRe = /[\:\/]/g;
    } else {
      /* ファイル名に使えない文字
       *   /
       */
      sepRe = /\//g;
    }
    
    return filename.replace(sepRe, "_");
  },
  
  /**
   * ディレクトリを作成する
   *
   * @param   {nsIFile} dir
   *          対象のディレクトリ
   */
  createDirectory: function(dir) {
    if (!dir.exists()) {
      dir.create(Ci.nsIFile.DIRECTORY_TYPE, this.PERM_DIRECTORY);
    }
  },
  
  /**
   * ユニークなディレクトリを作成する
   *
   * @param   {nsIFile} dir
   *          対象のディレクトリ
   */
  createUniqueDirectory: function(dir) {
    if (!dir.exists()) {
      dir.createUnique(Ci.nsIFile.DIRECTORY_TYPE, this.PERM_DIRECTORY);
    }
  },
  
  /**
   * file プロトコルをファイルに変換する
   *
   * @param   {string} uri
   *          file プロトコル
   * @returns {nsIFile}
   *          ファイル、もしくは
   */
  getFileFromURLSpec: function(uri) {
    return this._fileProtocolHandler.getFileFromURLSpec(uri);
  },
  
  /**
   * ファイルを file プロトコルに変換する
   *
   * @param   {nsIFile} file
   *          ファイル
   * @returns {string}
   *          file プロトコル
   */
  getURLSpecFromFile: function(file) {
    return this._fileProtocolHandler.getURLSpecFromFile(file);
  },
  
  /* ---- ディレクトリ取得 ---- */
  
  /**
   * Applications ディレクトリ
   * Darwin 専用
   * @type {nsIFile}
   */
  get applicationsDirectory() {
    try {
      return Services.dirsvc.get("LocApp", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  },
  
  /**
   * ユーザの Applications ディレクトリ
   * Darwin 専用
   * @type {nsIFile}
   */
  get userApplicationsDirectory() {
    try {
      return Services.dirsvc.get("UsrApp", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  },
  
  /**
   * Program Files ディレクトリ
   * WINNT 専用
   * @type {nsIFile}
   */
  get programFilesDirectory() {
    try {
      return Services.dirsvc.get("ProgF", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  },

  /**
   * 現在のユーザの Profile ディレクトリ
   * @type {nsIFile}
   */
  get profileDirectory() {
    try {
      return Services.dirsvc.get("ProfD", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  },
  
  /**
   * 現在のユーザの Desktop ディレクトリ
   * @type {nsIFile}
   */
  get desktopDirectory() {
    try {
      return Services.dirsvc.get("Desk", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  },
  
  /**
   * 現在のユーザの Temp ディレクトリ
   * @type {nsIFile}
   */
  get tempDirectory() {
    try {
      return Services.dirsvc.get("TmpD", Ci.nsIFile);
    } catch (e) {
    }
    
    return null;
  }
};
XPCOMUtils.defineLazyGetter(arFileUtil, "_fileProtocolHandler", function() {
  return Services.io.getProtocolHandler("file")
    .QueryInterface(Ci.nsIFileProtocolHandler);
});
