"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.defineLazyGetter(this, "arArrayUtil", function() {
  XPCOMUtils.importRelative(this, "arArrayUtil.jsm");
  return arArrayUtil;
});

let EXPORTED_SYMBOLS = ["arListView"];

/**
 * ドラッグアンドドロップ時のマーカー
 *
 * @param   {nsIDOMXULElement} node
 *          tree 要素
 */
function arListViewDnDMarker(tree) {
  let document = tree.ownerDocument;
  
  this._marker = document.createElement("panel");
  tree.parentNode.appendChild(this._marker);
  
  this._marker.style.mozAappearance = "none";
  this._marker.style.mozWindowShadow = "none";
  this._marker.style.width = "200px";
  this._marker.style.height = "1px";
  this._marker.style.border = "none";
  this._marker.style.margin = "0";
  this._marker.style.padding = "0";
  this._marker.setAttribute("level", "top");
  this._marker.setAttribute("noautofocus", "true");
  this._marker.setAttribute("noautohide", "true");
  
  let ns = "http://www.w3.org/1999/xhtml";
  let sep = document.createElementNS(ns, "div");
  sep.style.width = "100%";
  sep.style.height = "1px";
  sep.style.backgroundColor = "#000000";
  this._marker.appendChild(sep);
}
arListViewDnDMarker.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    if (this._marker) {
      this._marker.parentNode.removeChild(this._marker);
      this._marker = null;
    }
  },
  
  /**
   * マーカーを表示する
   *
   * @param   {number} x
   *          左上の X 座標 (スクリーン座標) [px]
   * @param   {number} y
   *          左上の Y 座標 (スクリーン座標) [px]
   * @param   {number} width
   *          幅 [px]
   */
  show: function(x, y, width) {
    this._marker.style.width = width + "px";
    this._marker.openPopupAtScreen(x, y, false);
    this._marker.moveTo(x, y);
  },
  
  /**
   * マーカーを消す
   */
  hide: function() {
    this._marker.hidePopup();
  },
  
  /* ==== private ==== */
  
  _marker: null, /** マーカー要素
                  *  @type {nsIDOMXULElement} */
  
};

/**
 * データのリストビューによる操作
 *
 * @param   {nsIDOMXULElement} tree
 *          tree 要素
 * @param   {nsIDOMXULElement} cols
 *          treecols 要素
 *          ソートを必要としない場合は null
 * @param   {nsIDOMXULElement} treechildren
 *          treechildren 要素
 * @param   {nsIDOMXULElement} newButton
 *          追加ボタン
 * @param   {nsIDOMXULElement} dupButton
 *          複製ボタン
 *          無い場合 null
 * @param   {nsIDOMXULElement} delButton
 *          削除ボタン
 * @param   {string} type
 *          ドラッグ中の行の MIME-Type
 * @param   {Array.<number>} ignoreColumns
 *          ドラッグアンドドロップの対象にしないカラム
 *            [number カラムの位置, ...]
 * @param   {boolean} isEditable
 *          tree が直接編集可能か
 * @param   {Array.<ANY>} datas
 *          表示するデータ
 *            [ANY 要素, ...]
 * @param   {arIListViewDataHandler} dataHandler
 *          データハンドラ
 */
function arListView(tree, cols, treechildren,
                    newButton, dupButton, delButton,
                    type, ignoreColumns, isEditable,
                    datas,
                    dataHandler) {
  this._tree = tree;
  this._cols = cols;
  this._treechildren = treechildren;
  this._newButton = newButton;
  this._dupButton = dupButton;
  this._delButton = delButton;
  this._type = type;
  this._ignoreColumns = ignoreColumns;
  this._isEditable = isEditable;
  this._datas = datas;
  this._dataHandler = dataHandler;
  
  this._sortAsc = [];
  this._marker = new arListViewDnDMarker(this._tree);
  
  this._init();
}
arListView.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    if (this._cols) {
      Array.forEach(
        this._cols.getElementsByTagName("treecol"),
        function(col) {
          col.removeEventListener("click", this.onTreeColClick, false);
        }.bind(this));
    }
    
    this._tree.removeEventListener("dragstart", this.onTreeDragStart, false);
    this._tree.removeEventListener("dragover", this.onTreeDragOver, false);
    this._tree.removeEventListener("dragend", this.onTreeDragEnd, false);
    
    this._tree.removeEventListener("select", this.onTreeSelect, false);
    if (this._isEditable) {
      this._tree.removeEventListener("change", this.onTreeChange, false);
    }
    
    this._newButton.removeEventListener("command", this.onNewButtonClick, false);
    if (this._dupButton) {
      this._dupButton.removeEventListener("command", this.onDupButtonClick, false);
    }
    this._delButton.removeEventListener("command", this.onDelButtonClick, false);
    
    this._tree = null;
    this._cols = null;
    this._treechildren = null;
    this._newButton = null;
    this._dupButton = null;
    this._delButton = null;
    this._type = null;
    this._ignoreColumns = null;
    this._dataHandler = null;
    this._datas = null;
    
    this._marker.destruct();
    this._marker = null;
  },
  
  /**
   * 選択中のデータを取得する
   *
   * @returns {Array}
   *            [ANY  選択中の要素, number  選択中の行のインデックス]
   */
  getSelectedItems: function() {
    return this._datas
      .map(function(item, i) {
        return [item, i];
      }.bind(this))
      .filter(function([item, i]) {
        return this._tree.view.selection.isSelected(i);
      }.bind(this));
  },
  
  /**
   * 非選択中のデータを取得する
   *
   * @returns {Array}
   *            [ANY  非選択中の要素, number  非選択中の行のインデックス]
   */
  getUnselectedItems: function() {
    return this._datas
      .map(function(item, i) {
        return [item, i];
      }.bind(this))
      .filter(function([item, i]) {
        return !this._tree.view.selection.isSelected(i);
      }.bind(this));
  },
  
  /**
   * 全てのデータを更新する
   *
   * @param   {Array.<ANY>} datas
   *          表示するデータ
   *            [ANY 要素, ...]
   */
  setDatas: function(datas) {
    this._datas = datas;
    
    /* tree を直接編集中かもしれないので不用意に要素を削除してはいけない */
    
    this._doNotCatchSelected = true;
    if (this._treechildren.childNodes.length > this._datas.length) {
      for (let count = this._treechildren.childNodes.length - datas.length;
           count > 0; count --) {
        this._treechildren.removeChild(this._treechildren.lastChild);
      }
    }
    this._doNotCatchSelected = false;
    
    this._datas.forEach(function(item, i) {
      if (i < this._treechildren.childNodes.length) {
        let listitem = this._treechildren.childNodes[i];
        this._changeRow(item, i);
      } else {
        this._addRow(item, -1, false);
      }
    }.bind(this));
    
    if (this._datas.length) {
      this.select(0);
    }
  },
  
  /**
   * データを追加する
   *
   * @param   {ANY} item
   *          対応する要素
   * @param   {number} pos
   *          追加する位置
   *          -1 にすると末尾
   * @param   {boolean} _select
   *          追加した行を選択するか
   */
  addData: function(item, pos, _select) {
    if (pos < 0 || pos > this._datas.length) {
      pos = this._datas.length;
    }
    
    this._datas
      = [...this._datas.slice(0, pos), item, ...this._datas.slice(pos)];
    
    this._addRow(item, pos, _select);
    
    try {
      this._dataHandler.onModifyDatas(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /**
   * 要素を選択する
   *
   * @param   {number} pos
   *          行のインデックス
   */
  select: function(pos) {
    if (pos < 0 && pos >= this._treechildren.childNodes.length) {
      return;
    }
    
    this._tree.view.selection.select(pos);
  },
  
  /**
   * 複数要素を要素を選択する
   *
   * @param   {Array.<number>} selectedIndices
   *          行のインデックス
   *            [number pos, ...]
   */
  selectRange: function(selectedIndices) {
    selectedIndices.forEach(function(pos, i) {
      if (pos < 0 && pos >= this._treechildren.childNodes.length) {
        return;
      }
      
      this._doNotCatchSelected = true;
      if (i == 0) {
        this._tree.view.selection.select(pos);
      } else {
        this._tree.view.selection.rangedSelect(pos, pos, true);
      }
      this._doNotCatchSelected = false;
    }.bind(this));
    
    let items = this.getSelectedItems();
    
    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /**
   * 選択イベントを強制的に発行する
   * イベント中の場合発行されないため
   */
  dispatchSelect: function() {
    let items = this.getSelectedItems();
    
    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /**
   * 選択中のデータを更新する
   *
   * @param   {Function} modifyFunc
   *          選択中のデータを更新する関数
   *          選択していない、または複数選択している場合は null
   *            function(item)
   *              *param   {ANY} item
   *                       選択中のデータ
   * @returns {ANY}
   *          選択中の要素
   */
  modifySelectedData: function(modifyFunc) {
    if (this._tree.view.selection.count != 1) {
      return null;
    }
    
    return this.modifyDataAt(this._tree.view.selection.currentIndex, modifyFunc);
  },
  
  /**
   * データを更新する
   *
   * @param   {Function} modifyFunc
   *          選択中のデータを更新する関数
   *          選択していない、または複数選択している場合は null
   *            function(item)
   *              *param   {ANY} item
   *                       選択中のデータ
   * @param   {number} index
   *          インデックス
   * @returns {ANY}
   *          対象の要素
   */
  modifyDataAt: function(index, modifyFunc) {
    if (index < 0 || index >= this._datas.length) {
      return null;
    }
    
    let item = this._datas[index];
    modifyFunc(item);
    
    let treerow = this._treechildren.childNodes[index].firstChild;
    
    this._dataHandler.dataToRow(item)
      .forEach(function(cellData, i) {
        let cell = treerow.childNodes[i];
        this._updateCell(cell, cellData);
      }.bind(this));
    
    return item;
  },
  
  /* ---- UI イベント ---- */
  
  /**
   * 追加ボタンをクリックしたイベント
   * 選択中のデータの次にデータを追加する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onNewButtonClick: function(event) {
    let pos = this._tree.view.selection.currentIndex + 1;
    
    let item = this._dataHandler.newData();
    
    this.addData(item, pos, true);
  },

  /**
   * 追加ボタンをクリックしたイベント
   * 選択中のデータの次にデータを追加する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDupButtonClick: function(event) {
    let items = this.getSelectedItems();
    
    let n = 0;
    
    let selectedIndices
      = items.map(function([origItem, origPos]) {
        let item = this._dataHandler.dupData(origItem);
        let pos = n + origPos + 1;
        
        this.addData(item, pos, false);
        
        n ++;
        
        return pos;
      }.bind(this));
    
    this.selectRange(selectedIndices);
  },
  
  /**
   * 削除ボタンをクリックしたイベント
   * 選択中のデータを削除する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onDelButtonClick: function(event) {
    let selectedItems = this.getSelectedItems();
    let unselectedItems = this.getUnselectedItems();
    
    selectedItems
      .map(function([item, i]) {
        return this._treechildren.childNodes[i];
      }.bind(this))
      .forEach(function(row) {
        row.parentNode.removeChild(row);
      }.bind(this));
    
    this._datas = unselectedItems.map(function([item, i]) item);
    
    try {
      this._dataHandler.onModifyDatas(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
    
    if (selectedItems.length) {
      this.select(selectedItems[0][1]);
    }
  },
  
  /**
   * ツリーのカラムをクリックしたイベント
   * そのカラムでソートする
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeColClick: function(event) {
    if (!this._cols) {
      return;
    }
    
    if (this._sortAsc.length == 0) {
      this._sortAsc
        = [false for (i of arArrayUtil.range(0, this._cols.childNodes.length))];
    }
    
    let pos = Array.indexOf(this._cols.getElementsByTagName("treecol"),
                            event.target);
    if (pos == -1) {
      return;
    }
    
    let descFunc = function(a, b) a[0] < b[0];
    let ascFunc  = function(a, b) a[0] > b[0];
    let sortFunc = descFunc;
    
    let asc = this._sortAsc[pos];
    Array.forEach(this._cols.childNodes,
                  function(col, i) {
                    if (i == pos) {
                      col.setAttribute("sortDirection",
                                       asc ? "ascending" : "descending");
                    } else {
                      col.setAttribute("sortDirection", "natural");
                    }
                  });
    if (this._sortAsc[pos]) {
      sortFunc = ascFunc;
    }
    this._sortAsc[pos] = !this._sortAsc[pos];
    
    this._doNotCatchSelected = true;
    this._datas
      = this._datas
      .map(function(item, i) {
        let key;
        
        try {
          key = this._dataHandler.getSortKeyFor(item, pos);
        } catch (e) {
          Cu.reportError(e + "\n" + e.stack);
        }
        
        return [key, item, this._treechildren.childNodes[i]];
      }.bind(this))
      .sort(sortFunc)
      .map(function([label, item, node]) {
        this._treechildren.appendChild(node);
        return item;
      }.bind(this));
    this._doNotCatchSelected = false;
    
    try {
      this._dataHandler.onModifyDatas(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /**
   * ツリーの行をドラッグ開始したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragStart: function(event) {
    if (event.target != this._treechildren) {
      return;
    }
    
    let row = {}, col = {}, child = {};
    
    this._tree.treeBoxObject.getCellAt(event.clientX, event.clientY,
                                       row, col, child);
    
    if (col.value) {
      if (this._ignoreColumns.indexOf(col.value.index) != -1) {
        return;
      }
    }
    
    let items = this.getSelectedItems();
    if (items.length == 0) {
      return;
    }
    let firstRow = this._treechildren.childNodes[items[0][1]];
    
    event.dataTransfer.mozSetDataAt(this._type, JSON.stringify(items), 0);
    event.dataTransfer.mozSetDataAt(this._type + "-node", firstRow, 1);
    event.stopPropagation();
  },
  
  /**
   * ツリーの上でドラッグしているイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragOver: function(event) {
    if (event.target != this._treechildren) {
      return;
    }
    
    let draggingItems = this._getDraggingItems(event.dataTransfer);
    if (!draggingItems) {
      this._currentRow = -1;
      return;
    }
    
    let row = {}, col = {}, child = {};
    
    this._tree.treeBoxObject.getCellAt(event.clientX, event.clientY,
                                       row, col, child);    
    
    this._currentRow = row.value;
    if (this._currentRow == -1) {
      this._currentRow = this._treechildren.childNodes.length - 1;
    }
    
    let tbo = this._tree.treeBoxObject;
    let rowHeight = tbo.rowHeight;
    let eventY
      = event.clientY - tbo.treeBody.boxObject.y
      - rowHeight * (this._currentRow - tbo.getFirstVisibleRow());
    
    let x = {}, y = {};
    tbo.getCoordsForCellItem(this._currentRow, tbo.columns[0], "cell",
                             x, y, {}, {});
    
    x.value += event.screenX - event.clientX + this._treechildren.boxObject.x;
    y.value += event.screenY - event.clientY + this._treechildren.boxObject.y;
    
    this._orientation = 0;
    if (row.value == -1 || eventY > rowHeight * 0.5) {
      y.value += rowHeight;
      this._orientation = 1;
    }
    this._marker.show(x.value, y.value, this._treechildren.boxObject.width);
    
    event.preventDefault();
    event.stopPropagation();
  },
  
  /**
   * ツリーの上にドロップしたイベント
   * ドロップした位置にデータを移動する
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeDragEnd: function(event) {
    if (this._currentRow == -1) {
      return;
    }
    
    this._marker.hide();
    
    let draggingItems = this._getDraggingItems(event.dataTransfer);
    if (!draggingItems || draggingItems.length == 0) {
      this._currentRow = -1;
      return;
    }
    
    let markerPos = this._currentRow + this._orientation;
    
    let nextPos = markerPos;
    while (draggingItems.some(function([item, pos]) pos == nextPos)) {
      nextPos ++;
    }
    let nextRow
      = let (nodes = this._treechildren.childNodes)
    ((nextPos < nodes.length) ? nodes[nextPos] : null);
    
    let insertPos
      = this._datas
      .filter(function(item, i)
              draggingItems.every(function([item, pos]) pos != i) &&
              i < nextPos).length;
    
    let tmpData
      = this._datas
      .filter(function(item, i)
              draggingItems.every(function([item, j]) i != j));
    let nextData = tmpData.splice(insertPos);
    
    this._datas
      = [...tmpData,
         ...draggingItems.map(function([item, i]) item),
         ...nextData];
    
    this._doNotCatchSelected = true;
    draggingItems
      .map(function([item, i]) {
        return this._treechildren.childNodes[i];
      }.bind(this))
      .forEach(function(row) {
        row.parentNode.insertBefore(row, nextRow);
      });
    this._doNotCatchSelected = false;
    
    try {
      this._dataHandler.onModifyDatas(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
    
    this.selectRange(draggingItems.map(function(_tmp, i) insertPos + i));
  },
  
  /**
   * ツリーの行を選択したイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeSelect: function(event) {
    if (this._doNotCatchSelected) {
      return;
    }
    
    let items = this.getSelectedItems();
    
    try {
      this._dataHandler.onSelect(items);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /**
   * ツリーのテキストが編集されたイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onTreeChange: function(event) {
    this._datas
      = this._datas.map(function(item, i) {
        let row = [];
        for (let j = 0, len = this._tree.columns.length; j < len; j ++) {
          let cell = {
            label: this._tree.view.getCellText(i, this._tree.columns[j])
          };
          row.push(cell);
        }
        
        return this._dataHandler.rowToData(row, item);
      }.bind(this));
    
    try {
      this._dataHandler.onModifyDatas(this._datas);
    } catch (e) {
      Cu.reportError(e + "\n" + e.stack);
    }
  },
  
  /* ==== private ==== */

  _sortAsc: null,  /** 各カラムについてソートが昇順であるか
                    *  @type {Array.<boolean>} */
  
  _marker: null, /** ドラッグアンドドロップ時のマーカー
                  *  @type {arListViewDnDMarker} */
  
  _doNotCatchSelected: false, /** onselect イベントをキャッチしない
                               *  @type {boolean} */
  
  /* ---- コンストラクタの引数保持 ---- */
  _tree: null,
  _cols: null,
  _treechildren: null,
  _newButton: null,
  _dupButton: null,
  _delButton: null,
  _type: null,
  _ignoreColumns: null,
  _isEditable: false,
  _dataHandler: null,
  _datas: null,
  
  /**
   * 初期化処理
   */
  _init: function() {
    this.onTreeColClick = this.onTreeColClick.bind(this);
    
    this.onTreeDragStart = this.onTreeDragStart.bind(this);
    this.onTreeDragOver = this.onTreeDragOver.bind(this);
    this.onTreeDragEnd = this.onTreeDragEnd.bind(this);

    this.onTreeSelect = this.onTreeSelect.bind(this);
    this.onTreeChange = this.onTreeChange.bind(this);
    
    this.onNewButtonClick = this.onNewButtonClick.bind(this);
    this.onDupButtonClick = this.onDupButtonClick.bind(this);
    this.onDelButtonClick = this.onDelButtonClick.bind(this);
    
    if (this._cols) {
      Array.slice(this._cols.getElementsByTagName("treecol"))
        .forEach(function(col) {
          col.addEventListener("click", this.onTreeColClick, false);
        }.bind(this));
    }
    
    this._tree.addEventListener("dragstart", this.onTreeDragStart, false);
    this._tree.addEventListener("dragover", this.onTreeDragOver, false);
    this._tree.addEventListener("dragend", this.onTreeDragEnd, false);

    this._tree.addEventListener("select", this.onTreeSelect, false);
    if (this._isEditable) {
      this._tree.addEventListener("change", this.onTreeChange, false);
    }
    
    this._newButton.addEventListener("command", this.onNewButtonClick, false);
    if (this._dupButton) {
      this._dupButton.addEventListener("command", this.onDupButtonClick, false);
    }
    this._delButton.addEventListener("command", this.onDelButtonClick, false);
    
    this.setDatas(this._datas);
  },
  
  /**
   * ドラッグ中の行データを取得する
   *
   * @param   {DataTransfer} dataTransfer
   *          ドラッグのデータ
   * @returns {Array.<[ANY,number]>}
   *            [[ANY 要素, number 行のインデックス], ...]
   */
  _getDraggingItems: function(dataTransfer) {
    for (let i = 0; i < dataTransfer.mozItemCount; i ++) {
      if (Array.indexOf(dataTransfer.mozTypesAt(i), this._type) != -1) {
        return JSON.parse(dataTransfer.mozGetDataAt(this._type, i));
      }
    }
    
    return null;
  },
  
  /**
   * 行を追加する
   *
   * @param   {ANY} item
   *          対応する要素
   * @param   {number} pos
   *          追加する位置 (行のインデックス)
   *          -1 にすると末尾
   * @param   {boolean} _select
   *          追加した行を選択するか
   */
  _addRow: function(item, pos, _select) {
    let document = this._tree.ownerDocument;

    let nextSibling = null;
    if (pos >= 0 && pos < this._treechildren.childNodes.length) {
      nextSibling = this._treechildren.childNodes[pos];
    }
    
    let treeitem = document.createElement("treeitem");
    this._doNotCatchSelected = true;
    this._treechildren.insertBefore(treeitem, nextSibling);
    this._doNotCatchSelected = false;
    
    let treerow = document.createElement("treerow");
    treeitem.appendChild(treerow);
    
    this._dataHandler.dataToRow(item)
      .forEach(function(cellData) {
        let cell = document.createElement("treecell");
        this._updateCell(cell, cellData);
        treerow.appendChild(cell);
      }.bind(this));
    
    if (_select) {
      this.select(pos);
    }
  },

  /**
   * 行を修正する
   *
   * @param   {ANY} item
   *          対応する要素
   * @param   {number} pos
   *          修正する位置 (行のインデックス)
   */
  _changeRow: function(item, pos) {
    let document = this._tree.ownerDocument;

    let treeitem = null;
    if (pos >= 0 && pos < this._treechildren.childNodes.length) {
      treeitem = this._treechildren.childNodes[pos];
    } else {
      return;
    }
    
    let cells = treeitem.getElementsByTagName("treecell");
    
    this._dataHandler.dataToRow(item)
      .forEach(function(cellData, i) {
        let cell = cells[i];
        
        this._updateCell(cell, cellData);
      }.bind(this));
  },
  
  /**
   * セルの表示を更新する
   *
   * @param   {nsIDOMXULElement} cell
   *          セル
   * @param   {object} cellData
   *          セルのデータ
   */
  _updateCell: function(cell, cellData) {
    if ("label" in cellData) {
      if (!cell.hasAttribute("label") ||
          cell.getAttribute("label") != cellData.label) {
        cell.setAttribute("label", cellData.label);
      }
    } else {
      if (cell.hasAttribute("label")) {
        cell.removeAttribute("label");
      }
    }
    
    if ("src" in cellData) {
      if (!cell.hasAttribute("src") ||
          cell.getAttribute("src") != cellData.src) {
        cell.setAttribute("src", cellData.src);
      }
    } else {
      if (cell.hasAttribute("src")) {
        cell.removeAttribute("src");
      }
    }
    
    if ("className" in cellData) {
      if (!("className" in cell) ||
          cell.className != cellData.className) {
        cell.className = cellData.className;
      }
    } else {
      if ("className" in cell &&
          cell.className != "") {
        cell.className = "";
      }
    }
  }
};
