"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arPrototype.jsm");
XPCOMUtils.importRelative(this, "arTimer.jsm");
XPCOMUtils.defineLazyGetter(this, "arArrayUtil", function() {
  XPCOMUtils.importRelative(this, "arArrayUtil.jsm");
  return arArrayUtil;
});
XPCOMUtils.defineLazyGetter(this, "arStylesheetLoader", function() {
  XPCOMUtils.importRelative(this, "arStylesheetLoader.jsm");
  return arStylesheetLoader;
});

let EXPORTED_SYMBOLS = ["arOverlayLoader"];

/**
 * 要素に対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arElementOverlay(owindow, node) {
  this._owindow = owindow;
  this._node = node;
}
arElementOverlay.prototype = {
  /**
   * 対象の要素
   * @type {nsIDOMXULElement}
   */
  get node() {
    return this._node;
  },
  
  /**
   * イベントハンドラを追加する
   *
   * @param   {string} type
   *          イベントの種類
   * @param   {Function} fn
   *          イベントハンドラ
   */
  addEventListener: function(type, fn) {
    this._owindow.addEventListener(this._node, type, fn);
    return this;
  },
  
  /**
   * 要素を追加する
   *
   * @param   {string} nodeName
   *          要素名
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @param   {Constructor} prototype
   *          コンストラクタ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addElement: function(nodeName, attrs, oncommand, prototype) {
    return this._owindow.addElement(this._node, nodeName, attrs, oncommand, prototype);
  },
  
  /**
   * label 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addLabel: function(attrs, oncommand) {
    return this.addElement("label", attrs, oncommand);
  },

  /**
   * textbox 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addTextBox: function(attrs, oncommand) {
    return this.addElement("textbox", attrs, oncommand);
  },

  /**
   * button 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addButton: function(attrs, oncommand) {
    return this.addElement("button", attrs, oncommand);
  },

  /**
   * menupopup 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addMenuPopup: function(attrs, oncommand) {
    return this.addElement("menupopup", attrs, oncommand, arMenuPopupOverlay);
  },

  /**
   * panel 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addPanel: function(attrs, oncommand) {
    return this.addElement("panel", attrs, oncommand);
  },
  
  /**
   * grid 要素を追加する
   *
   * @param   {object} attrs
   *          grid 要素の属性
   * @param   {object} params
   *          行や列
   * @returns {arGridOverlay}
   *          要素に対するオーバーレイ
   */
  addGrid: function(attrs, params) {
    return this._owindow.addGrid(this._node, attrs, params);
  },

  /* ==== private ==== */
  
  _owindow: null, /** ウィンドウのオーバーレイ
                   *  @type {arWindowOverlay} */
  _node: null     /** 対象の要素
                   *  @type {nsIDOMXULElement} */
};

/**
 * ツールバーに対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arToolbarOverlay(owindow, node) {
  arElementOverlay.call(this, owindow, node);
}
arToolbarOverlay.prototype = {
  /**
   * toolbaritem 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addItem: function(attrs, oncommand) {
    return this._addToolbarElement("toolbaritem", attrs, oncommand);
  },

  /**
   * toolbarbutton 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addButton: function(attrs, oncommand) {
    return this._addToolbarElement("toolbarbutton", attrs, oncommand);
  },
  
  /* ==== private ==== */

  /**
   * ツールバーパレットに要素を追加する
   *
   * @param   {string} nodeName
   *          要素名
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  _addToolbarElement: function(nodeName, attrs, oncommand) {
    let onode = this._owindow.addElement(null, nodeName, attrs, oncommand);
    let node = onode.node;
    let window = this._owindow.window;
    
    let toolbox = arArrayUtil.mapfind(
      [
        "navigator-toolbox",
        "mail-toolbox"
        ],
      function(id) window.document.getElementById(id),
      function(n) n);
    if (!toolbox) {
      Cu.reportError("toolbox not found.");
      return onode;
    }
    
    toolbox.palette.appendChild(node);
    
    let [toolbar, currentset, pos] = arArrayUtil.mapfind(
      window.document.getElementsByTagName("toolbar"),
      function(toolbar) {
        let currentset = toolbar.getAttribute("currentset").split(",");
        return [toolbar, currentset, currentset.indexOf(node.id)];
      },
      function([toolbar, currentset, pos]) pos != -1) ||
      [null, null, -1];
    if (!toolbar) {
      return onode;
    }
    
    let before = (pos < currentset.length - 1)
    ? this._findBeforeNode(toolbar, currentset, pos)
    : null;
        
    if (!toolbar.toolbox) {
      /* Personal Titlebar で追加したツールバーは toolbox を持たないので
       * パレットに追加したものは insertItem で追加できない */
      node.setAttribute("removable", "true");
      try {
        toolbar.insertBefore(node, before);
      } catch (e) {
        /* 他のアドオンと操作が競合して失敗するかもしれないので
         * その場合末尾に追加する */
        toolbar.insertBefore(node, null);
      }
          
      return onode;
    }
        
    try {
      toolbar.insertItem(node.id, before);
    } catch (e) {
      /* 他のアドオンと操作が競合して失敗するかもしれないので
       * その場合末尾に追加する */
      toolbar.insertItem(node.id, null);
    }
    
    return onode;
  },

  /**
   * ツールバーの項目の直前の要素を探す
   *
   * @param   {nsIDOMXULElement} toolbar
   *          対象の項目の要素
   * @param   {Array} toolbar
   *          ツールバーの currentset 属性をカンマで分割した配列
   * @param   {number} pos
   *          対象の項目の位置
   * @returns {nsIDOMXULElement}
   *          ツールバーの項目の直前の要素
   *          見付からなければ null
   */
  _findBeforeNode: function(toolbar, currentset, pos) {
    return arArrayUtil.mapfind(
      currentset.slice(pos + 1),
      function(beforeId) {
        if (beforeId == "separator" ||
            beforeId == "spring" ||
            beforeId == "spacer") {
          /* 複数出現可能な id を持たない要素は出現位置から取得する */
          let beforePos = currentset.slice(0, pos)
          .filter(function(n) n == beforeId)
          .length;
          
          let nodes = Array.filter(toolbar.getElementsByTagName("toolbar" + beforeId),
                                   function(n) n.parentNode == toolbar);
          if (beforePos < nodes.length) {
            return nodes[beforePos];
          }
        
          return null;
        }
      
        let before = null;
        try {
          before = toolbar.querySelector("#" + beforeId);
        } catch (e) {
          /* Collusion 等で不正な文字列が ID として使用されており
           * そのまま使用すると querySelector が例外を投げる */
          before = toolbar.ownerDocument.getElementById(beforeId);
        }
        if (before) {
          if (before.parentNode != toolbar) {
            /* forecastfox-toolbar は初期状態で status-bar 内に居るが
             * 設定により後から addon-bar に移動する事があり
             * 状況によっては status-bar 内の forecastfox-toolbar が
             * 取得されてしまう事がある
             * この場合 insertItem の引数としては適切ではない */
            return null;
          }
        
          return before;
        }
        
        return null;
      },
      function(n) n) || null;
  }
};
arPrototype.extend(arToolbarOverlay, arElementOverlay);

/**
 * statusbar 要素に対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arStatusbarOverlay(owindow, node) {
  arElementOverlay.call(this, owindow, node);
}
arStatusbarOverlay.prototype = {
  /**
   * statusbarpanel 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addPanel: function(attrs, oncommand) {
    return this.addElement("statusbarpanel", attrs, oncommand);
  }
};
arPrototype.extend(arStatusbarOverlay, arElementOverlay);

/**
 * menupopup 要素に対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arMenuPopupOverlay(owindow, node) {
  arElementOverlay.call(this, owindow, node);
}
arMenuPopupOverlay.prototype = {
  /**
   * menuitem 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addItem: function(attrs, oncommand) {
    return this.addElement("menuitem", attrs, oncommand);
  },

  /**
   * menuseparator 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addSeparator: function(attrs, oncommand) {
    return this.addElement("menuseparator", attrs, oncommand);
  },

  /**
   * menu 要素と menupopup 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          menupopup 要素に対するオーバーレイ
   */
  addMenu: function(attrs, oncommand) {
    let omenu = this.addElement("menu", attrs, oncommand);
    return this._owindow.addElement(omenu.node, "menupopup", {}, null, arMenuPopupOverlay);
  }
};
arPrototype.extend(arMenuPopupOverlay, arElementOverlay);

/**
 * urlbar-icons 要素に対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arURLBarIconsOverlay(owindow, node) {
  arElementOverlay.call(this, owindow, node);
}
arURLBarIconsOverlay.prototype = {
  /**
   * image 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addIcon: function(attrs, oncommand) {
    return this.addElement("image", attrs, oncommand);
  }
};
arPrototype.extend(arURLBarIconsOverlay, arElementOverlay);

/**
 * grid 要素に対するオーバーレイ
 *
 * @param   {arWindowOverlay} owindow
 *          オーバーレイ
 * @param   {nsIDOMXULElement} node
 *          対象の要素
 */
function arGridOverlay(owindow, node) {
  arElementOverlay.call(this, owindow, node);
}
arGridOverlay.prototype = {
  /**
   * row 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addRow: function(attrs, oncommand) {
    if (!this._orows) {
      this._orows = this.addElement("rows");
    }
    return this._orows.addElement("row", attrs, oncommand);
  },
  
  /**
   * column 要素を追加する
   *
   * @param   {object} attrs
   *          要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addColumn: function(attrs, oncommand) {
    if (!this._ocolumns) {
      this._ocolumns = this.addElement("columns");
    }
    return this._ocolumns.addElement("column", attrs, oncommand);
  },

  /* ==== private ==== */
  
  _orows: null,   /** rows 要素のオーバーレイ
                   *  @type {arElementOverlay} */
  _ocolumns: null /** columns 要素のオーバーレイ
                   *  @type {arElementOverlay} */
};
arPrototype.extend(arGridOverlay, arElementOverlay);

/**
 * オーバーレイ情報
 */
function arWindowOverlayInfo() {
  this.events = [];
  this.nodes = [];
}
arWindowOverlayInfo.prototype = {
  events: null, /** 登録したイベント
                  *  @type {Array.<[nsIDOMXULElement,string,Function]>} */
  nodes: null,  /** 追加した要素
                  *  @type {Array.<nsIDOMXULElement>} */
  
  /**
   * 破棄
   */
  destruct: function() {
    this.events = null;
    this.nodes = null;
  },

  /**
   * オーバーレイを解除する
   */
  unload: function() {
    this.events.forEach(function([node, type, fn]) {
      try {
        node.removeEventListener(type, fn, false);
      } catch (e) {
        /* 対象が既に閉じている等 */
      }
    });
    this.events = [];
    this.nodes.forEach(function(node) {
      if (node.parentNode) {
        try {
          node.parentNode.removeChild(node);
        } catch (e) {
          /* 対象が既に閉じている等 */
        }
      }
    });
    this.nodes = [];
  }
};

/**
 * ウィンドウに対するオーバーレイ
 *
 * @param   {nsIXULWindow} window
 *          対象のウィンドウ
 */
function arWindowOverlay(window, oinfo) {
  this._window = window;
  this._oinfo = oinfo;
  
  let $ = function(id) {
    let node = window.document.getElementById(id);
    if (!node) {
      Cu.reportError("not found: " + id);
    }
    return node;
  };
  let $$ = function(selector) {
    let node = window.document.querySelector(selector);
    if (!node) {
      Cu.reportError("not found: " + selector);
    }
    return node;
  };
  
  XPCOMUtils.defineLazyGetter(this, "fileMenu", function() {
    return new arMenuPopupOverlay(this, $("menu_FilePopup"));
  });
  XPCOMUtils.defineLazyGetter(this, "toolsMenu", function() {
    let menu = $("menu_ToolsPopup");
    if (!menu) {
      menu = $("taskPopup");
    }
    return new arMenuPopupOverlay(this, menu);
  });
  XPCOMUtils.defineLazyGetter(this, "contentMenu", function() {
    return new arMenuPopupOverlay(this, $("contentAreaContextMenu"));
  });
  XPCOMUtils.defineLazyGetter(this, "frameContentMenu", function() {
    return new arMenuPopupOverlay(this, $$("#frame > menupopup"));
  });
  XPCOMUtils.defineLazyGetter(this, "toolbar", function() {
    return new arToolbarOverlay(this);
  });
  XPCOMUtils.defineLazyGetter(this, "statusBar", function() {
    return new arStatusbarOverlay(this, $("status-bar"));
  });
  XPCOMUtils.defineLazyGetter(this, "mainPopupSet", function() {
    return new arMenuPopupOverlay(this, $("mainPopupSet"));
  });
  XPCOMUtils.defineLazyGetter(this, "URLBarIcons", function() {
    return new arURLBarIconsOverlay(this, $("urlbar-icons"));
  });
};
arWindowOverlay.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    this._window = null;
    this._oinfo = null;
  },
  
  /**
   * 対象のウィンドウ
   * @type {nsIXULWindow}
   */
  get window() {
    return this._window;
  },
  
  /**
   * 要素のオーバーレイを取得する
   *
   * @param   {string} id
   *          ID
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  getElement: function(id) {
    let node = this._window.document.getElementById(id);
    if (!node) {
      Cu.reportError("not found: " + id);
    }
    return new arElementOverlay(this, node);
  },
  
  /**
   * スタイルシートを登録する
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   */
  loadCSS: function(uriSpec) {
    arStylesheetLoader.load(uriSpec, Ci.nsIStyleSheetService.AGENT_SHEET);
  },
  
  /**
   * イベントハンドラを追加する
   *
   * @param   {nsIDOMXULElement} node
   *          対象の要素
   * @param   {string} type
   *          イベントの種類
   * @param   {Function} fn
   *          イベントハンドラ
   */
  addEventListener: function(node, type, fn) {
    if (!node) {
      return;
    }
    
    this._oinfo.events.push([node, type, fn]);
    node.addEventListener(type, fn, false);
  },
  
  /**
   * 要素を追加する
   *
   * @param   {nsIDOMXULElement} parentNode
   *          親要素
   * @param   {string} nodeName
   *          要素名
   * @param   {object} attrs
   *          grid 要素の属性
   * @param   {Function} oncommand
   *          command イベントのハンドラ
   * @param   {Constructor} prototype
   *          コンストラクタ
   * @returns {arElementOverlay}
   *          要素に対するオーバーレイ
   */
  addElement: function(parentNode, nodeName, attrs, oncommand, prototype) {
    let node = this._window.document.createElement(nodeName);
    this._oinfo.nodes.push(node);
    if (attrs) {
      if ("id" in attrs) {
        node.id = attrs.id;
      }
      for (let name in attrs) {
        let value = attrs[name];
        if (name == "id" ||
            name == "insertbefore" ||
            name == "insertafter" ||
            name == "POSITION") {
          continue;
        }
        node.setAttribute(name, value);
      }
    }
    if (oncommand) {
      this.addEventListener(node, "command", oncommand);
    }
    
    if (parentNode) {
      let before = null;
      if (attrs) {
        if ("insertbefore" in attrs) {
          before = arArrayUtil.mapfind(
            attrs["insertbefore"].split(","),
            function(id) parentNode.querySelector("#" + id),
            function(n) n && n.parentNode == parentNode) || null;
        } else if ("insertafter" in attrs) {
          let after = arArrayUtil.mapfind(
            attrs["insertafter"].split(","),
            function(id) parentNode.querySelector("#" + id),
            function(n) n && n.parentNode == parentNode) || null;
          before = after ? after.nextSibling : null;
        } else if ("POSITION" in attrs) {
          let pos = attrs["POSITION"];
          if (pos == "first") {
            before = parentNode.firstChild;
          } else if (pos == "last") {
            before = null;
          }
        }
      }
      parentNode.insertBefore(node, before);
    }
    
    if (prototype) {
      return new prototype(this, node);
    }
    return new arElementOverlay(this, node);
  },
  
  /**
   * グリッドを追加する
   *
   * @param   {nsIDOMXULElement} parentNode
   *          親要素
   * @param   {object} attrs
   *          grid 要素の属性
   * @param   {object} params
   *          行や列
   * @returns {arGridOverlay}
   *          grid 要素に対するオーバーレイ
   */
  addGrid: function(parentNode, attrs, params) {
    let onode = this.addElement(parentNode, "grid", attrs, null, arGridOverlay);
    
    for (let name in params) {
      let value = params[name];
      if (name == "columns") {
        for (let attrs of value) {
          onode.addColumn(attrs);
        }
      } else if (name == "rows") {
        onode.addRow(attrs);
      }
    }
    
    return onode;
  },
  
  /* ==== private ==== */
  
  _window: null /** 対象のウィンドウ
                 *  @type {nsIXULWindow} */
};

/**
 * オーバーレイの適用と解除
 */
let arOverlayLoader = {
  /**
   * オーバーレイをロードする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {Function} loadFunc
   *          オーバーレイ登録関数
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   */
  loadInto: function(window, loadFunc, isStartup) {
    let oinfo = new arWindowOverlayInfo();
    let owindow = new arWindowOverlay(window, oinfo);
    
    let f = function() {
      loadFunc(window, owindow);
      owindow.destruct();
    };
    
    if (isStartup) {
      /* スタートアップイベント中に insertItem を呼ぶと
       * ツールバーが強制的に表示されるため遅延させる */
      arTimer.setTimeout(f, 100);
    } else {
      f();
    }
    
    this._overlays.push([window, loadFunc, oinfo]);
  },
  
  /**
   * オーバーレイをアンロードする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {Function} loadFunc
   *          オーバーレイ登録関数
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   */
  unloadFrom: function(window, loadFunc, isShutdown) {
    this._overlays = this._overlays.filter(function([w, f, oinfo]) {
      if (w == window && f == loadFunc) {
        /* CSS の登録は load 時で良いが、解除はアドオン全体の終了時だけ */
        oinfo.unload();
        oinfo.destruct();
        return false;
      }
      return true;
    }.bind(this));
  },
  
  /* ==== private ==== */
  
  _overlays: [] /** 適用したウィンドウとオーバーレイ情報
                 *  @type {Array.<[Window,Function,arWindowOverlayInfo]>} */
};
