"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

let EXPORTED_SYMBOLS = ["arPathUtil"];

/**
 * パス情報
 */
function arPathInfo() {
  this.pathComponents = [];
}
arPathInfo.prototype = {
  type: 0,                    /** パスの種類
                               *  @type {number}
                               *    arPathUtil.TYPE_* */
  
  sep: "/",                   /** パスの区切り
                               *  @type {string} */
  
  pathComponents: null,       /** パスの要素
                               *  @type {Array.<string>} */
  
  isMHTML: undefined,         /** mhtml: か
                               *  @type {boolean} */
  mhtmlFilename: undefined,   /** mhtml: のファイル名
                               *  @type {string} */
  mhtmlFragment: undefined,   /** mhtml: のフラグメント
                               *  @type {string} */
  
  driveLetter: undefined,     /** Windows のドライブ文字
                               *  @type {string} */
  extraSlash: undefined,      /** Windows の file の余分なスラッシュ
                               *  @type {string} */
  
  volumeLabel: undefined,     /** Mac OS 9 のボリュームラベル
                               *  @type {string} */
  
  data: undefined,            /** data: のデータ
                               *  @type {string} */
  
  scheme: undefined,          /** URI のスキーム
                               *  @type {string} */
  host: undefined,            /** URI のホスト
                               *  @type {string} */
  query: undefined,           /** URI のクエリ
                               *  @type {string} */
  fragment: undefined,        /** URI のフラグメント
                               *  @type {string} */
  
  /**
   * 複製する
   *
   * @returns {arPathInfo}
   */
  clone: function() {
    let info = new arPathInfo();
    
    info.type = this.type;
    info.sep = this.sep;
    info.pathComponents = this.pathComponents.slice(0);
    info.isMHTML = this.isMHTML;
    info.mhtmlFilename = this.mhtmlFilename;
    info.mhtmlFragment = this.mhtmlFragment;
    info.driveLetter = this.driveLetter;
    info.extraSlash = this.extraSlash;
    info.volumeLabel = this.volumeLabel;
    info.data = this.data;
    info.scheme = this.scheme;
    info.host = this.host;
    info.query = this.query;
    info.fragment = this.fragment;
    
    return info;
  }
};

/**
 * 仮想パス操作ユーティリティ
 * ローカルのパスや valid な URI ではない可能性のあるパスを操作する
 */
let arPathUtil = {
  /** パスの種類のマスク
   *  @type {number} */
  MASK_WINDOWS:                  0x000010,
  MASK_MACOX9:                   0x000020,
  MASK_URI:                      0x000100,
  MASK_ABSOLUTE:                 0x001000,
  MASK_RELATIVE:                 0x002000,
  MASK_DATA:                     0x010000,
  
  /** パスの種類
   *  @type {number} */
  TYPE_UNIX_LOCAL:               0x001001,
  TYPE_WINDOWS_LOCAL:            0x001011,
  TYPE_WINDOWS_NETWORK:          0x001012,
  TYPE_MACOS9_LOCAL:             0x001021,
  TYPE_URI:                      0x001101,
  TYPE_URI_NO_SCHEME:            0x001102,
  TYPE_URI_NO_SLASH:             0x011103,
  TYPE_WINDOWS_LOCAL_FILE_URI:   0x001111,
  TYPE_WINDOWS_NETWORK_FILE_URI: 0x001112,
  TYPE_RELATIVE:                 0x002001,
  
  /**
   * パスを連結、解決する
   *
   * @param   {string} baseDir
   *          親ディレクトリのパス
   *          末尾にセパレータが無くてもディレクトリ名として扱われる事に注意
   * @param   {string} sub
   *          相対パス、もしくは絶対パス
   * @returns {string}
   *          連結したパス
   */
  resolve: function(baseDir, sub) {
    if (!baseDir) {
      return sub;
    }
    
    let baseInfo = this.getPathInfo(baseDir);
    let subInfo = this.getPathInfo(sub);
    
    if (baseInfo.isMHTML &&
        subInfo.type == arPathUtil.TYPE_RELATIVE) {
      baseInfo.mhtmlFilename = sub;
      return this.createPath(baseInfo, false);
    }
    
    if (baseInfo.type & arPathUtil.MASK_DATA ||
        subInfo.type & arPathUtil.MASK_DATA) {
      return sub;
    }
    
    if (subInfo.type == arPathUtil.TYPE_URI_NO_SCHEME &&
        baseInfo.scheme != undefined) {
      return baseInfo.scheme + ":" + sub;
    }
    
    if (subInfo.type & arPathUtil.MASK_ABSOLUTE &&
        subInfo.type != arPathUtil.TYPE_UNIX_LOCAL) {
      /* TYPE_UNIX_LOCAL は URI のパスでもあるのでそのまま返さない */
      return sub;
    }
    
    if (subInfo.type == arPathUtil.TYPE_UNIX_LOCAL) {
      baseInfo.pathComponents = subInfo.pathComponents;
    } else {
      baseInfo.pathComponents
        = baseInfo.pathComponents.concat(subInfo.pathComponents);
    }
    
    baseInfo.query = subInfo.query;
    baseInfo.fragment = subInfo.fragment;
    
    this._simplifyPathComponents(baseInfo);
    
    return this.createPath(baseInfo, false);
  },
  
  /**
   * パスを比較しやすいように単純化する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {string}
   *          単純化したパス
   */
  unique: function(path) {
    let info = this.getPathInfo(path);
    let path = "";
    
    switch (info.type) {
      case arPathUtil.TYPE_WINDOWS_LOCAL: {
        path = ("file:///" + info.driveLetter.toLowerCase() + ":/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathUtil.TYPE_WINDOWS_NETWORK: {
        path = ("file://///" + info.host + "/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathUtil.TYPE_WINDOWS_LOCAL_FILE_URI: {
        path = ("file:///" + info.driveLetter.toLowerCase() + ":/"
                + info.pathComponents.join("/")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
      case arPathUtil.TYPE_WINDOWS_NETWORK_FILE_URI: {
        path = ("file://///" + info.host + "/"
                + info.pathComponents.join("/")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
      case arPathUtil.TYPE_URI: {
        path = (info.scheme + "://" + info.host + "/"
                + info.pathComponents.join("/")
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
      case arPathUtil.TYPE_URI_NO_SCHEME: {
        path = ("//" + info.host + "/"
                + info.pathComponents.join("/")
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
      case arPathUtil.TYPE_URI_NO_SLASH: {
        path = (info.scheme + ":" + info.data
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
      case arPathUtil.TYPE_UNIX_LOCAL: {
        path = ("/"
                + info.pathComponents.join("/"));
        break;
      }
      case arPathUtil.TYPE_MACOS9_LOCAL: {
        path = (info.volumeLabel + ":"
                + info.pathComponents.join(":"));
        break;
      }
      case arPathUtil.TYPE_RELATIVE: {
        path = (info.pathComponents.join("/")
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
        break;
      }
    }
    
    if (info.isMHTML) {
      path = ("mhtml:" + path
              + ((info.mhtmlFilename != undefined)
                 ? ("!" + info.mhtmlFilename) : "")
              + ((info.mhtmlFragment != undefined)
                 ? ("#" + info.mhtmlFragment) : ""));
    }
    
    return path;
  },
  
  /**
   * 親ディレクトリのパスを取得する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {string}
   *          親ディレクトリのパス
   */
  getBaseDir: function(path) {
    let info = this.getPathInfo(path);
    
    if (info.type & arPathUtil.MASK_DATA) {
      return path;
    }
    if (info.pathComponents.length) {
      info.pathComponents.pop();
      info.pathComponents.push("");
    }
    info.fragment = undefined;
    info.query = undefined;
    info.mhtmlFragment = undefined;
    
    return this.createPath(info);
  },
  
  /**
   * パス情報を取得する
   *
   * @param   {string} path
   *          対象のパス
   * @returns {arPathInfo}
   *          パス情報
   */
  getPathInfo: function(path) {
    let info = new arPathInfo();
    let re;
    
    /* -- IE の mhtml: -- */
    
    re = path.match(/^mhtml:([^!#]*)(?:!([^#]*))?(#(.*))?$/);
    if (re) {
      /**
       * mhtml:******!fuga.html
       * mhtml:******!fuga.html#section2
       */
      path = re[1];
      info.isMHTML = true;
      info.mhtmlFilename = re[2];
      if (re[3]) {
        info.mhtmlFragment = re[4];
      }
      /* path について処理を続ける */
    }
    
    /* -- 絶対パス -- */
    
    re = path.match(/^([A-Za-z]):([\/\\])(.*)$/);
    if (re) {
      /**
       * C:/hoge/fuga.html
       * C:\hoge\fuga.html
       */
      info.type = arPathUtil.TYPE_WINDOWS_LOCAL;
      info.driveLetter = re[1];
      info.sep = re[2];
      info.pathComponents = re[3].split(info.sep);
      return info;
    }
    
    re = path.match(/^\\\\([^\\]+)\\(.*)$/);
    if (re) {
      /**
       * \\host1\hoge\fuga.html
       */
      info.type = arPathUtil.TYPE_WINDOWS_NETWORK;
      info.sep = "\\";
      info.host = re[1];
      info.pathComponents = re[2].split(info.sep);
      return info;
    }
    
    re = path.match(/^file:\/\/(\/?)([A-Za-z]):([\/\\])([^#]*)(#(.*))?$/);
    if (re) {
      /**
       * file://C:/hoge/fuga.html
       * file:///C:\hoge\fuga.html#section2
       */
      info.type = arPathUtil.TYPE_WINDOWS_LOCAL_FILE_URI;
      info.extraSlash = re[1];
      info.driveLetter = re[2];
      info.sep = re[3];
      info.pathComponents = re[4].split(info.sep);
      if (re[5]) {
        info.fragment = re[6];
      }
      return info;
    }
    
    re = path.match(/^file:\/\/(\/{1,3})([^\/]+)\/([^#]*)(#(.*))?$/);
    if (re) {
      /**
       * file://///host1/hoge/fuga.html
       * file://///host1/hoge/fuga.html#section2
       */
      info.sep = "/";
      info.type = arPathUtil.TYPE_WINDOWS_NETWORK_FILE_URI;
      info.extraSlash = re[1];
      info.host = re[2];
      info.pathComponents = re[3].split(info.sep);
      if (re[4]) {
        info.fragment = re[5];
      }
      return info;
    }
    
    re = path.match(/^([A-Za-z0-9\-]+):\/\/([^\/]*)\/([^\?#]*)(\?([^#]*))?(#(.*))?$/);
    if (re) {
      /**
       * http://host1/hoge/fuga.html
       * http://host1/hoge/fuga.cgi?name=value#section2
       */
      info.type = arPathUtil.TYPE_URI;
      info.sep = "/";
      info.scheme = re[1];
      info.host = re[2];
      info.pathComponents = re[3].split(info.sep);
      if (re[4]) {
        info.query = re[5];
      }
      if (re[6]) {
        info.fragment = re[7];
      }
      return info;
    }

    re = path.match(/^(about|data):([^#]*)(#(.*))?$/);
    if (re) {
      /**
       * about:blank
       * data:text/html,hogehoge#section2
       */
      info.type = arPathUtil.TYPE_URI_NO_SLASH;
      info.sep = "/";
      info.scheme = re[1];
      info.data = re[2];
      if (re[3]) {
        info.fragment = re[4];
      }
      return info;
    }
    
    re = path.match(/^\/\/([^\/]*)\/([^\?#]*)(\?([^#]*))?(#(.*))?$/);
    if (re) {
      /**
       * //host1/hoge/fuga.html
       * //host1/hoge/fuga.cgi?name=value#section2
       */
      info.type = arPathUtil.TYPE_URI_NO_SCHEME;
      info.sep = "/";
      info.host = re[1];
      info.pathComponents = re[2].split(info.sep);
      if (re[3]) {
        info.query = re[4];
      }
      if (re[5]) {
        info.fragment = re[6];
      }
      return info;
    }
    
    re = path.match(/^\/(.*)$/);
    if (re) {
      /**
       * /hoge/fuga.html
       */
      info.type = arPathUtil.TYPE_UNIX_LOCAL;
      info.sep = "/";
      info.pathComponents = re[1].split(info.sep);
      return info;
    }
    
    re = path.match(/^([^:]+):(.+)$/);
    if (re) {
      /**
       * Macintosh HD:hoge:fuga.html
       */
      info.type = arPathUtil.TYPE_MACOS9_LOCAL;
      info.sep = ":";
      info.volumeLabel = re[1];
      info.pathComponents = re[2].split(info.sep);
      return info;
    }
    
    /* -- 相対パス -- */
    
    re = path.match(/^\.\.?([\\\/:]).*$/);
    if (re) {
      /**
       * ./hoge/fuga.html
       * ..\hoge\fuga.html
       * ..:hoge:fuga.html (あるのかどうか知らない)
       */
      info.type = arPathUtil.TYPE_RELATIVE;
      info.sep = re[1];
      info.pathComponents = path.split(info.sep);
      return info;
    }
    
    info.type = arPathUtil.TYPE_RELATIVE;
    
    /* **** FIXME: この方法では誤検出もありうる **** */
    if (path.indexOf("/") != -1) {
      /* とりあえず最もありうる / を最優先にする */
      info.sep = "/";
    } else if (path.indexOf("\\") != -1) {
      info.sep = "\\";
    } else if (path.indexOf(":") != -1) {
      info.sep = ":";
    } else {
      info.sep = "/";
    }
    
    re = path.match(/^([^\?#]*)(\?([^#]*))?(#(.*))?$/);
    if (re) {
      /**
       * hoge/fuga.cgi?name=value#section2
       */
      info.pathComponents = re[1].split(info.sep);
      if (re[2]) {
        info.query = re[3];
      }
      if (re[4]) {
        info.fragment = re[5];
      }
    } else {
      /**
       * hoge/fuga.html
       */
      info.pathComponents = path.split(info.sep);
    }
    return info;
  },
  
  /**
   * パス情報からパスを生成する
   *
   * @param   {arPathInfo} info
   *          対象のパス情報
   * @param   {boolean} ignoreMHTML
   *          isMHTML を無視するか
   * @returns {arPathInfo}
   *          パス
   */
  createPath: function(info, ignoreMHTML) {
    if (!ignoreMHTML && info.isMHTML) {
      return ("mhtml:" + this.createPath(info, true)
              + ((info.mhtmlFilename != undefined)
                 ? ("!" + info.mhtmlFilename) : "")
              + ((info.mhtmlFragment != undefined)
                 ? ("#" + info.mhtmlFragment) : ""));
    }
    
    switch (info.type) {
      case arPathUtil.TYPE_WINDOWS_LOCAL: {
        return (info.driveLetter + ":" + info.sep
                + info.pathComponents.join(info.sep));
      }
      case arPathUtil.TYPE_WINDOWS_NETWORK: {
        return ("\\\\" + info.host + info.sep
                + info.pathComponents.join(info.sep));
      }
      case arPathUtil.TYPE_WINDOWS_LOCAL_FILE_URI: {
        return ("file://" + info.extraSlash
                + info.driveLetter + ":" + info.sep
                + info.pathComponents.join(info.sep)
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
      case arPathUtil.TYPE_WINDOWS_NETWORK_FILE_URI: {
        return ("file://" + info.extraSlash
                + info.host + info.sep
                + info.pathComponents.join(info.sep)
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
      case arPathUtil.TYPE_URI: {
        return (info.scheme + "://" + info.host + info.sep
                + info.pathComponents.join(info.sep)
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
      case arPathUtil.TYPE_URI_NO_SCHEME: {
        return ("//" + info.host + info.sep
                + info.pathComponents.join(info.sep)
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
      case arPathUtil.TYPE_URI_NO_SLASH: {
        return (info.scheme + ":" + info.data
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
      case arPathUtil.TYPE_UNIX_LOCAL: {
        return (info.sep
                + info.pathComponents.join(info.sep));
      }
      case arPathUtil.TYPE_MACOS9_LOCAL: {
        return (info.volumeLabel + info.sep
                + info.pathComponents.join(info.sep));
      }
      case arPathUtil.TYPE_RELATIVE: {
        return (info.pathComponents.join(info.sep)
                + ((info.query != undefined) ? ("?" + info.query) : "")
                + ((info.fragment != undefined) ? ("#" + info.fragment) : ""));
      }
    }
    
    return "";
  },
  
  /* ==== private ==== */
  
  /**
   * パスの要素を簡略化する
   *
   * @param   {arPathInfo} info
   *          パス情報
   */
  _simplifyPathComponents: function(info) {
    let isDir = false;
    if (info.pathComponents.length &&
        info.pathComponents[info.pathComponents.length - 1] == "") {
      isDir = true;
    }
    
    info.pathComponents
      = info.pathComponents.reduce(function(pathComponents, component) {
        if (component == ".") {
          return pathComponents;
        } else if (component == "..") {
          return pathComponents.slice(0, pathComponents.length - 1);
        } else if (component) {
          return pathComponents.concat([component]);
        }
        return pathComponents;
      }, []);
    if (isDir) {
      info.pathComponents.push("");
    }
  }
};
