"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arUconv.jsm");
XPCOMUtils.defineLazyGetter(this, "arFileUtil", function() {
  XPCOMUtils.importRelative(this, "arFileUtil.jsm");
  return arFileUtil;
});

let EXPORTED_SYMBOLS = ["arPrefClass"];

/**
 * 設定操作と既定値
 */
function arPrefClass() {
  this._defPrefs = new Map();
  this._prefGetter = new Map();
  this._prefSetter = new Map();
  this._observers = [];
}
arPrefClass.prototype = {
  prefix: "extensions.", /** 設定のプレフィックス
                          *  @type {string} */
  
  /**
   * 設定名一覧
   * @type {Array.<string>}
   */
  get names() {
    return [name for ([name,] of this._defPrefs)];
  },
  
  /**
   * 設定値の型を返す
   *
   * @param   {string} name
   *          設定名
   * @returns {string}
   *          設定値の型
   * @throws  {Error("UNSUPPORTED_PREFERENCE")}
   *          未知の設定名
   */
  typeof: function(name) {
    return (typeof(this.getDefPref(name)));
  },
  
  /* ---- 既定値 ---- */
  
  /**
   * 既定値を返す
   *
   * @param   {string} name
   *          設定名
   * @returns {string}
   *          設定値の型
   * @throws  {Error("UNSUPPORTED_PREFERENCE")}
   *          未知の設定名
   */
  getDefPref: function(name) {
    if (!this._defPrefs.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    
    return this._defPrefs.get(name);
  },
  
  /* ---- ローカルのリスナ ---- */
  
  /**
   * 設定を監視する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {Function} func
   *          設定が変更された時に実行する関数
   * @param   {ANY} owner
   *          登録者
   * @param   {boolean} run
   *          登録と同時に関数を実行するか
   */
  addListener: function(domain, func, owner, run) {
    this.addGlobalListener(this.prefix + domain, func, owner, run);
  },
  
  /**
   * 設定が変更されたイベントのリスナを削除する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {Function} func
   *          設定が変更された時に実行する関数
   */
  removeListener: function(domain, func) {
    this.removeGlobalListener(this.prefix + domain, func);
  },

  /* ---- グローバルのリスナ ---- */
  
  /**
   * 設定を監視する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {Function} func
   *          設定が変更された時に実行する関数
   * @param   {ANY} owner
   *          登録者
   * @param   {boolean} run
   *          登録と同時に関数を実行するか
   */
  addGlobalListener: function(domain, func, owner, run) {
    let observer = {
      QueryInterface: XPCOMUtils.generateQI([Ci.nsIObserver]),
      
      /* ==== nsIObserver インターフェース ==== */
      
      /**
       * 通知を受けたイベント
       *
       * @param   {nsISupports} subject
       *          変更やアクションのあったオブジェクト
       * @param   {string} topic
       *          イベントの種類
       * @param   {string} data
       *          イベントのデータ
       */
      observe: function(subject, topic, data) {
        if (topic == "nsPref:changed") {
          func();
        }
      }
    };
    
    Services.prefs.addObserver(domain, observer, false);
    
    this._observers.push([domain, func, owner, observer]);
    
    if (run) {
      func();
    }
  },

  /**
   * 設定が変更されたイベントのリスナを削除する
   *
   * @param   {string} domain
   *          対象のドメイン
   * @param   {Function} func
   *          設定が変更された時に実行する関数
   */
  removeGlobalListener: function(domain, func) {
    this._observers = this._observers.filter(function([d, f, o, ob]) {
      if (d == domain && f == func) {
        Services.prefs.removeObserver(d, ob);
        return false;
      }
      return true;
    }.bind(this));
  },

  /**
   * 設定が変更されたイベントのリスナを削除する
   *
   * @param   {ANY} owner
   *          登録者
   */
  removeAllListeners: function(owner) {
    this._observers = this._observers.filter(function([d, f, o, ob]) {
      if (o == owner) {
        Services.prefs.removeObserver(d, ob);
        return false;
      }
      return true;
    }.bind(this));
  },

  /* ---- 設定操作 ---- */
  
  /**
   * 設定値を読み込む
   *
   * @param   {string} name
   *          設定名
   * @returns {(boolean|string|number)}
   *          設定値
   *          無ければ既定値を返す
   * @throws  {Error("UNSUPPORTED_PREFERENCE")}
   *          未知の設定名
   */
  getPref: function(name) {
    if (!this._prefGetter.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    return this._prefGetter.get(name)();
  },
  
  /**
   * 設定値を保存する
   *
   * @param   {string} name
   *          設定名
   * @param   {(boolean|string|number)} value
   *          設定値
   * @throws  {Error("UNSUPPORTED_PREFERENCE")}
   *          未知の設定名
   */
  setPref: function(name, value) {
    if (!this._prefSetter.has(name)) {
      throw new Error("UNSUPPORTED_PREFERENCE: " + name);
    }
    return this._prefSetter.get(name)(value);
  },

  /* ---- ファイルから設定読み込み ---- */
  
  /**
   * ファイルからの設定の読み込みを開始する
   *
   * @param   {nsIFile} file
   *          読み込む対象のファイル
   *          null にすると既定値を書いたファイル扱い
   * @returns {MakumakuPrefFileInputObject}
   *          設定入力オブジェクト
   *          ファイルが不正な場合は null
   *            実体は {Map.<string,(boolean|string|number)>}
   *              <string 設定名, boolean/string/number 設定値>
   */
  initPrefFromFile: function(file) {
    if (file == null) {
      return new Map();
    }
    
    let text = fromUTF8(arFileUtil.readData(file));
    
    try {
      let map = JSON.parse(text);
      return new Map([[name, map[name]] for (name in map)]);
    } catch (e) {
      Cu.reportError("File format error\n"
                     + e + "\n" + e.stack);
      return null;
    }
  },
  
  /**
   * ファイルからの設定の読み込みを終了する
   *
   * @param   {MakumakuPrefFileInputObject} input
   *          設定入力オブジェクト
   */
  closePrefFromFile: function(input) {
    /* 特にやる事はない */
  },
  
  /**
   * ファイルから設定値を読み込む
   *
   * @param   {MakumakuPrefFileInputObject} input
   *          設定入力オブジェクト
   * @param   {string} name
   *          設定名
   * @returns {(boolean|string|number)}
   *          設定値
   *          無ければ既定値を返す
   * @throws  {Error("UNSUPPORTED_PREFERENCE")}
   *          未知の設定名
   */
  getPrefFromFile: function(input, name) {
    if (input.has(name)) {
      return input.get(name);
    }
    
    return this.getDefPref(name);
  },
  
  /* ---- ファイルへ設定保存 ---- */

  /**
   * ファイルへの設定の保存を開始する
   *
   * @param   {nsIFile} file
   *          保存対象のファイル
   * @returns {MakumakuPrefFileOutputObject}
   *          設定出力オブジェクト
   */
  initPrefToFile: function(file) {
    return {
      map: new Map(),
      stream: arFileUtil.openOutputStream(file)
    };
  },
  
  /**
   * ファイルへの設定の保存を終了する
   *
   * @param   {MakumakuPrefFileOutputObject} output
   *          設定出力オブジェクト
   */
  closePrefToFile: function(output) {
    let obj = {};
    for (let [name, value] of output.map) {
      obj[name] = value;
    }
    
    let data = toUTF8(JSON.stringify(obj, null, "  "));
    
    output.stream.write(data, data.length);
    output.stream.close();
    
    output.map = null;
    output.stream = null;
  },
  
  /**
   * 設定を保存する
   * 
   * @param   {MakumakuPrefFileOutputObject} output
   *          設定出力オブジェクト
   * @param   {string} name
   *          設定名
   * @param   {(boolean|string|number)} value
   *          既定値
   */
  setPrefToFile: function(output, name, value) {
    output.map.set(name, value);
  },
  
  /* ==== private ==== */
  
  /** 既定値
   *  @type {Map.<string,(boolean|string|number)>}  */
  _defPrefs: null,

  /** 設定値読み込み関数マップ
   *  @type {Map.<string,Function>}  */
  _prefGetter: null,

  /** 設定値保存関数マップ
   *  @type {Map.<string,Function>}  */
  _prefSetter: null,
  
  _observers: null,  /** 設定が変更されたイベントのオブザーバ
                      *  @type {Array.<[string,Function,ANY,nsIObserver]>}
                      *   [[string ドメイン, Function 実行する関数, ANY 登録者
                      *     nsIObserver オブザーバ], ...] */

  /**
   * 既定値を作成する
   *
   * @param   {object.<string,(boolean|string|number)>} prefs
   *          既定値
   */
  _setDefPrefs: function(prefs) {
    this._defPrefs = new Map([[name, prefs[name]] for (name in prefs)]);
    this._createDefaultBranch();
    this._createPrefGetterSetter();
  },

  /**
   * デフォルト設定を作成する
   */
  _createDefaultBranch: function() {
    let branch = Services.prefs.getDefaultBranch("");
    for (let [name, value] of this._defPrefs) {
      switch (typeof(value)) {
        case "boolean": {
          branch.setBoolPref(this.prefix + name, value);
          break;
        }
        case "string": {
          branch.setCharPref(this.prefix + name, toUTF8(value));
          break;
        }
        case "number": {
          branch.setIntPref(this.prefix + name, value);
          break;
        }
      }
    }
  },
  
  /**
   * 設定操作関数マップを作成する
   */
  _createPrefGetterSetter: function() {
    for (let [name, value] of this._defPrefs) {
      let pname = this.prefix + name;
      switch (typeof(value)) {
        case "boolean": {
          this._prefGetter.set(name, function() {
            return Services.prefs.getBoolPref(pname);
          });
          this._prefSetter.set(name, function(value) {
            return Services.prefs.setBoolPref(pname, value);
          });
          break;
        }
        case "string": {
          this._prefGetter.set(name, function() {
            return fromUTF8(Services.prefs.getCharPref(pname));
          });
          this._prefSetter.set(name, function(value) {
            return Services.prefs.setCharPref(pname, toUTF8(value));
          });
          break;
        }
        case "number": {
          this._prefGetter.set(name, function() {
            return Services.prefs.getIntPref(pname);
          });
          this._prefSetter.set(name, function(value) {
            return Services.prefs.setIntPref(pname, value);
          });
          break;
        }
      }
    }
  }
};
