"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arShutdownEvent.jsm");

let EXPORTED_SYMBOLS = ["arStylesheetLoader"];

/**
 * スタイルシート適用
 */
let arStylesheetLoader = {
  /**
   * スタイルシートを登録する
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   * @param   {number} type
   *          スタイルシートの種類
   *            Ci.nsIStyleSheetService.AGENT_SHEET
   *            Ci.nsIStyleSheetService.USER_SHEET
   */
  load: function(uriSpec, type) {
    if (this._sheets.some(function([u, t]) u == uriSpec && t == type)) {
      /* 多重登録はしない */
      return;
    }
    let uri = Services.io.newURI(uriSpec, null, null);
    this._styleSheetService.loadAndRegisterSheet(uri, type);
    this._sheets.push([uriSpec, type]);
  },
  
  /**
   * スタイルシートの登録を解除する
   * 登録したものはシャットダウン時に自動的に解除される
   *
   * @param   {string} uriSpec
   *          スタイルシートの URI
   * @param   {number} type
   *          スタイルシートの種類
   *            Ci.nsIStyleSheetService.AGENT_SHEET
   *            Ci.nsIStyleSheetService.USER_SHEET
   */
  unload: function(uriSpec, type) {
    let uri = Services.io.newURI(uriSpec, null, null);
    if (this._styleSheetService.sheetRegistered(uri, type)) {
      this._styleSheetService.unregisterSheet(uri, type);
    }
    
    this._sheets
      = this._sheets.filter(function([u, t]) u != uriSpec && t != type);
  },
  
  /* ==== arIShutdownEventListener ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    this._sheets.slice().forEach(function([uriSpec, type]) {
      this.unload(uriSpec, type);
    }.bind(this));
  },
  
  /* ==== private ==== */
  
  _sheets: [],              /** 登録したスタイルシート
                             *  @type {Array.<[string,number]>}
                             *    [[string シートの URI,
                             *      number シートの種類]...] */
  _styleSheetService: null, /** スタイルシートサービス
                             *  @type {nsIStyleSheetService} */
  
  /**
   * 初期化処理
   */
  init: function() {
    this._styleSheetService
      = Cc["@mozilla.org/content/style-sheet-service;1"]
      .getService(Ci.nsIStyleSheetService);
    
    arShutdownEvent.addListener(this);
  }
};
arStylesheetLoader.init();
