"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;

let EXPORTED_SYMBOLS = ["arUconv",
                        "toUTF8", "fromUTF8"];

/**
 * 文字列変換器
 */
let arUconv = {
  /**
   * Unicode から変換する
   *
   * @param   {string} text
   *          変換する文字列
   * @param   {string} charset
   *          変換先の文字コード
   * @returns {string}
   *          変換した文字列
   */
  fromUnicode: function(text, charset) {
    try {
      let converter
        = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
        .createInstance(Ci.nsIScriptableUnicodeConverter);
      converter.charset = charset;
      return converter.ConvertFromUnicode(text) + converter.Finish();
    } catch (e) {
    }
    
    return text;
  },
  
  /**
   * Unicode から変換する
   * 変換できないものは実体参照に置き換える
   *
   * @param   {string} text
   *          変換する文字列
   * @param   {string} charset
   *          変換先の文字コード
   * @returns {string}
   *          変換した文字列
   */
  fromUnicodeWithEntity: function(text, charset) {
    let text2
      = this.toUnicode(this.fromUnicode(text, charset), charset);
    
    let result = "";
    for (;;) {
      let p = text2.indexOf("?");
      if (p == -1) {
        result += text2;
        break;
      }
      
      if (text.charAt(p) != "?") {
        /* 変換できなかったため ? になった */
        
        result += text2.substr(0, p);
        
        let code1 = text.charCodeAt(p);
        if (code1 >= 0xd800 && code1 <= 0xdbff) {
          /* サロゲートペア */
          let code2 = text.charCodeAt(p + 1);
          let code = (code1 - 0xd800) * 0x400 + (code2 - 0xdc00) + 0x10000;
          
          result += "&#" + code + ";";
          
          text = text.substr(p + 2);
          text2 = text2.substr(p + 1);
        } else {
          result += "&#" + code1 + ";";
          text = text.substr(p + 1);
          text2 = text2.substr(p + 1);
        }
      } else {
        /* 本当に ? だった */
        result += text2.substr(0, p + 1);
        text = text.substr(p + 1);
        text2 = text2.substr(p + 1);
      }
    }
    
    return this.fromUnicode(result, charset);
  },
  
  /**
   * Unicode に変換する
   *
   * @param   {string} text
   *          変換する文字列
   * @param   {string} charset
   *          変換元の文字コード
   * @returns {string}
   *          変換した文字列
   */
  toUnicode: function(text, charset) {
    try {
      let converter
        = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
        .createInstance(Ci.nsIScriptableUnicodeConverter);
      converter.charset = charset;
      return converter.ConvertToUnicode(text);
    } catch (e) {
    }
    
    return text;
  }
};

/**
 * Unicode から UTF-8 に変換する
 *
 * @param   {string} text
 *          Unicode の文字列
 * @returns {string}
 *          UTF-8 の文字列
 */
function toUTF8(text) {
  return arUconv.fromUnicode(text, "utf-8");
}

/**
 * UTF-8 から Unicode に変換する
 *
 * @param   {string} text
 *          UTF-8 の文字列
 * @returns {string}
 *          Unicode の文字列
 */
function fromUTF8(text) {
  return arUconv.toUnicode(text, "utf-8");
}
