"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arArrayUtil.jsm");

let EXPORTED_SYMBOLS = ["arWinRegUtil"];

/**
 * Windows レジストリ操作ユーティリティ
 */
let arWinRegUtil = {
  /**
   * HKEY_LOCAL_MACHINE 内の値を取得する
   *
   * @param   {Array.<string>} paths
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {(string|number)}
   *          値
   */
  getHKLMValue: function(paths, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE,
                          paths, name);
  },
  
  /**
   * HKEY_CURRENT_USER 内の値を取得する
   *
   * @param   {Array.<string>} paths
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {(string|number)}
   *          値
   */
  getHKCUValue: function(paths, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_CURRENT_USER,
                          paths, name);
  },
  
  /**
   * HKEY_CLASSES_ROOT 内の値を取得する
   *
   * @param   {Array.<string>} paths
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {(string|number)}
   *          値
   */
  getHKCRValue: function(paths, name) {
    return this._getValue(Ci.nsIWindowsRegKey.ROOT_KEY_LOCAL_MACHINE,
                          paths, name);
  },
  
  /* ==== private ==== */
  
  /**
   * 指定されたルートキー内の値を取得する
   *
   * @param   {number} rootKey
   *          ルートキー
   *            Ci.nsIWindowsRegKey.ROOT_KEY_*
   * @param   {Array.<string>} paths
   *          レジストリのパス
   * @param   {string} name
   *          パス内の名前
   * @returns {(string|number)}
   *          値
   */
  _getValue: function(rootKey, paths, name) {
    let regKey
      = Cc["@mozilla.org/windows-registry-key;1"]
      .createInstance(Ci.nsIWindowsRegKey);
    regKey.open(rootKey, paths[0], Ci.nsIWindowsRegKey.ACCESS_READ);
    
    let subRegKeys
      = arArrayUtil.scanl(
        paths.slice(1),
        function(regKey, path)
        (regKey
         ? (regKey.hasChild(path)
            ? regKey.openChild(path, Ci.nsIWindowsRegKey.ACCESS_READ)
            : null)
         : null),
        regKey);
    
    let targetKey = subRegKeys[subRegKeys.length - 1];
    
    let value = null;
    if (targetKey) {
      switch (targetKey.getValueType(name)) {
        case Ci.nsIWindowsRegKey.TYPE_STRING: {
          value = targetKey.readStringValue(name);
          break;
        }
        case Ci.nsIWindowsRegKey.TYPE_BINARY: {
          value = targetKey.readBinaryValue(name);
          break;
        }
        case Ci.nsIWindowsRegKey.TYPE_INT: {
          value = targetKey.readIntValue(name);
          break;
        }
        case Ci.nsIWindowsRegKey.TYPE_INT64: {
          value = targetKey.readInt64Value(name);
          break;
        }
      }
    }
    
    subRegKeys.forEach(function(key) (key && key.close()));
    regKey.close();
    
    return value;
  }
};
