"use strict";

let { classes: Cc, interfaces: Ci, results: Cr, utils: Cu } = Components;
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
XPCOMUtils.importRelative(this, "arShutdownEvent.jsm");
XPCOMUtils.defineLazyGetter(this, "arOverlayLoader", function() {
  XPCOMUtils.importRelative(this, "arOverlayLoader.jsm");
  return arOverlayLoader;
});

let EXPORTED_SYMBOLS = ["arWindowHookLoader"];

/**
 * ウィンドウイベントリスナ
 *
 * @implements nsIWindowMediatorListener
 */
function arWindowListener(loader) {
  this._loader = loader;
}
arWindowListener.prototype = {
  QueryInterface:    XPCOMUtils.generateQI([Ci.nsIWindowMediatorListener]),
  
  /**
   * 破棄
   */
  destruct: function() {
    this._loader = null;
  },
  
  /* ==== nsIWindowMediatorListener インターフェース ==== */
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  onOpenWindow: function(window) {
    window = window.QueryInterface(Ci.nsIInterfaceRequestor)
      .getInterface(Ci.nsIDOMWindow);
    
    let loadFunc = function() {
      window.removeEventListener("load", loadFunc, false);
      
      this._loader.onLoad(window);
    }.bind(this);
    
    window.addEventListener("load", loadFunc, false);
  },
  
  /**
   * ウィンドウを閉じたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  onCloseWindow: function(window) {
    /* onOpenWindow と window が一致しないので使用しない */
  },
  
  /**
   * ウィンドウを開いたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {string} title
   *          タイトル
   */
  onWindowTitleChange: function(window, title) {
  },
  
  /* ==== private ==== */
  
  _loader: null /** ウィンドウフックローダー
                 *  @type {arWindowHookLoader} */
};

/**
 * ウィンドウフックローダー
 *
 * @param   {object.<string,object>} params
 *          パラメータ
 *            <string  対象の URI,
 *             {
 *               overlay:  {Function}
 *                         オーバーレイ登録関数
 *               props:    {object}
 *                         ウィンドウに付与するプロパティ
 *               genHooks: {Function}
 *                         フックを生成する関数
 *                           function()
 *                             *returns {Array.<arIWindowHook>}
 *               hooks:    {Array.<arIWindowHook>}
 *                         フック
 *             }>
 * @param   {Array.<string>} dialogs
 *          ダイアログの URI
 * @implements arIStartupEventListener
 * @implements arIShutdownEventListener
 */
function arWindowHookLoader(params, dialogs) {
  this._params = params;
  this._dialogs = dialogs;
  
  this._listener = new arWindowListener(this);
  this._registered = new Map();
  
  this.onUnload = this.onUnload.bind(this);
}
arWindowHookLoader.prototype = {
  /**
   * 破棄
   */
  destruct: function() {
    if (this._listener) {
      this._listener.destruct();
      this._listener = null;
    }
    
    this._params = null;
    this._dialogs = null;
    this._listener = null;
  },
  
  /**
   * ウィンドウがロードされたイベント
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   */
  onLoad: function(window) {
    this._loadInto(window, false, -1);
  },
  
  /**
   * ウィンドウに登録するアンロードイベント
   *
   * @param   {Event} event
   *          対象のイベント
   */
  onUnload: function(event) {
    this._unloadFrom(event.target.defaultView, false, -1);
  },
  
  /* ==== arIStartupEventListener インターフェース ==== */
  
  /**
   * スタートアップイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onStartup: function(data, reason) {
    for (let window of XPCOMUtils.IterSimpleEnumerator(
      Services.wm.getEnumerator(null), Ci.nsIDOMWindow)) {
      this._loadInto(window, true, reason);
    }
    
    Services.wm.addListener(this._listener);
    
    arShutdownEvent.addListener(this);
  },
  
  /* ==== arIShutdownEventListener インターフェース ==== */
  
  /**
   * シャットダウンイベント
   *
   * @param   {object} data
   *          Bootstrap データ
   * @param   {number} reason
   *          理由
   */
  onShutdown: function(data, reason) {
    Services.wm.removeListener(this._listener);
    
    for (let window of XPCOMUtils.IterSimpleEnumerator(
      Services.wm.getEnumerator(null), Ci.nsIDOMWindow)) {
      this._unloadFrom(window, true, reason);
      
      if (this._dialogs.indexOf(window.document.documentURI) != -1) {
        window.close();
      }
    }
    
    this.destruct();
  },
  
  /* ==== private ==== */
  
  _registered: null, /** 組み込みリスト
                      *  @type {Map.<nsIXULWindow,object>} */
  _listener: null,   /** ウィンドウイベントリスナ
                      *  @type {arWindowListener} */
  
  _params: null,     /** パラメータ
                      *  @type {object} */
  _dialogs: null,    /** ダイアログの URI
                      *  @type {Array.<string>} */
  
  /**
   * ウィンドウにフックをロードする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {boolean} isStartup
   *          スタートアップイベントか
   * @param   {number} reason
   *          スタートアップの理由
   */
  _loadInto: function(window, isStartup, reason) {
    let location = window.document.documentURI;
    if (!(location in this._params)) {
      return;
    }
    
    let param = this._params[location];
    
    if ("overlay" in param && param.overlay) {
      arOverlayLoader.loadInto(window, param.overlay, isStartup);
    }
    
    if ("props" in param) {
      for (let name in param.props) {
        window[name] = param.props[name];
      }
    }
    
    let list;
    if ("genHooks" in param) {
      list = param.genHooks().map(function(n) new n(window));
    } else if ("hooks" in param) {
      list = param.hooks.map(function(n) new n(window));
    }
    
    list.forEach(function(item) item.load(isStartup, reason));

    this._registered.set(window, list);
    
    window.addEventListener("unload", this.onUnload, false);
  },
  
  /**
   * ウィンドウからフックをアンロードする
   *
   * @param   {nsIXULWindow} window
   *          対象のウィンドウ
   * @param   {boolean} isShutdown
   *          シャットダウンイベントか
   * @param   {number} reason
   *          シャットダウンの理由
   */
  _unloadFrom: function(window, isShutdown, reason) {
    if (window.document) {
      let location = window.document.documentURI;
      if (location in this._params) {
        let param = this._params[location];
        
        if ("props" in param) {
          for (let name in param.props) {
            delete window[name];
          }
        }
        
        if ("overlay" in param && param.overlay) {
          arOverlayLoader.unloadFrom(window, param.overlay, isShutdown);
        }
      }
    }
    
    window.removeEventListener("unload", this.onUnload, false);
    
    if (this._registered.has(window)) {
      let list = this._registered.get(window);
      
      list.forEach(function(item) {
        item.unload(isShutdown, reason);
        item.destruct();
      });
    }
    this._registered.delete(window);
  }
};
